Package: hespdiv
Type: Package
Title: Hierarchical Spatial Data Subdivision into Topologically
        Contiguous Units
Version: 1.2.10
Authors@R: c(
    person(given = "Liudas",
           family = "Daumantas",
           role = c("aut", "cre"),
           email = "liudas.daumantas@chgf.vu.lt"),
    person(given = "Andrej",
           family = "Spiridonov",
           role = "aut",
           ),
    person("Joseph", "O'Rourke",
           role = c("ctb", "cph"),
           comment = "Author and copyright holder of point-in-polygon code in src/pip.c"),
    person("Min", "Xu",
           role = "ctb",
           comment = "Contributor to point-in-polygon code in src/pip.c"))
Maintainer: Liudas  Daumantas <liudas.daumantas@chgf.vu.lt>
Description: Implementation of the HespDiv framework for hierarchical spatial subdivision of geographical occurrence data. The main function hespdiv() performs iterative spatially constrained subdivision of a study area to identify topologically contiguous clusters in geographic space using user-defined or preset subdivision methods. Additional functions provide tools for analysing subdivision results, visualizing hierarchical spatial structures, and evaluating robustness through sensitivity analyses and statistical testing. Some examples use the optional HDData data package, which is available from GitHub at Liudas-Dau/hespdiv_data. The methodology is described in Daumantas and Spiridonov (2024) <doi:10.1111/pala.12702>.
License: GPL-3
Encoding: UTF-8
RoxygenNote: 7.3.3
OS_type: windows
LazyData: true
URL: https://doi.org/10.1111/pala.12702,
        https://github.com/Liudas-Dau/hespdiv,
        https://github.com/Liudas-Dau/hespdiv_data
Imports: pracma, DescTools, ggplot2, viridis, ggrepel, grDevices,
        RColorBrewer, rgl, scales, graphics, igraph, gridExtra, grid,
        gridGraphics, stats, magick, future.apply, future, rlang
Suggests: HDData
Depends: R (>= 4.0)
NeedsCompilation: yes
Packaged: 2026-05-18 07:22:53 UTC; liuda
Author: Liudas Daumantas [aut, cre],
  Andrej Spiridonov [aut],
  Joseph O'Rourke [ctb, cph] (Author and copyright holder of
    point-in-polygon code in src/pip.c),
  Min Xu [ctb] (Contributor to point-in-polygon code in src/pip.c)
Repository: CRAN
Date/Publication: 2026-05-21 13:10:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(plot,nullhespdiv)
S3method(print,hespdiv)
S3method(print,nullhespdiv)
export(blok3d)
export(change_base)
export(create_gif)
export(cross_comp)
export(dendro)
export(depend_splits)
export(get_data)
export(group_effect)
export(hespdiv)
export(hsa)
export(hsa_detailed)
export(hsa_quant)
export(hsa_sample_constrained)
export(nulltest)
export(plot_cs_hsa)
export(plot_hespdiv)
export(plot_hsa)
export(plot_hsa_q)
export(poly_scheme)
export(polypop)
export(remove_splits)
export(taxon_effect)
importFrom(RColorBrewer,brewer.pal.info)
importFrom(future,availableCores)
importFrom(future,multisession)
importFrom(future,plan)
importFrom(future.apply,future_lapply)
importFrom(ggplot2,aes)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,element_rect)
importFrom(ggplot2,geom_path)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_segment)
importFrom(ggplot2,geom_text)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,ggtitle)
importFrom(ggplot2,guide_colourbar)
importFrom(ggplot2,guide_legend)
importFrom(ggplot2,guides)
importFrom(ggplot2,scale_size_area)
importFrom(ggplot2,scale_size_continuous)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_set)
importFrom(ggplot2,theme_void)
importFrom(ggrepel,geom_label_repel)
importFrom(grDevices,adjustcolor)
importFrom(grDevices,chull)
importFrom(grDevices,colors)
importFrom(grDevices,dev.list)
importFrom(grDevices,dev.new)
importFrom(grDevices,dev.off)
importFrom(grDevices,dev.set)
importFrom(grDevices,png)
importFrom(graphics,boxplot)
importFrom(graphics,hist)
importFrom(graphics,layout)
importFrom(graphics,lines)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(graphics,text)
importFrom(grid,grid.draw)
importFrom(grid,grid.grabExpr)
importFrom(grid,grid.newpage)
importFrom(gridExtra,arrangeGrob)
importFrom(gridGraphics,grid.echo)
importFrom(igraph,E)
importFrom(igraph,V)
importFrom(igraph,ends)
importFrom(igraph,graph_from_data_frame)
importFrom(igraph,layout_as_tree)
importFrom(igraph,vcount)
importFrom(magick,image_animate)
importFrom(magick,image_read)
importFrom(magick,image_write)
importFrom(pracma,poly_center)
importFrom(pracma,polyarea)
importFrom(rgl,identify3d)
importFrom(rgl,lines3d)
importFrom(rgl,open3d)
importFrom(rgl,par3d)
importFrom(rgl,persp3d)
importFrom(rgl,plot3d)
importFrom(rgl,points3d)
importFrom(rgl,polygon3d)
importFrom(rgl,rgl.ids)
importFrom(rgl,rgl.pop)
importFrom(rgl,rgl.snapshot)
importFrom(rgl,rotate3d)
importFrom(rgl,rotationMatrix)
importFrom(rgl,text3d)
importFrom(rgl,view3d)
importFrom(rlang,.data)
importFrom(scales,alpha)
importFrom(stats,aggregate)
importFrom(stats,density)
importFrom(stats,na.omit)
importFrom(stats,quantile)
importFrom(stats,setNames)
importFrom(stats,spline)
importFrom(utils,combn)
importFrom(viridis,scale_color_viridis)
useDynLib(hespdiv, R_point_in_polygon_sp)

# HespDiv

`hespdiv` is an R package implementing the **HespDiv framework for hierarchical spatial data subdivision**.
The package provides tools for identifying topologically contiguous spatial clusters and analysing hierarchical regionalization structures in palaeobiogeography and related spatial datasets.

HespDiv represents a suitable framework for identifying hierarchically nested,
spatially contiguous regions (geobiomes *sensu* Spiridonov & Eldredge (2024)) from taxonomic occurrence data.

The theoretical foundations of HespDiv are laid out in my PhD dissertation (link to be added!) and in:

Daumantas, L. & Spiridonov, A. (2024).
hespdiv: an R package for spatially constrained, hierarchical and contiguous regionalization in palaeobiogeography.
Palaeontology, 67(3), e12702.
https://doi.org/10.1111/pala.12702

# Installation
You can install the development version from GitHub:
```{r eval = FALSE}
if (!require("devtools")) install.packages("devtools")
devtools::install_github("Liudas-Dau/hespdiv")
```
# Example and walkthrough
A full demonstration of the workflow is provided in the vignette:

[**“HespDiv Walkthrough: Case of US Miocene Mammals”**](https://htmlpreview.github.io/?https://github.com/Liudas-Dau/hespdiv/blob/master/inst/Walkthrough.html)

The vignette illustrates how to:

- prepare spatial occurrence data
- perform hierarchical spatial subdivision
- analyse subdivision results
- visualise hierarchical regionalization patterns

# Example datasets

The walkthrough uses example datasets distributed in a separate package:

```{r eval = FALSE}
# install devtools is you don't have it!
devtools::install_github("Liudas-Dau/hespdiv_data")
```

# Citation
If you use *hespdiv* in your research, please cite:

Daumantas, L. & Spiridonov, A. (2024).
hespdiv: an R package for spatially constrained, hierarchical and contiguous regionalization in palaeobiogeography.
Palaeontology, 67(3), e12702.
https://doi.org/10.1111/pala.12702

#' Visualize Constrained HespDiv Sensitivity Analysis Results
#'
#' @description
#' Displays the alternative (subsampled) \code{hespdiv} subdivisions and the
#' basal (original) \code{hespdiv} subdivision on one or multiple plots,
#' illustrating how the split-lines vary across different ranks. Additionally,
#' for each alternative split-line (which is defined by a start and end coordinate),
#' the function aggregates identical endpoints and overlays their counts on the plot.
#'
#' @param obj An object of class \code{hsa_constrained}, the output
#'   of \code{\link{hsa_sample_constrained}}.
#' @param type An integer indicating the type of plot. Defaults to 1.
#'   \describe{
#'     \item{\code{1}}{A single plot overlaying all alternative lines
#'       and the basal lines, colored or width-coded by rank.}
#'     \item{\code{2}}{Multiple plots, each corresponding to a specific
#'       rank. Alternative lines are displayed in a user-defined color
#'       (default \code{"lightyellow3"}), with the basal line highlighted
#'       in another color.}
#'   }
#' @param rank Integer. Optional. When \code{type = 2}, if provided the function
#'   will only generate plots for the specified rank. If \code{NULL} (default),
#'   plots for all unique ranks will be produced.
#' @param col_basal Character or numeric specifying the color of basal split-lines
#'   (default \code{"gray20"}).
#' @param main Character. Title for the plot(s).
#' @param col_boundary Character or numeric specifying the color of the outer
#'   (first) polygon boundary (default \code{7}).
#' @param col_alternatives Character or numeric specifying the color of alternative
#'   split-lines (default \code{"lightyellow3"}).
#' @param max_lwd Numeric. The maximum line width for the highest-ranked split-line
#'   (default \code{2.5}).
#' @param min_lwd Numeric. The minimum line width for the lowest-ranked split-line
#'   (default \code{0.75}).
#' @param alpha_alt Numeric in the range \code{[0, 1]}. The transparency of
#'   alternative split-lines (default \code{0.6}).
#' @return
#'   \code{NULL}. The function is called for its side effect of generating
#'   one or more plots.
#'
#' @details
#' \itemize{
#'   \item In \code{type = 1}, the function creates a single plot showing all
#'         alternative split-lines overlaid on the first polygon boundary,
#'         plus all basal split-lines of the \code{hespdiv} basis object.
#'   \item In \code{type = 2}, the function creates separate plots, each focusing
#'         on polygons of a specific rank, drawing alternative lines in the user-specified
#'         color (with transparency) and the basal line in another color or line width.
#'         If a specific \code{rank} is provided, only that rank is plotted.
#'   \item In both cases, after drawing the alternative split-lines the function
#'         aggregates their endpoints (start and end coordinates) and overlays the
#'         count at each unique coordinate using \code{text()}.
#' }
#'
#' @importFrom graphics plot lines text
#' @importFrom grDevices adjustcolor
#' @family functions for hespdiv sensitivity analysis
#' @family HespDiv visualization options
#' @export
plot_cs_hsa <- function(obj,
                        type = 1,
                        rank = NULL,
                        col_basal = "gray20",
                        main,
                        col_boundary = 7,
                        col_alternatives = "lightyellow3",
                        max_lwd = 2.5,
                        min_lwd = 0.75,
                        alpha_alt = 0.6) {
  # 1) Basic checks
  if (!inherits(obj, "hsa_constrained")) {
    stop("`obj` must be an `hsa_constrained` object.", call. = FALSE)
  }
  if (!type %in% c(1, 2)) {
    warning("Unsupported `type` provided; defaulting to `type = 1`.", call. = FALSE)
    type <- 1
  }

  # 2) Retrieve the 'Basis' and the 'Alternatives'
  basis <- obj$Basis
  alt_list <- obj$Alternatives

  if (!("poly.stats" %in% names(basis))) {
    stop("The `basis` object (`obj$Basis`) must have `poly.stats`.", call. = FALSE)
  }
  if (!all(names(alt_list) %in% rownames(basis$poly.stats))) {
    stop(
      "Mismatch between `names(obj$Alternatives)` and `rownames(basis$poly.stats)`.",
      call. = FALSE
    )
  }

  ranks <- basis$poly.stats[names(alt_list), "rank"]
  unique_ranks <- unique(ranks)

  # If a specific rank is provided (for type==2) then use only that rank
  if (type == 2 && !is.null(rank)) {
    if (!rank %in% unique_ranks) {
      warning(
        "Provided rank ", rank, " was not found. Nothing will be plotted.",
        call. = FALSE
      )
      return(invisible(NULL))
    }
    unique_ranks <- rank
  }

  pol.n <- length(alt_list)
  n.runs <- if (pol.n > 0) length(alt_list[[1]]) else 0

  alt_col_transparent <- adjustcolor(col_alternatives, alpha.f = alpha_alt)

  # Helper function to overlay counts for unique endpoints
  overlay_density_endpoints <- function(alt_coords) {
    if (length(alt_coords) == 0) return()
    # Each element in alt_coords is a matrix with two rows (start, end)
    # Combine all endpoints into one matrix (each row = one endpoint)
    all_points <- do.call(rbind, alt_coords)
    # Create a unique key for each point by concatenating its coordinates
    keys <- apply(all_points, 1, function(row) paste(row, collapse = ","))
    counts <- table(keys)
    unique_keys <- names(counts)
    for (key in unique_keys) {
      coords <- as.numeric(strsplit(key, split = ",")[[1]])
      text(coords[1], coords[2], labels = counts[[key]], col = "blue", cex = 0.8)
    }
  }

  # 3) Plot logic
  if (type == 1) {
    # ---- TYPE = 1 ----
    if (!is.null(basis$polygons.xy) && length(basis$polygons.xy) >= 1) {
      plot(basis$polygons.xy[[1]], type = 'l', col = col_boundary, main = main)
      points(obj$Basis$call.info$Call_ARGS$xy.dat,pch=19,cex = 0.25)
    } else {
      stop(
        "`basis$polygons.xy[[1]]` is not available for plotting the boundary.",
        call. = FALSE
      )
    }

    # alt_coords will store the endpoints of each alternative split-line
    alt_coords <- list()

    lw_vec <- seq(max_lwd, min_lwd, length.out = length(unique_ranks))

    # A) Plot alternative lines and collect endpoints
    for (i in seq_len(pol.n)) {
      this_rank <- ranks[i]
      lw_poly <- if (length(lw_vec) > 1) lw_vec[this_rank] else lw_vec[1]
      for (run_idx in seq_len(n.runs)) {
        alt_obj <- alt_list[[i]][[run_idx]]
        if (!is.null(alt_obj$split.lines) && length(alt_obj$split.lines) >= 1) {
          line_coords <- alt_obj$split.lines[[1]]
          lines(line_coords, col = alt_col_transparent, lwd = lw_poly)
          # Record the start and end points of this split-line
          endpoints <- line_coords[c(1, nrow(line_coords)), , drop = FALSE]
          alt_coords[[length(alt_coords) + 1]] <- endpoints
        }
      }
    }

    # B) Plot basal lines
    if (!is.null(basis$split.lines)) {
      for (i in seq_along(basis$split.lines)) {
        lines(basis$split.lines[[i]], lwd = 0.5, col = col_basal)
      }
    }

    # C) Overlay the density (number of split-lines) at each unique endpoint
    overlay_density_endpoints(alt_coords)

  } else {
    # ---- TYPE = 2 ----
    for (rk in unique_ranks) {
      alt_coords <- list()
      ids <- which(ranks == rk)
      main_txt <- paste("Rank =", rk)

      if (!is.null(basis$polygons.xy) && length(basis$polygons.xy) >= 1) {
        plot(basis$polygons.xy[[1]], type = 'l', col = col_boundary, main = main_txt)
        points(obj$Basis$call.info$Call_ARGS$xy.dat,pch=19,cex = 0.25)
      } else {
        stop(
          "`basis$polygons.xy[[1]]` is not available for plotting the boundary.",
          call. = FALSE
        )
      }

      # A) Plot alternative lines for polygons of this rank and collect endpoints
      for (poly_idx in ids) {
        for (run_idx in seq_len(n.runs)) {
          alt_obj <- alt_list[[poly_idx]][[run_idx]]
          if (!is.null(alt_obj$split.lines) && length(alt_obj$split.lines) >= 1) {
            line_coords <- alt_obj$split.lines[[1]]
            lines(line_coords, col = alt_col_transparent, lwd = 1.2)
            endpoints <- line_coords[c(1, nrow(line_coords)), , drop = FALSE]
            alt_coords[[length(alt_coords) + 1]] <- endpoints
          }
        }
        if (!is.null(basis$split.lines) && length(basis$split.lines) >= poly_idx) {
          lines(basis$split.lines[[poly_idx]], lwd = 0.5, col = "red")
        }
      }

      all_lines <- seq_along(basis$split.lines)
      others <- setdiff(all_lines, ids)
      for (ln_idx in others) {
        lines(basis$split.lines[[ln_idx]], lwd = 0.5, col = col_basal)
      }

      overlay_density_endpoints(alt_coords)
    }
  }

  invisible(NULL)
}

#' Create data frame of x and y intervals of selected polygon segments
#' @param rule logic vector indicating which polygon segments are selected
#' @param cut_poly data.frame of a polygon
#' @author Liudas Daumantas
#' @noRd
.create_ints <- function(rule,cut_poly){
  IDs <- which(rule)
  data.frame(x1 = cut_poly[IDs,1],
             y1 = cut_poly[IDs,2],
             x2 = cut_poly[IDs+1,1],
             y2 = cut_poly[IDs+1,2])
}

#' Indicate whether the provided polygon is above zero
#' @param lowest_up_intervals data frame of filtered polygon segments are the
#' nearest to the y axis.
#' @param polygon data frame of a polygon
#' @author Liudas Daumantas
#' @noRd
.is_pol_up <- function(lowest_up_intervals,polygon) {
  .point.in.polygon(point.x = lowest_up_intervals[2,1],
                   point.y = lowest_up_intervals[2,2]-lowest_up_intervals[2,2]/2,
                   pol.x = polygon[,1],
                   pol.y = polygon[,2]) != 0

}

#' Convert polygon segments data.frame into points data.frame
#' @param inters data frame of x and y intervals of polygon segments
#' @noRd
.ints_to_pnts <- function(inters) {
  idx<-cbind(rep(1:nrow(inters),each=2),rep(c(1,3),nrow(inters)))
  idy<-cbind(rep(1:nrow(inters),each=2),rep(c(2,4),nrow(inters)))
  pol <- round(data.frame(x=inters[idx],y=inters[idy]),6)
  if (any(duplicated(pol))) {
    pol <- pol[-which(duplicated(pol)),]
  }
  if (pol[1,2] != 0){
    pol <- rbind(data.frame(x=0,y=0),pol)
  }
  if (pol[nrow(pol),2] != 0) {
    pol <- rbind(pol,data.frame(x=pol$x[nrow(pol)],y=0))
  }
  return(pol)
}


#' Insert given rows to a given data frame at given locations
#' @param df data frame object where rows will be inserted
#' @param newrows data frame of rows to be inserted
#' @param r numeric vector of indeces that show where the rows should be placed
#' @author Liudas Daumantas
#' @noRd
.insert_filter_Rows <- function(df, newrows, r) {
  df <- rbind(df,newrows)
  df <- df[order(c(1:(nrow(df)-nrow(newrows)),r+0.5)),]
  x3 <- dynGet(
    x = "x3",
    ifnotfound = stop("`x3` was not found in the calling environment.", call. = FALSE),
    minframe = 1,
    inherits = TRUE
  )

  if (x3 == 0) {
    df <- df[df$x >= 0, ]
  } else {
    df <- df[df$x <= x3, ]
  }
  row.names(df) <- 1:nrow(df)
  return(df)
}

#' Instead of indicated polygon vertexes, add new vertexes
#' with interpolated y coordinates
#'
#' @description Function removes indicated polygon vertexes. Instead of the
#' removed vertexes it adds new vertexes at provided x coordinates
#' for each mismatching pair of adjacent indicators of removal.
#' @param changes_log logical indicator, showing which vertexes are to be
#' removed from the polygon. Adjacent mismatches (TRUE FALSE and FALSE TRUE)
#' indicate where in the modified polygon vertexes should be added.
#' @param x3 an x coordinate at which y coordinates will be interpolated for
#' added vertexes.
#' @param polygon A data frame containing coordinates of a polygon vertexes.
#' @return  A data frame containing coordinates of vertexes of a modified
#'  polygon
#' @author Liudas Daumantas
#' @noRd
.find_y <- function(changes_log,x3,polygon){
  change_id0 <- which( changes_log[ -1 ] - changes_log[ -length(polygon$x) ]
                       !=0 )
  if (change_id0[1] == 1 & changes_log[1] ) {
    change_id0 <- change_id0[-1]
  } else {
    if (change_id0[ length(change_id0) ] == length(polygon$x) - 1 &
        changes_log[length(polygon$x) ] ) {
      change_id0 <- change_id0[ -length(change_id0) ]
    }
  }

  new_pts <- data.frame(x = rep(x3, length(change_id0)),
                        y = .pt_on_line(x1 = polygon$x[change_id0],
                                       x2 = polygon$x[ change_id0 + 1 ],
                                       y1 = polygon$y[change_id0],
                                       y2 = polygon$y[ change_id0 + 1 ],
                                       x3 = rep(x3,length(change_id0)) )
  )


  return( .insert_filter_Rows(polygon, new_pts, change_id0) )
}

#' Remove parts of a polygon that are outside of x range
#' of a split line
#'
#' @description Function removes polygon vortexes that are outside of x range
#' of a split line, as well as interpolates and adds vertexes where polygon
#' segments from deleted vertexes enters the allowed x range.
#' @param polygon A data frame containing coordinates of a polygon vertexes, an
#' output of .split_poly function that is rotated so that the split line is
#' horizontal, starting at x=0, y=0 and ending at some positive x, y = 0.
#' The provided polygon can be either closed or open.
#' @return A data frame containing the coordinates of vertexes of a provided
#' polygon that are inside an x range of a split line.
#' @author Liudas Daumantas
#' @noRd
.cut_by_x_margins <- function(polygon) {
  # XXXX atein uzdaras polygonas- galima atverti iskart nestestavus
  if (all(polygon[1,]==polygon[nrow(polygon),])){
    polygon <- polygon[-nrow(polygon),]
  }

  pol1 <- .find_y(polygon$x <= 0, x3= 0,polygon)
  return (.find_y(pol1$x >= pol1$x[ length(pol1$x) ],
                  x3 = pol1$x[ length(pol1$x) ], pol1) )
}

#' Filter segments of a polygon that are not OK (out of allowed y range,
#' turns left or are not the nearest to the Y axis)
#'
#' @param intervals a data frame of x and y intervals of a polygon segments.
#' These segments are pre-filtered by the allowed x ranges.
#' @return  A data frame containing coordinates of vertexes of a modified
#'  polygon
#' @author Liudas Daumantas
#' @noRd
.segment_filter <- function(intervals) {
  if (nrow(intervals)>1){
    intervals <- round(intervals,6)
    # XXXX CHeezz Liudai! @.@ :O...  intervals[,c(1,3)] is enough for range | ARBA NE, JEI SVARBU, KAD BUTU MAZENIS KAIREJE
    int_x_ranges <- cbind(t(apply(intervals[,c(1,3)],1,range)),1:nrow(intervals))
    # kaip su tom nelygybem lygybem? Po @.@ atrodo gerai pirmos dvi salygos,
    # nes intervalai lyg ir turi buti su tais paciais endpointas,
    # tai kai kurie endpointai logiska, kad sutampa, bet abieju pradzios
    # negali sutapti (int_x_ranges pirmas stulpelis mazesnieji),
    # kaip ir abieju pabaigos.
    # Tuo tarpu trecia salyga atrodo pertekline, nes neisivaizduoju, kur imanoma
    # kad vienos atkarpos pirmas endpointas butu kairiau kitos ir tuo paciu jos
    # antras endpointas butu desiniau kitos. Tai tarsi diktuoja, kad endpointai
    # nesiliecia. Tas gali butu imanoma tik tada, jei intervalai butu prafiltruoti.
    # Ir taip, jie filtruoti su create_ints() ir is_x_shifted(), bet tai
    # nufiltruoja tik tas vietas, kur x nesikeicia
    # Taigi, endopointai turetu visur tureti bent viena lygybe. Tikrinu: Gavosi butent taip.
    # Taip pat, reikalinga butu ir kita salyga, jei trecia yra reikalinga -
    # kur viena o atkarpa butu praryta, o ne apziotu kaip numato dabar paskutine
    # salyga. Taip trecia reikalinga ir reikalinga butent tais atvejais, kai
    # persidengia negretimos atkarpos. TY. persidengti gali artimos atkarpos, jei
    # suka linija ten is kur atejo, bet gali ir toliau uzklisti.
    # Ketvirta salyga kazkodel pertekline.... Pasirodo gal ir viskas buvo gerai :-(
    # ISVADA: palieku kol kas papildoma salyga, veliau reiks su ja ir be jos padaryt
    # ir pazet ar ji reikalinga, ar ne. NES - be trecios neveikia, o ketvirta neitakoja
    # sio atvejo. Taigi, trecia ir ketvirta neanalogiska, tik nera atveju ketvirtos
    overlap_log <-  apply(int_x_ranges[,1:2], 1,function(o) {
      (o[1] >= int_x_ranges[,1] & o[1] < int_x_ranges[,2] ) |
        ( o[2] > int_x_ranges[,1] & o[2] <= int_x_ranges[,2] ) |
        (o[1] <= int_x_ranges[,1] & o[2] >= int_x_ranges[,2]) |
        (o[1] >= int_x_ranges[,1] & o[2] <= int_x_ranges[,2])
    })
    diag(overlap_log) <- FALSE
    # stulpeliai atitinka obs. (eilutes int_x_ranges). eilutes - su kuo persidengia siu obs eilutes
    # noverlap_int_log - indikatoriai intervalu, kurie nepersidengia su niekuo
    # overlap_int_ranges - intervalai, kurie persidengia su kitais intervalais
    noverlap_int_log <- apply(!overlap_log,2,all)
    if (any(noverlap_int_log)) {
      overlap_int_ranges <- int_x_ranges[-which(noverlap_int_log),]
    } else {
      overlap_int_ranges <- int_x_ranges
    }
    # Jei nera persidengianciu intervalu, grazinam nekoreguotus intervalus atitransformuotus i taskus
    if (nrow(overlap_int_ranges)==0) {
      return(.ints_to_pnts(intervals))
    }
    # Isruosiuoti intervalu endpoitai, be duplikatu (taskai isilgai x)
    x_end_points <- sort(unique(c(overlap_int_ranges[,1:2])))
    # intervalai isrusiuoti x didejimo tvarka,o ne taip kaip poligonas nupaisytas
    x_ranges <- cbind(x_end_points[-length(x_end_points)],x_end_points[-1])
    # is end_points sukurtuose x_ranges gali buti intervalai, kur persidengimo
    # su overlap_int_ranges nera

    # TOKIU INTERVALO ISMETIMO KODAS (letesnis sprendimas, patestavus):
    #id_interval_gap <- which((overlap_int_ranges[-1,3] - overlap_int_ranges[-nrow(overlap_int_ranges),3]) != 1)
    #if (length(id_interval_gap)>0){
      # imanoma, kad overlap_int_ranges gapinis intervalas nera butinai mazesnio x. x_ranges isrusiuotas, overlap_int_ranges ne
      # id_interval_gap + 1
      # PROBLEMA: gali gautis ir taip, kad x_ranges intervalai isrusiuoti
      # visai nebera salia tie, kurie buvo salia su overlap_int_ranges ir ties
      # rastu luziu, pvz. veliau ar pries poligono atkarpai patekus i ta gapini
      # intervala. tokiu atveju sprendimas butu, intervalus ismetines is x_ranges
      # pagal x_ranges_int_id rezultata (jei sarasas - randi tuscius id ir
      # istrini atitinkamus x_ranges, tada kartoji x_ranges_int_id)
      # Taciau kita vertus, tas gapinis intervalas atsiranda del to, kad
      # tame intervale nera persidengianciu intervalu, tai gal ir negali isiterpti
      # niekas i ji
      # gali buti daugiau nei vienas id_del
      #gap_del <- data.frame(x1 = numeric(length(id_interval_gap)),
      #                      x2 = numeric(length(id_interval_gap)))
      #for (i in 1:length(id_interval_gap)){
      #  if (overlap_int_ranges[id_interval_gap[i],2] <
      #      overlap_int_ranges[id_interval_gap[i]+1,1]){
      #    gap_del[i,] <- c(overlap_int_ranges[id_interval_gap[i],2],
      #                     overlap_int_ranges[id_interval_gap[i]+1,1])
      #  } else {
      #gap_del[i,] <-  c(overlap_int_ranges[id_interval_gap[i]+1,2],
      #                    overlap_int_ranges[id_interval_gap[i],1])
#}
#      x_ranges <- x_ranges[-(which(x_ranges[,1] == gap_del[i,1] &
#                        x_ranges[,2] == gap_del[i,2])),]
#      }
#    }
    x_ranges_int_id <-  apply(x_ranges, 1,function(o) {
      # su kuriais poligono nerusiuotais persidengianciais intervalais, persidengia
      # kiekvienas rusiuotas persidengiantis intervalas (o arba x_ranges obs, eilute)
      # su savimi paciu aisku ir plius gal dar kazkas dar (nes intervalai,
      # kurie tik su savimi persidengia nufiltruoti --> diag(overlap_log) <- FALSE,
      # taciau overlap_int_ranges isdalies persidengiantys intervalai gali
      # tureti x_ranges atkarpas, kurios nepersidengia su daugiau niekuo.
      # Pastariosios gaunamos isrusiavus persidengianciu intervalu endpointus ir
      # is ju padarius intervalus)
      # cia vel galimai truksta dar ketvirtos salygos, kur o intervalas yra prarytas
      # Iterpiu salyga, bet niekas nesikeicia. BUTINA BUS PRATIKRINTI
      ids <- which(
        (o[1] >= overlap_int_ranges[,1] & o[1] < overlap_int_ranges[,2] ) |
        (o[2] > overlap_int_ranges[,1] & o[2] <= overlap_int_ranges[,2] ) |
        (o[1] <= overlap_int_ranges[,1] & o[2] >= overlap_int_ranges[,2]) |
          (o[1] >= overlap_int_ranges[,1] & o[2] <= overlap_int_ranges[,2]))
      # is visu su o intervalu persidengianciu nerusiuotu intervalu
      # atrenkam ID to nerusiuoto intervalo, kurio y, o intervalo (su kuriuo
      # persidengia) viduryje yra arciausiai nulio. Taigi, gaunam
      # overlap_int_ranges id, kurie yra prarusiuoti ir kuriu Y galima naudoti
      # x_ranges intervalams priskirti
      overlap_int_ranges[ids[which.min(abs(.pt_on_line(
        x1 = intervals$x1[overlap_int_ranges[ids,3]],
        x2 = intervals$x2[overlap_int_ranges[ids,3]],
        y1 = intervals$y1[overlap_int_ranges[ids,3]],
        y2 = intervals$y2[overlap_int_ranges[ids,3]],
        x3 = rep(o[1]+(o[2]-o[1])/2,length(ids)))))],3]
    })
    if (is.list(x_ranges_int_id)) { # then this side should not be the one
      # needed. Could be improved by returning indicator and moving on to
      # the next side. If both sides gets indicator, then abort
      x_ranges<- x_ranges[-which(unlist(lapply(x_ranges_int_id, function(o)
        {length(o)<1}))),]
     x_ranges_int_id <-  apply(x_ranges, 1,function(o) {
        # su kuriais poligono nerusiuotais persidengianciais intervalais, persidengia
        # kiekvienas rusiuotas persidengiantis intervalas (o arba x_ranges obs, eilute)
        # su savimi paciu aisku ir plius gal dar kazkas dar (nes intervalai,
        # kurie tik su savimi persidengia nufiltruoti --> diag(overlap_log) <- FALSE,
        # taciau overlap_int_ranges isdalies persidengiantys intervalai gali
        # tureti x_ranges atkarpas, kurios nepersidengia su daugiau niekuo.
        # Pastariosios gaunamos isrusiavus persidengianciu intervalu endpointus ir
        # is ju padarius intervalus)
        # cia vel galimai truksta dar ketvirtos salygos, kur o intervalas yra prarytas
        # Iterpiu salyga, bet niekas nesikeicia. BUTINA BUS PRATIKRINTI
        ids <- which(
          (o[1] >= overlap_int_ranges[,1] & o[1] < overlap_int_ranges[,2] ) |
          (o[2] > overlap_int_ranges[,1] & o[2] <= overlap_int_ranges[,2] ) |
          (o[1] <= overlap_int_ranges[,1] & o[2] >= overlap_int_ranges[,2]) |
          (o[1] >= overlap_int_ranges[,1] & o[2] <= overlap_int_ranges[,2]))
        # is visu su o intervalu persidengianciu nerusiuotu intervalu
        # atrenkam ID to nerusiuoto intervalo, kurio y, o intervalo (su kuriuo
        # persidengia) viduryje yra arciausiai nulio. Taigi, gaunam
        # overlap_int_ranges id, kurie yra prarusiuoti ir kuriu Y galima naudoti
        # x_ranges intervalams priskirti
        overlap_int_ranges[ids[which.min(abs(.pt_on_line(
          x1 = intervals$x1[overlap_int_ranges[ids,3]],
          x2 = intervals$x2[overlap_int_ranges[ids,3]],
          y1 = intervals$y1[overlap_int_ranges[ids,3]],
          y2 = intervals$y2[overlap_int_ranges[ids,3]],
          x3 = rep(o[1]+(o[2]-o[1])/2,length(ids)))))],3]
      })
    }
    inters <- data.frame(x1 = x_ranges[,1],y1 = .pt_on_line(
      x1 = intervals[x_ranges_int_id,1],
      x2 = intervals[x_ranges_int_id,3],
      y1 = intervals[x_ranges_int_id,2],
      y2 = intervals[x_ranges_int_id,4],
      x3 = x_ranges[,1]), x2 = x_ranges[,2],
      y2 = .pt_on_line(
        x1 = intervals[x_ranges_int_id,1],
        x2 = intervals[x_ranges_int_id,3],
        y1 = intervals[x_ranges_int_id,2],
        y2 = intervals[x_ranges_int_id,4],
        x3 = x_ranges[,2]))
    if (any(noverlap_int_log)) {
      # turim x_ranges x intervalus, kurie tvarkingai isrusiuoti. Kiekvienam
      # tokiam intervalui, zimone kuria poligono atkarpa naudoti
      # (turim x_ranges_int_id)
      #
      # suklijuojam x_ranges_int_id ( overlap_int_ranges pateiktus teisinga eiles
      # tvarka kiekvienam x_ranges) su intervalais, kurie su niekuo nepersidenge
      # PASTARIEJI NERUSIUOTI, NETVARKINGI
      #
      # MANO SIULYMAS: X_RANGES X INTERVALAMS SUGENERUOTI Y VERTES IR GAUTI
      # GERAI RUSIUOTAS ATKARPAS PERSIDENGIMU SRITYJE
      # TADA SUJUNGI SU ATKARPOMIS, KURIOS NEOVERLAPINA, IR NUKREIPTOS I DESINE
      # GALIAUSIAI ISRUSIUOTI PAGAL INTERVALU PRADZIAS.
      # ARBA JEI DARYT PANASIAI KAIP BUVO, TAI:
      # GAUNAM NEOVERLAPINANCIU INTERVALU x_RANGES
      #
      inters_nover <- t(apply(intervals[which(noverlap_int_log),],1,function(o){
        id <- which.min(c(o[1],o[3]))
        if (id == 1 ) o else o[c(3,4,1,2)]
      }))
      inters <- rbind(inters,inters_nover)
      id_ord <- order(inters[,1])
      inters <- inters[id_ord,]
     # x_ranges_int_id <- c(x_ranges_int_id,which(noverlap_int_log))

      # grazina tvarka intervalu, kuri isrusiuota pagal intervalu pradzias.
      # Jei intervalai nepersidengiantys tai turetu issirusiuoti taip, kad
      # gretimu atkarpu x intervalai sugultu salia, einant is kaires i desine.
      #
      # ID PERRUSIUOJAMI PAGAL poligono atkarpu kairinius taskus
     # x_ranges_int_id_order <- order(sapply(x_ranges_int_id,
           #                                 function(o){
          #                                    min(intervals[
         #                                       o,c(1,3)])}))
      # naudojant gauta tvarka, persusiuojami intervalu indeksai
      # gaunam rusiavimo indeksus skirtus perrusiuoti intervalus
      #

     # x_ranges_int_id <- x_ranges_int_id[x_ranges_int_id_order]
      # x_ranges sudedamas su analogiskos strukturos nepersidengianciais
      # intervalais ir isrusiuojami pagal x_ranges_int_id_order tvarka
      #  nesuprantu, kodel tik.
      #x_ranges <- rbind(x_ranges,t(apply(intervals[which(noverlap_int_log),
       #                                            c(1,3)]
        #                                 ,1,range)))[x_ranges_int_id_order,]

    }
   # apsukam intervalus (cia kaip ir aisku) (galiMAI NEBUTINAS VEIKSMAS)
   # intervals_selected <- t(apply(intervals[x_ranges_int_id,],1,
    #                              function(o){
     #                               id <- which.min(c(o[1],o[3]))
      #                              if (id == 1 ) o else o[c(3,4,1,2)]
       #                           }))

    #intervals_selected <- intervals_selected[paste(x_ranges_int_id),]
    # kiekienam x intervalui, paimam atrinkta intervala poligono ir panaudojam
    # sugeneruoti taskus
  #  inters <- data.frame(x1 = x_ranges[,1],y1 = .pt_on_line(
   #   x1 = intervals_selected[,1],
    #  x2 = intervals_selected[,3],
     # y1 = intervals_selected[,2],
    #  y2 = intervals_selected[,4],
    #  x3 = x_ranges[,1]), x2 = x_ranges[,2],
    #  y2 = .pt_on_line(
    #    x1 = intervals_selected[,1],
    #    x2 = intervals_selected[,3],
    #    y1 = intervals_selected[,2],
    #    y2 = intervals_selected[,4],
    #    x3 = x_ranges[,2]))

  } else{
    inters <- intervals
  }
  in.pol <- .ints_to_pnts(inters)

  return(in.pol)
}


#' Remove left-turning segments of a polygon divided with horizontal split line
#'
#' @description This function from a polygon derived with .split_poly
#' function extracts the maximum area part that is directly above or below
#' split line (depending on the which side of divided polygon is provided) and
#' that has no left-turning segments.
#' @param polygon A data frame containing coordinates of a polygon, output of
#' .split_poly function that is rotated so that the split line is horizontal,
#' starting at x=0, y0 and ending at some positive x, y = 0.
#' The provided polygon can be either closed or open.
#' @return A list of two values:
#' `[[1]]` A data frame containing the coordinates of the extracted polygon part.
#' The returned polygon is open and it is not strictly inside a provided
#' polygon, since it inherits vertical & right-turning segments that are
#' directly above or below the split line.
#' `[[2]]` -1 or +1 integer indicating the side of the extracted polygon in
#' relation to the split line (+1 above, -1 below)
#' @author Liudas Daumantas
#' @examples #Creating a data.frame of an irregular polygon
#' poly<-data.frame(x=c(0,-1,12,0,5,12,12,7,-5,
#' 11,-3,10,-10,-10,15,15,-1,13,13,-7,-7,0),
#' y=c(0,5,10,4,4,3,-3,-10,-11,-4,-4,0,-2,20,20,-20,-20,-15,15,15,2,0))
#' plot(poly,type='o',main='Horizontal Split Line')
#' #Horizontal Split Line
#' lines(poly[c(1,12),1],poly[c(1,12),2],lty='dotted')
#' #Dividing a polygon
#' upper_half <- split_poly(polygon = poly, split_ids = c(1,12), min_id = 1,
#' trivial_side = F, poli_side = T)
#' lower_half <- split_poly(polygon = poly, split_ids = c(1,12), min_id = 1,
#' trivial_side = F, poli_side = F)
#' lines(upper_half,col=4)
#' lines(lower_half,col=2)
#' #Filtering left turning segments
#' upper_half_lf <- .inner_poly_hull(upper_half)
#' lower_half_lf <- .inner_poly_hull(lower_half)
#' lines(upper_half_lf[[1]],col=4,lwd=3)
#' lines(lower_half_lf[[1]],col=2,lwd=3)
#' @noRd
.inner_poly_hull<-function(polygon){

  if ( all(polygon[-1,1]-polygon[-nrow(polygon),1]>=0 ) ){
    return(
      list(polygon,sign(polygon[2,2]))
    )
  }
  v_cutted_poly <- .cut_by_x_margins(polygon)

  x_shift <- v_cutted_poly[,1][-1] -
    v_cutted_poly[,1][-nrow(v_cutted_poly)]

  if (all(x_shift>=0)){

    if (!(all(v_cutted_poly$y <=0) | all(v_cutted_poly$y >=0))){
      if(all(x_shift[1:2] == c(0,0))){
        v_cutted_poly <- v_cutted_poly[-2,]
      }
      if(all(x_shift[(length(x_shift)-1):length(x_shift)] ==
             c(0,0))){
        v_cutted_poly <- v_cutted_poly[-(nrow(v_cutted_poly)-1),]
      }
    }
    return(
      list(v_cutted_poly,sign(v_cutted_poly[2,2]))
    )
  }
  is_x_shift <- x_shift != 0

  if (all(v_cutted_poly$y <=0) | all(v_cutted_poly$y >=0) ) {

    side <- sign(v_cutted_poly$y[2])
    int <- .create_ints(is_x_shift,v_cutted_poly)

    return(
      list(.segment_filter(int),side)
    )

  } else {

    upper_int <- .create_ints(is_x_shift & v_cutted_poly[,2][-1] >=0 &
                                v_cutted_poly[,2][-nrow(v_cutted_poly)] >=0
                              ,v_cutted_poly)

    lowest.up.intervals <- .segment_filter(upper_int)

    if (is.null(lowest.up.intervals)){
      side <- FALSE
    } else {
      side <- .is_pol_up(lowest.up.intervals,polygon)
    }

    if (side) {

      return(
        list(lowest.up.intervals, 1)
      )

    } else {

      lower_int <- .create_ints(is_x_shift & v_cutted_poly[,2][-1] <=0 &
                                  v_cutted_poly[,2][-nrow(v_cutted_poly)] <=0,
                                v_cutted_poly)
      return(
        list(.segment_filter(lower_int),-1)
      )
    }
  }
}

#' Plot hespdiv results
#'
#' @description This function is used to plot the results obtained with the \code{hespidv}
#' function. The plot showcases subdivisions of the study area by split-lines,
#' visualizing their performances or rank with colors or line widths. Additionally,
#' it can display the spatial distribution of observations and number of
#' observations in each location.
#' @param obj A hespdiv object.
#' @param type A character. Either "width" or "color" (default "color").
#' Determines whether quality of split-lines is expressed by line width or
#' color.
#' @param performance logical. TRUE - display split-line performance,
#' FALSE - rank. Displaying rank makes the spatial dendrogram clearer.
#' @param n.loc A Boolean value. Would you like to visualize the number of observations
#' at each location? Only possible, when there are localities with more than
#' one observation. If the type is 'color,' the number of observations is
#' expressed through point sizes. Otherwise, they are expressed using color in
#' a logarithmic scale.
#' @param legend_title A character value that indicates the title of the legend
#' for the split-lines. The default is built according to the method information
#' available in "obj$call.info".
#' @param title A character that indicates the title of the plot.
#' @param subtitle A character that indicates the subtitle of the plot.
#' @param seed An integer value that indicates seed used to randomize the colors
#' of the split-lines. Only meaningful, when argument \code{type = "width"}.
#' Try setting a different value, if colors of parallel split-lines or nearby
#' labels look too similar or to increase the general appeal of the graph.
#' @param pnts.col A character or numeric vector providing color codes
#' for data points.
#' @return A ggplot object.
#' @importFrom ggrepel geom_label_repel
#' @importFrom viridis scale_color_viridis
#' @importFrom ggplot2 aes geom_point geom_path guides guide_legend ggtitle scale_size_area guide_colourbar scale_size_continuous theme_set theme element_rect element_blank element_blank
#' @importFrom stats aggregate na.omit
#' @importFrom rlang .data
#' @details The return ggplot object can be edited as any other ggplot objects
#' by removing undesired elements, changing theme or overlying the plot
#' with additional elements.
#' @examples
#' plot_hespdiv(example_hespdiv)
#' plot_hespdiv(example_hespdiv, type = "width")
#' plot_hespdiv(example_hespdiv, n.loc = TRUE)
#' @author Liudas Daumantas
#' @family HespDiv visualization options
#' @export
plot_hespdiv <- function(obj, type = "color",n.loc = FALSE, performance = TRUE,
                           legend_title = NULL, title = NULL,subtitle = NULL,
                           pnts.col = NULL,seed = 10){
  xy.dat <- obj$call.info$Call_ARGS$xy.dat
  type <- .arg_check("type",type,c("width","color"))
  if (type == "width" & !performance){
    stop(
      "There is currently no option to display rank using line widths.",
      call. = FALSE
    )
  }
  if (n.loc){
    xy_df <- xy.dat
    xy_df$n <- 1
    uni.loc.n <- stats::aggregate(n~., data = xy_df ,FUN = sum)
    uni.loc.n <- uni.loc.n[order(uni.loc.n$n,decreasing = FALSE),]
    if (nrow(uni.loc.n) == nrow(xy_df)){
      n.loc <- FALSE
      warning(
        "All observations are from unique locations. Using `n.loc = FALSE`.",
        call. = FALSE
      )
    }
    if (type == "color") {
      if (is.null(pnts.col)) {
        pnts.col <- rep(1, nrow(uni.loc.n))
      } else {
        if (length(pnts.col) != nrow(uni.loc.n))
          stop(
            "Length of `pnts.col` is not equal to the number of unique locations.",
            call. = FALSE
          )
        }
    } else {
      if (!is.null(pnts.col)){
        stop(
          "Conflicting arguments: `pnts.col` is not NULL when `n.loc = TRUE` and ",
          "`type = \"width\"`.\n",
          "Cannot map two variables to the same aesthetic.",
          call. = FALSE
        )
      }
    }
  } else {
    if (is.null(pnts.col)) {
      pnts.col <- rep(1, nrow(xy.dat))
    } else {
      if (length(pnts.col) != nrow(xy.dat))
        stop(
          "Length of `pnts.col` is not equal to the number of observations.",
          call. = FALSE
        )
    }
  }

  split.stats <- obj$split.stats
  maximize <- obj$call.info$METHOD$maximize
  if (is.null(legend_title)){
    if (!performance){
      legend_title <- "Rank"
    } else {
    if (obj$call.info$METHOD$method.type == "custom"){
      legend_title <- obj$call.info$METHOD$metric
    } else {
      if (obj$call.info$METHOD$metric == "sorensen"){
        legend_title <- "S\u00f8rensen-Dice\ncoefficient"
      } else {
        if (obj$call.info$METHOD$metric == "morisita"){
          legend_title <- paste0("Morisita\nSimilarity")
        } else {
          if (obj$call.info$METHOD$metric == "pielou"){
            legend_title <- paste0("Pielou\nentropy\nreduction\n")
          } else {
            if (obj$call.info$METHOD$metric == "horn.morisita"){
              legend_title <- paste0("Morisita-\nHorn\nSimilarity")
            }
          }
        }
      }
    }
    }
  }
  if (performance){
    key <- "performance"
  } else {
    key <- "rank"
    split.stats[,"rank"] <- as.integer(split.stats[,"rank"])
  }
  if (!maximize & performance){
    split.stats[,"performance"] <- -obj$split.stats[,"performance"]
  }
  ord <- order(split.stats[,key], decreasing = FALSE)
  split.stats <- split.stats[ord,]
  split.lines <- lapply(ord,function(id){obj$split.lines[[id]]})
  df <- Reduce(rbind,split.lines)
  npt.in.split <- as.numeric(lapply(split.lines,nrow))
  if (type == "width"){
    size <- rep(split.stats[,key], times = npt.in.split)
  } else {
    color <- rep(split.stats[,key], times = npt.in.split)
  }
  df$group <- factor(rep(1:length(split.lines),times=npt.in.split))

  base <- ggplot2::ggplot(obj$polygons.xy[[1]],ggplot2::aes(.data$x,.data$y)) +
    geom_path(data= obj$call.info$Call_ARGS$study.pol,aes(.data$x,.data$y), linewidth=0.5,
                     lineend = "round",linejoin = "round",color = "gray20",alpha = 0.5)+
    geom_path(data= obj$polygons.xy[[1]],ggplot2::aes(.data$x,.data$y), linewidth=0.5,
              lineend = "round",linejoin = "round",color = 1) +
    ggplot2::labs(x = "x", y = "y")
    if (!is.null(title) | !is.null(subtitle)){
      base <- base + ggplot2::ggtitle(title, subtitle = subtitle)
    }

  if (n.loc){
    scale_id <- 2
    if (type == "color"){
      base <- base + ggplot2::geom_point(data=uni.loc.n,ggplot2::aes(x=.data$x,y=.data$y,size=.data$n),
                                pch =rep(1,nrow(uni.loc.n)),color = pnts.col) +
        ggplot2::guides(size=ggplot2::guide_legend(title=paste0("Number of", "\nobservations" ,
       "\nin a location"),title.hjust = 0.5,label.position = "left",
       label.hjust = 1))+
        ggplot2::scale_size_area(max_size = 8)
    } else {
      base <- base + ggplot2::geom_point(data=uni.loc.n,ggplot2::aes(x=.data$x,y=.data$y,color=.data$n),
                                pch =rep(19,nrow(uni.loc.n)),size =2) +
        viridis::scale_color_viridis(guide ="colourbar",trans = "log") +
        ggplot2::guides(color = ggplot2::guide_colourbar(
          title = paste0("Number of", "\nobservations" ,
                         "\nin a location"),label.position = "left",
          label.hjust = 1, title.hjust = 0.5,title.vjust = 1))

      base$scales$scales[[1]]$limits <-  range(log(uni.loc.n$n))
      base$scales$scales[[1]]$labels <- round(exp(as.numeric(stats::na.omit(
        base$scales$scales[[1]]$get_breaks() ))))
      base$scales$scales[[1]]$breaks <- stats::na.omit(
        exp(base$scales$scales[[1]]$get_breaks() ))

    }

  } else {
    scale_id <- 1
  base <- base + ggplot2::geom_point(data = xy.dat, mapping = ggplot2::aes(.data$x,.data$y),
    pch=16,color=pnts.col) +
    ggplot2::geom_path(data= obj$polygons.xy[[1]],ggplot2::aes(.data$x,.data$y), linewidth=.5,
                           lineend = "round",linejoin = "round")
  }
  if (type == "width"){
    color <- .generate_cols(nrow(split.stats), seed)
    df$color <- rep(color, times=npt.in.split)
    df$size <- size
    base<-base + ggplot2::geom_path(data = df, ggplot2::aes(.data$x,.data$y,group=.data$group,
                                                            linewidth = .data$size),
                           color=df$color) +
      ggplot2::scale_size_continuous(range = c(0.5,2))
    size.l <- seq(0.5,2,0.5)
    base <- base + ggplot2::guides(linewidth = ggplot2::guide_legend(override.aes =
                                                list(size = size.l))) +
      ggplot2::guides(linewidth=ggplot2::guide_legend(title=legend_title, title.hjust = 0.5))
  } else {
    df$color <- color
    if (performance) {
    base<-base + ggplot2::geom_path(data = df, aes(.data$x,.data$y,group=.data$group, color=.data$color),linewidth = 2) +
      viridis::scale_color_viridis(guide ="colourbar") +
      ggplot2::guides(color = ggplot2::guide_colourbar(title = legend_title, title.hjust = 0.5,
                                   label.position = "left",label.hjust = 1))
    } else {
      base<-base + ggplot2::geom_path(
        data = df, aes(.data$x,.data$y,group=.data$group, color=.data$factor(color)), linewidth = 2) +
        viridis::scale_color_viridis(labels = levels(factor(color)),
                                     discrete = TRUE) +
        ggplot2::guides(
          color = ggplot2::guide_legend(
            title = legend_title,
            title.hjust = 0.5,              # Center the title
            label.position = "left",        # Position labels on the left of the bar
            label.hjust = 1,                # Align labels horizontally
          )
        )

    }
  }

  if (!maximize & performance) {
      base$scales$scales[[scale_id]]$limits <-  range(split.stats[,key], na.rm = TRUE)
      base$scales$scales[[scale_id]]$breaks <-
        as.numeric(na.omit( base$scales$scales[[scale_id]]$get_breaks() ))
      base$scales$scales[[scale_id]]$labels <-
        -base$scales$scales[[scale_id]]$breaks
}
  if (type == "color" & performance) {
    if (maximize){
      base$scales$scales[[scale_id]]$limits <-  range(split.stats[,key], na.rm = TRUE)
    }
    self <- base$scales$scales[[scale_id]]
    if (any(is.na(split.stats[, key]) | is.nan(split.stats[, key]))){

      # Identify which rows are NA or NaN
      na_idx <- is.na(split.stats[, key]) | is.nan(split.stats[, key])

      # Replace NAs with the lower limit of the scale (or any valid placeholder)
      perf_vals_no_na <- split.stats[, key]
      perf_vals_no_na[na_idx] <- mean(split.stats[, key], na.rm = TRUE)

      # Rescale and map to colors
      x <- self$rescale(self$oob(perf_vals_no_na, range = self$limits), self$limits)
      color <- self$palette(x)

      # Finally, assign a neutral color (e.g. grey) for any originally NA/NaN
      color[na_idx] <- "grey50"
    } else {
      x <- self$rescale(self$oob(split.stats[,key], range = self$limits), self$limits)
      color <- self$palette(x)
    }
  }
  if (!performance) {
    color <- base$scales$scales[[scale_id]]$
      palette(n = length(unique(split.stats$rank)))
    names(color) <-  unique(split.stats$rank)
    color <- color[split.stats$rank]
    }
  base <- base +
    ggplot2::theme_bw()  +
    ggplot2::theme(
      legend.key = ggplot2::element_blank(),
      legend.background = ggplot2::element_blank(),
      legend.title = ggplot2::element_text(hjust = 0.5),
      panel.grid = ggplot2::element_blank(),
      panel.background = ggplot2::element_rect(
        colour = "black",
        linewidth = 0.5,
        fill = "white"
      )
    )

  mid.pt <- data.frame(x=numeric(),y=numeric())
  for (a in 1:length(obj$split.lines)){
    if(nrow(split.lines[[a]])!=2){
      mid.pt <- rbind(mid.pt,split.lines[[a]][
        round(length(split.lines[[a]]$x)/2,0),])} else {
          mid.pt<-rbind(mid.pt,data.frame(x=mean(split.lines[[a]]$x),
                                          y=mean(split.lines[[a]]$y)))
        }
  }
  if (is.null(split.stats$p.val)){
    base<-base + ggrepel::geom_label_repel(data=mid.pt, aes(.data$x,.data$y),alpha=rep(3/5, nrow(mid.pt)),
                                  label = paste0(ord,") ",
                                                round(obj$split.stats[ord,"performance"],2)),
                                  fill  = color, size = 4,
                                  direction="both",fontface='bold',
                                  colour  = rep(1, nrow(mid.pt)))
  } else {
    base<-base + ggrepel::geom_label_repel(data=mid.pt, aes(.data$x,.data$y),alpha=rep((3/5), (nrow(mid.pt))),
                                  label = paste(ord,") p = ",
                                                split.stats$p.val,
                                                "\n  Div. qual. = ",
                                                round(split.stats$delta.E,2)),
                                  fill = unique(color),
                                  size = 3.5,direction="both",fontface='bold')
  }

  base
}

#' Create a gif of blok3D
#'
#' @description  Functions creates a gif of rotating polygons displayed in a currently actively rgl device produced with blok3D
#' @param output_file name and path of output file
#' @param frames integer. Number of frames.
#' @param angle_per_frame numeric. Angle to rotate per frame in degrees.
#' @param fps integer. Frames per second.
#' @return No return value. Called for the side effect of saving
#' a GIF of a rotating 3D plot.
#' @author Liudas Daumantas
#' @family HespDiv visualization options
#' @importFrom rgl par3d rotationMatrix view3d rotate3d rgl.snapshot
#' @importFrom magick image_read image_animate image_write
#' @note You can adjust the size of rgl device window to control the size of gif.
#' @export
create_gif <- function(output_file, frames = 90, angle_per_frame = 5,
                       fps = 10) {

  oldpar3d <- rgl::par3d(no.readonly = TRUE)
  on.exit(rgl::par3d(oldpar3d), add = TRUE)

  angles <- rep(angle_per_frame * pi / 180, frames)
  # Directory for temporary files
  temp_dir <- tempdir()
  file_list <- character(frames)  # Store file paths for each frame
  rgl::par3d(userMatrix = rgl::rotationMatrix(-45 * pi / 180, 45, 0, 1))
  # Render each frame with rotation
  for (i in seq_len(frames)) {

    rgl::view3d(userMatrix = rgl::rotate3d(rgl::par3d("userMatrix"),
                                 angles[i], 0, 0, 1))

    # Save the frame as a PNG
    file_list[i] <- file.path(temp_dir, sprintf("frame_%03d.png", i))
    rgl::rgl.snapshot(file_list[i])
  }

  # Combine frames into a GIF
  images <- magick::image_read(file_list)  # Read all frames into magick
  animation <- magick::image_animate(images, fps = fps)  # Create animation
  magick::image_write(animation, path = output_file)  # Save as GIF

  # Clean up temporary files
  unlink(file_list)

  message("GIF saved to: ", output_file)
}

#' Plot a nullhespdiv object
#'
#' @description Plot method for a \code{nullhespdiv} object.
#'
#' @method plot nullhespdiv
#' @param x A \code{nullhespdiv} object.
#' @param ... Additional arguments passed to \code{graphics::boxplot()}.
#'
#' @return No return value. Called for the side effect of plotting a
#'   \code{nullhespdiv} object.
#'
#' @author Liudas Daumantas
#' @family HespDiv visualization options
#' @importFrom graphics boxplot points
#' @examples
#' test_results <- nulltest(example_hespdiv)
#' plot(test_results)
#' @export
plot.nullhespdiv <- function(x, ...) {
  args <- list(...)

  if (is.null(args$ylim)) {
    args$ylim <- range(c(x[[1]]$performance, unlist(x[[2]])), na.rm = TRUE)
  }

  if (is.null(args$main)) {
    args$main <- paste0("n = ", nrow(x[[2]]))
  }

  do.call(
    graphics::boxplot,
    c(list(x = x[[2]]), args)
  )

  graphics::points(
    1:ncol(x[[2]]),
    x[[1]]$performance,
    col = 2,
    pch = 19,
    cex = 1.5
  )
}

#' Hierarchically subdivide spatial data
#'
#' This function is an implementation of spatial data analysis method HespDiv.
#' It performs hierarchical spatial data subdivision by recursively
#' dividing the data using random split-lines, evaluating their comparison
#' values (how well they separate data), and using the best to perform
#' subdivisions.
#
#' @param data An R object that contains the data to be analyzed. The required data
#' structure depends on the selected method (e.g., character vector of taxa names).
#' @param xy.dat  A data.frame containing the coordinates for observations in the
#' \code{data}. This parameter can be ignored if \code{data} is a data frame or
#'  matrix with columns \code{x} or \code{y}.
#' @param n.split.pts integer. The number of split-points - 1. These points are
#' used in creating straight split-lines (see details). The total number of straight
#' split-lines generated can be obtained by \code{sum(1:n.split.pts)}.
#' Increasing the value of \code{n.split.pts} leads to an increase in both the
#' computation time and the fit to the data.
#' @param same.n.split logical. Should the number of split-points
#' (\code{n.split.pts}) remain the same within each lower-ranked polygon?
#' Choosing \code{TRUE} would result in a higher density of straight split-lines
#' within lower-ranked polygons, whereas \code{FALSE} would preserve the same
#' split-line density. Thus, essentially you are choosing between cross-scale
#' and fixed-scale analysis.
#' @param method Character. Name or abbreviation of a preset method:
#'   \code{"sorensen"}, \code{"pielou"}, \code{"morisita"}, or
#'   \code{"horn.morisita"}. Internally determines values for
#'   \code{compare.f}, \code{generalize.f}, and \code{maximize}.
#' @param generalize.f Function. Optional function used in custom methods to
#'   prepare input for the \code{compare.f} function; see Details. It must have
#'   a single argument, for example \code{x}, to which a spatially filtered
#'   subset of \code{data} can be assigned; see Note for exceptions.
#' @param compare.f function. Only required in custom methods. Employed to
#' quantify the comparison value of a split-line. For this purpose, the
#' \code{compare.f} function requires two arguments where the outputs of the
#' \code{generalize.f} function can be assigned (see notes for exceptions).
#' @param maximize logical. Only required in custom methods. Determines whether
#' the split-line comparison value should be maximized or minimized during the
#' optimization process.
#' @param N.crit number. Minimum required number of observations that should
#' be present in each polygon obtained by a split-line in order for it to meet
#' this criterion.
#' @param N.rel.crit number from 0 to 0.5. Each polygon obtained with a
#' split-line must have at least such proportion of observations to pass this
#' criterion. Equation of the proportion:  (Number of observations in 1st/2nd
#' resulting polygon) / (Number of observations in the polygon being subdivided)
#' @param N.loc.crit number. Minimum required number of different locations that
#' should be present in each polygon obtained by a split-line in order for it to
#' meet this criterion.
#' @param N.loc.rel.crit number from 0 to 0.5. Each polygon obtained with a
#' split-line must have at least such proportion of different locations to pass this
#' criterion. Equation of the proportion:  (Number of different locations in 1st/2nd
#' resulting polygon) / (Number of different locations in the polygon being subdivided)
#' @param S.crit number from 0 to 1. Each polygon obtained with a
#' split-line must have at least such area proportion to pass this
#' criterion. Equation of the proportion:  (Area of 1st/2nd resulting polygon) /
#' (Area of the first polygon). The first polygon is the provided study area
#' or convex hull of observation locations.
#' @param S.rel.crit number from 0 to 0.5. Each polygon obtained with a
#' split-line must have at least such area proportion to pass this
#' criterion. Equation of the proportion:  (Area of 1st/2nd resulting polygon) /
#' (Area of the polygon being subdivided).
#' @param Q.crit number. The threshold for a split-line comparison value to be
#' considered acceptable for a subdivision. When \code{maximize = TRUE}, higher
#' values of the comparison value indicate a better subdivision. Conversely, when
#' \code{maximize = FALSE}, lower values of the comparison value indicate a
#' better subdivision.
#' @param c.splits logical. When set to TRUE, the algorithm will explore
#' nonlinear split-lines in addition to straight split-lines in order to
#' find the optimal subdivision.
#' @param c.Q.crit number.  The threshold for a split-line comparison value to
#' be considered acceptable for generating nonlinear split-lines. It is
#' recommended to use the default value, which does not impose a performance
#' requirement, unless you have a clear understanding of the potential
#' improvements that nonlinear split-lines can achieve over straight split-lines.
#'  If \code{maximize = TRUE}, a suggested value for \code{c.Q.crit} is
#'  \code{Q.crit} minus the maximum potential improvement.
#' @param c.crit.improv number. The threshold for the improvement in a
#' split-line comparison value required for a nonlinear split-line to be
#' selected instead of a straight split-line for subdivision. The default value
#' of 0 means that even if a nonlinear split-line performs equally to a
#' straight split-line, the straight split-line will still be chosen.
#' @param c.X.knots integer. Specifies the number of columns in a network of
#' spline knots used to generate nonlinear split-lines. These knots are evenly
#' distributed along the straight split-line. Adjusting the value of
#' \code{c.X.knots} controls the degree of wiggliness (number of turns) in the
#' resulting nonlinear  split-lines.
#' @param c.Y.knots integer. specifies the number of  rows in a network of
#' spline knots used to generate nonlinear split-lines. These knots are
#' distributed regularly along lines orthogonal to the straight
#' split-line. Adjusting the value of \code{c.Y.knots} controls the number
#' of amplitudes tested in each wiggle of a spline, influencing the shape of
#' nonlinear split-lines.
#' @param c.max.iter.no integer. The maximum number of iterations allowed through
#' the network of spline knots when searching for the optimal shape of a
#' nonlinear split-line. Setting a higher value, such as \code{+Inf}, increases
#' the chances of converging to the best possible curve. It is recommended to
#' use higher values when \code{c.fast.optim = FALSE}, as in this case, only a
#' single spline knot can be selected per iteration.
#' @param c.fast.optim logical. Determines when spline knots are selected. If
#' \code{TRUE}, the algorithm selects the first knot that generates a curve with
#' a better comparison value, ensuring faster convergence. If set to
#' \code{FALSE}, the algorithm completes a full iteration through the network
#' of spline knots, which may result in slower convergence but potentially
#' better overall performance.
#' @param c.corr.term number from 0.01 to 0.2. A correction term for nonlinear
#' split-lines that intersect the boundary of the polygon. Smaller values
#' (default is 0.05) are recommended, as they determine the extent to
#' which the outlying interval of the generated spline, which crosses the
#' polygon boundary, should be shifted away from the boundary and inside the
#' polygon in a direction orthogonal to the straight split-line. This shift is
#' specified as a proportion of the polygon height where the spline intersects
#' the polygon boundary.
#' @param study.pol A data frame with two columns, \code{x} and \code{y}. Rows
#'   should correspond to vertices of a polygon that defines the study area
#'   encompassing all observation locations in \code{xy.dat}. If not provided
#'   (default is \code{NULL}), the convex hull of \code{xy.dat} is used as the
#'   study area polygon.
#' @param use.chull logical. If \code{study.pol} is provided, you can use
#' \code{use.chull = TRUE} (which is default) to use it only for visualization.
#' @param tracing Optional character vector of length two controlling diagnostic
#'   tracing output. The first element specifies the tracing level and must be
#'   one of \code{"best"}, \code{"main"}, or \code{"all"}. The second element
#'   specifies the traced object and must be one of \code{"curves"},
#'   \code{"straight"}, or \code{"both"}. The default, \code{NULL}, produces no
#'   tracing output.
#' @param pnts.col character or numeric. Specifies the color of observations
#' in a plot. The argument is used when \code{tracing} is not NULL. If
#' \code{pnts.col = NULL}, observations will not be displayed.
#' @param display logical. Display a simple plot of results at the end of
#' computations?
#' @param pacific.region logical (default is FALSE). When set to TRUE, indicates
#' that the study area is crossed by the 180th meridian, such as being within
#' the Pacific Ocean. In this case, the coordinates of  \code{xy.dat} and
#' \code{study.pol} are transformed to eliminate the artificial abrupt change
#' in x-coordinate values at the 180th meridian.
#' @param .do_recurse Logical. Controls recursion (for internal use only). Default is `TRUE`.
#' Normal users should not modify this parameter.
#' @return A \code{hespdiv} object, which is a list with seven elements:
#' \describe{
#'   \item{\code{poly.stats}}{
#'     A data frame containing information about polygons established by the
#'     selected split-lines. Its columns are:
#'     \itemize{
#'       \item \code{rank}: The rank of a polygon. It corresponds to the rank of the
#'       split-line that produced the polygon and to the polygon's position in the
#'       hierarchical structure of the subdivision.
#'       \item \code{plot.id}: The ID assigned to the polygon. It corresponds to the
#'       order in which the polygon was processed during the \code{hespdiv} analysis.
#'       \item \code{root.id}: The ID of the parent polygon whose subdivision
#'       resulted in the current polygon.
#'       \item \code{n.splits}: The number of straight split-lines evaluated in an
#'       attempt to subdivide the polygon. This count excludes split-lines that crossed
#'       the polygon boundary or did not meet area, sample size, or location-number criteria.
#'       \item \code{n.obs}: The number of observations inside the polygon.
#'       \item \code{mean}: The average comparison value of the straight split-lines
#'       used in the attempted subdivision of the polygon. This value reflects the
#'       general spatial heterogeneity of the data.
#'       \item \code{sd}: The standard deviation of the comparison values of the
#'       straight split-lines used in the attempted subdivision. It indicates
#'       the extent of anisotropy or variation in spatial heterogeneity within the polygon.
#'       \item \code{str.best}: The comparison value of the best straight split-line
#'       produced within the polygon.
#'       \item \code{str.z.score}: The z-score of the comparison value of the best
#'       straight split-line within the polygon. It indicates how outstanding the best
#'       straight split-line is relative to other evaluated straight subdivisions.
#'       \item \code{has.split}: Logical. Indicates whether a subdivision was established
#'       in the polygon.
#'       \item \code{is.curve}: Logical. Indicates whether the established subdivision
#'       was obtained using a curve. If \code{has.split} is \code{FALSE}, this value is \code{NA}.
#'       \item \code{crv.best}: The same as \code{str.best}, but for nonlinear split-lines.
#'       \item \code{crv.z.score}: The same as \code{str.z.score}, but for nonlinear split-lines.
#'       \item \code{c.improv}: The improvement in comparison value achieved by using
#'       nonlinear split-lines instead of straight split-lines.
#'     }
#'   }
#'   \item{\code{split.stats}}{
#'     A data frame containing information about established split-lines.
#'     Its columns can be interpreted similarly to those in \code{poly.stats},
#'     but from the perspective of split-lines rather than polygons. For example,
#'     \code{rank} is the rank of the split-line, not of the polygon.
#'     The \code{performance} column contains the comparison value of the split-line
#'     and corresponds to either \code{str.best} or \code{crv.best} in \code{poly.stats},
#'     depending on the value of \code{is.curve}. The same applies to the
#'     \code{z.score} column.
#'   }
#'   \item{\code{split.lines}}{
#'     A list of data frames containing coordinates of established split-lines.
#'     The order of split-lines in this list corresponds to their order in
#'     \code{split.stats}.
#'   }
#'   \item{\code{polygons.xy}}{
#'     A list of data frames containing coordinates of polygons established by
#'     the split-lines. The order of polygons in this list corresponds to their
#'     order in \code{poly.stats}.
#'   }
#'   \item{\code{poly.obj}}{
#'     A list of polygon objects, that is, outputs of \code{generalize.f} for each polygon.
#'     The order of elements in this list corresponds to the row order of
#'     \code{poly.stats} and the polygon order in \code{polygons.xy}.
#'   }
#'   \item{\code{call.info}}{
#'     Information about the \code{hespdiv} call, including the method and arguments used.
#'   }
#'   \item{\code{str.difs}}{
#'     A list containing comparison values of evaluated straight split-lines for
#'     each polygon. The elements of this list correspond to the rows of
#'     \code{poly.stats}.
#'   }
#' }
#' @note Please note that if you use the \code{method} argument, the arguments
#' \code{generalize.f}, \code{compare.f}, and \code{maximize} are determined
#' internally and should not be provided. Therefore, you should only assign
#' values to these arguments when using a custom method, not a predefined one.
#' Additionally, you can ignore the \code{generalize.f} argument even when
#' applying custom methods. If \code{generalize.f} is set to NULL (default), the
#' data remains unchanged, as \code{generalize.f} acts as an identity function.
#' Hence, \code{generalize.f} is only an optional argument that allows to omit
#' the transformation or generalization step in \code{compare.f} function,
#' simplifying it.
#'
#' Both \code{generalize.f} and \code{compare.f} inherit environments from
#' parent functions: \code{hespdiv()} and \code{.spatial_div()}. This
#' allows them to use additional variables from those environments, such as
#' \code{xy.dat}, \code{samp.xy}, \code{id1}, and \code{id2}.
#'
#' There is a small possibility that a nonlinear split-line may cross a polygon
#' boundary. If the result contains such a split-line, or if an error related to
#' this issue occurs, slightly change one of the arguments, such as
#' \code{n.split.pts}, and re-run \code{hespdiv()}.
#'
#' @details
#' \subsection{The Algorithm}{
#' \bold{1) Split-point Placement:} The function places a predetermined number of
#' split-points (\code{n.split.pts + 1}) along the perimeter of the study area
#' (\code{study.pol}) or the convex hull of observation locations
#' (\code{xy.dat}) if a study area polygon is not provided. These split-points
#' are evenly spaced, resulting in a distance between points equal to
#' 1/(n.split.pts + 1) fraction (1/16 by default) of a polygon circumference.
#'
#' \bold{2) Straight Split-lines:} Straight split-lines are generated by connecting
#' the split-points. The total number of straight split-lines generated is equal
#' to the value of \code{(n.split.pts + 1) * n.split.pts / 2} or to
#' \code{choose(n.split.pts + 1, 2)}. This holds true only in the first iteration when
#' \code{same.n.split} is set to \code{FALSE} or in all iterations when
#' \code{same.n.split} is set to \code{TRUE}. Note that the total number of
#' split-lines generated will not be equal to the number of split-lines
#' evaluated since split-lines that cross polygon boundary or do not pass
#' sample size, area or location number subdivision criteria are not evaluated.
#'
#' \bold{3) Subdivisions:} Each split-line spatially divides the data and study area
#' polygon into two subsets.
#'
#' \bold{4) Criteria:} Both subsets are then checked to see if they meet
#' sample size, area and location number criteria.
#'
#' \bold{5) Obtaining comparison values:} Subsets that meet the criteria are
#' compared using the \code{generalize.f} and \code{compare.f} functions to
#' obtain a comparison value. First, each subset is passed to
#' \code{generalize.f} to obtain a generalization value, such as the Pielou
#' evenness index for the \code{"pielou"} method; see the \strong{Note} section
#' for clarification. These values are then used by \code{compare.f} to compare
#' the subsets. In this way, each split-line that passed all subdivision criteria
#' is assigned a comparison value.
#'
#' \bold{6) Best Straight Split-line Selection:} The best performing straight
#' split-line is determined based on whether the \code{maximize} argument is
#' set to \code{TRUE} or \code{FALSE}. If \code{maximize} is \code{TRUE}, the
#' best split-line is the one with the highest comparison value; if
#' \code{maximize} is \code{FALSE}, the best split-line has the lowest
#' comparison value.
#'
#' The combination of the \code{generalize.f}, \code{compare.f}, and
#' \code{maximize} arguments can be provided, enabling the creation of custom
#' methods, or it can be determined internally based on the chosen preset
#' subdivision methods (see below).
#'
#' \bold{7) Nonlinear Split-lines:} If the best straight split-line meets
#' the quality criteria specified by the \code{c.Q.crit}, it serves as the basis
#' for generating variously shaped curves (nonlinear split-lines). These curves
#' are produced using splines, which are mathematical functions that can create
#' smooth and flexible curves.
#'
#' To generate the curves, a number of knots (control points) for the splines are
#' distributed evenly along a set of lines orthogonal to the best straight
#' split-line. These orthogonal lines are also evenly distributed along the
#' straight split-line itself. The number of knots and lines used can be
#' adjusted through the parameters \code{c.Y.knots} and \code{c.X.knots},
#' respectively.
#'
#' The algorithm then iterates through this network of knots, considering
#' different combinations of knots, to produce curves. By varying the selection
#' and arrangement of knots, different shapes of curves are generated.
#'
#'
#' \bold{8) Nonlinear Split-line Evaluation and Selection:} Curves are then
#' processed in the same manner as straight split-lines in steps 3 to 6.
#'
#' \bold{9) Final Split-line Selection and Establishment of Subdivision:} If the best
#' curve outperforms the best straight split-line by a margin of
#' \code{c.crit.improv}, the best split-line becomes nonlinear. Otherwise, it
#' remains straight. The split-line must also satisfy the criteria established
#' by the \code{Q.crit} argument to be used for subdivision.
#'
#' \bold{10) Recursive Iteration:} The process described above is iteratively applied,
#' resulting in a collection of selected split-lines with their performance
#' values. These split-lines hierarchically subdivide space and data, forming
#' polygons of various shapes.
#' }
#' \subsection{Preset Methods}{
#' Preset methods and their combinations of \code{compare.f},
#' \code{generalize.f}, and \code{maximize}:
#' \describe{
#'   \item{\code{"sorensen"}}{
#'     Functions calculate the Sorensen similarity index
#'     (Sorensen 1948). \code{maximize = FALSE}. Values vary from 0 to 1.
#'   }
#'   \item{\code{"pielou"}}{
#'     Functions calculate the mean proportional decrease in
#'     Pielou evenness (Pielou 1966) that occurs after polygon subdivision.
#'     \code{maximize = TRUE}. Values vary from 0 to 1.
#'   }
#'   \item{\code{"morisita"}}{
#'     Functions calculate the Morisita overlap index
#'     (Morisita 1959). \code{maximize = FALSE}. Values vary from 0 to 1.
#'   }
#'   \item{\code{"horn.morisita"}}{
#'     Functions calculate the Morisita overlap index, as modified by
#'     Horn (1966). \code{maximize = FALSE}. Values vary from 0 to 1.
#'   }
#' }
#' All preset methods currently available are specifically designed for
#' bioregionalization purposes. These methods require two key inputs: the
#' coordinates of fossil taxon occurrences (\code{xy.dat}) and the names or IDs
#' of taxa (\code{data}). These names or IDs should be structured as
#' character or numeric vectors, with each element corresponding to a row in the
#' \code{xy.dat} data frame. Each method compares fossil communities on
#' opposite sides of a split-line, aiming to minimize similarity or
#' maximize difference. The output yields biogeographical provinces with a
#' hierarchical structure.
#' }
#' @author Liudas Daumantas
#' @importFrom grDevices chull
#' @importFrom pracma polyarea
#' @references Horn, H. S. (1966). Measurement of" overlap" in comparative ecological studies. The American Naturalist, 100(914), 419-424.
#' @references Morisita, M. (1959). Measuring of interspecific association and similarity between assemblages. Mem Fac Sci Kyushu Univ Ser E Biol, 3, 65-80.
#' @references Pielou, E. C. (1966). The measurement of diversity in different types of biological collections. Journal of theoretical biology, 13, 131-144.
#' @references Sorensen, T. A. (1948). A method of establishing groups of equal amplitude in plant sociology based on similarity of species content and its application to analyses of the vegetation on Danish commons. Biol. Skar., 5, 1-34.
#' @examples
#' # Simulated data with a known boundary at x = 0.45 to illustrate boundary detection
#'
#' study.area <- data.frame(
#'   x = c(0, 0.8, 1, 0.6, 0, 0),
#'   y = c(0, 0, 0.4, 1.1, 1.1, 0)
#' )
#'
#' set.seed(852)
#' # Simulate 100 occurrence coordinates
#' N <- 100
#' xy.dat_arg <- data.frame(
#'   x = rpois(N, 4) / 10,
#'   y = rpois(N, 4) / 10
#' )
#'
#' xy.dat_arg <- xy.dat_arg[order(xy.dat_arg$x), ]
#'
#' # Simulate a boundary at x = 0.45
#' n_left <- sum(xy.dat_arg$x < 0.45)
#'
#' set.seed(1)
#' common_data <- letters[1:5]
#'
#' left_data <- sample(
#'   c(common_data, LETTERS[1:10]),
#'   size = n_left,
#'   replace = TRUE,
#'   # common-endemic probability ratio 3:4
#'   prob = c(rep(3, 5), rep(4, 10))
#' )
#'
#' right_data <- sample(
#'   c(common_data, LETTERS[11:20]),
#'   size = N - n_left,
#'   replace = TRUE,
#'   # common-endemic probability ratio 3:4
#'   prob = c(rep(3, 5), rep(4, 10))
#' )
#'
#' data_arg <- c(left_data, right_data)
#'
#' # Apply hespdiv
#' r <- hespdiv(
#'   data = data_arg,
#'   xy.dat = xy.dat_arg,
#'   n.split.pts = 6,  # small value used here for illustration
#'   method = "sor",   # subdivision minimizing Sorensen-Dice similarity
#'   S.crit = 0.3,     # minimum area size is 30% of study area
#'   study.pol = study.area,
#'   use.chull = FALSE
#' )
#'
#' plot_hespdiv(r, n.loc = TRUE) + ggplot2::geom_vline(xintercept = 0.45)
#'
#' # Detected split-line performance
#' r$split.stats$performance
#'
#' # Sorensen-Dice similarity across the true simulated boundary
#' 2 * length(intersect(left_data, right_data)) /
#'   (length(unique(left_data)) + length(unique(right_data)))
#' @export

hespdiv<-function(data,
                  xy.dat = NULL,
                  n.split.pts = 15,
                  same.n.split = TRUE,
                  method = 'horn.morisita',
                  generalize.f = NULL,
                  compare.f = NULL,
                  maximize = NULL,
                  N.crit = 1,
                  N.rel.crit = 0.2,
                  N.loc.crit = 1,
                  N.loc.rel.crit = 0.2,
                  S.crit = 0.05,
                  S.rel.crit = 0.2,
                  Q.crit = NULL,
                  c.splits = TRUE,
                  c.Q.crit = NULL,
                  c.crit.improv = 0,
                  c.X.knots = 5,
                  c.Y.knots = 10,
                  c.max.iter.no = +Inf,
                  c.fast.optim = FALSE,
                  c.corr.term = 0.05,
                  study.pol = NULL,
                  use.chull = TRUE,
                  tracing = NULL,
                  pnts.col = 1,
                  display = FALSE,
                  pacific.region = FALSE,
                  .do_recurse = TRUE){

  if (is.null(xy.dat)){
    if (class(data) %in% c("data.frame", "matrix")){
      if (any(colnames(data) == 'x') & any(colnames(data) == 'y')){
        if (sum(colnames(data) %in% c('x','y')) != 2){
          stop(
            "`data` must not contain more than one pair of columns named `x` and `y`.",
            call. = FALSE
          )
        }
        xy.dat <- as.data.frame(data[,c('x','y')])
        data <- data[,-which(colnames(data) %in% c('x','y'))]
      } else {
        stop(
          "`xy.dat` was not provided, but is required when `data` is not a data frame ",
          "containing `x` and `y` columns.",
          call. = FALSE
        )
      }
    } else {
      stop(
        "`xy.dat` was not provided, but is required when `data` is not a data frame ",
        "containing `x` and `y` columns.",
        call. = FALSE
      )
    }
  } else {
    if (any(colnames(data) == 'x') & any(colnames(data) == 'y')){
      stop(
        "`data` contains `x` and `y` columns while `xy.dat` is also provided. ",
        "Remove one of these coordinate sources from the input.",
        call. = FALSE
      )
    }
    if (!is.data.frame(xy.dat)){
      if (is.matrix(xy.dat)){
        xy.dat <- as.data.frame(xy.dat)
      } else {
        stop("`xy.dat` must be a data frame.", call. = FALSE)
      }
    }
    if (ncol(xy.dat) !=2 )
      stop(
        "`xy.dat` must have exactly two columns named `x` and `y`.",
        call. = FALSE
      )

    if (any(colnames(xy.dat) != c('x','y')))
      stop(
        "`xy.dat` must contain columns `x` and `y` in that order.",
        call. = FALSE
      )

    if (is.matrix(data) | is.data.frame(data)){
      if (nrow(data) != nrow(xy.dat))
        stop(
          "`xy.dat` must have one row for each observation in `data`.",
          call. = FALSE
        )
    } else {
      if (length(data) != nrow(xy.dat))
        stop(
          "`xy.dat` must have one row for each observation in `data`.",
          call. = FALSE
        )
    }
  }
  args <- sapply(ls(),get,environment())
  n.split.pts <- n.split.pts + 1
  c.Y.knots <- c.Y.knots + 2
  c.X.knots <- c.X.knots + 2

  if (!is.null(tracing)) {
    if (!is.vector(tracing) || length(tracing) != 2) {
      stop("`tracing` must be a vector with exactly two elements.", call. = FALSE)
    }

    trace.object <- tracing[1]
    trace.level <- tracing[2]
  }

  if (exists("trace.level")){
    if (!is.null(trace.object)){
      trace.object <- .arg_check("trace.object", trace.object,
                                 c("straight", "curve", "both"))
    }
    if (!is.null(trace.level)){
      trace.level <- .arg_check("trace.level", trace.level,c("all","main","best"))
    }
  } else {
    trace.object <- NULL
    trace.level <- NULL
  }

  if (is.null(compare.f)){
    if (is.null(method)){
      stop(
        "Neither `method` nor `compare.f` is specified. ",
        "Please select a preset method or provide a custom method.",
        call. = FALSE
      )
    }
    method.type <- "preset"
    method <- .arg_check("metric", method, names(
      .get_methods()[['biozonation']]))
    if (method == "pielou") {
      variant <- '1'
      metric <- "pielou"
      method <- 'biozonation'
      similarity <- FALSE
      maximize <- TRUE
      if(!is.vector(data))
        stop(
          "`data` must be a vector when using method \"", method, "\".",
          call. = FALSE
        )

      generalize.f <- function(plot.dat){
        x <- table(plot.dat)
        p <- x/sum(x)
        -sum(log(p)*p)/log(length(p))
      }
      compare.f <- function(x,y) {
        base.eveness <- poly.obj[[testid]]
        (1 - mean(c(x, y)) / base.eveness) # proportional change in mean eveness
      }

    } else {
      if (method == "sorensen"){
        similarity <- TRUE
        variant <- '1'
        metric <- 'sorensen'
        method <- 'biozonation'
        maximize <- FALSE
        if(!is.vector(data))
          stop(
            "`data` must be a vector when using method \"", method, "\".",
            call. = FALSE
          )

        generalize.f <- function(plot.dat){
          unique(plot.dat)
        }
        compare.f <- function(x,y) {
          sum_total <- length(x) + length(y)  # Total taxa in both samples
          shared_taxa <- length(intersect(x, y)) * 2  # Shared taxa counted twice for Sørensen
          sorensen_similarity <- shared_taxa / sum_total  # Sørensen similarity formula
        }
      } else {
        if (method == "morisita"){
          similarity <- TRUE
          metric <- "morisita"
          maximize <- FALSE
          variant <- '1'
          method <- 'biozonation'
          if(!is.vector(data))
            stop(
              "`data` must be a vector when using method \"", method, "\".",
              call. = FALSE
            )

          generalize.f <- function(plot.dat){
            plot.dat
          }

          compare.f <- function(x,y) {
            all_sp <- unique(c(x,y))
            x_f <- factor(x,levels = all_sp)
            y_f <- factor(y,levels = all_sp)
            (2*sum(table(x_f) * table(y_f)))/
              (length(x) * length(y) *
                 ((sum(table(x_f)*(table(x_f)-1)) /
                     (length(x)* (length(x)-1))) +
                    (sum(table(y_f)*(table(y_f)-1)) /
                       (length(y)* (length(y)-1)))))
          }
        } else {
          if (method == "horn.morisita"){
            similarity <- TRUE
            metric <- "horn.morisita"
            maximize <- FALSE
            variant <- '2'
            method <- 'biozonation'
            if(!is.vector(data))
              stop(
                "`data` must be a vector when using method \"", method, "\".",
                call. = FALSE
              )

            generalize.f <- function(plot.dat){
              plot.dat
            }

            compare.f <- function(x,y) {
              all_sp <- unique(c(x,y))
              x_f <- factor(x,levels = all_sp)
              y_f <- factor(y,levels = all_sp)
              (2*sum(table(x_f) * table(y_f))) /
                (length(x) * length(y) *
                   ( sum(table(x_f)^2) / (length(x)^2)  +
                       sum(table(y_f)^2) / (length(y)^2)))
            }
          }
        }
      }
    }
    args$maximize <- maximize
    args$generalize.f <- generalize.f
    args$compare.f <- compare.f
  } else {
    if (!is.null(method)){
      stop(
        "Only one of `compare.f` and `method` can be provided; the other must be NULL.",
        call. = FALSE
      )
    }
    if (is.null(generalize.f)) {
      generalize.f <- function(plot.dat) plot.dat
      args$generalize.f <- generalize.f
      }
    method.type <- "custom"
    variant <- NULL
    metric <- NULL
    similarity <- NULL
    if (is.null(maximize)){
      stop("Argument `maximize` must be provided.", call. = FALSE)
    }
  }

  if (is.data.frame(data) | is.matrix(data)){
    .slicer <- .slicer.table
  } else {
    if (is.list(data)) {
      .slicer <- .slicer.list
    } else {
      if (!is.vector(data))
        warning(
          "`hespdiv()` does not recognize the structure of `data`.\n",
          "The function will attempt to subset `data` using `[]`.",
          call. = FALSE
        )
      .slicer <- .slicer.vect
    }
  }


  if (!maximize){
    if (is.null(Q.crit)) Q.crit <- +Inf
    if (is.null(c.Q.crit)) c.Q.crit <- Q.crit
    .comp <- function(x,criteria){ x < criteria}
    c.sign <- "<"
    .minormax <- min
    .which_minormax <- which.min
  } else {
    if (is.null(Q.crit)) Q.crit <- -Inf
    if (is.null(c.Q.crit)) c.Q.crit <- Q.crit

    .comp <- function(x,criteria){ x > criteria}
    c.sign <- ">"
    .minormax <- max
    .which_minormax <- which.max
  }
  args$c.Q.crit <- c.Q.crit
  args$Q.crit <- Q.crit
  if ((c.splits == FALSE & Q.crit != c.Q.crit) |
      ((Q.crit > c.Q.crit) & maximize) |
      ((Q.crit < c.Q.crit) & !maximize) ){
    if (c.splits == FALSE & Q.crit != c.Q.crit & !c.Q.crit %in% c(+Inf,-Inf)){
      warning(
        "Since `c.splits` is FALSE, `c.Q.crit` was set equal to `Q.crit`.",
        call. = FALSE
      )
      c.Q.crit <- Q.crit
    }
    if ((Q.crit > c.Q.crit) & !maximize){
      warning(
        "`Q.crit` should be less than or equal to `c.Q.crit` when optimization is ",
        "reached by minimizing the metric.",
        call. = FALSE
      )
    }
    if ((Q.crit < c.Q.crit) & maximize){
      warning(
        "`Q.crit` should be greater than or equal to `c.Q.crit` when optimization is ",
        "reached by maximizing the metric.",
        call. = FALSE
      )
    }
  }
  if (pacific.region){
    if (!is.null(study.pol)) {
      ind <- study.pol[,1] >= -180 & study.pol[,1] <= 0
      study.pol[ind,1] <- study.pol[ind,1] + 360
    }
    ind <- xy.dat$x >= -180 & xy.dat$x <= 0
    xy.dat$x[ind] <- xy.dat$x[ind] + 360
  }

  if (use.chull) {
    if (!is.null(study.pol)) {
      study.pol.viz <- study.pol
    }
    ch <- chull(xy.dat$x, xy.dat$y)
    study.pol <- data.frame(x = xy.dat$x[c(ch,ch[1])],
                            y = xy.dat$y[c(ch,ch[1])])
  } else {
    if (is.null(study.pol)){
      stop(
        "`study.pol` must be provided unless `use.chull = TRUE`.",
        call. = FALSE
      )
    }
    study.pol <- .clean_poly(study.pol)
  }
  if (!exists("study.pol.viz")){
    study.pol.viz <- NULL
  }
  if (!same.n.split){
    dst.pts <- .calc.perim(study.pol) / n.split.pts
  } else {
    dst.pts <- NULL
  }


  rims <- list(study.pol)

  poly.info <- data.frame(root.id = numeric(),
                          n.splits = numeric(),
                          n.obs = numeric(),
                          mean = numeric(), # mean spatial heterogeneity irrespective of split-line position
                          sd = numeric(), # anizotropy of heterogeneity based on straight split-lines
                          str.best = numeric(),
                          str.z.score = numeric(), # level of outstandingness. Are there other competetive candidate splits?,
                          has.split = logical(),
                          is.curve = logical(),
                          crv.best = numeric(),
                          crv.z.score = numeric(),
                          c.improv = numeric())

  poly.obj <- list()
  str.split.quals <- list()

  S.cond <- round(abs(pracma::polyarea(study.pol$x,study.pol$y)) * S.crit,2)
  splits <- numeric()

  e <- environment()
  environment(.spatial_div) <- e
  ### obtaining results:
  .spatial_div(data,xy.dat,root.id=0, recursive = .do_recurse)

  ### results obtained. Formatting them.
  if (length(splits)>0) {
    plot.id <- 1:nrow(poly.info)
    names(poly.obj) <- plot.id
    names(rims) <- plot.id
    poly.info <- data.frame(plot.id = plot.id,poly.info)

    names(str.split.quals) <- 1:length(str.split.quals)


    # which best is not clear, when no split was above c.Q.crit
    # (c.Q.crit is lower performance boundary, Q.crit - upper;
    # if c.Q.crit is beaten with straight split-line, but not Q.crit,
    # then it is possible that non-linear split-line will beat Q.crit;
    # If no split-line beats c.Q.crit, the str.best value in poly.info will be
    # c.Q.crit, but not the best performance of straight split-line)
    if (any(poly.info$str.best == c.Q.crit, na.rm = TRUE)){

      plot.id.with.ns <- which(poly.info$n.splits != 0)
      poly.info$str.best[-plot.id.with.ns] <- NA
      poly.info$str.best[plot.id.with.ns] <- ifelse(poly.info$str.best[
        plot.id.with.ns] == c.Q.crit,
        unlist(lapply(str.split.quals[plot.id.with.ns],.minormax)),
        poly.info$str.best[plot.id.with.ns])
    }
    if (!c.splits){
      poly.info <- poly.info[,-c(10:13)]
    } else {
      # if no data about curves, then we don't know how curves compare to straight
      poly.info$is.curve <- ifelse(is.na(poly.info$crv.best),NA,
                                   poly.info$is.curve)
      # is no improvement was made by the curve, then we dont know the performance
      # of the best curve since worse values than straight-split performance
      # are not saved.
      poly.info$crv.best <- ifelse(poly.info$c.improv == 0,NA,
                                   poly.info$crv.best)
      poly.info$crv.z.score <- ifelse(poly.info$c.improv == 0,NA,
                                      poly.info$crv.z.score)
    }
    if (any(poly.info$has.split)){
      names(splits) <- 1:sum(poly.info$has.split)
    }

    plot.id <- which(poly.info$has.split)
    e <- environment()
    result <- .format_result(e)
    result$poly.stats <- data.frame(rank = .split.rank(
      result$poly.stats),
      result$poly.stats)
    result$split.stats <- data.frame(
      rank = result$poly.stats[result$split.stats$plot.id,"rank"],
      result$split.stats)

    if (display){
      .visualise_splits.end(pnts.col, xy.dat, rims, study.pol.viz)
    }

    return(print.hespdiv(result))
  } else {
    base::message("No split-lines were established.\nTry loosening the split-line selection criteria.")
  }
}
.slicer.table <- function(sample,ind) sample[ind,]
.slicer.list <- function(sample,ind) sample[[ind]]
.slicer.vect <- function(sample,ind) sample[ind]

.format_result <- function(e){
  if (e$method.type == "custom"){
    return(.format_result.general(e))
  } else{
    if (e$metric == "pielou"){
      return(.form.bioz.Pielou(.format_result.general(e)))
    } else {
      return(.format_result.general(e))
    }
  }
}
.structurise <- function(e,split.stats,split.lines){

  structure(list(
    split.lines = split.lines,
    polygons.xy = e$rims,
    poly.stats = e$poly.info,
    poly.obj = structure(e$poly.obj),
    split.stats = structure(split.stats),
    call.info = list(METHOD =
                       list(method = e$method, metric = e$metric,
                            variant = e$variant, similarity = e$similarity,
                            maximize = e$maximize, method.type = e$method.type),
                     Call_ARGS = e$args),
    str.difs = e$str.split.quals
  ),
  class = "hespdiv"
  )
}
.form.bioz.Pielou <- function(result){
  parent.Pe <- unlist(result$poly.obj)
  result$poly.stats$parent.Pe <- parent.Pe
  if (!is.null(result$split.stats) ){
    result$split.stats$parent.Pe <- parent.Pe[result$split.stats$plot.id]
    result$split.stats$delta.Pe <- -result$split.stats$parent.Pe *
      result$split.stats$performance/100
  }
  result
}
.format_result.general <- function(e){
  if (length(e$plot.id) == 0 ){
    split.stats <- NULL
    split.lines <- NULL
  } else {
    split.lines <- e$splits
    if (e$c.splits){
      split.stats <- data.frame(
        plot.id = e$plot.id,
        n.splits = e$poly.info$n.splits[e$plot.id],
        n.obs = e$poly.info$n.obs[e$plot.id],
        mean = e$poly.info$mean[e$plot.id],
        sd = e$poly.info$sd[e$plot.id],
        z.score = apply(e$poly.info[e$plot.id,],1,
                        function(o) if (o[[10]]) {o[[12]]} else {o[[8]]}),
        performance = apply(e$poly.info[e$plot.id,],1,
                            function(o) if (o[[10]]) {o[[11]]} else {o[[7]]}),
        is.curve = e$poly.info$is.curve[e$plot.id]
      )
    } else {
      split.stats <- data.frame(
        plot.id = e$plot.id,
        n.splits = e$poly.info$n.splits[e$plot.id],
        n.obs = e$poly.info$n.obs[e$plot.id],
        mean = e$poly.info$mean[e$plot.id],
        sd = e$poly.info$sd[e$plot.id],
        z.score = e$poly.info[e$plot.id,"str.z.score"],
        performance = e$poly.info[e$plot.id,"str.best"]
      )
    }
  }
  .structurise(e,split.stats,split.lines)
}

.arg_check <- function(name, given,NAMES){
  matched.i <- pmatch(tolower(given), tolower(NAMES))
  if(is.na(matched.i))
    stop(
      "Invalid value for `", name, "`: \"", given, "\".\n",
      "Please select a valid option: ",
      paste(sprintf("\"%s\"", NAMES), collapse = ", "),
      call. = FALSE
    )
  NAMES[matched.i]
}

.get_methods <- function(){
  list(
    "biozonation" = list(
      "pielou" = c("1"),
      "morisita" = c("1"),
      "sorensen" = c("1"),
      "horn.morisita" = c("2")
    )
  )
}
# Calculate rank of the split-line
#' @noRd
.split.rank <- function(poly.stats){

  roots <- poly.stats$root.id
  ranks.id <- numeric(length(roots))
  for (split in seq(length(roots))){
    split_rank <- 1
    split_root <- roots[split]
    while (split_root != 0) {
      split_rank <- split_rank + 1
      split_root <- poly.stats$root.id[
        which(poly.stats$plot.id == split_root)]
    }
    ranks.id[split] <- split_rank
  }
  ranks.id
}
#  get ids of observations within polygon
#' @noRd
.get_ids <- function(polygon, xy_dat) {
  which(.point.in.polygon(pol.x = polygon[,1],
                             pol.y = polygon[,2],
                             point.x = xy_dat$x,
                             point.y = xy_dat$y) != 0 )
}

#' Find the curve that divides the polygon best (main function)
#'
#' @description function prepares the data to start the searching procedure of the curve that provides the best spatial separation.
#' As curve are generated using splines, matrix of knot coordinates are prepared. Also, polygon is rotated so that split line would be horizontal and at Y = 0, and start at X = 0.
#' @param poly.x a vector of x coordinates of a polygon perimeter points
#' @param poly.y a vector of y coordinates of a polygon perimeter points
#' @param min.x.id index of split line vertex in poly.x and poly.y objects that has lower x coordinate
#' @param max.x.id index of split line vertex in poly.x and poly.y objects that has lower y coordinate
#' @param b slope of a split line
#' @param samp.dat spatial subset of data
#' @param c.X.knots number of spline knots along the split line
#' @param c.Y.knots number of spline knots orthogonal to the split line
#' @param N.crit minimum number of observations required to establish
#' subdivision of a polygon.
#' @param N.rel.crit number from 0 to 0.5. The minimum allowed observation
#' proportion in polygon containing less data.
#' @param N.loc.crit Subdivision stopping criteria - number of unique locations.
#' Only meaningful when there are observations from the same location. Default
#' 0.2.
#' @param N.loc.rel.crit number from 0 to 0.5. The minimum allowed unique
#' locations proportions in polygon containing less locations.
#' @param S.cond minimum area required to establish subdivision of a plot
#' @param S.rel.crit number from 0 to 0.5. The minimum allowed area
#' proportion of a smaller polygon. Default 0.2.
#' @param S_org area of the polygon being divided.
#' @param c.max.iter.no maximum number of allowed iterations through the spline
#' knots.
#' @param c.fast.optim Logical (default TRUE). Determines when spline knots
#' are selected: TRUE - when first better curve than before is obtained, FALSE -
#' when the best curve is obtained (after all spline knots are checked).
#' @param trace.level Sting indicating the trace level. Can be one of the
#' following: "best", "main", "all"
#' @param trace.object String that informs what graphical information to show.
#' Can be either "curves", "straight" or "both".
#' @param c.corr.term term that defines how much the a problematic spline
#' will be corrected (in terms of proportion of polygon width where spline
#' intersects the polygon boundary) if the spline is not contained within the
#' plot. Small values recommended (default is 0.05).
#' @param pnts.col Color of data points, default is 1. Argument is used when
#' \code{trace} > 0. If set to NULL, data points will not be displayed.
#' @param straight.qual Quality of the best straight split-line.
#' @return A list of two elements: 1) curve in shape of a spline that produces
#' the best data separation; 2) quality of the division
#' @noRd
.curvial_split <- function(poly.x,poly.y,min.x.id,max.x.id,b,
                           samp.dat,c.X.knots=20,c.Y.knots=20,
                           N.crit,
                           N.rel.crit,N.loc.crit,N.loc.rel.crit,S.cond,
                           S.rel.crit, S_org, c.max.iter.no, c.fast.optim,
                           c.corr.term,
                           trace.object = trace.object,
                           trace.level = trace.level,
                           pnts.col = pnts.col,
                           straight.qual){
  #nustatau, kuris padalinimo linijos id yra kairej, kuris desinej
  # length of a split line
  #AE<-sqrt(sum((c(poly.x[min.x.id],
  #                poly.y[min.x.id])-c(poly.x[max.x.id],poly.y[max.x.id]))^2))
  #randu kampa, kuriuo reikia paversti poligona, kad padalinimo linija butu
  # horizontali
  teta<-atan(b)
  #paverciu poligona ir duomenis
  rot.pol.cords <- DescTools::Rotate(x=poly.x,y=poly.y,mx=poly.x[min.x.id],
                          my=poly.y[min.x.id],theta=-teta)
  rot.dat.cords <- DescTools::Rotate(x=samp.xy$x,y=samp.xy$y,mx=poly.x[min.x.id],
                          my=poly.y[min.x.id],theta=-teta)

  #pastumiu duomenis ir poligona, kad padalinimo linijos kairinis taskas butu
  # koordinaciu sistemos pradzioje
  rot.dat.cords <- data.frame(x = rot.dat.cords$x-poly.x[min.x.id],
                              y = rot.dat.cords$y-poly.y[min.x.id])
  #rot.data<-data.frame(data[,1],x = rot.dat.cords$x-poly.x[min.x.id],
   #                    y = rot.dat.cords$y-poly.y[min.x.id])
  #sukuriu duomenu masyva pasukto poligono
  rot.poli<-data.frame(x= rot.pol.cords$x-poly.x[min.x.id],
                       y= rot.pol.cords$y-poly.y[min.x.id])
  #padalinimo linijos pabaigoje y turetu buti 0, bet del paklaidu jis gali buti
  # nelygus 0 ir filtracija gali blogai ivykti - nunulinu
  rot.poli[-min.x.id,2] <- rot.poli[-min.x.id,2] - rot.poli[max.x.id,2] # removing error at the
  # end of the split line.
  AE <- rot.poli[max.x.id,1]
  #"atidarau" poligona
  # rot.poli<-rot.poli[-nrow(rot.poli),]
  #nufiltruoju virsutine ir apatine poligono dali, t
  #askai tampa organizuoti palei poligono kreive is kaires i desine
  rot.poli.up<-.close_poly(.split_poly(
    polygon = rot.poli,
    min_id = 1,
    split_ids = c(min.x.id,max.x.id),
    trivial_side = TRUE,
    poli_side = TRUE))
  poli.up <- .close_poly(.split_poly(
    polygon = data.frame(x = poly.x, y = poly.y),
    min_id = 1,
    split_ids = c(min.x.id,max.x.id),
    trivial_side = TRUE,
    poli_side = TRUE))

  rot.poli.do<-.close_poly(.split_poly(
    polygon = rot.poli,
    min_id = 1,
    split_ids = c(min.x.id,max.x.id),
    trivial_side = TRUE,
    poli_side = FALSE))
  poli.do <- .close_poly(.split_poly(
    polygon = data.frame(x = poly.x, y = poly.y),
    min_id = 1,
    split_ids = c(min.x.id,max.x.id),
    trivial_side = TRUE,
    poli_side = FALSE))

  #randu vidinio pataisyto poligono apatine ir virsutine dali
  bottom.inner.poli <-
    .inner_poly_hull(polygon = rot.poli.do[!duplicated(rot.poli.do),])
  upper.inner.poli <-
    .inner_poly_hull(polygon = rot.poli.up[!duplicated(rot.poli.up),])
  if( bottom.inner.poli[[2]] == 1 ){

    change <- bottom.inner.poli
    bottom.inner.poli <- upper.inner.poli
    upper.inner.poli <- change

    change <- rot.poli.do
    rot.poli.do <- rot.poli.up
    rot.poli.up <- change

    change <- poli.up
    poli.up <- poli.do
    poli.do <- change

  }

  # The following steps are needed to identify points that lie exactly
  # on the contour of polygon, because after rotation small errors are
  # introduced and because of them these points might be detected as lying
  # oustide polygon. For this reason, some points might be not considered
  # when searching for the best curvial split line. Thus, the ids of these
  # points are found, and added, if missed, to remove this bug.
  #
  {
  # unr_ids: filter point ids by unrotated  up & bottom poly
  unr_ids_up <- .get_ids(polygon = poli.up, xy_dat = samp.xy)
  unr_ids_do <- .get_ids(polygon = poli.do, xy_dat = samp.xy)
  # r_ids: filter point ids by rotated up & bottom poly
  r_ids_up <- .get_ids(polygon = rot.poli.up, xy_dat = rot.dat.cords)
  r_ids_do <- .get_ids(polygon = rot.poli.do, xy_dat = rot.dat.cords)

  # m_ids: find unr_ids that are missing in r_ids
  m_ids_up <- unr_ids_up[which(!unr_ids_up %in% r_ids_up)]
  m_ids_do <- unr_ids_do[which(!unr_ids_do %in% r_ids_do)]
  # theoretically m_ids are points on polygon perimeter

  # if m_ids is not empty  (filter.all?):
  # split_pt_ids find ids of data points that are on a straight split line
  # - point in poly -  3, when up or bottom, but 1 when whole.

  if (length(m_ids_up) > 0 | length(m_ids_do) > 0 ){
    ids.on.pol.updo <-
    unique(c(which(.point.in.polygon(pol.x = poli.up[,1],
                               pol.y = poli.up[,2],
                               point.x = samp.xy$x,
                               point.y = samp.xy$y) == 3),
    which(.point.in.polygon(pol.x = poli.do[,1],
                               pol.y = poli.do[,2],
                               point.x = samp.xy$x,
                               point.y = samp.xy$y) == 3)))
    split_pt_ids <- ids.on.pol.updo[
      ids.on.pol.updo %in%
        which(.point.in.polygon(pol.x = poly.x,
                                   pol.y = poly.y,
                                   point.x = samp.xy$x,
                                   point.y = samp.xy$y) == 1)]
    # theoretically split_pt_ids are points on straight split_line, but not
    # on polygon perimeter.
    # This solution should be added to .get_ids, when filter.all = FALSE
    #
    # if split_pt_ids is not empty:
    #  remove split_pt_ids  from m_ids
    if (length(split_pt_ids) > 0){
      if (any(split_pt_ids %in% m_ids_up))
        m_ids_up <- m_ids_up[-which(split_pt_ids == m_ids_up)]
      if (any(split_pt_ids %in% m_ids_do))
        m_ids_do <- m_ids_do[-which(split_pt_ids == m_ids_do)]}

  }

  # if m_ids not empty:
  #  inside curvi_split & curvi_split_fast, when using .get_ids in
  #  .curve_quality, additionally add m_ids to the output of ids.
  m_ids <- list(m_ids_up, m_ids_do)
  }
  rot.poli.do <- rot.poli.do[-nrow(rot.poli.do),]
  rot.poli.up <- rot.poli.up[-nrow(rot.poli.up),]
  #pasirupinam splino knot'ais pradiniais

  split.line.x <- seq(0,AE,length.out = c.X.knots)
  split.line.x[c.X.knots] <- AE
  #ant pataisyto poligono reikia rasti Y koordinates duotom x koordinatem,
  # atitinkanciom padalinimo linijos atkarpu galus
  up.y <- numeric(c.X.knots)
  do.y <- numeric(c.X.knots)
  for(i in 2:(length(split.line.x)-1)){
    up.y[i]<-.y.online(x=upper.inner.poli[[1]]$x,y=upper.inner.poli[[1]]$y,
                       x3=split.line.x[i])
    do.y[i]<-.y.online(x=bottom.inner.poli[[1]]$x,y=bottom.inner.poli[[1]]$y,
                       x3=split.line.x[i])
  }
  #randam ploti pataisyto poligono
  range <- up.y-do.y
  #nustatom i kiek daliu dalinsim polygona pagal Y koordinate
  B <- seq(0,1,length.out = c.Y.knots)
  #sugeneruojam matrica su testuojamu spline knotu y koordinatem.
  #Kiekvienas stulpelis atitinka skirtinga X verte ant padalinimo linijos
  #spit.line.x, kiekviena eilute atitinka skirtinga poligono pjuvio ties x
  # koordinate dali - virsutines eilutes apatine poligono dalis
  knot.y.matrix <- matrix(B, c.Y.knots, 1) %*% matrix(range, 1, c.X.knots) +
    matrix(rep(do.y, each = c.Y.knots), c.Y.knots, c.X.knots)
  # if (testid == 17){
  #   trace.object <- "curve"
  #   trace.level <- "main"
  # }
  .visualise_splits.curve_start(what = trace.object,
                                rot.dat.cords, pnts.col, rot.poli, AE, c.X.knots,
                                split.line.x, c.Y.knots, knot.y.matrix)

  #randam geriausia padalinimo kreive
  if (c.fast.optim){
  environment(.curvi_split.fast) <- environment()

  best_curvi_split <- .curvi_split.fast(
    knot.y.matrix,split.line.x,Xup=upper.inner.poli[[1]]$x,
    Xdown=bottom.inner.poli[[1]]$x,Yup=upper.inner.poli[[1]]$y,
    Ydown=bottom.inner.poli[[1]]$y,
    N.crit,
    N.rel.crit,N.loc.crit,N.loc.rel.crit,S.cond,
    S.rel.crit,S_org,c.Y.knots,c.X.knots,rot.poli.up,rot.poli.do,
    rot.dat.cords,c.max.iter.no = c.max.iter.no, c.corr.term = c.corr.term,
    trace.object = trace.object, trace.level = trace.level,
    straight.qual = straight.qual,samp.dat = samp.dat,
    m_ids = m_ids
    )
  } else {
    environment(.curvi_split) <- environment()
    best_curvi_split <- .curvi_split(
      knot.y.matrix,split.line.x,Xup=upper.inner.poli[[1]]$x,
      Xdown=bottom.inner.poli[[1]]$x,Yup=upper.inner.poli[[1]]$y,
      Ydown=bottom.inner.poli[[1]]$y,
      N.crit,
      N.rel.crit,N.loc.crit,N.loc.rel.crit,S.cond,
      S.rel.crit,S_org,c.Y.knots,c.X.knots,rot.poli.up,rot.poli.do,
      rot.dat.cords,c.max.iter.no = c.max.iter.no, c.corr.term = c.corr.term,
      trace.object = trace.object, trace.level = trace.level,
      straight.qual = straight.qual,samp.dat = samp.dat,
      m_ids = m_ids
    )
  }

# Up --> Yup, Down --> Ydown; xp --> x; yp --> y.  visur pakeist
  best.curve <- data.frame(x=best_curvi_split[[1]]$x+poly.x[min.x.id],
                           y=best_curvi_split[[1]]$y+poly.y[min.x.id])
  best.curve <- DescTools::Rotate(x=best.curve$x,y=best.curve$y,
                                  mx=poly.x[min.x.id], my=poly.y[min.x.id],
                                  theta=teta)
  return(list(best.curve, best_curvi_split[[2]]))
}

#' Find the curve that divides the polygon best (recursive function)
#'
#' @description At this stage of an algorithm, polygon is rotated so that the split line is horizontal and positioned along X axis (at Y = 0), and split line starts at X = 0.
#' Function iterates trough knot matrix several times (maximum allowed number of times is defined by c.max.iter.no argument) and tries different splines to seperate data in space.
#' Iteration through knots starts from left to right along the split line. For each position along the split line several positions orthogonal to the split line are tried.
#' The best orthogonal positions are estimated from a spline model (Split Quality ~ Y coordinate of a knot for a given X coordinate)
#' So the best Y knot coordinate is provided for each knot along X. Thus, knots along the split line are fixated, but along Y are not. When one iteration is complete, the next
#' one starts in different direction (e.g. from right to left along the split line).
#'
#' @param knot.y.matrix a matrix with Y coordinates in columns for each knot along the split line
#' @param split.line.x a vector of x coordinates for each knot along the split line
#' @param Xup X coordinates of upper part (above split line) of standartized polygon
#' @param Xdown X coordinates of lower part (below split line) of standartized polygon
#' @param Yup Y coordinates of upper part (above split line) of standartized polygon
#' @param Ydown Y coordinates of lower part (below split line) of standartized polygon
#' @param N.crit minimum number of observations required to establish
#' subdivision of a polygon.
#' @param N.rel.crit number from 0 to 0.5. The minimum allowed observation
#' proportion in polygon containing less data.
#' @param N.loc.crit Subdivision stopping criteria - number of unique locations.
#' Only meaningful when there are observations from the same location. Default
#' 0.2.
#' @param N.loc.rel.crit number from 0 to 0.5. The minimum allowed unique
#' locations proportions in polygon containing less locations.
#' @param S.cond minimum area required to establish subdivision of a plot
#' @param S.rel.crit number from 0 to 0.5. The minimum allowed area
#' proportion of a smaller polygon. Default 0.2.
#' @param S_org area of the polygon being divided.
#' @param c.Y.knots number of spline knots orthogonal to the split line
#' @param c.X.knots number of spline knots along the split line
#' @param rot.poli.up rotated, but not standartized, open upper polygon
#' @param rot.poli.do rotated, but not standartized, open lower polygon
#' @param rot.dat.cords rotated coordinates of data
#' @param c.max.iter.no allowed number of iterations trough columns of spline
#' knots
#' @param c.corr.term term that defines how much the a problematic spline
#' will be corrected (in terms of proportion of polygon width where spline
#' intersects the polygon boundary) if the spline is not contained within the
#' plot. Small values recommended (default is 0.05).
#' @param trace.level Sting indicating the trace level. Can be one of the
#' following: "best", "main", "all"
#' @param trace.object String that informs what graphical information to show.
#' Can be either "curves", "straight" or "both".
#' @param straight.qual Quality of the best straight split-line.
#' @param samp.dat spatial subset of data
#' @param m_ids list of data point IDs from up and bottom polygon, respectively,
#' that lie on the perimeter of polygon, but are not correctly filtered with
#' .get_ids and therefore should be added manually to the output of .get_ids.
#' @return A list of two elements: 1) rotated (not suitable for the original polygon) curve in shape of a spline that produces the best data separation; 2) quality of the division
#' @importFrom stats spline
#' @author Liudas Daumantas
#' @noRd
.curvi_split.fast<-function(knot.y.matrix,split.line.x,Xup,Xdown,Yup,Ydown,
                            N.crit,N.rel.crit,N.loc.crit,N.loc.rel.crit,S.cond,
                            S.rel.crit,S_org,c.Y.knots,c.X.knots,rot.poli.up,
                            rot.poli.do,rot.dat.cords,c.max.iter.no = c.max.iter.no,
                            c.corr.term,trace.object = trace.object,
                            trace.level = trace.level, teta = teta, straight.qual,
                            samp.dat, m_ids){
  best.y.knots <- numeric(c.X.knots)
  y.cord.quality <- numeric(c.Y.knots-2)
  any.qual <- numeric()
  best.y.coords <- numeric(c.X.knots-2)
  best.qual <- straight.qual
  best.old.curve <- data.frame(x = numeric(0), y = numeric(0))
  best.y <- 0
  knot.y.matrix <- knot.y.matrix[-c(1,c.Y.knots),-c(1,c.X.knots)]
  knot.state <- !logical(c.X.knots-2)
  counter <- 0
  it <- 1
  SS <- list(NaN,NaN)

  while(it <= c.max.iter.no) {
    if (all(!knot.state)) break
    if(it%%2==1){
      if(it==1){
        seq.of.x.knots <- (1:(c.X.knots-2))[knot.state]
      } else {
        seq.of.x.knots <- (2:(c.X.knots-2))[knot.state[-1]]
      }} else {
        seq.of.x.knots <- ((c.X.knots-3):1)[rev(knot.state)[-1]]
      }
    for (l in seq.of.x.knots) {

      for (k in 1:(c.Y.knots-2)) {
        #isbandom vis kita Y koordinate knotui su duota x koordinate
        best.y.knots[1+l] <- knot.y.matrix[k,l]
        #sugeneruojam splina per knotus
        curve <- stats::spline(split.line.x,best.y.knots,n=100)
        #iteratyviai darom, kol nebemeta klaidos:
        #patikrinam ar poligono viduj kreive, gaunam pataisos tasku koordinates
        #skeliam split.line.x ir best.y.knots i tiek daliu, kiek reikia ir
        # reikiamose vietose pridedam koordinates
        #sukuriam nauja split.line.x best.y.knots varianta
        #kartojam splina
        curve <- .spline_corrections(curve,Xup,Xdown,Yup,Ydown,best.y.knots,
                                     split.line.x,c.corr.term=c.corr.term)
        counter <- counter + 1
        if (!is.null(curve)){
          .visualise_splits.try_curve(what = trace.object,level = trace.level,
                                      counter, curve)

          #ivertinam poligono padalinimo su sugeneruota kreive kokybe
          environment(.curve_quality) <- environment()
          SS <- .curve_quality(curve=curve,rot.poli.up, rot.poli.do,
                               rot.dat.cords,samp.dat,N.crit,
                               N.rel.crit,N.loc.crit,N.loc.rel.crit,S.cond,
                               S.rel.crit,S_org, m_ids)
        } else {
          SS[[1]] <- NaN
          SS[[2]] <- "Too complex geometry"
        }

        if ( SS[[2]] != "" ){
          y.cord.quality[k] <- NaN
          message <- SS[[2]]
          .visualise_splits.bad_curve(what = trace.object,level = trace.level,
                                      curve, message)
        } else{
          y.cord.quality[k] <- SS[[1]]
          any.qual <- c(any.qual,SS[[1]])
          if (.comp(SS[[1]],best.qual)){
            .visualise_splits.good_curve(what = trace.object,
                                         level = trace.level,
                                         curve, SS, best.old.curve)
            best.old.curve <- curve
            best.qual <- SS[[1]]
            .visualise_splits.selected_knot(what = trace.object,
                                            level = trace.level,
                                            x = split.line.x[l+1],
                                            y = knot.y.matrix[k,l],
                                            old.knot.y = best.y.coords[l])
            best.y.coords[l] <- knot.y.matrix[k,l]
            knot.state <- rep(TRUE, c.X.knots-2)
          } else {
            message <- paste0("Poor split quality.\n","Obtained: ",
                              round(SS[[1]],2),
                              "\nRequired: ",c.sign, round(best.qual,2))
            .visualise_splits.bad_curve(what = trace.object,level = trace.level,
                                        curve, message)
          }
        }
        #fiksuojam verte
      }
      knot.state[l] <- FALSE
      #sugeneruojam spline, kurio X asyje yra knotu Y koordinates, o Y asyje
      #padalinimo kreives kokybe
      #skirta interpoliuojant aproksimuoti tarpiniu Y koordinaciu vertes
      if (!(sum(!is.nan(y.cord.quality)) <= 3 |
            length(unique(y.cord.quality)) <= 2)){
        proj <- stats::spline(knot.y.matrix[,l], y.cord.quality, n=100)
        if (.comp(.minormax(proj$y), best.qual)){
          test.knots <- best.y.knots
          test.knots[1+l] <- proj$x[.which_minormax(proj$y)]
          curve.test <- stats::spline(split.line.x,test.knots,n=100)
          curve.test <- .spline_corrections(curve.test,Xup,Xdown,Yup,Ydown,test.knots,
                                            split.line.x,c.corr.term = c.corr.term)
          if (!is.null(curve.test)){
            .visualise_splits.try_int_curve(what = trace.object,
                                            level = trace.level,
                                            curve = curve.test,
                                            x = split.line.x[l+1],
                                            y = test.knots[1+l])
            environment(.curve_quality) <- environment()
            SS <- .curve_quality(curve.test, rot.poli.up, rot.poli.do,
                                 rot.dat.cords, samp.dat, N.crit,
                                 N.rel.crit,N.loc.crit,N.loc.rel.crit,S.cond,
                                 S.rel.crit,S_org, m_ids)
          } else {
            SS[[2]] <- "Too complex geometry"
            SS[[1]] <- NaN
          }
          if (SS[[2]] == ""){
            if (.comp(SS[[1]], best.qual)){
              .visualise_splits.good_curve(what = trace.object,
                                           level = trace.level,
                                           curve.test, SS,best.old.curve)
              best.old.curve <- curve.test
              .visualise_splits.interpol_knot(what = trace.object,
                                              level = trace.level,
                                              x = split.line.x[l+1],
                                              y = proj$x[.which_minormax(proj$y)],
                                              old.knot.y = best.y.coords[l])
              best.y.coords[l] <- test.knots[1+l]
              knot.state <- rep(TRUE, c.X.knots-2)
              knot.state[l] <- FALSE
              best.qual <- SS[[1]]
              any.qual <- c(any.qual,SS[[1]])
              #issaugom, kaip geriausia Y koordinate X koordinates knotui ta, kuri turi
              # didziausia kokybe
              best.y.knots[1+l] <- proj$x[.which_minormax(proj$y)]
            } else {
              SS[[2]] <- paste0("Poor split quality.\n","Obtained: ",
                                round(SS[[1]],2),
                                "\nRequired: ",c.sign, round(best.qual,2))
            }
          }
          if (SS[[2]] != "")
            .visualise_splits.bad_int_curve(what = trace.object,
                                            level = trace.level,
                                            curve = curve.test,
                                            x = split.line.x[l+1],
                                            y = test.knots[1+l],
                                            SS[[2]])
        }
      }
      best.y.knots[1+l] <- best.y.coords[l]
    }
    it <- it + 1
  }
  #siame etape kiekvienam X koordinates knotui turime po geriausia Y koordinate
  #sugeneruojam per siuos knotus kreive
  if (all(best.y.knots == 0 )) {
    .visualise_splits.no_best_curve(what = trace.object)
    return(list(data.frame(x = c(0,split.line.x[c.X.knots+2]),
                           y = c(0,0)),straight.qual))
  } else {
    curve.final <- stats::spline(split.line.x, best.y.knots, n=100)
    #kreive gali islisti is uz polygono (pvz. kokybes dideli iverciai buvo
    # gauti pataisius duotus knotus)
    #taigi, kreive dar karta bandoma pataisyti, jei reikia
    curve.final <- .spline_corrections(curve.final,Xup,Xdown,Yup,Ydown,best.y.knots,
                                       split.line.x,c.corr.term = c.corr.term)
    #ivertinam galutines padalinimo kreives kokybe
    environment(.curve_quality) <- environment()
    SS <- .curve_quality(curve.final, rot.poli.up, rot.poli.do, rot.dat.cords,
                         samp.dat, N.crit,
                         N.rel.crit,N.loc.crit,N.loc.rel.crit,S.cond,
                         S.rel.crit, S_org, m_ids)
    .visualise_splits.best_curve(what = trace.object,
                                 curve.final, SS)
    #grazinam padalinimo kreive, jos kokybes iverti ir visu kitu lokaliai
    # geriausiu kreiviu ivercius - SSk
    #SSk tik tam, kad pasizeti, ar tikrai grizta pati geriausia kreive
    if(length(any.qual)>0){
      if (any(.comp(any.qual,SS[[1]]))){
        warning(
          "Intermediate curves performed better than the final curve.",
          call. = FALSE
        )
      }
    }
    return(list(curve.final,SS[[1]]))
  }
}


.curvi_split<-function(knot.y.matrix,split.line.x,Xup,Xdown,Yup,Ydown,N.crit,
                       N.rel.crit,N.loc.crit,N.loc.rel.crit,S.cond,
                       S.rel.crit,S_org,c.Y.knots,c.X.knots,rot.poli.up,
                       rot.poli.do,rot.dat.cords,c.max.iter.no = c.max.iter.no,
                       c.corr.term,trace.object = trace.object,
                       trace.level = trace.level, teta = teta,
                       straight.qual,samp.dat,m_ids){
  best.y.knots <- numeric(c.X.knots)
  y.cord.quality <- rep(NA,c.Y.knots-2)
  any.qual <- numeric()
  best.y.coords <- numeric(c.X.knots-2)
  best.y.coord <- 0
  best.col.id <- 0
  old.col.id <- 0
  SS <- list(NaN,NaN)
  best.qual <- straight.qual
  best.old.curve <- data.frame(x=numeric(0),y=numeric(0))
  knot.y.matrix <- knot.y.matrix[-c(1,c.Y.knots),-c(1,c.X.knots)]
  counter <- 0
  valid.ids <- numeric()
  it <- 1
  x.seq <- (1:(c.X.knots-2))
  while(it <= c.max.iter.no) {
    if (it != 1){
      if (best.col.id == old.col.id){
        break
      } else {
        .visualise_splits.selected_knot(what = trace.object,
                                        level = trace.level,
                                        x = split.line.x[best.col.id+1],
                                        y = best.y.coord,
                                        old.knot.y = best.y.coords[best.col.id])
        best.y.coords[best.col.id] <-  best.y.coord
        best.y.knots <- c(0,best.y.coords,0)
        old.col.id <- best.col.id
        x.seq <- (1:(c.X.knots-2))[-best.col.id]
      }
    }
    for (l in x.seq) {

      for (k in 1:(c.Y.knots-2)) {

        #isbandom vis kita Y koordinate knotui su duota x koordinate
        best.y.knots[1+l] <- knot.y.matrix[k,l]
        #sugeneruojam splina per knotus
        curve <- stats::spline(split.line.x,best.y.knots,n=100)
        #iteratyviai darom, kol nebemeta klaidos:
        #patikrinam ar poligono viduj kreive, gaunam pataisos tasku koordinates
        #skeliam split.line.x ir best.y.knots i tiek daliu, kiek reikia ir
        # reikiamose vietose pridedam koordinates
        #sukuriam nauja split.line.x best.y.knots varianta
        #kartojam splina
        # if (testid == 1 & counter == 55){
        #   browser()
        #   trace.level = "all"; trace.object = "curve"}
        curve <- .spline_corrections(curve,Xup,Xdown,Yup,Ydown,best.y.knots,
                                     split.line.x,c.corr.term=c.corr.term)
        counter <- counter + 1

        if (!is.null(curve)){
          .visualise_splits.try_curve(what = trace.object,level = trace.level,
                                      counter, curve)


          #ivertinam poligono padalinimo su sugeneruota kreive kokybe
          environment(.curve_quality) <- environment()
          SS <- .curve_quality(curve=curve,rot.poli.up, rot.poli.do,
                               rot.dat.cords, samp.dat, N.crit,
                               N.rel.crit,N.loc.crit,N.loc.rel.crit,S.cond,
                               S.rel.crit,S_org,m_ids)
        } else {
          SS[[1]] <- NaN
          SS[[2]] <- "Too complex geometry"
        }



        if ( SS[[2]] != "" ){
          y.cord.quality[k] <- NaN
          message <- SS[[2]]
          .visualise_splits.bad_curve(what = trace.object,level = trace.level,
                                      curve, message)
        } else {
          #fiksuojam verte
          y.cord.quality[k] <- SS[[1]]
          any.qual <- c(any.qual,SS[[1]])
          if (.comp(SS[[1]], best.qual)){

            .visualise_splits.good_curve(what = trace.object,
                                         level = trace.level,
                                         curve, SS, best.old.curve)
            best.old.curve <- curve
            best.qual <- SS[[1]]
            best.y.coord <- knot.y.matrix[k,l]
            best.col.id <- l
            #.visualise_splits.selected_knot(what = trace.object,
            #         level = trace.level,
            #          x = split.line.x[l+1],
            #           y = knot.y.matrix[k,l],
            #            old.knot.y = best.y.coords[l])
            #best.y.coords[l] <- knot.y.matrix[k,l]
          } else {
            message <- paste0("Poor split quality.\n","Obtained: ",
                              round(SS[[1]],2),
                              "\nRequired: ",c.sign, round(best.qual,2))
            .visualise_splits.bad_curve(what = trace.object,level = trace.level,
                                        curve, message)
          }
        }
      }
      #sugeneruojam spline, kurio X asyje yra knotu Y koordinates, o Y asyje
      #padalinimo kreives kokybe
      #skirta interpoliuojant aproksimuoti tarpiniu Y koordinaciu vertes
      if (!(sum(!is.nan(y.cord.quality)) <= 3 |
            length(unique(y.cord.quality)) <= 2)){
        proj <- stats::spline(knot.y.matrix[,l], y.cord.quality, n=100)
        if (.comp(.minormax(proj$y), best.qual)){
          test.knots <- best.y.knots
          test.knots[1+l] <- proj$x[.which_minormax(proj$y)]
          curve.test <- stats::spline(split.line.x,test.knots,n=100)
          curve.test<-.spline_corrections(curve.test,Xup,Xdown,Yup,Ydown,test.knots,
                                          split.line.x,c.corr.term = c.corr.term)
          if (!is.null(curve.test)){
            .visualise_splits.try_int_curve(what = trace.object,
                                            level = trace.level,
                                            curve = curve.test,
                                            x = split.line.x[l+1],
                                            y = test.knots[1+l])
            environment(.curve_quality) <- environment()
            SS <- .curve_quality(curve.test,rot.poli.up, rot.poli.do, rot.dat.cords,
                                 samp.dat,N.crit,
                                 N.rel.crit,N.loc.crit,N.loc.rel.crit,S.cond,
                                 S.rel.crit,S_org, m_ids)
          } else {
            SS[[2]] <- "Too complex geometry"
            SS[[1]] <- NaN
          }
          if (SS[[2]] == ""){
            if (.comp(SS[[1]], best.qual)){
              .visualise_splits.good_int_curve(what = trace.object,
                                               level = trace.level,
                                               curve = curve.test,
                                               SS,
                                               best.old.curve,
                                               x = split.line.x[l+1],
                                               y = test.knots[1+l])
              best.old.curve <- curve.test
              any.qual <- c(any.qual,SS[[1]])
              #.visualise_splits.interpol_knot(what = trace.object,
              # level = trace.level,
              # x = split.line.x[l+1],
              # y = proj$x[which.max(proj$y)],
              # old.knot.y = best.y.coords[l])
              #best.y.coords[l] <- proj$x[which.max(proj$y)]
              best.y.coord <- test.knots[1+l]
              best.col.id <- l
              best.qual <- SS[[1]]
            } else {
              SS[[2]] <- paste0("Poor split quality.\n","Obtained: ",
                                round(SS[[1]],2),
                                "\nRequired: ",c.sign, round(best.qual,2))
            }
            #issaugom, kaip geriausia Y koordinate X koordinates knotui ta, kuri turi
            # didziausia kokybe
            #best.y.knots[1+l] <- proj$x[which.max(proj$y)]
          }
          if (SS[[2]] != "")
            .visualise_splits.bad_int_curve(what = trace.object,
                                            level = trace.level,
                                            curve = curve.test,
                                            x = split.line.x[l+1],
                                            y = test.knots[1+l],
                                            SS[[2]])
        } else {
          # best.y.knots[1+l] <- best.y.coords[l]
        }
      }
      best.y.knots <- c(0,best.y.coords,0)
    }

    it <- it + 1
  }
  if (it > c.max.iter.no & best.col.id != old.col.id ){
    .visualise_splits.selected_knot(what = trace.object,
                                    level = trace.level,
                                    x = split.line.x[best.col.id+1],
                                    y = best.y.coord,
                                    old.knot.y = best.y.coords[best.col.id])
    best.y.coords[best.col.id] <-  best.y.coord
    best.y.knots <- c(0,best.y.coords,0)
  }
  #siame etape kiekvienam X koordinates knotui turime po geriausia Y koordinate
  #sugeneruojam per siuos knotus kreive
  if (all(best.y.knots == 0 )) {
    .visualise_splits.no_best_curve(what = trace.object)
    return(list(data.frame(x = c(0,split.line.x[c.X.knots+2]),
                           y = c(0,0)),straight.qual))
  } else {
    curve.final <- stats::spline(split.line.x, best.y.knots, n=100)
    #kreive gali islisti is uz polygono (pvz. kokybes dideli iverciai buvo
    # gauti pataisius duotus knotus)
    #taigi, kreive dar karta bandoma pataisyti, jei reikia
    curve.final <- .spline_corrections(curve.final,Xup,Xdown,Yup,Ydown,best.y.knots,
                                       split.line.x,c.corr.term = c.corr.term)
    #ivertinam galutines padalinimo kreives kokybe
    environment(.curve_quality) <- environment()
    SS <- .curve_quality(curve.final, rot.poli.up, rot.poli.do, rot.dat.cords,
                         samp.dat, N.crit,
                         N.rel.crit,N.loc.crit,N.loc.rel.crit,S.cond,
                         S.rel.crit, S_org, m_ids)
    .visualise_splits.best_curve(what = trace.object,
                                 curve.final, SS)
    #grazinam padalinimo kreive, jos kokybes iverti ir visu kitu lokaliai
    # geriausiu kreiviu ivercius - SSk
    #SSk tik tam, kad pasizeti, ar tikrai grizta pati geriausia kreive
    if(length(any.qual)>0){
      if (any(.comp(any.qual,SS[[1]]))){
        warning(
          "Intermediate curves performed better than the final curve.",
          call. = FALSE
        )
      }
    }
    return(list(curve.final,SS[[1]]))
  }
}

#' Evaluate the quality of curve in terms of data spatial separation
#'
#' @description function forms two polygons from a provided curve (upper and lower),
#' filters data using each polygon and compares them.
#'
#' @param curve a matrix with Y coordinates in columns for each knot along the split line
#' @param rot.poli.up rotated, but not standartized, open upper polygon
#' @param rot.poli.do rotated, but not standartized, open lower polygon
#' @param rot.dat.cords rotated coordinates of data that is analyzed
#' @param samp.dat spatial subset of data
#' @param N.crit minimum number of observations required to establish
#' subdivision of a polygon.
#' @param N.rel.crit number from 0 to 0.5. The minimum allowed observation
#' proportion in polygon containing less data.
#' @param N.loc.crit Subdivision stopping criteria - number of unique locations.
#' Only meaningful when there are observations from the same location. Default
#' 0.2.
#' @param N.loc.rel.crit number from 0 to 0.5. The minimum allowed unique
#' locations proportions in polygon containing less locations.
#' @param S.cond minimum area required to establish subdivision of a plot
#' @param S.rel.crit number from 0 to 0.5. The minimum allowed area
#' proportion of a smaller polygon. Default 0.2.
#' @param S_org area of the polygon being divided.
#' @param m_ids list of data point IDs from up and bottom polygon, respectively,
#' that lie on the perimeter of polygon, but are not correctly filtered with
#' .get_ids and therefore should be added manually to the output of .get_ids.
#' @return A list of two elements: 1) rotated (not suitable for the original polygon) curve in shape of a spline that produces the best data separation; 2) quality of the division
#' @author Liudas Daumantas
#' @importFrom pracma polyarea
#' @noRd
.curve_quality<-function(curve,rot.poli.up,rot.poli.do,rot.dat.cords,
                         samp.dat,N.crit,
                         N.rel.crit,N.loc.crit,N.loc.rel.crit,S.cond,
                         S.rel.crit,S_org, m_ids){
  #sudarom poligonus padalintus kreive
  I.poli <- data.frame(x = c(rot.poli.up$x, rev(curve$x)[-1]),
                       y = c(rot.poli.up$y, rev(curve$y)[-1]))
  II.poli <- data.frame(x = c(rot.poli.do$x, rev(curve$x)[-1]),
                        y = c(rot.poli.do$y, rev(curve$y)[-1]))
  #atrenkam taskus patenkancius i skirtingas poligono dalis, padalintas kreive
  id1 <- c(.get_ids(I.poli, rot.dat.cords), m_ids[[1]])
  id2 <- c(.get_ids(II.poli,rot.dat.cords), m_ids[[2]])
  if (any(duplicated(id1)) | any(duplicated(id2))){
    warning(
      "Duplicated filtered points in curvilinear splits.",
      call. = FALSE
    )
  }
  #atliekam plotu palyginima, t.y. padalinimo gerumo ivertinima, jei kreive
  #netenkina minimaliu kriteriju - padalinimo kokybe nuline
  message <- ""
  if (N.crit > 0){
    good <- length(id1) > N.crit & length(id2) > N.crit
    if (!good){
      return(list(NA, paste0("Not enough observations.",
                             "\nObtained: ", length(id1), ' and ', length(id2),
                             "\nRequired: >", N.crit) ))
    }
  }

  if (N.rel.crit > 0){
    good <- length(id1) / nrow(rot.dat.cords) > N.rel.crit &
      length(id2) / nrow(rot.dat.cords) > N.rel.crit
    if (!good){
      return(list(NA,paste0("Too low proportion of observations.\nObtained: ",
                            round(length(id1) / nrow(rot.dat.cords), 2),
                            ' and ', round(length(id2) / nrow(rot.dat.cords),
                                           2), "\nRequired: >", N.rel.crit) ))
    }
  }

  if (N.loc.crit > 0 | N.loc.rel.crit > 0){
    n.uni.1 <- nrow(unique(rot.dat.cords[id1,]))
    n.uni.2 <- nrow(unique(rot.dat.cords[id2,]))
    if (N.loc.crit > 0){
      good <- n.uni.1 > N.loc.crit &
        n.uni.2 > N.loc.crit
      if (!good){
        return(list(NA, paste0("Not enough locations.",
                               "\nObtained: ", n.uni.1,
                               ' and ', n.uni.2,
                               "\nRequired: >", N.loc.crit) ))
      }
    }
    if (N.loc.rel.crit > 0){
      n.uni <- nrow(unique(rot.dat.cords))
      good <- n.uni.1 / n.uni > N.loc.rel.crit &
        n.uni.2 / n.uni > N.loc.rel.crit
      if (!good){
        return(list(NA, paste0("Too low proportion of locations.",
                               "\nObtained: ", round(n.uni.1 / n.uni,2),
                               ' and ', round(n.uni.2 / n.uni, 2),
                               "\nRequired: >", N.loc.rel.crit) ))
      }
    }
  }

  if (S.crit > 0 | S.rel.crit > 0){
    S1 <- abs(pracma::polyarea(I.poli$x,I.poli$y))
    S2 <- abs(pracma::polyarea(II.poli$x,II.poli$y))
    if (S.crit > 0) {
      good <- S1 > S.cond & S2 > S.cond
      if (!good){
        return(list(NA, paste0("One of the areas was too small.\n","Obtained: ",
                               round(S1,2), ' and ', round(S2,2),
                               "\nRequired: >", S.cond) ))
      }
    }
    if (S.rel.crit > 0) {
      good <- S1 / S_org  > S.rel.crit &
        S2 / S_org > S.rel.crit
      if (!good){
        message <- paste0("Too low proportion of area.\n","Obtained: ",
                          round(S1 / S_org,2), ' and ',
                          round(S2 / S_org, 2),
                          "\nRequired: >", S.rel.crit)
        return(list(NA, paste0("Too low proportion of area.\n","Obtained: ",
                               round(S1 / S_org,2), ' and ',
                               round(S2 / S_org, 2),
                               "\nRequired: >", S.rel.crit) ))
      }
    }
  }

  environment(.dif_fun) <- environment()
  SS <- .dif_fun(.slicer(samp.dat,id1),.slicer(samp.dat,id2))
  list(SS, message)
}

#' Correct the curve
#'
#' @description function corrects the curve if it is not contained within the polygon.
#' It does so by finding intervals of a curve that cross the boundaries of a polygon,
#' then it adds additional knots at the middle of these intervals inside the polygon.
#' Finally, curve is generated again using new knots, but it may still have some outlying intervals.
#' Thus, the procedure is repeated recursively until no outlying intervals are left.
#' @param curve a curve (output of spline function from stats package)
#' @param Xup X coordinates of upper part (above split line) of standartized polygon
#' @param Xdown X coordinates of lower part (below split line) of standartized polygon
#' @param Yup Y coordinates of upper part (above split line) of standartized polygon
#' @param Ydown Y coordinates of lower part (below split line) of standartized polygon
#' @param best.y.knots a vector of Y coordinates of knots of a spline that were used to produce the curve.
#' Each Y coordinate is dedicated for corresponding elements of split.line.x, that is X coordinates of spline knots.
#' @param split.line.x a vector of x coordinates for each knot along the split line
#' @param c.corr.term term that defines how much the a problematic spline
#' will be corrected (in terms of proportion of polygon width where spline
#' intersects the polygon boundary) if the spline is not contained within the
#' plot. Small values recommended (default is 0.05).
#' @param count. number of recursions.
#' @return a curve (output of a spline function from stats package)
#' @author Liudas Daumantas
#' @noRd
.spline_corrections<-function(curve,Xup,Xdown,Yup,Ydown,best.y.knots,
                             split.line.x,c.corr.term,count. = 1){

  corrected.coords<-.y_corrections(spline.x = curve$x,spline.y =curve$y ,
                                  Xup = Xup,Xdown = Xdown,Yup = Yup,
                                  Ydown = Ydown,
                                  c.corr.term = c.corr.term)
  if (length(dim(corrected.coords)) > 1){
    if (length(
      apply(corrected.coords, 1, function(xy)
        which(xy[1] == split.line.x & xy[2] == best.y.knots)))
      > 0 | count. == 25)
      return(NULL)
    count. <- count. +1
    ind<-numeric()
    k<-0
    corrected.split.line.x<-split.line.x
    corrected.best.y.knots<-best.y.knots
    for (a in seq(length(corrected.coords[,1]))){
      for (i in seq(length(split.line.x))){
        if(split.line.x[i]>=corrected.coords[a,1]){
          ind<-i-1
          corrected.split.line.x<-c(corrected.split.line.x[c(1:(ind+k))],
                                    corrected.coords[a,1],
                                    corrected.split.line.x[
                                      c((ind+k+1):
                                          length(corrected.split.line.x))
                                    ])
          corrected.best.y.knots<-c(corrected.best.y.knots[c(1:(ind+k))],
                                    corrected.coords[a,2],
                                    corrected.best.y.knots[
                                      c((ind+k+1):
                                          length(corrected.best.y.knots))
                                    ])
          k<-k+1
          break
        }
      }
    }
    curve <- tryCatch(stats::spline(corrected.split.line.x,corrected.best.y.knots,n=100),
                    warning = function(cond) NULL)
    if (is.null(curve)){
      return(curve)
    }

    best.y.knots<-corrected.best.y.knots
    split.line.x<-corrected.split.line.x
    return(.spline_corrections(curve =curve ,Xup =Xup ,Xdown =Xdown ,Yup=Yup,
                              Ydown = Ydown,best.y.knots=best.y.knots,
                              split.line.x = split.line.x,
                              c.corr.term = c.corr.term, count. = count.) )
  } else {
    return(curve)
  }
}



#' Find coordinates for new knots
#'
#' @description function checks if there are curve intervals that cross the boundary of a polygon.
#' If present, for each interval it calculates the middle X coordinate.
#' Then, Y coordinates are produced for each X coordinate in a way so that each produced Y are within the polygon.
#' The Y coordinates are calculated as follows: 1) width of the polygon is calculated
#' at X coordinate (distance in Y coord. between upper and lower boundaries of the
#' standartized polygon at X coord.); 2) the c.corr.term is then multiplied by the calculated width
#' to get the correction size in Y units; 3) correction is then added to the Y of lower polygon boundary
#' if curve crosses the boundary there, or is subtracted from the upper boundary Y if else.
#'
#' @param spline.x x coordinates of a curve
#' @param spline.y y coordinates of a curve
#' @param Xup X coordinates of upper part (above split line) of standartized polygon
#' @param Xdown X coordinates of lower part (below split line) of standartized polygon
#' @param Yup Y coordinates of upper part (above split line) of standartized polygon
#' @param Ydown Y coordinates of lower part (below split line) of standartized polygon
#' @param best.y.knots a vector of Y coordinates of knots of a spline that were used to produce the curve.
#' Each Y coordinate is dedicated for corresponding elements of split.line.x, that is X coordinates of spline knots.
#' @param c.corr.term term that defines how much the a problematic spline
#' will be corrected (in terms of proportion of polygon width where spline
#' intersects the polygon boundary) if the spline is not contained within the
#' plot. Small values recommended (default is 0.05).
#' @return If there are curve intervals that cross the boundary of a polygon,
#' then function returns a data frame of x and y coordinates that should be combined with coordinates of other knots used to produce the curve.
#' If no such intervals are present function returns "FALSE".
#' @author Liudas Daumantas
#' @noRd
.y_corrections<-function(spline.x,spline.y,Xup,Xdown,Yup,Ydown,
                        c.corr.term){
  #tikrinam, ar kreive polygone, nuimam po viena taska krastini, nes jie ant poligono ribos
  curve.in.polygon <- .point.in.polygon(
    point.x = spline.x,
    point.y = spline.y,
    pol.x = c(Xup,rev(Xdown)),
    pol.y = c(Yup,rev(Ydown))
    )[-c(1,length(spline.x))]

  #jei kreive ne polygone
  #v2 jei neatsivelgti,kuris taskas labiausiai nutoles nuo poligono
  if (any(curve.in.polygon!=1)) {
    #pasiziurim, kurie kreives taskai nepatenka i poligona
    ind<-which(curve.in.polygon!=1)
    #patikrinam, kiek yra ind verciu, jei viena, reiska, tik vienas taskas iseina is poligono, isisaugom taska
    if (length(ind)==1){
      vid.x<-spline.x[ind+1]
      vid.y<-spline.y[ind+1]
    } else { # JEI NE - gali buti kelios kreives atkarpos, nepatenkancios i poligona
      #surandam centrus intervalu, kurie nepatenka i poligona
      min.id<-ind[1]
      vid.x<-numeric()
      vid.y<-numeric()
      for (i in 1:(length(ind)-1)){
        #jei randam luzi indeksuose, reiskia priejom naujo intervalo pradzia,
        #fiksuojam buvusio intervalo viduri ir atnaujiname intervalo pradzia
        if(ind[i+1]-ind[i]>1){
          x3<-spline.x[min.id+1]+(spline.x[ind[i]+1]-spline.x[min.id+1])/2
          vid.x<-c(vid.x,x3)
          #y3 ant splino
          y3.spline<-.y.online(x=spline.x,y=spline.y,x3=x3)
          vid.y<-c(vid.y,y3.spline)
          #fiksuojam naujo intervalo pradzia
          min.id<-ind[i+1]}
      }
      # jei naujo intervalo neprieita, reiskia yra vienas intervalas - issisaugom jo centra
      if (min.id==ind[1]){
        if(ind[1]==1){
          vid.x<-spline.x[1]+(spline.x[ind[length(ind)]+1]-spline.x[1])/2
        } else {
          vid.x<-spline.x[ind[1]+1]+(spline.x[ind[length(ind)]+1]-spline.x[ind[1]+1])/2
        }
        vid.y<-.y.online(x=spline.x,y=spline.y,x3=vid.x)
      } else {# jei rasta intervalu, tai ufiksuoti ju viduriai, bet paskutinio intervalo vidurys neuzfiksuotas
        x3<-spline.x[min.id+1]+(spline.x[ind[length(ind)]+1]-spline.x[min.id+1])/2
        vid.x<-c(vid.x,x3)
        #y3 ant splino
        y3.spline<-.y.online(x=spline.x,y=spline.y,x3=x3)
        vid.y<-c(vid.y,y3.spline)
      }
    }
    #siame etape turim tasku koordinates, kuriuos reikia "nustumti" i poligona ir pakartoti splina
    #poligonas skaidytas i dvi dalis, zemiau 0 ir virs 0.
    #surandam y3 ant poligono virsaus ir apacios, x3 vietoje
    y3.up<-numeric()
    y3.down<-numeric()
    RANGES<-numeric()
    corrected.y<-numeric()
    for (i in 1:length(vid.x)){
      # are there any vertical lines in standartized polygon?
      # if so, assume that y coords of more distant (from split line)
      # vertical vertices that are the same as of vertices closer to the line.
      # crude solution...
      Yup <- .verts.y(Xup,Yup)
      Ydown <- .verts.y(Xdown,Ydown)
      y3.up<-c(y3.up,.y.online(x=Xup,y=Yup,x3=vid.x[i]))
      y3.down<-c(y3.down,.y.online(x3=vid.x[i],x=Xdown,y=Ydown))
      #surandam atstuma tarp y3 up ir down
      RANGES<-c(RANGES,y3.up[i]-y3.down[i])
      #paziurim, per kuria puse kreives atkarpa iseina uz poligono (per virsu ar ne?)
      #pataisom y3 taip, kad jis atsidurtu poligone, 5% atstumu nuo krasto poligono
      if (vid.y[i]>=0){
        corrected.y<-c(corrected.y,y3.up[i]-c.corr.term*RANGES[i])
      } else {
        corrected.y<-c(corrected.y,y3.down[i]+c.corr.term*RANGES[i])
      }
    }
    return(data.frame(vid.x,corrected.y))
  } else {
    return(FALSE)}
}

.verts.y <- function(x,y) {
  verts <- which(x[-1]-x[-length(x)] == 0)
  if (length(verts) > 0 ){
    # if so, assume that y coords of more distant (from split line)
    # vertical vertices that are the same as of vertices closer to the line.
    for (vert in verts){
      min.y.vert <- y[c(vert,vert+1)][which.min(abs(y[c(vert,vert+1)]))]
      y[c(vert,vert+1)] <- min.y.vert
    }
  }
  y
}

#' Print the results of nullhespdiv object
#'
#' @method print nullhespdiv
#' @description  print method for nullhespdiv class object
#' @param x A nullhespdiv class object
#' @param ... other arguments
#' @return x
#' @author Liudas Daumantas
#' @export
print.nullhespdiv <- function(x, ...){
  print(x[[1]])
  invisible(x)
  }

#' Investigate Group Effects on Split-Line Performance
#'
#' Re-evaluates split-line performance within each level of a grouping factor and
#' tests how much each group influences the detected split-lines in a \code{hespdiv} object.
#'
#' @description
#' For every level of \code{group}, the function:
#' \enumerate{
#'   \item subsets \code{data} and \code{xy.dat} to the focal group's observations;
#'   \item recomputes split-line performance using \code{compare.f} and \code{generalize.f};
#'   \item runs two contribution assessments:
#'     \itemize{
#'       \item \emph{group-removal}: removes the group's observations and recomputes performance;
#'       \item \emph{group-permutation (locality-block shuffle)}: shuffles intact within-locality assemblages of the focal group
#'             among the set of localities where the group occurs within the split’s parent polygons (non-group observations fixed).
#'     }
#'   \item optionally re-plots the object for that group via \code{\link{plot_hespdiv}} (with a group-specific subtitle).
#' }
#'
#' @details
#' \strong{Within-group recomputation (agreement).}
#' For each split, the group's observations are partitioned by the two child polygons and
#' performance is recomputed as \code{compare.f(generalize.f(pol1), generalize.f(pol2))}.
#' If the group's points fall on only one side, \code{within$est} is set to \code{maxdif};
#' if absent from both sides, it is \code{NA}.
#'
#' \strong{Elimination test.}
#' All observations of the focal group are removed from both polygons and performance is recomputed on the remaining data.
#' If the group's points fall on only one side, \code{elim$est} is computed normally;
#' if absent from both sides, it is \code{NA} (if desired, you could identify these cases afterwards from \code{n_per_pol}, changing to baseline performance and zero delta).
#'
#' \strong{Permutation test (locality-block shuffle).}
#' Localities are defined by identical coordinate pairs among the focal group's occurrences.
#' For each permutation, whole localities (blocks) are reassigned via a one-to-one mapping within the split’s parent polygons.
#' For speed, polygon membership of unique locality coordinates is precomputed once per split and then locality memberships are shuffled in each permutation.
#' If the group's points fall on only one side, \code{perm$est} is set to NA;
#' if absent from both sides, it is also \code{NA}.
#'
#' \strong{Baseline and deltas.}
#' The baseline vector is \code{baseline <- obj$split.stats$performance}.
#' Let \code{maximize <- obj$call.info$Call_ARGS$maximize}.
#' Deltas are defined so that \emph{positive values always indicate a positive contributor} (removal/permutation worsens performance):
#' \itemize{
#'   \item if \code{maximize == FALSE} (lower is better; e.g. similarity): \code{delta = est - baseline};
#'   \item if \code{maximize == TRUE}  (higher is better; e.g. distance):  \code{delta = baseline - est}.
#' }
#'
#' @param obj   A \code{hespdiv} object.
#' @param group A factor (or coercible to factor) giving the group label for each observation (row) in
#'              \code{obj$call.info$Call_ARGS$xy.dat}.
#' @param perm.n Integer (\eqn{>0}). Number of permutations in the locality-block permutation test. Default \code{999}.
#' @param maxdif Numeric. The performance value representing a maximal between-polygons difference for the chosen metric.
#'               If \code{NULL} and the metric is one of \code{c("pielou","morisita","sorensen","horn.morisita")},
#'               \code{maxdif} is set automatically. Otherwise, it must be provided.
#' @param plot  Logical. If \code{TRUE} (default), stores a \code{\link{plot_hespdiv}} output per group with
#'              \code{performance = TRUE} and a group-specific subtitle.
#' @param ...   Additional arguments passed to \code{\link{plot_hespdiv}}. Note that \code{obj},
#'              \code{performance}, and \code{subtitle} are set internally.
#'
#' @return A list of class \code{"group_effect"} with elements:
#' \itemize{
#'   \item \code{within = list(est, delta)} where \code{est} and \code{delta} are \code{[n_splits x n_groups]} matrices;
#'   \item \code{elim   = list(est, delta)} as above (group removed);
#'   \item \code{perm   = list(est, delta)} where each is a nested list \code{[[split]][[group]]} of numeric vectors (length \code{perm.n}).  If permutations are uninformative (e.g., one-sided/absent), the corresponding entries are \code{NA};
#'   \item \code{baseline}: the original performance vector;
#'   \item \code{n_per_pol}: \code{data.frame} with columns \code{split.id}, \code{group}, \code{pol.id}, \code{n};
#'   \item \code{plots}: list of \code{plot_hespdiv} outputs by group (or \code{NULL} if \code{plot = FALSE}).
#' }
#'
#' @section Notes:
#' \itemize{
#'   \item Localities are defined by exact duplicate coordinate pairs among the group's observations
#'         (harmonise coordinates upstream if needed).
#'   \item The locality-block permutation preserves the \emph{number of group-present localities per polygon};
#'         only identities of the locality blocks are shuffled. The number of group occurrences per side can vary if block sizes differ.
#' }
#'
#' @family functions for hespdiv post-processing
#' @seealso \code{\link{plot_hespdiv}}
#' @importFrom stats setNames
#' @examples
#' data <- example_hespdiv$call.info$Call_ARGS$data
#' # See the hespdiv() example; lowercase letters represent widespread taxa:
#' endemic_l <- !data %in% letters[1:5]
#' group <- factor(ifelse(endemic_l, "endemic", "widespread"))
#'
#' gr_ef <- group_effect(obj = example_hespdiv, group = group, plot = FALSE)
#'
#' barplot(
#'   gr_ef$within$est,
#'   main = "Split-line similarity within each group",
#'   ylab = "Performance"
#' )
#'
#' barplot(
#'   gr_ef$within$delta,
#'   main = "Group contribution to similarity within each group",
#'   ylab = "Contribution"
#' )
#' abline(h = 0)
#'
#' barplot(
#'   gr_ef$elim$est,
#'   main = "Split-line similarity with group eliminated",
#'   ylab = "Performance"
#' )
#'
#' barplot(
#'   gr_ef$elim$delta,
#'   main = "Group contribution to similarity with group eliminated",
#'   ylab = "Contribution"
#' )
#' abline(h = 0)
#'
#' boxplot(
#'   gr_ef$perm$est$`1`,
#'   main = "Split-line similarity with group spatially permuted",
#'   ylab = "Performance"
#' )
#'
#' boxplot(
#'   gr_ef$perm$delta$`1`,
#'   main = "Group contribution to similarity with group spatially permuted",
#'   ylab = "Contribution"
#' )
#' abline(h = 0)
#' @export
group_effect <- function(obj, group, perm.n = 999, maxdif = NULL, plot = TRUE, ...) {

  if (!inherits(obj, "hespdiv")) {
    stop("`obj` must be a `hespdiv` object.", call. = FALSE)
  }

  if (length(group) != nrow(obj$call.info$Call_ARGS$xy.dat)) {
    stop(
      "`group` length must match the number of rows in `xy.dat`.",
      call. = FALSE
    )
  }

  if (length(perm.n) != 1L || !is.finite(perm.n) || perm.n < 1 ||
      perm.n %% 1 != 0) {
    stop("`perm.n` must be a positive integer.", call. = FALSE)
  }
  perm.n <- as.integer(perm.n)

  # metric settings
  if (is.null(maxdif)) {
    supported_metrics <- c("pielou", "morisita", "sorensen", "horn.morisita")
    if (obj$call.info$METHOD$metric %in% supported_metrics) {
      if (obj$call.info$METHOD$metric == "pielou") {maxdif <- 1} else {maxdif <- 0}
    } else stop("Argument `maxdif` must be provided when using a custom method.", call. = FALSE)
  }

  baseline <- obj$split.stats$performance
  maximize <- isTRUE(obj$call.info$Call_ARGS$maximize)

  # factorise group
  group <- as.factor(group)
  group_levels <- levels(group)
  group.n <- length(group_levels)
  n_splits <- length(obj$split.lines)

  # containers
  within_est <- matrix(NA_real_, nrow = n_splits, ncol = group.n,
                       dimnames = list(rownames(obj$split.stats), group_levels))
  elim_est   <- matrix(NA_real_, nrow = n_splits, ncol = group.n,
                       dimnames = list(rownames(obj$split.stats), group_levels))

  perm_est <- create_nested_list(list(rownames(obj$split.stats), group_levels))  # [[split]][[group]] -> numeric vector

  N <- group.n * n_splits * 2
  n_per_pol <- data.frame(
    split.id = integer(N),
    group    = character(N),
    pol.id   = integer(N),
    n        = integer(N),
    stringsAsFactors = FALSE
  )

  plots <- vector("list", length = group.n)
  names(plots) <- group_levels

  # slicer & data
  dat_in_obj <- obj$call.info$Call_ARGS$data
  xy_in_obj  <- obj$call.info$Call_ARGS$xy.dat
  if (is.data.frame(dat_in_obj) || is.matrix(dat_in_obj)) {
    .slicer <- .slicer.table
  } else if (is.list(dat_in_obj)) {
    .slicer <- .slicer.list
  } else {
    .slicer <- .slicer.vect
  }

  for (group_id in seq_len(group.n)) {
    gr <- group_levels[group_id]
    idx_group <- which(group == gr)

    data_sub <- .slicer(dat_in_obj, idx_group)
    xy_sub   <- xy_in_obj[idx_group, , drop = FALSE]

    for (split.id in seq_len(n_splits)) {

      pol_ids <- which(obj$poly.stats$root.id == obj$split.stats$plot.id[split.id])
      pol_id1 <- obj$poly.stats$plot.id[pol_ids[1]]
      pol_id2 <- obj$poly.stats$plot.id[pol_ids[2]]
      child_pol1 <- obj$polygons.xy[[as.character(pol_id1)]]
      child_pol2 <- obj$polygons.xy[[as.character(pol_id2)]]
      base <- baseline[split.id]

      # group points by child polygon (relative to xy_sub)
      split.ids1 <- .get_ids(child_pol1, xy_sub)
      split.ids2 <- .get_ids(child_pol2, xy_sub)

      # record counts
      offset_group <- (group_id - 1L) * (2L * n_splits)
      offset_split <- (split.id - 1L) * 2L
      row1 <- offset_group + offset_split + 1L
      row2 <- offset_group + offset_split + 2L
      n_per_pol[row1, ] <- list(split.id,
                                gr,
                                pol_id1,
                                length(split.ids1))
      n_per_pol[row2, ] <- list(split.id,
                                gr,
                                pol_id2,
                                length(split.ids2))

      # if absent from both sides
      if (n_per_pol[row1,"n"] == 0L && n_per_pol[row2,"n"] == 0L) {
        within_est[split.id, group_id] <- NA # cannot be defined
        elim_est  [split.id, group_id] <- NA # by definition will not change
        perm_est [[ split.id ]][[group_id]] <- NA # by definition will not change
      } else {
        # full-data indices per child polygon
        split.ids1_all <- .get_ids(child_pol1, xy_in_obj)
        split.ids2_all <- .get_ids(child_pol2, xy_in_obj)

        split.ids1_all_gr   <- split.ids1_all[split.ids1_all %in% idx_group]
        split.ids2_all_gr   <- split.ids2_all[split.ids2_all %in% idx_group]
        split.ids1_all_nogr <- split.ids1_all[!(split.ids1_all %in% idx_group)]
        split.ids2_all_nogr <- split.ids2_all[!(split.ids2_all %in% idx_group)]
        all_gr_ids_in_parent <- c(split.ids1_all_gr, split.ids2_all_gr)

        # elimination (remove group)
        elim_est[split.id, group_id] <- obj$call.info$Call_ARGS$compare.f(
          obj$call.info$Call_ARGS$generalize.f(.slicer(dat_in_obj, split.ids1_all_nogr)),
          obj$call.info$Call_ARGS$generalize.f(.slicer(dat_in_obj, split.ids2_all_nogr))
        )

        # --- Permutation: early skip if one side is empty ------------------------------

        # Check whether the focal group has occurrences on *both* sides of the split-line.
        # If one side has zero group observations, there is nothing meaningful to permute.
        has_both_sides <- (n_per_pol[row1, "n"] > 0L && n_per_pol[row2, "n"] > 0L)

        if (!has_both_sides) {
          # If the group occurs on only one side (or neither),
          # permutations are uninformative → store NA for this split/group combination.
          perm_est[[split.id]][[group_id]] <- NA

        } else {
          ## boundary-aware locality-block shuffle
          # Proceed with permutation test only if the group occupies both sides.

          # Extract coordinates of *all* occurrences of the focal group
          # that lie within the parent polygon of the current split.
          coords_agp <- xy_in_obj[all_gr_ids_in_parent, , drop = FALSE]

          # Identify unique locality coordinates
          uni_loc <- unique(coords_agp)

          # Map each observation to its unique locality index
          # These indeces allow to link locality-level info with occurrences:
          # locality_level_var[obs_to_loc] -> get locality info for each
          # all_gr_ids_in_parent
          obs_to_loc <- match(
            paste(coords_agp[[1]], coords_agp[[2]]),
            paste(uni_loc[[1]],   uni_loc[[2]])
          )
          # --- Determine polygon membership for each locality --------------------------
          # Identify which unique locality coordinates fall within the first and second
          # child polygons of this split. Localities can belong to both polygons (boundary cases).
          uni_loc_ids_in_1 <- .get_ids(child_pol1, uni_loc)
          uni_loc_ids_in_2 <- .get_ids(child_pol2, uni_loc)

          # convenience var:
          nloc <- nrow(uni_loc)

          # Create logical vectors marking polygon membership per locality.
          logi_loc_in_pol_1 <- logical(nloc)
          logi_loc_in_pol_1[uni_loc_ids_in_1] <- TRUE

          logi_loc_in_pol_2 <- logical(nloc)
          logi_loc_in_pol_2[uni_loc_ids_in_2] <- TRUE


          # Preallocate a numeric vector to store the recomputed performance
          # for each permutation iteration.
          res <- numeric(perm.n)

          # --- Main permutation loop ---------------------------------------------------
          # Each iteration permutes whole locality blocks among available positions,
          # preserving the number of blocks per polygon but randomising their identities.
          for (p in seq_len(perm.n)) {

            # Generate a random permutation of locality indices 1:nloc.
            perm_levels <- sample.int(nloc)

            # target locality index for each observation after permutation
            tgt_loc  <- perm_levels[obs_to_loc]

            # per-observation membership (after permutation)
            to_pol1 <- logi_loc_in_pol_1[tgt_loc]
            to_pol2 <- logi_loc_in_pol_2[tgt_loc]

            # observation ids for each polygon
            g_id1 <- all_gr_ids_in_parent[to_pol1]
            g_id2 <- all_gr_ids_in_parent[to_pol2]

            # Combine the non-group (fixed) observations with the permuted group observations
            # to form new assemblages for the two child polygons.
            dat1 <- .slicer(dat_in_obj, c(split.ids1_all_nogr, g_id1))
            dat2 <- .slicer(dat_in_obj, c(split.ids2_all_nogr, g_id2))

            # Recompute split-line performance for this permutation:
            #   - generalize.f() aggregates each polygon’s data
            #   - compare.f() computes between-polygons dissimilarity or distance.
            res[p] <- obj$call.info$Call_ARGS$compare.f(
              obj$call.info$Call_ARGS$generalize.f(dat1),
              obj$call.info$Call_ARGS$generalize.f(dat2)
            )
          }

          # Store the full vector of permutation results (length = nperm)
          # in the nested list corresponding to this split and group.
          perm_est[[split.id]][[group_id]] <- res
        }



        # within-group (agreement)
        if (n_per_pol[row1,"n"] == 0L || n_per_pol[row2,"n"] == 0L) {
          within_est[split.id, group_id] <- maxdif
        } else {
          dat_pol1 <- .slicer(data_sub, split.ids1)
          dat_pol2 <- .slicer(data_sub, split.ids2)
          within_est[split.id, group_id] <- obj$call.info$Call_ARGS$compare.f(
            obj$call.info$Call_ARGS$generalize.f(dat_pol1),
            obj$call.info$Call_ARGS$generalize.f(dat_pol2)
          )
        }
      }
    } # split loop

    if (plot) {
      obj_copy <- obj
      obj_copy$split.stats$performance <- within_est[, group_id]
      obj_copy$call.info$Call_ARGS$data <- data_sub
      obj_copy$call.info$Call_ARGS$xy.dat <- xy_sub
      plots[[group_id]] <- plot_hespdiv(obj_copy, performance = TRUE, subtitle = paste("Group:", gr), ...)
    } else {
      plots[[group_id]] <- NULL
    }
  } # group loop

  # ----- build deltas (consistent sign) -----
  make_delta <- function(est_mat, baseline, maximize) {
    d <- sweep(est_mat, 1L, baseline, FUN = "-")
    if (maximize) d <- -d
    d
  }
  within_delta <- -make_delta(within_est, baseline, maximize)
  elim_delta   <- make_delta(elim_est,   baseline, maximize)

  # perm deltas: transform each vector vs the split-specific baseline
  perm_delta <- create_nested_list(list(rownames(obj$split.stats), group_levels))
  for (split.id in seq_len(n_splits)) {
    base <- baseline[split.id]
    for (group_id in seq_len(group.n)) {
      v <- perm_est[[split.id]][[group_id]]
      if (is.null(v) || (length(v) == 1L && is.na(v))) {
        perm_delta[[split.id]][[group_id]] <- NA
      } else {
        d <- v - base
        if (maximize) d <- -d
        perm_delta[[split.id]][[group_id]] <- d
      }
    }
  }

  result <- list(
    within   = list(est = within_est, delta = within_delta),
    elim     = list(est = elim_est,   delta = elim_delta),
    perm     = list(est = perm_est,   delta = perm_delta),
    baseline = baseline,
    n_per_pol = n_per_pol,
    plots     = plots
  )
  class(result) <- "group_effect"
  return(result)
}

#' @noRd
create_nested_list <- function(levels) {
  if (length(levels) == 1) {
    return(setNames(replicate(length(levels[[1]]), list(), simplify = FALSE), levels[[1]]))
  }
  setNames(lapply(levels[[1]], function(x) create_nested_list(levels[-1])), levels[[1]])
}

#' Identify dependent split-lines and polygons
#'
#' @description  For a provided split-line id, function returns a list of
#' all offspring polygons and split-lines, if there are any.
#'
#' @return a list:
#' \enumerate{
#'   \item Dependent split-line IDs, if any
#'   \item Dependent polygon IDs
#' }
#' @param obj hespdiv object
#' @param id id of a split-line
#' @export
depend_splits <- function(obj, id){
  offspring <- .dep_split(obj, id) # obtain dependent split ids and polygon ids

  # add offspring to a total vector of offspring:
  all_depend_splits <- offspring$splits
  all_depend_pols <- offspring$polygons

  # create a vector of split-line still needed to check for offspring
  unchecked <- offspring$splits

  # while there are unchecked offspring split-lines, obtain offspring of each
  # offspring
  while (length(unchecked) > 0){
    new <- numeric()
    for (split.id in unchecked){
      offspring <- .dep_split(obj, split.id)
      new <- c(new, offspring$splits)
      all_depend_splits <- c(all_depend_splits, offspring$splits)
      all_depend_pols <- c(all_depend_pols, offspring$polygons)

    }
    unchecked <- new # reset to new list of offspring split-lines
  }
  return(list(
    splits = all_depend_splits,
    polygons = all_depend_pols
  ))
}

# Identify dependent splits and polygons if any
#' @noRd
.dep_split <- function(obj, id){
  # Step 1: identify produced polygons
  offspring_id <-
    which(obj$poly.stats$root.id == obj$split.stats$plot.id[id])
  # Step 2: does each polygon has a split-line?
  depend_l <- obj$poly.stats[offspring_id, "has.split"]
  # if any TRUE, return split ids + polygon ids; else return only polygon ids
  if (any(depend_l)) {
    list(splits = which(obj$split.stats$plot.id %in% offspring_id[depend_l]),
         polygons = offspring_id)
  } else {
    list(splits = numeric(),
         polygons = offspring_id)
  }
}

#' Constrained HespDiv Sensitivity Analysis by Subsampling
#'
#' @description
#' Conduct a constrained sensitivity analysis on a \code{hespdiv} object by
#' repeatedly subsampling observations within each polygon. Each subsample
#' is used to call \code{hespdiv} with recursion disabled (i.e., single-split only).
#'
#' @details
#' For each polygon in the \code{hespdiv} object, this function draws \code{subsample_factor}
#' of the data (by default 70%), creating multiple random subsamples (\code{n.runs}).
#' These are processed in chunks (as given by \code{chunk_size}) and runs in each chunk being
#' parallelized to manage memory usage.
#'
#' @param obj A \code{hespdiv} object.
#' @param n.runs Integer. The number of subsampling runs to perform (default: 100).
#' @param subsample_factor Numeric proportion of data to subsample within each polygon (0 to 1].
#'   For example, 0.7 means 70% of the data in each polygon are retained.
#' @param RAM Integer. Approximate amount of RAM in GB to guide how many parallel
#'   workers to use. The function uses up to 80\% of your available CPU cores
#'   but also caps the number of workers at \code{RAM}.
#' @param chunk_size Integer. Number of runs submitted per batch. Parallelism is
#'   controlled by \code{workers}; chunking mainly controls memory/queue size.
#' @param load_prop Numeric value (0,1]. Specifies the proportion of available
#'   CPU cores or RAM to be used for setting up parallel workers. For example,
#'   \code{load_prop = 0.8} uses 80% of the available resources. If both
#'   \code{load_prop} and \code{RAM} are provided, the number of workers will
#'   be the minimum based on the constraints imposed by both. Defaults to
#'   \code{0.8} if not provided.
#' @param workers A number of parrallel workers. Determines the number of hespdiv runs to be processed in parallel.
#'
#' @return A \code{hsa_constrained} class object, a list with two elements:
#' \itemize{
#'   \item \strong{Alternatives}
#'   A named list corresponding to each polygon where each entry is another
#'   list of \code{hespdiv} results for each subsample run.
#'
#'   \item \strong{Basis}
#'   The original \code{hespdiv} object (\code{obj}).
#' }
#'
#' @family functions for hespdiv sensitivity analysis
#' @family functions for hespdiv results post-processing
#'
#' @seealso
#' \code{\link{hespdiv}} for details on the main function.
#' \CRANpkg{future.apply}
#'
#' @importFrom future.apply future_lapply
#' @importFrom future plan multisession availableCores
#' @importFrom pracma polyarea
#' @export
hsa_sample_constrained <- function(obj,
                                   n.runs = 100,
                                   subsample_factor = 0.7,
                                   RAM = NULL,
                                   load_prop = NULL,
                                   chunk_size = 8,
                                   workers = NULL) {
  # ---- Extract and Validate Basic Args ----
  if (!is.numeric(subsample_factor) || subsample_factor < 0 || subsample_factor > 1) {
    stop(
      "`subsample_factor` must be in the interval (0, 1].",
      call. = FALSE
    )
  }
  if (!inherits(obj, "hespdiv")) {
    stop("`obj` must be of class `hespdiv.`", call. = FALSE)
  }
  if (!is.numeric(n.runs) || n.runs < 1 || n.runs %% 1 || length(n.runs) != 1)
    stop("`n.runs` must be a positive integer.", call. = FALSE)
  if (!is.numeric(chunk_size) || chunk_size < 1 || chunk_size %% 1 ||
      length(chunk_size) != 1)
    stop("`chunks` must be a positive integer.", call. = FALSE)

  # Extract call_args AFTER we confirm obj is a hespdiv
  call_args <- obj$call.info$Call_ARGS
  if (!call_args$same.n.split) {
    stop("This function currently works only when `same.n.split` is TRUE.",
         call. = FALSE)
  }

  # ---- Extract Core Information ----
  dat_in_obj <- call_args$data

  # The bounding polygon's area
  big_poly_area <- pracma::polyarea(x = obj$polygons.xy[[1]][,1],
                                    y = obj$polygons.xy[[1]][,2])
  S_crit_abs <- call_args$S.crit * abs(big_poly_area)

  # Decide on .slicer function
  if (is.data.frame(dat_in_obj) || is.matrix(dat_in_obj)) {
    .slicer <- .slicer.table
  } else if (is.list(dat_in_obj)) {
    .slicer <- .slicer.list
  } else {
    .slicer <- .slicer.vect
  }

  # ---- Prepare Output Structure ----
  split_ids <- obj$split.stats$plot.id
  hespdivs  <- vector("list", length(split_ids))
  names(hespdivs) <- split_ids

  # ---- Parallel Setup ----
  if (is.null(workers)) {
    num_cores <- future::availableCores()

    # if load_prop is NULL, default to 0.8 (or 1, or any sensible default)
    if (is.null(load_prop)) {
      load_prop <- 0.8
    }
    if (!is.numeric(load_prop) || load_prop <= 0 || load_prop > 1) {
      stop("`load_prop` must be a numeric value in the interval (0, 1].",
           call. = FALSE)
    }
    safe_cores <- max(1, floor(num_cores * load_prop))

    # if RAM is NULL, treat it as infinite
    if (is.null(RAM)) {
      RAM <- Inf
    } else {
      if (!is.numeric(RAM) || length(RAM) != 1 || RAM < 1) {
        stop("`RAM` must be either NULL or a positive integer", call. = FALSE)
      }
    }

    safe_workers <- max(1, min(safe_cores, RAM))
  } else {
    if (!is.numeric(workers) || workers < 1) {
      stop("`workers` must be a positive integer.", call. = FALSE)
    }
    safe_workers <- workers
  }
  message("Using ", safe_workers, " parallel workers.")

  # Configure future
  future::plan(future::multisession, workers = safe_workers)
  options(future.globals.maxSize = 1024^3 * 2)  # 2 GB limit for serialized objects

  # ---- Main Loop: Iterate Over Polygons ----
  for (i in seq_along(split_ids)) {
    pol.id <- split_ids[i]

    # Subset data to this polygon
    polygon <- obj$polygons.xy[[as.character(pol.id)]]
    ids <- .get_ids(polygon, call_args$xy.dat)
    xy.dat <- call_args$xy.dat[ids, ]
    data_for_poly <- .slicer(dat_in_obj, ids)

    l <- nrow(xy.dat)

    # Generate random subsets (indices) for each run
    indices_list <- lapply(seq_len(n.runs), function(.) {
      sample.int(l, size = round(l * subsample_factor), replace = FALSE)
    })

    # Split the runs into chunks of a given size
    run_indices <- split(seq_len(n.runs), ceiling(seq_len(n.runs) / chunk_size))

    # Process each chunk in parallel
    result_list <- lapply(run_indices, function(runs_for_this_chunk) {
      future_lapply(runs_for_this_chunk, function(run) {
        # Subset data for this run
        l_xy_dat <- xy.dat[indices_list[[run]], , drop = FALSE]
        l_data   <- .slicer(data_for_poly, indices_list[[run]])

        # Calculate S.crit for this polygon area
        poly_area <- abs(pracma::polyarea(x = polygon[,1],
                                          y = polygon[,2]))
        S_crit_local <- S_crit_abs / poly_area

        # Call hespdiv with single-split recursion
        hespdiv(
          data         = l_data,
          xy.dat       = l_xy_dat,
          n.split.pts  = call_args$n.split.pts,
          same.n.split = call_args$same.n.split,  # must be TRUE
          method       = NULL,
          generalize.f = call_args$generalize.f,
          compare.f    = call_args$compare.f,
          maximize     = call_args$maximize,
          N.crit       = call_args$N.crit,
          N.rel.crit   = call_args$N.rel.crit,
          N.loc.crit   = call_args$N.loc.crit,
          N.loc.rel.crit = call_args$N.loc.rel.crit,
          S.crit       = S_crit_local,  # local area-based
          S.rel.crit   = call_args$S.rel.crit,
          Q.crit       = call_args$Q.crit,
          c.splits     = call_args$c.splits,
          c.Q.crit     = call_args$c.Q.crit,
          c.crit.improv= call_args$c.crit.improv,
          c.X.knots    = call_args$c.X.knots,
          c.Y.knots    = call_args$c.Y.knots,
          c.max.iter.no= call_args$c.max.iter.no,
          c.fast.optim = call_args$c.fast.optim,
          c.corr.term  = call_args$c.corr.term,
          study.pol    = polygon,
          use.chull    = FALSE,
          tracing      = NULL,
          pnts.col     = 1,
          display      = FALSE,
          pacific.region = call_args$pacific.region,
          .do_recurse  = FALSE
        )
      })
    })

    # Flatten the results from chunks into a single list for this polygons
    hespdivs[[i]] <- unlist(result_list, recursive = FALSE)
  }

  # ---- Return 'hsa_constrained' object ----
  structure(
    list(Alternatives = hespdivs, Basis = obj),
    class = "hsa_constrained"
  )
}

#' Interpolate point on polygon
#' @param x a numeric vector sorted in increasing order (x coordinates)
#' @param y a numeric vector associated with x (y coordinates)
#' @param x3 a numeric value, in the range of x. y coordinate will be obtained
#' for this x coordiante by linear interpolation.
#' @author Liudas Daumantas
#' @noRd

.y.online <- function(x,y,x3) {
  ind<-.binar_search(x,x3)
  if (x[ind[2]] == x3 ){
    if (x[ind[2]+1] == x3){
      return(y[which.min(c(abs(y[ind[2]+1]),abs(ind[2])))])
    }
    return(y[ind[2]])
  }
  y[ind[1]]+(x3-x[ind[1]])*(y[ind[2]]-y[ind[1]])/(x[ind[2]]-x[ind[1]])
}

#' @importFrom RColorBrewer brewer.pal.info
#' @importFrom grDevices colors
#' @noRd
.generate_cols <- function(n,seed){
  set.seed(seed)
  #sp.n <- length(unique(data[,-which(colnames(data) %in% c('x','y'))]))
  if (n<=74){
    qual_col_pals <- RColorBrewer::brewer.pal.info[
      RColorBrewer::brewer.pal.info$category == 'qual',]
    col_vector <- unlist(mapply(
      RColorBrewer::brewer.pal, qual_col_pals$maxcolors,
      rownames(qual_col_pals)))
    if (n <=27) {
    palete <- sample(col_vector[
      -c(3:4,5,15,12,10,8,1,21,23,25,29,30:44,19,13,27,28, 45,22,18,48,24,51,
         58,62,63,64,65,59,70,71,73,74)], n, replace = FALSE)
    } else {
      palete <- sample(col_vector, n, replace = FALSE)
    }
  } else{
    color <- grDevices::colors()[grep('gr(a|e)y', grDevices::colors(), invert = T)]
    palete <- sample(color,n)
  }
  palete
}

#' Plot hespdiv dendrogram
#'
#' @description This function displays a dendrogram of polygons produced by hespdiv split-lines. Branch length is proportional to difference. If performance of split-lines is a similarity measure, it is internally converted to difference.
#' @param obj A hespdiv object.
#' @param poly.scheme ggplot2 object produced with poly_scheme function. Provide if you want identical colors for polygons in both plots.
#' @param color color vector used for dendrogram nodes and branches.
#' @param performance.col color vector used for text, displaying difference values between polygons
#' @param labels.col color vector used for text, displaying polygon IDs.
#' @param offset.factor numeric value used to scale the offset distance of displayed polygon IDs and performance values from a dendrogram node. Adjust experimentally, if you don't like the current distance.
#' @param arrange logical. Plot hespdiv dendrogram above polygon scheme?
#' @param label.size size of labels.
#' @param grob logical. Convert plot to grob? Must be true, if you want to arrange polygon scheme and the dendrogram in a single plot.
#' @return A `grob` or `TableGrob` object if `grob = TRUE`, otherwise `NULL`.
#' @importFrom igraph graph_from_data_frame V E ends vcount layout_as_tree
#' @importFrom gridExtra arrangeGrob
#' @importFrom grid grid.grabExpr grid.draw grid.newpage
#' @importFrom gridGraphics grid.echo
#' @author Liudas Daumantas
#' @note If you want to transform similarity to difference externally, before applying dendro, change maximize to TRUE in the call info of obj.
#' @family HespDiv visualization options
#' @examples
#' scheme <- poly_scheme(example_hespdiv)
#' # notice the colors in used in scheme:
#' scheme
#' # Dendrogram visualization of polygons, using colors from scheme
#' dendro(example_hespdiv, poly.scheme = scheme, arrange = FALSE, grob = FALSE)
#' @export
dendro <- function(obj, poly.scheme = NULL, color = 1, performance.col = "blue",
                   labels.col = 1, offset.factor = 1, arrange = TRUE,
                   grob = TRUE, label.size = 0.5){
  if (arrange & !grob) {
    stop(
      "Can only arrange the polygon scheme and dendrogram in a single plot when ",
      "`grob = TRUE`.",
      call. = FALSE
    )
  }

  if (grob) {

    # Wrapper function to call .base_dendro
    .call_base <- function() {
      .base_dendro(obj, poly.scheme = poly.scheme, color = color,
                   performance.col = performance.col, labels.col = labels.col,
                   offset.factor = offset.factor, label.size = label.size)
    }

    dendro_grob <- grid::grid.grabExpr(
      gridGraphics::grid.echo(.call_base),
      wrap = TRUE,
      offscreen = TRUE
    )

    if (arrange && !is.null(poly.scheme)) {
      out <- gridExtra::arrangeGrob(dendro_grob, poly.scheme, ncol = 1)
    } else {
      out <- dendro_grob
    }

    grid::grid.newpage()
    grid::grid.draw(out)
  } else {
    .base_dendro(obj, poly.scheme = poly.scheme, color = color,
                 performance.col = performance.col, labels.col = labels.col,
                 offset.factor = offset.factor, label.size)
  }

}

# generate the dendrogram using base R graphics
#' @noRd
.base_dendro <- function(obj, poly.scheme = poly.scheme,
                         color = color,
                         performance.col = performance.col,
                         labels.col = labels.col,
                         offset.factor = offset.factor,
                         label.size){
  # extract essential data
  pols <- obj$poly.stats[, c("plot.id", "root.id", "has.split")]
  pols$performance <- ifelse(
    test = obj$poly.stats$has.split,
    yes = obj$split.stats$performance[match(obj$poly.stats$plot.id,
                                            obj$split.stats$plot.id)],
    no = NA
  )

  if (!is.null(poly.scheme)) {
    pols$color <- poly.scheme$layers[[4]]$aes_params$colour
  } else {
    if (is.null(color)){
      stop("Either `poly.scheme` or `color` must be provided.", call. = FALSE)
    }
    pols$color <- color
  }


  # Step 1: Calculate branch lengths for child nodes
  if (!obj$call.info$Call_ARGS$maximize) {
    if (all(pols$performance >= 0, na.rm = TRUE) &
        all(pols$performance <= 1, na.rm = TRUE)) {
      # assuming that maximum theoretical value is 1
      pols$branch_length <- 1 - pols$performance
    } else {
      if (any(pols$performance < 0, na.rm = TRUE)) {
        # removing negative values
        pols$performance <- pols$performance - min(pols$performance)
      }
      # Assuming that maximum similarity is the one recorded
      pols$branch_length <- max(pols$performance) - pols$performance
    }
  } else {
    if (any(pols$performance < 0, na.rm = TRUE)) {
      # removing negative values
      pols$performance <- pols$performance - min(pols$performance)
    }
    pols$branch_length <- pols$performance
  }

  # Step 2: Create edges with branch lengths
  edges <- data.frame(
    from = pols$root.id[pols$root.id != 0],  # Parent cluster
    to = pols$plot.id[pols$root.id != 0],    # Child cluster
    weight = pols$branch_length[match(pols$root.id[pols$root.id != 0],
                                      pols$plot.id)]  # Branch length
  )

  # Step 3: Build the graph with weights and vertex attributes
  g <- igraph::graph_from_data_frame(edges, directed = TRUE, vertices = pols)

  # Add `plot_id` and `color` attributes to vertices
  igraph::V(g)$plot_id <- pols$plot.id
  igraph::V(g)$color <- pols$color

  # Assign edge colors based on the child node's color
  vertex_map <- stats::setNames(seq_along(igraph::V(g)), names(igraph::V(g)))

  igraph::E(g)$color <- igraph::V(g)$color[
    vertex_map[as.character(igraph::ends(g,igraph::E(g))[, 2])]]

  # Step 4: Calculate node positions manually
  end_nodes_id <- obj$poly.stats$plot.id[!obj$poly.stats$has.split]
  x_kords <- rep(NaN, nrow(obj$poly.stats))
  names(x_kords) <- obj$poly.stats$plot.id
  x_kords[as.character(end_nodes_id)] <- 1:length(end_nodes_id)
  y_koords <- rep(NaN, nrow(obj$poly.stats))
  names(y_koords) <- obj$poly.stats$plot.id
  end_node_parents <- sapply(end_nodes_id, .find_pol_parents, obj = obj, simplify = FALSE)


  y_koords["1"] <- 0
  node_parents <- sapply(obj$poly.stats$plot.id[-1], .find_pol_parents, obj = obj)
  y_koords[as.character(obj$poly.stats$plot.id[-1])] <-
    unlist(lapply(node_parents, function(ids,obj) {
      sum(pols$branch_length[which(pols$plot.id %in% ids)])
    },
    obj = obj))




  siblings <- which(duplicated(end_node_parents))
  while(length(siblings) > 0){
    parent_id <- sapply(end_node_parents[siblings], `[`, 1)
    oldparents <- lapply(end_node_parents[siblings], `[`, -1)
    x_kords[as.character(parent_id)] <-
      (x_kords[as.character(end_nodes_id[siblings])] +
         x_kords[as.character(end_nodes_id[siblings-1])]) / 2

    end_nodes_id[c(siblings-1)] <- parent_id
    end_nodes_id <- end_nodes_id[-siblings]
    end_node_parents[siblings -1] <- oldparents
    end_node_parents <- end_node_parents[-siblings]
    siblings <- which(duplicated(end_node_parents))
  }

  layout <- as.matrix(data.frame(x_kords,-y_koords))
  ylimit <-  c(min(layout[, 2] - 0.1 * offset.factor * stats::dist(range(layout[, 2]) * 1.1) /
                     length(unique(obj$split.stats$rank))), 0)


  # Step 5: Prepare data for lateral-first edges
  edge_segments <- data.frame(
    x0 = layout[vertex_map[as.character(igraph::ends(g, igraph::E(g))[, 1])], 1],  # Parent x
    y0 = layout[vertex_map[as.character(igraph::ends(g, igraph::E(g))[, 1])], 2],  # Parent y
    x1 = layout[vertex_map[as.character(igraph::ends(g, igraph::E(g))[, 2])], 1],  # Child x
    y1 = layout[vertex_map[as.character(igraph::ends(g, igraph::E(g))[, 2])], 2],  # Child y
    weight = igraph::E(g)$weight  # Edge weight for segment length
  )
  # Step 6: Plot the dendrogram with lateral-first edges
  plot(
    NULL,
    xlim = range(layout[, 1]),  # Set x-axis limits
    ylim = ylimit,  # Set y-axis limits
    xlab = "",          # Label for x-axis
    ylab = "",            # Label for y-axis
    main = "",
    xaxt = "n",                 # Suppress default x-axis
    yaxt = "n",                       # Suppress default y-axis
    bty = "n"                         # Remove the box around the plot
  )

  # Draw lateral-first edges
  for (i in 1:nrow(edge_segments)) {
    # Horizontal edge from parent to the x-position of the child
    lines(
      x = c(edge_segments$x0[i], edge_segments$x1[i]),
      y = c(edge_segments$y0[i], edge_segments$y0[i]),
      col = igraph::E(g)$color[i]
    )
    # Vertical edge from horizontal line down to the child node
    lines(
      x = c(edge_segments$x1[i], edge_segments$x1[i]),
      y = c(edge_segments$y0[i], edge_segments$y1[i]),
      col = igraph::E(g)$color[i]
    )
    # Get parent x- and y-coordinates
    parent_x <- edge_segments$x0[i]
    parent_y <- edge_segments$y0[i]
    # Add segment length only for left branches
    if (edge_segments$x1[i] < edge_segments$x0[i]) {
      graphics::text(
        x = parent_x,
        y = parent_y - 0.2 * offset.factor * stats::dist(range(layout[, 2]) *
                                                           1.1) /
          length(unique(obj$split.stats$rank)),  # Slightly below the parent node
        labels = round(edge_segments$weight[i], 2),
        col = performance.col,
        cex = label.size
      )
    }
  }

  # Add nodes (circles)
  points(
    layout[, 1],
    layout[, 2],
    pch = 21,
    bg = igraph::V(g)$color,
    cex = 2
  )

  # Add cluster labels below the nodes
  graphics::text(
    layout[, 1],
    layout[, 2] - 0.1 * offset.factor * stats::dist(range(layout[, 2]) * 1.1) /
      length(unique(obj$split.stats$rank)),  # Slightly below the nodes
    labels = igraph::V(g)$plot_id,
    col = labels.col,
    pos = NULL,  # Place them directly below
    cex = label.size
  )

  rounded_y_ticks <- round(seq(min(layout[, 2]), max(layout[, 2]), length.out = 5), 2)

  labels <- -rounded_y_ticks


  graphics::axis(
    side = 2,                                # Left-side Y-axis
    at = rounded_y_ticks,                    # Rounded tick positions
    labels = -rounded_y_ticks,                # Rounded tick labels
    las = 1,                                 # Rotate labels to be horizontal
    line = 1                              # Adjust spacing of axis to prevent overlap
  )
}

#' @noRd
.find_pol_parents <- function(obj,id){
  parent_id <- obj$poly.stats$root.id[which(obj$poly.stats$plot.id == id)]
  ids <- parent_id
  while (parent_id != 1){
    parent_id <- obj$poly.stats$root.id[which(obj$poly.stats$plot.id ==
                                                parent_id)]
    ids <- c(ids, parent_id)
  }
  ids
}

#' Main recursive hespdiv inner helper function
#'
#' @description  During each recursive iteration this function fits straight and
#' curvi-linear split-lines, then it uses the better of the two to separate
#' data in space, saves information about the polygon and split-lines,
#' and calls itself again using the extracted data samples.
#' @param samp.dat a spatial subset of \code{data} that lies entirely within
#' some polygon, produced using split-line subdivisions of original study area.
#' At first iteration, while no subdivisions are established,
#' \code{samp.dat = data}.
#' @param samp.xy a spatial subset of \code{xy.dat}
#' @param root.id An id of recursive iteration that produced the \code{samp.dat}.
#' An id of parent-polygon.
#' @param recursive logical.
#' @return No return. Function updates variables in \code{hespdiv} environment.
#' @author Liudas Daumantas
#' @note Function inherits the environment of \code{hespdiv} function.
#' @importFrom pracma poly_center
#' @noRd

.spatial_div <- function(samp.dat, samp.xy, root.id, recursive = TRUE){

  #testuojamas plotas
  testid <- length(rims)

  assign(x = "poly.obj" ,
         value = do.call(c,list(poly.obj,list(generalize.f(samp.dat)))),
         envir = e)
  perim_pts <- .perimeter_pts(polygon = rims[[testid]],n.pts = n.split.pts,
                              dst.pts = dst.pts)

  .visualise_splits.start(what = trace.object,
                          pnts.col, xy.dat, rims,perim_pts,
                          testid, study.pol.viz)

  #testavimui pjuviai paruosiami

  environment(.dif_fun) <- e
  if (S.rel.crit > 0){
    S_org <- abs(pracma::polyarea(x=rims[[testid]][,1],y=rims[[testid]][,2]))
  } else {
    S_org <- NULL
  }

  pairs_pts <- .pair_pts(perim_pts[[1]],polygon = perim_pts[[2]])
  maxdif <- c.Q.crit # first split minimum quality. P.crit
  any.split <- numeric()
  maxid <- 0
  cond <- TRUE
  if (nrow(pairs_pts) != 0){
    #pjaustymo ir testavimo ciklas
    for (i in 1:nrow(pairs_pts)){
      if (!cond) .visualise_splits.bad_straight(what = trace.object,
                                                level = trace.level,
                                                pairs_pts, message, i-1)
      cond <- TRUE
      virs <- .close_poly(
        open.poly =
          .split_poly(
            polygon = perim_pts[[2]],
            min_id = 1,
            split_ids = as.numeric(pairs_pts[i,c(6:7)]),
            trivial_side = TRUE,
            poli_side = TRUE
          ))
      po <- .close_poly(
        open.poly =
          .split_poly(
            polygon = perim_pts[[2]],
            min_id = 1,
            split_ids = as.numeric(pairs_pts[i,6:7]),
            trivial_side = TRUE,
            poli_side = FALSE
          ))

      # padalinami duomenys i dvi dalis pagal pjuvio koordinates
      id1 <- .get_ids(po, samp.xy)
      id2 <- .get_ids(virs, samp.xy)

      .visualise_splits.try_straight(what = trace.object, level = trace.level,
                                     pairs_pts, i)

      if (N.crit > 0){
        cond <- length(id1) > N.crit & length(id2) > N.crit
        if (!cond){
          message <- paste0("Not enough observations.",
                            "\nObtained: ", length(id1), ' and ', length(id2),
                            "\nRequired: >", N.crit)
          next
        }
      }

      if (N.rel.crit > 0){
        cond <- length(id1) / nrow(samp.xy) > N.rel.crit &
          length(id2) / nrow(samp.xy) > N.rel.crit
        if (!cond){
          message <- paste0("Too low proportion of observations.\nObtained: ",
                            round(length(id1) / nrow(samp.xy), 2),
                            ' and ', round(length(id2) / nrow(samp.xy), 2),
                            "\nRequired: >", N.rel.crit)
          next
        }
      }

      if (N.loc.crit > 0 | N.loc.rel.crit > 0){
        n.uni.1 <- nrow(unique(samp.xy[id1,]))
        n.uni.2 <- nrow(unique(samp.xy[id2,]))
        if (N.loc.crit > 0){
          cond <- n.uni.1 > N.loc.crit &
            n.uni.2 > N.loc.crit
          if (!cond){
            message <- paste0("Not enough locations.",
                              "\nObtained: ", n.uni.1,
                              ' and ', n.uni.2,
                              "\nRequired: >", N.loc.crit)
            next
          }
        }
        if (N.loc.rel.crit > 0){
          n.uni <- nrow(unique(samp.xy))
          cond <- n.uni.1 / n.uni > N.loc.rel.crit &
            n.uni.2 / n.uni > N.loc.rel.crit
          if (!cond){
            message <- paste0("Too low proportion of locations.",
                              "\nObtained: ", round(n.uni.1 / n.uni,2),
                              ' and ', round(n.uni.2 / n.uni, 2),
                              "\nRequired: >", N.loc.rel.crit)
            next
          }
        }
      }

      if (S.crit > 0 | S.rel.crit > 0 ){
        SpjuvioI <- abs(pracma::polyarea(x=virs[,1],y=virs[,2]))
        SpjuvioII <- abs(pracma::polyarea(x=po[,1],y=po[,2]))
        if (S.crit > 0) {
          cond <- SpjuvioI > S.cond & SpjuvioII > S.cond
          if (!cond){
            message <- paste0("One of the areas was too small.\n","Obtained: ",
                              round(SpjuvioI,2), ' and ', round(SpjuvioII,2),
                              "\nRequired: >", S.cond)
            next
          }
        }
        if (S.rel.crit > 0) {
          cond <- SpjuvioI / S_org  > S.rel.crit &
            SpjuvioII / S_org > S.rel.crit
          if (!cond){
            message <- paste0("Too low proportion of area.\n","Obtained: ",
                              round(SpjuvioI / S_org,2), ' and ',
                              round(SpjuvioII / S_org, 2),
                              "\nRequired: >", S.rel.crit)
            next
          }
        }
      }
      #nupiesiam padalinima ir paskaiciuojam kokybe
      environment(.dif_fun) <- environment()
      Skirtumas <- .dif_fun(.slicer(samp.dat,id1), .slicer(samp.dat,id2))
      any.split <- c(any.split,Skirtumas)
      #Paskaiciuojam plotus padalintu bloku
      if (.comp(Skirtumas,maxdif) ){

        .visualise_splits.good_straight(what = trace.object,
                                        level = trace.level,
                                        pairs_pts, Skirtumas, i,
                                        maxid)

        #Jei padalinimas patenkina minimalias saligas ir yra
        # geresnis nei pries tai - pasizymim ir issisaugom ji
        maxdif <- Skirtumas
        maxid <- i
      } else {
        message <- paste0("Poor split quality.\n","Obtained: ",
                          round(Skirtumas,2),
                          "\nRequired: ",c.sign, round(maxdif,2))
        cond <- FALSE
      }
    }
    if (!cond) .visualise_splits.bad_straight(what = trace.object,
                                              level = trace.level,
                                              pairs_pts, message, i)
    assign(x = "str.split.quals" ,
           value = do.call(c,list(str.split.quals,list(any.split))),
           envir = e)

    if(length(any.split) > 0){

      mean.dif <- mean(any.split) # mean spatial heterogeneity
      sd.dif <- sd(any.split) # NA if 1 split
      # sd(any.split) -> anisotropy of heterogeneity
      # if all are equal, then complete randomness: no anisotropy - no splits present
      if( all(any.split==any.split[1]) & length(any.split) > 1  ){
        message(
          "All tested splits were of equal quality in polygon ", testid,
          ". A random split that meets the subdivision criteria was selected as best."
        )
      }
    } else {
      mean.dif <- NA # negalejom ivertinti ne vieno padalinimo, taigi performance lygu max.
      sd.dif <- NA # P.crit
    }
    assign(x = "poly.info" ,value = rbind(poly.info, data.frame(
      root.id = root.id,
      n.splits = length(any.split),
      n.obs = nrow(samp.xy),
      mean = mean.dif, # NA if 0
      sd = sd.dif, # NA if 1 or 0 split
      str.best = maxdif,
      str.z.score = (maxdif - mean.dif) / sd.dif, # NA if 1 or 0 split
      has.split = FALSE,
      is.curve = FALSE,
      crv.best = NA,
      crv.z.score = NA,
      c.improv = NA)
    )
    , envir = e)
    # duomenu saugojimas
    #Jei rastas tinkamas padalinimas - ieskom geriausios padalinimo kreives,
    #issaugom duomenis ir ziurim ar galima skaidyti toliau
    if (maxid>0) {
      # skirtas sumazinti poligona pateikiama?
      del.id <- unique(unlist(pairs_pts[,6:7])) # visi prideti taskai (atrinkti, taip kad nesidubliuotu)
      dub.id <- which(duplicated(perim_pts[[2]][-nrow(perim_pts[[2]]),]))
      except <- unique(c(
        unlist(apply(pairs_pts[maxid,6:7],2,function(o) which(del.id %in% o))),
        which(del.id == 1), # kurie prideti taskai yra pirmas taskas arba geriausias
        which(apply(perim_pts[[2]][del.id,],1,function(o)
          any(o[1] == perim_pts[[2]][dub.id,1] &
                o[2] == perim_pts[[2]][dub.id,2]) ))
      )) # kurie taskai yra kartu ir dubliai su
      # poligono taskais.

      del.id <- del.id[-except] # pasalinam situos taskus is del.id

      if (length(dub.id) > 0){
        if(any(c(pairs_pts[maxid,6:7]) %in% dub.id)) {
          except.id <- which(dub.id %in% pairs_pts[maxid,6:7])
          dubs <- dub.id[-except.id]
          except.id <- dub.id[except.id]
          alter.ids <- which(perim_pts[[2]]$x.poly %in%
                               perim_pts[[2]]$x.poly[except.id] &
                               perim_pts[[2]]$y.poly %in%
                               perim_pts[[2]]$y.poly[except.id])
          alter.ids <- alter.ids[ !alter.ids %in%
                                    c(pairs_pts[maxid,6],pairs_pts[maxid,7])]
          dub.id <- c(dubs, alter.ids)
        }
      }
      del.id <- unique(c(del.id, dub.id))
      if (length(del.id) > 0){
        perim_pts[[2]] <- perim_pts[[2]][-del.id,] # istrinam del.id is poligono
        min.x.id <- pairs_pts[maxid,6] - sum(del.id<pairs_pts[maxid,6])
        max.x.id <- pairs_pts[maxid,7] - sum(del.id<pairs_pts[maxid,7]) # pakoreguojam min.x.id ir max.x.id pagal istrintu
        # tasku kieki
      } else {
        min.x.id <- pairs_pts[maxid,6]
        max.x.id <- pairs_pts[maxid,7]
      }

      if ( c.splits) {
        .visualise_splits.best_straight(what = trace.object,
                                        pairs_pts, maxid, maxdif)

        environment(.curvial_split) <- environment()
        best.curve <- .curvial_split(

          poly.x = perim_pts[[2]]$x.poly,
          poly.y = perim_pts[[2]]$y.poly,
          min.x.id = min.x.id,
          max.x.id = max.x.id,
          b = pairs_pts[maxid,5],
          samp.dat = samp.dat,
          c.X.knots = c.X.knots,
          c.Y.knots = c.Y.knots,
          N.crit = N.crit,
          N.rel.crit = N.rel.crit,
          N.loc.crit = N.loc.crit,
          N.loc.rel.crit = N.loc.rel.crit,
          S.cond = S.cond,
          S.rel.crit = S.rel.crit,
          S_org = S_org,
          c.max.iter.no = c.max.iter.no,
          c.fast.optim = c.fast.optim,
          c.corr.term = c.corr.term,
          trace.object = trace.object,
          trace.level = trace.level,
          pnts.col = pnts.col,
          straight.qual = maxdif

        )
        poly.info[testid,"crv.best"] <- best.curve[[2]]
        poly.info[testid,"crv.z.score"] <- (best.curve[[2]] - mean.dif) / sd.dif
        poly.info[testid,"c.improv"] <- abs(best.curve[[2]] - maxdif)
        if(.comp(best.curve[[2]],maxdif) &
           poly.info[testid,"c.improv"] >= c.crit.improv){
          poly.info[testid,"is.curve"] <- TRUE
          maxdif <- best.curve[[2]]
        }
      }
      if (!.comp(maxdif, Q.crit) ){
        maxid <- 0
      }
    }


    if (maxid > 0){ # save the split and perform new splits if TRUE

      poly.info[testid,"has.split"] <- TRUE
      assign(x = "poly.info" ,value = poly.info, envir = e)

      if(poly.info[testid,"is.curve"]) {
        best.splitl <- data.frame(x = best.curve[[1]]$x, y = best.curve[[1]]$y)
      } else {
        best.splitl <- data.frame(x=as.numeric(c(pairs_pts[maxid,c(1,3)])),
                                  y=as.numeric(c(pairs_pts[maxid,c(2,4)])))
      }
      .visualise_splits.best_split(what = trace.object,
                                   best.splitl, maxdif)

      assign(x = "splits" ,value = do.call(c,list(splits,list(best.splitl))),
             envir = e)
      up.pol <- .close_poly(
        open.poly =
          .split_poly(
            polygon = data.frame(x=perim_pts[[2]][,1][-nrow(perim_pts[[2]])],
                                 y=perim_pts[[2]][,2][-nrow(perim_pts[[2]])]),
            split_ids = c(min.x.id, max.x.id),
            min_id = 1,
            trivial_side = TRUE,
            poli_side = TRUE
          ),
        close.line = best.splitl
      )
      up.ids <- .get_ids(up.pol, samp.xy)
      up.xy <- .slicer.table(samp.xy, up.ids)
      up.dat <- .slicer(samp.dat, up.ids)

      do.pol <- .close_poly(
        open.poly =
          .split_poly(
            polygon = data.frame(x=perim_pts[[2]][,1][-nrow(perim_pts[[2]])],
                                 y=perim_pts[[2]][,2][-nrow(perim_pts[[2]])]),
            split_ids = c(min.x.id, max.x.id),
            min_id = 1,
            trivial_side = TRUE,
            poli_side = FALSE
          ),
        close.line = best.splitl
      )
      do.ids <- .get_ids(do.pol, samp.xy)
      do.xy <- .slicer.table(samp.xy, do.ids)
      do.dat <- .slicer(samp.dat, do.ids)

      assign(x = "rims" ,value = do.call(c,list(rims,list(up.pol))) ,envir = e)

      if (recursive){
        .spatial_div(up.dat,up.xy, root.id = testid)
      }

      assign(x = "rims" ,value = do.call(c,list(rims,list(do.pol))) ,envir = e)

      if (recursive){
        .spatial_div(do.dat,do.xy, root.id = testid)
      }

    } # else no adequate split
  } else {
    assign(x = "str.split.quals" ,
           value = do.call(c,list(str.split.quals,list(any.split))),
           envir = e)
    assign(x = "poly.info" ,value = rbind(poly.info, data.frame(
      root.id = root.id,
      n.splits = 0,
      n.obs = nrow(samp.xy),
      mean = NA, # NA if 0
      sd = NA, # NA if 1 or 0 split
      str.best = NA,
      str.z.score = NA, # NA if 1 or 0 split
      has.split = FALSE,
      is.curve = FALSE,
      crv.best = NA,
      crv.z.score = NA,
      c.improv = NA))
      , envir = e)
    if (!is.null(trace.level))
      message(
        "No adequate pair of points on the perimeter of the polygon.\n",
        "The polygon may be too irregular, subdivisions too detailed, ",
        "or `n.pts` too low."
      )
  }
}



#' Calculate difference between two data sets
#'
#' @description  Function combines generalize.f and compare.f functions, to
#' estimate the difference between two data sets obtained from opposite sides
#' of a split-line. This difference reflects the ability of a split-line to
#' spatially separate data. Thus, it is considered to reflect the quality of a
#' split-line.
#' @param dat1 a spatial subset of \code{samp.dat} that is located at opposite
#' side of a split-line than \code{dat2} is.
#' \code{samp.dat = data}.
#' @param dat2 a spatial subset of \code{samp.dat} that is located at opposite
#' side of a split-line than \code{dat1} is.
#' @return Numeric value, expressing the difference between two data sets and
#' the quality of a split-line.
#' @author Liudas Daumantas
#' @note Function inherits the environment of \code{hespdiv} function. Also,
#' it forces this environment to be inherited by generalize.f and compare.f
#' functions.
#' @noRd
.dif_fun <- function(dat1,dat2) {
  environment(generalize.f) <- environment()
  environment(compare.f) <- environment()
  compare.f( generalize.f(dat1), generalize.f(dat2) )
}

#' Visualise split lines.
#'
#' @description Helper functions that adds the selected graphical information to
#' the existing plot. After second dot (e.g. '.start') is defined the situation
#' when the function is called.
#' @param what String that informs what graphical information to show. Can be
#' either "curve", "straight" or "both".
#' @author Liudas Daumantas
#' @importFrom grDevices dev.set dev.list dev.set dev.new
#' @importFrom graphics lines points
#' @noRd
.visualise_splits.start <- function(what,
                                    pnts.col, xy.dat, rims, perim_pts,
                                    testid,study.pol.viz) {
  if (!is.null(what)){

    plot(NULL,xlim = range(rims[[testid]]$x),ylim = range(rims[[testid]]$y),col=0,
         xlab = "x coordinate", ylab = "y coordinate")
    if (!is.null(study.pol.viz)){
      graphics::lines(study.pol.viz[,1:2], lwd=2)
    }
    graphics::lines(rims[[1]])
    graphics::points(perim_pts[[1]],pch=17,col="darkgreen",cex = 1)
    if(!is.null(pnts.col)){
      graphics::points(xy.dat$x, xy.dat$y, col=pnts.col,pch=19,cex=0.5)
    }
    centras <- pracma::poly_center(rims[[testid]][,1],rims[[testid]][,2])
    points(centras[1],centras[2],col=2,pch=19,cex=0.65)
    points(centras[1],centras[2],col=2,cex=1.5)

    if (testid>1) {
      for (i in 2:c(testid)){
        graphics::lines(x=rims[[i]][,1],y=rims[[i]][,2],col=1,lwd=2)
      }}
    readline(prompt = cat(paste0('\nGoing to test polygon No.: '
                                 ,testid,
                                 "\n\nPress enter to continue...\n")))
  }}
.visualise_splits.end <- function(pnts.col, xy.dat, rims, study.pol.viz) {
  if(!is.null(pnts.col)){
    plot(xy.dat$x, xy.dat$y, col=pnts.col,xlab = "x coordinate",
         ylab = "y coordinate" ,pch=19, ylim = range(rims[[1]]$y),
         xlim = range(rims[[1]]$x))
  } else {
    plot(NULL, ylim = range(rims[[1]]$y),
         xlim = range(rims[[1]]$x), col=0,
         xlab = "x coordinate", ylab = "y coordinate")
  }
  if (!is.null(study.pol.viz)){
    graphics::lines(study.pol.viz[,1:2], lwd=2)
  }
  lines(rims[[1]])

  if (length(rims)>1) {
    for (i in 2:length(rims)){
      graphics::lines(x=rims[[i]][,1],y=rims[[i]][,2],col=21,lwd=2)
    }}
}
.visualise_splits.best_straight <- function(what,
                                            pairs_pts, maxid, maxdif) {
  if (!is.null(what)){
    graphics::lines(pairs_pts[maxid,c(1,3)],pairs_pts[maxid,c(2,4)],lwd=2)
    readline(prompt = cat(paste0('\nThe best straight split-line',
                                 ' is displayed.\nIt\'s quality is: ',
                                 round(maxdif,2),
                                 "\n\nPress enter to continue...\n")))
  }
}
.visualise_splits.curve_start <- function(what,
                                          rot.dat.cords, pnts.col, rot.poli, AE,
                                          c.X.knots, split.line.x, c.Y.knots,
                                          knot.y.matrix) {
  if (!is.null(what)){
    if (what != 'straight') {
      grDevices::dev.new()
      if(!is.null(pnts.col)){
        plot(rot.dat.cords$x, rot.dat.cords$y, col=pnts.col,
             xlab = "shifted & rotated X coordinate",
             ylab = "shifted & rotated Y coordinate" ,pch=19,
             xlim = range(rot.poli$x),ylim = range(rot.poli$y),cex = 0.5)
      } else {
        plot(NULL,xlim = range(rot.poli$x),ylim = range(rot.poli$y),col=0,
             xlab = "shifted & rotated X coordinate",
             ylab = "shifted & rotated Y coordinate")
      }
      graphics::lines(rot.poli)

      graphics::lines(c(0,AE),c(0,0),col = "lightblue")
      graphics::points(split.line.x[c(-1, - length(split.line.x))],
             rep(0,length(split.line.x)-2),
             pch=8,cex=2,col="blue")
      for (i in 1:c.X.knots){
        graphics::points(rep(split.line.x[i],c.Y.knots-2),
               knot.y.matrix[c(-1,-nrow(knot.y.matrix)),i])
      }
    }
  }
}

.visualise_splits.no_best_curve <- function(what){
  if (!is.null(what)){
    if (what != 'straight'){
      invisible(grDevices::dev.set(grDevices::dev.list()[length(grDevices::dev.list())]))
      readline(prompt = cat(paste0('\nNo better nonlinear split-line',
                                   ' was found.',
                                   "\nNow the window with nonlinear",
                                   " split-lines will close.",
                                   "\n\nPress enter to continue...\n")))
      grDevices::dev.off(grDevices::dev.list()[length(grDevices::dev.list())])
    }
  }
}

.visualise_splits.best_curve <- function(what,
                                         curve.final, SS) {
  if (!is.null(what)){
    if (what != 'straight'){
      invisible(grDevices::dev.set(grDevices::dev.list()[length(grDevices::dev.list())]))
      graphics::lines(curve.final,lwd=2)
      readline(prompt = cat(paste0('\nThe best curvial split-line',
                                   ' is displayed.\nIt\'s quality is: ',
                                   round(SS[[1]],2),
                                   "\nNow the window with nonlinear",
                                   " split-lines will close.",
                                   "\n\nPress enter to continue...\n")))
      grDevices::dev.off(grDevices::dev.list()[length(grDevices::dev.list())])
    }
  }
}
.visualise_splits.good_curve <- function(what, level,
                                         curve, SS, best.old.curve) {
  if (!is.null(what)){
    if (level != 'best' & what != 'straight'){
      invisible(dev.set(dev.list()[length(dev.list())]))
      graphics::lines(best.old.curve,lwd=3, col = 'gray70')
      graphics::lines(curve,lwd=2, col = '#56B4E9')
      readline(prompt = cat(paste0('\nThe displayed curvial split-line',
                                   ' is so far the best.',
                                   '\nIt\'s quality is: ',
                                   round(SS[[1]],2),
                                   "\n\nPress enter to continue...\n")))


    }}}
.visualise_splits.try_int_curve <- function(what, level,
                                            curve, x, y) {
  if (!is.null(what)){
    if (what != "straight" & level == "all"){
      invisible(grDevices::dev.set(grDevices::dev.list()[length(grDevices::dev.list())]))
      graphics::lines(curve, col = '#E69F00')
      graphics::points(x,y,col="#E69F00", pch = 8,cex= 2 )
      readline(prompt = cat(paste0("\nPotentially better curve was found by using",
                                   " knot, interpolated from other knot performances",
                                   ".\n\nPress enter to",
                                   " test it...\n")))
    }}}
.visualise_splits.good_int_curve <- function(what, level,
                                         curve, SS, best.old.curve,
                                         x, y) {
  if (!is.null(what)){
    if (level != 'best' & what != 'straight'){
      invisible(grDevices::dev.set(grDevices::dev.list()[length(grDevices::dev.list())]))
      graphics::lines(best.old.curve,lwd=3, col = 'gray70')
      graphics::lines(curve,lwd=2, col = '#56B4E9')
      if (level == "main"){
        graphics::points(x,y,col="#E69F00", pch = 8,cex= 2 )
      }
      readline(prompt = cat(paste0(
        '\nThe curve produced by using interpolated, potentially better knot',
                                   ' was indeed the best so far.',
                                   '\nIt\'s quality is: ',
                                   round(SS[[1]],2),
                                   "\n\nPress enter to continue...\n")))


    }}
}
.visualise_splits.bad_int_curve <- function(what, level,
                                            curve, x, y, message) {
  if (!is.null(what)){
    if (level == 'all' & what != 'straight'){
      invisible(grDevices::dev.set(grDevices::dev.list()[length(grDevices::dev.list())]))
      graphics::lines(curve, col = '#999999')
      graphics::points(x, y, col="gray70", pch = 8,cex = 2 )
      readline(prompt = cat(paste0("\nThe interpolated knot did not produce",
                                   ' a better curve.',
                                   ' \nReason: ',message,
                                   "\n\nPress enter to continue...\n")))

    }}
}
.visualise_splits.selected_knot <- function(what, level,
                                            x, y, old.knot.y) {
  if (!is.null(what)){
    if (level != 'best' & what != 'straight'){

      invisible(grDevices::dev.set(grDevices::dev.list()[length(grDevices::dev.list())]))
      graphics::points(x,y,col="#D55E00", pch = 8,cex= 2 )
      graphics::points(x,old.knot.y,col="gray70", pch = 8,cex= 2 )
      cat(paste0('\nSelected knot is displayed.\n'))

    }}}

.visualise_splits.interpol_knot<- function(what = trace.object,
                                level = trace.level,
                                x , y, old.knot.y){
  if (!is.null(what)){
    if (level != 'best' & what != 'straight'){
      invisible(grDevices::dev.set(grDevices::dev.list()[length(grDevices::dev.list())]))
      graphics::points(x,y,col="purple", pch = 8,cex= 2 )
      graphics::points(x,old.knot.y,col="gray70", pch = 8,cex= 2 )
      cat(paste0('\nInterpolated knot is displayed.\n'))

    }}}


.visualise_splits.good_straight <- function(what, level,
                                            pairs_pts, Skirtumas, i, maxid) {
  if (!is.null(what)){
    if (level != 'best' & what != 'curve'){
      graphics::lines(x= pairs_pts[maxid,c(1,3)], y = pairs_pts[maxid,c(2,4)],lwd=3,
                      col = 'gray70')
      graphics::lines(x= pairs_pts[i,c(1,3)], y = pairs_pts[i,c(2,4)],lwd=2,
                      col = '#56B4E9')
      readline(prompt = cat(paste0('\nThe displayed straigth split-line',
                                   ' is so far the best.',
                                   '\nIt\'s quality is: ',
                                   round(Skirtumas,2),
                                   "\n\nPress enter to continue...\n")))

    }}}
.visualise_splits.try_straight <- function(what, level, pairs_pts, i) {
  if (!is.null(what)){
    if (what != "curve" & level == "all"){
      graphics::lines(x = pairs_pts[i,c(1,3)], y = pairs_pts[i,c(2,4)],
            col = "#E69F00")
      readline(prompt = cat(paste0("\nTesting straight split-line",
                                   " No. ", i,".\n\nPress enter to",
                                   " continue...\n")))
    }}}
.visualise_splits.bad_straight <- function(what, level,
                                           pairs_pts, message, i) {
  if (!is.null(what)){
    if (what != "curve" & level == "all"){

      readline(prompt = cat(paste0("\nSplit-line was not selected.",
                                   '\nReason: ',message,
                                   "\n\nPress enter to continue...\n")))
      graphics::lines(x = pairs_pts[i,c(1,3)], y = pairs_pts[i,c(2,4)],
            col = "#999999")
    }}}
.visualise_splits.try_curve <- function(what, level,
                                        counter, curve) {
  if (!is.null(what)){
    if (what != "straight" & level == "all"){
      invisible(grDevices::dev.set(grDevices::dev.list()[length(grDevices::dev.list())]))
      graphics::lines(curve, col = '#E69F00')
      readline(prompt = cat(paste0("\nTesting curvial split-line",
                                   " No. ", counter,
                                   ".\n\nPress enter to",
                                   " continue...\n")))
    }}}
.visualise_splits.bad_curve <- function(what, level,
                                        curve, message) {
  if (!is.null(what)){
    if (what != "straight" & level == "all"){

      readline(prompt = cat(paste0("\nSplit-line was not selected.",
                                   ' \nReason: ',message,
                                   "\n\nPress enter to continue...\n")))
      invisible(grDevices::dev.set(grDevices::dev.list()[length(grDevices::dev.list())]))
      graphics::lines(curve, col = '#999999')
    }}}
.visualise_splits.best_split <- function(what,
                                         best.splitl, maxdif) {
  if (!is.null(what)){
    graphics::lines(best.splitl,col = "gold", lwd=2)
    readline(prompt = cat(paste0('\nThe best split-line',
                                 ' that will be used to',
                                 ' divide polygon',
                                 ' is displayed.',
                                 '\nIt\'s quality is: ',
                                 round(maxdif,2),
                                 "\n\nPress enter to continue...\n")))
  }
}

#' Print a hespdiv object
#'
#' @description
#' Formats and prints the rounded \code{split.stats} data frame from a
#' \code{hespdiv} object.
#'
#' @method print hespdiv
#' @param x A \code{hespdiv} object.
#' @param ... Additional arguments, currently ignored.
#'
#' @return Invisibly returns \code{x}.
#'
#' @author Liudas Daumantas
#' @export
print.hespdiv <- function(x, ...){
  dat <- x$split.stats
  if (!inherits(x,"hespdiv"))
    stop("`x` must be a `hespdiv` object.", call. = FALSE)
  if (x$call.info$METHOD$method.type == "custom"){
    type <- "Custom"
    cat(paste0("\n","Information about the split-lines:", "\n\n",type,
               " method was used.\n\n"))
  } else {
    type <- "Preset"
    if (x$call.info$METHOD$method == "biozonation"){
      method <- "Spatial biozonation"
      if (x$call.info$METHOD$metric == "sorensen") {
        metric <- paste0("So",
        "rensen-Dice coefficient")
        variant <- x$call.info$METHOD$variant
        if (variant == '1'){
          variant <- "ref. Sorensen, T. A. (1948)"
        }
      } else {
        if (x$call.info$METHOD$metric == "morisita"){
          metric <- paste0("Morisita Overlap index")
          variant <- x$call.info$METHOD$variant
          if (variant == '1'){
            variant <- "ref. Morisita, M. (1959)"
          }
        } else {
          if (x$call.info$METHOD$metric == "pielou"){
            metric <- paste0("Proportion of mean reduction in Pielou entropy")
            variant <- x$call.info$METHOD$variant
            if (variant == '1'){
              variant <- paste0("(1 - \u0394Pe / Pe.base)(",
              "ref. in hespdiv documentation)")
            }
          } else {
            if (x$call.info$METHOD$metric == "horn.morisita"){
              metric <- paste0("Morisita Overlap index")
              variant <- x$call.info$METHOD$variant
              if (variant == '2'){
                variant <- "Horn modification, ref. Horn (1966)"
              }
            }
          }
        }
      }
    }

  if (x$call.info$METHOD$similarity){
    similarity <- "Similarity"
  } else {
    similarity <- "Distance"
  }
  if (variant == "" | is.null(variant)){
    ending <- ""
  } else {
    ending <- paste0(", ",variant)
  }

  metric.row <- paste0("Metric: ", metric," (",
                       similarity,")", ending, ".")
  nm<- nchar(metric.row)

  cat(paste0("\n","Information about the split-lines:", "\n\n",type,
             " method was used.\n",
             "Method: ",method,
             ".\n", metric.row,"\n",paste0(rep("-",nm),collapse = ""),"|\n"))
  }
  if (is.null(dat)){
    cat("No Split-lines were found.", "\n")
  } else {
    id <- which(colnames(dat) == "is.curve")
    dat[,-id] <- round(dat[,-id],2)
    print(dat)
    if (any(names(dat)=="n.m.rez")){
      cat("\n", "Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1")
    }
  }
  invisible(x)
}

#' Generate points on a perimeter
#'
#' This function generates regularly spaced points on a perimeter of a polygon. Their number is specified by arguments n.pts or dst.pts.
#' @param polygon A data frame of 2 columns (x,y) that contain coordinates of polygon vertices. Both, closed and open polygons, are accepted.
#' @param n.pts A number of regularly spaced points to be generated on a perimeter of a polygon. If n.pts is not specified, then it is calculated according to the argument dst.pts.
#' @param dst.pts A distance along a polygon perimeter between adjacent points to be generated on a perimeter of a polygon. If dst.pts is not specified, then it is calculated according to the argument n.pts.
#' @return A list of 2 elements:
#' \describe{
#'   \item{\code{per_pts}}{A data frame of 4 columns, providing the information about the generated points on a perimeter of a polygon. This data frame is used as an input in \code{.pair_pts} function.}
#'   \itemize{
#'   \item \code{x} - X coordinates of generated points.
#'   \item \code{y} - Y coordinates of generated points.
#'   \item \code{ID} - An ID that reflects the relative location of a point along a perimeter of a polygon in relation to other generated points and polygon vertices.
#'   \item \code{segment.no} = A vector indicating the ID of a polygon segment on which a generated point is located. It helps to indentify points located on the same
#' polygon segment.
#'   }
#'   \item{\code{full.poly}}{ A data frame that contains coordinates of the provided polygon vertices and generated points. \code{coords[,"ID"]} can be used to
#' extract rows of generated points.This data frame is used as an input in \code{.curvial_split} function.}
#' }
#' @note If both, n.pts and dst.pts, are specified, then points are generated according to n.pts.
#' @author Liudas Daumantas
#' @examples #Creating data frame of a polygon
#' poly<- data.frame(X=c(3.38,3.30,1.70,0.78,-0.06,-2.30,-2.94,-3.97,-1.61,
#' -0.39,0.68,1.28,1.60,3.38),
#'                   Y=c(-0.12,-0.31,-2.73,-3.22,-3.29,-2.19,-1.62,0.94,3.10,
#'                   3.00,2.91,2.49,2.20,-0.12))
#' plot(poly,type='o')
#' #Generating 10 points on a polygon perimeter
#' a <- .perimeter_pts(poly,n.pts = 10)
#' #location of points
#' points(a[[1]][,-3],col=2,pch=19)
#' #ID of points
#' text(x=a[[1]][,1],y=a[[1]][,2],a[[1]][,3],-0.3)
#' @noRd


.perimeter_pts<-function (polygon,n.pts=NULL,dst.pts=NULL){
  if (all(polygon[1,]!=polygon[nrow(polygon),])){
    polygon<-.close_poly(polygon)
  }
  x<-polygon[,1]
  y<-polygon[,2]
  perimetras<-0
  for( i in 1:c(length(x)-1)){
    perimetras<-perimetras+sqrt((x[i]-x[i+1])^2+(y[i]-y[i+1])^2)
  }
  if (!is.null(dst.pts)) {
    n.pts<-round(perimetras/dst.pts,0)
  } else {
    dst.pts<-perimetras/n.pts
  }
  passed<-0
  generated_x_pt<-x[1]
  generated_y_pt<-y[1]
  ID<-1
  i=2
  segment.no<-i
  x.poly<-x[1]
  y.poly<-y[1]
  lastx<-generated_x_pt
  lasty<-generated_y_pt
  while (length(generated_x_pt)<n.pts){
    m<-sqrt((lastx-x[i])^2+(lasty-y[i])^2)
    if (m+passed>=dst.pts){
      kof<-(dst.pts-passed)/m
      if (lastx > x[i]) {
        X<-lastx-(lastx-x[i])*kof
      } else {
        X<-lastx+(x[i]-lastx)*kof
      }
      if (lasty > y[i]){
        Y<-lasty-(lasty-y[i])*kof
      } else {
        Y<-lasty+(y[i]-lasty)*kof
      }
      generated_x_pt<-c(generated_x_pt,X)
      generated_y_pt<-c(generated_y_pt,Y)
      x.poly<-c(x.poly,X)
      y.poly<-c(y.poly,Y)
      ID<-c(ID,length(x.poly))
      lasty<-Y
      lastx<-X
      segment.no<-c(segment.no,i)
      passed<-0
    } else {
      passed<-passed+m
      x.poly<-c(x.poly,x[i])
      y.poly<-c(y.poly,y[i])
      lasty<-y[i]
      lastx<-x[i]
      i<-i+1
    }
  }

  coords<-data.frame(x=generated_x_pt,y=generated_y_pt,ID,segment.no)
  full.poly<-rbind(data.frame(x.poly,y.poly),data.frame(x.poly=x[i:length(x)],y.poly=y[i:length(x)]))
  return(list(per_pts=coords,full.poly))
}
#' Calculate perimeter of a polygon
#' @noRd
.calc.perim <- function(polygon){
  perimetras <- 0
  for( i in 1:c(nrow(polygon)-1)){
    perimetras<-perimetras+sqrt((polygon[,1][i]-polygon[,1][i+1])^2+
                                  (polygon[,2][i]-polygon[,2][i+1])^2)
  }
  perimetras
}

#' Calculate a polygon-object cross-comparison matrix
#'
#' @description
#' Computes a cross-comparison matrix for polygon objects from a \code{hespdiv}
#' object. The matrix quantifies similarity or dissimilarity among polygon
#' objects and can be used for further analyses, such as clustering, either
#' directly or after transformation.
#'
#' @param obj A \code{hespdiv} object.
#'
#' @return
#' A numeric matrix containing pairwise comparison values among the
#' \code{hespdiv} polygon objects stored in \code{obj$poly.obj}.
#'
#' @details
#' The \code{cross_comp()} function uses the \code{compare.f} function from
#' \code{obj$call.info$Call_ARGS} to perform pairwise comparisons of
#' \code{hespdiv} polygon objects stored in \code{obj$poly.obj}. The result is
#' a cross-comparison matrix.
#'
#' @note
#' Polygon cross-comparison is currently not available for the \code{"pielou"}
#' method. It is also not supported for custom methods whose \code{compare.f}
#' function relies on variables from environments other than the function's own
#' arguments.
#'
#' @family functions for hespdiv results post-processing
#' @author Liudas Daumantas
#' @export

cross_comp <- function(obj){
  if (!inherits(obj, "hespdiv")) {
    stop("`obj` must be a `hespdiv` object.", call. = FALSE)
  }

  if (obj$call.info$METHOD$metric == "pielou") {
    # should pass original data through generalize.f
    # and compare only non-overlapping polygons
    stop(
      "Polygon cross-comparison is not yet available for the Pielou method.",
      call. = FALSE
    )
  } else {
    compare.f <- obj$call.info$Call_ARGS$compare.f
  }
  comp.mat <- matrix(NA,nrow = length(obj$poly.obj),ncol =length(obj$poly.obj) )
  for (pol.id in seq(length(obj$poly.obj))){
    comp.mat[,pol.id] <- unlist(lapply(obj$poly.obj,FUN = compare.f,
                                obj$poly.obj[[pol.id]]))
  }
  comp.mat
}

#' Schematic plot of hespdiv polygons
#'
#' This function generates a schematic visualization of subdivided territory.
#' It highlights the location of each polygon by displaying their centroids,
#' ID labels, and punctuated lines that connect the polygon centroids with the
#' split-lines that created them. This visualization represents the spatial
#' arrangement of hespdiv polygons within the territory in 2D.
#'
#' @param obj A hespdiv object.
#' @param segment A Boolean value. Display the punctuated lines joining the
#' polygon centroids with the split-lines?
#' @param id A Boolean value. Display the IDs of polygons?
#' @param seed An integer value that determines the random set of colors used
#' in visualization of split-lines and polygons.
#' @return A ggplot object.
#' @family HespDiv visualization options
#' @author Liudas Daumantas
#' @importFrom ggplot2 ggplot geom_path theme_set theme theme_void element_blank geom_point geom_segment geom_text aes
#' @importFrom rlang .data
#' @note A much clearer way to visualize polygons is by using the \code{blok3D}
#' function, with \code{height = "rank"}. However, a 3D plot is less suitable
#' option for papers.
#' @examples
#' poly_scheme(example_hespdiv)
#'
#' @export
poly_scheme <- function(obj,segment = TRUE, id = TRUE, seed = 1){

  split.stats <- obj$split.stats
  ord <- order(split.stats[,"performance"], decreasing = FALSE)
  split.stats <- split.stats[ord,]
  split.lines <- lapply(ord,function(id){obj$split.lines[[id]]})
  df <- Reduce(rbind,split.lines)
  npt.in.split <- as.numeric(lapply(split.lines,nrow))
  df$group <- factor(rep(1:length(split.lines),times=npt.in.split))

  base <- ggplot2::ggplot(obj$polygons.xy[[1]],aes(.data$x,.data$y)) +
    ggplot2::geom_path(data= obj$polygons.xy[[1]],aes(.data$x,.data$y), linewidth=0.5,
              lineend = "round",linejoin = "round",color = 1)
  scale_id <- 1
  color <- .generate_cols(nrow(split.stats), seed)
  df$color <- rep(color, times=npt.in.split)
  base<-base + ggplot2::geom_path(data = df, aes(.data$x,.data$y),group=df$group, color=df$color,linewidth = 1)+
    ggplot2::theme_void()  +
    ggplot2::theme(panel.grid = ggplot2::element_blank(),panel.background =
                     ggplot2::element_blank())
  centrai <- as.data.frame(Reduce(rbind,lapply(obj$polygons.xy, FUN = function(o){
    pracma::poly_center(x = o[,1],y = o[,2])})))
  names(centrai) <- c("x1","y1")

  base <- base + ggplot2::geom_point(data=centrai,aes(.data$x1,.data$y1),shape = 20,size=3,
                                     color = c(1,
                                               color[
                                                 match(obj$poly.stats$root.id[-1],
                                                       split.stats$plot.id)]))
  if (segment){
    id.cord <- which.min(apply(obj$polygons.xy[[1]],1, FUN = function(o){
      sqrt(sum((o - unlist(centrai[1,]))^2))
    }))
    split.mid.p <- as.data.frame(rbind(unlist(obj$polygons.xy[[1]][id.cord,]),
                                       Reduce(rbind,lapply(obj$split.lines,function(o){if (nrow(o)==2){
                                         if (o[1,1]-o[2,1]!=0){
                                           c(mean(o[,1]),.pt_on_line(x1 = o[1,1],x2 = o[2,1], y1 = o[1,2],y2 = o[2,2],
                                                                     x3 = mean(o[,1])))} else {
                                                                       c(o[1,1],mean(o[,2]))
                                                                     }
                                       } else {
                                         c(o[round(nrow(o)/2,0),1],o[round(nrow(o)/2,0),2])
                                       }}))))
    names(split.mid.p) <- c("x2","y2")
    segments <- cbind(centrai,
                      split.mid.p[c(1,match(obj$poly.stats$root.id[-1],
                                            obj$split.stats$plot.id)+1),])
    base <- base + ggplot2::geom_segment(data=segments, linetype = 3,
                                aes(x=.data$x1,y=.data$y1,xend = .data$x2,yend =.data$y2),
                                color=c(1,color[match(obj$poly.stats$root.id[-1],
                                                      split.stats$plot.id)]))}
  if (id){
    base <- base + ggplot2::geom_text(data=centrai,
                             aes(x=.data$x1,y=.data$y1,label=obj$poly.stats$plot.id),
                             nudge_y = 0.03 * diff(range(obj$polygons.xy[[1]]$y, na.rm = TRUE)),
                             color=c(1,color[match(obj$poly.stats$root.id[-1],
                                                   split.stats$plot.id)]))}
  base
}

#' Remove polygons from rgl device
#'
#' This function allows you to interactively select and remove unwanted polygons
#' from a 3D plot created with the \code{blok3d} function.
#'
#' @param obj The hespdiv object used to create the currently active rgl
#' device with the \code{blok3d} function.
#' @param height A character value that indicates the height co-ordinate.
#' @return No return value. Called for the interactive modification of a
#' plot created by \code{blok3d}
#'
#' @family HespDiv visualization options
#' @author Liudas Daumantas
#' @importFrom pracma poly_center
#' @importFrom rgl identify3d rgl.ids rgl.pop
#' @export
polypop <- function(obj,height){
  height <- .arg_check("height",height, c("mean","sd","best","z.score",
                                          "str.best", "str.z.score","rank"))
  poly.stats <- obj$poly.stats
  if (height == "best"){
    poly.stats$best <- ifelse(is.na(obj$poly.stats$is.curve),
                              obj$poly.stats$str.best,
                              ifelse(obj$poly.stats$is.curve,
                                     obj$poly.stats$crv.best,
                                     obj$poly.stats$str.best))
  } else {
    if (height == "z.score" ){
      poly.stats$z.score <- ifelse(is.na(obj$poly.stats$is.curve),
                                   obj$poly.stats$str.z.score,
                                   ifelse(obj$poly.stats$is.curve,
                                          obj$poly.stats$crv.z.score,
                                          obj$poly.stats$str.z.score))
    }
  }
  if (height == "rank") {
    ZZ <- data.frame(zmin = obj$poly.stats$rank-1, zmax = obj$poly.stats$rank)
    del.id <- numeric()
  } else {
    ZZ <- .Zcoords(poly.stats,height)
    del.id <- which(ZZ[,1]==ZZ[,2])
  }
  if (length(del.id) != 0){
    true.ids <- poly.stats$plot.id[-del.id]
  } else {
    true.ids <- poly.stats$plot.id
  }
  basic.id<-5
  OIDS <- seq(length(true.ids))

  while(length(OIDS)>1){
    centrai <- data.frame(x=rep(NaN,length(OIDS)),
                          y=rep(NaN,length(OIDS)))
    for (i in seq(length(OIDS))) {
      centrai[i,] <- pracma::poly_center(
        obj$polygons.xy[[true.ids[OIDS[i]]]][,1],
        obj$polygons.xy[[true.ids[OIDS[i]]]][,2])
    }

    XO <- centrai[,1]
    YO <- centrai[,2]

    ZO <- ZZ[true.ids[OIDS],2]
    LABS <- poly.stats[true.ids[OIDS],"plot.id"]
    message("Select the centers of polygons you wish to remove.\n")
    pts <- rgl::identify3d(x=XO, y=YO, z=ZO, labels = LABS)
    if(length(pts)==1) {
      id.start<-basic.id+((pts-1)*5)
      OIDS<-OIDS[-which(OIDS==OIDS[pts])]
      id.finito<-id.start+4
      ids<-id.finito:id.start
      ids<-rgl::rgl.ids( type = "shapes", subscene = NA )[ids,1]
      rgl::rgl.pop(type = "lights",id=ids)
    } else {
      if (length(pts>1)){
        pts<-sort(pts,T)
        for (i in 1:length(pts)) {
          id.start<-basic.id+((pts[i]-1)*5)
          OIDS<-OIDS[-which(OIDS==OIDS[pts[i]])]
          id.finito<-id.start+4
          ids<-id.finito:id.start
          ids<-rgl::rgl.ids( type = "shapes", subscene = NA )[ids,1]
          rgl::rgl.pop(type = "lights",id=ids)
        }
      } else {message("Nothing was selected\n")}
    }

  }
  message("Nothing to remove - there is only one polygon.\n")
}

#' Quantify the stability of hespdiv clusters
#'
#' @description This function evaluates the stability of the basal subdivision
#' clusters using hespdiv sensitivity analysis results. It does so by
#' calculating Jaccard similarities between the observations of basal
#' subdivision clusters and the observations of alternative subdivision
#' clusters. For each basal cluster, the function identifies the most similar
#' analog cluster within each alternative subdivision. The stability of
#' each basal cluster can be assessed by examining the distribution of
#' similarity values with their corresponding analog clusters. If a highly
#' similar cluster reappears in multiple alternative subdivisions, it
#' indicates that the basal cluster is stable.
#' @param obj An object of class \code{hsa}.
#' @param probs A numeric vector of probabilities with values in the range
#' \eqn{0 \le p \le 1}. This argument is used to calculate quantiles of Jaccard similarity
#' values.
#' @return A list containing three data frames:
#' \describe{
#'   \item{jaccard.quantiles}{
#'     Quantiles of Jaccard similarities between the basal cluster and
#'     the analog clusters from alternative subdivisions.
#'   }
#'   \item{jaccard.similarity}{
#'     Jaccard similarity values between the basal cluster and the analog
#'     cluster from each alternative subdivision.
#'   }
#'   \item{analog.clusters}{
#'     IDs of the hespdiv polygons that produced the analog clusters
#'     in each alternative subdivision.
#'   }
#' }
#' @details
#' If a base subdivision cluster obtains a distribution of high similarity
#' values, it is considered stable and existing. Low analog-cluster similarity
#' values may indicate that a base cluster is an artifact of the
#' \code{hespdiv()} computation.
#'
#' The more technical description of how \code{hsa_quant} works:
#' \describe{
#' \item{Obtaining alternative hespdiv clusters:}{ The function filters the \code{xy.dat} coordinates of the basal subdivision using all the polygons of alternative subdivisions, obtaining alternative hespdiv clusters.}
#' \item{Quantifying Jaccard similarity:}{ The function measures the Jaccard overlap index between the observations of the basal subdivision clusters and the observations of the alternative clusters. }
#' \item{Identification of analog clusters and value assignments:}{ Each basal hespdiv cluster from each alternative subdivision is assigned the ID of the cluster that produced the maximum Jaccard similarity value, along with the corresponding similarity value.}
#' }
#' The purpose of the \code{hsa_quant} function is to address situations where hespdiv
#' polygons, despite having different geometry and location, may filter nearly
#' identical sets of observations, leading to similar hespdiv clusters. This can occur when the spatial
#' coverage of observations is incomplete and irregular, or when the boundaries
#' between hespdiv polygons are expected to be open, soft, or fuzzy, such as in
#' the case of boundaries between bioregions. In such cases, visual hespdiv
#' sensitivity analysis alone may show irregular and non-converging
#' distributions of split-lines. However, \code{hsa_quant} can reveal that these
#' irregular polygons are based on nearly identical clusters of observations,
#' indicating a strong spatial structure within the analyzed data. Conversely,
#' if the observations within these clusters significantly differ, it indicates
#' that the basal clusters are specific to the hespdiv parameters used and
#' likely lack ontological meaning.
#'
#' Thus, by analyzing the similarity values between clusters of observations, \code{hsa_quant}
#' facilitates the assessment of the stability and reliability of basal
#' subdivision clusters, aiding in evaluating their significance.
#'
#' @note You can use the hsa_quant function to track the evolution of hespdiv
#' subdivisions over time by providing correctly formatted input. For
#' instance, you can obtain the basal subdivision for time bin 1 using the
#' hespdiv function. Then, using the hsa function, you can specify the paired
#' xy.dat and data from time bin 2. The resulting hsa object can be inputted
#' into \code{hsa_quant} The \code{hsa_quant} result will then provide insights into
#' extinctions, speciations, fusions, and splits of hespdiv polygons/clusters
#' that occur between time bin 1 and 2. This allows for the analysis of changes
#' and dynamics in hespdiv subdivisions over time.
#' @family functions for hespdiv sensitivity analysis
#' @family functions to evaluate hesdpiv cluster stability
#' @family functions for hespdiv results post-processing
#' @author Liudas Daumantas
#' @importFrom stats quantile density
#' @importFrom graphics plot hist
#' @export

hsa_quant <- function(obj, probs = c(0.05,0.5,0.95)){
  if (!inherits(obj,"hsa"))
    stop(
      "`obj` must be an `hsa` object produced by `hsa()` or `hsa_detailed()`.",
      call. = FALSE
    )

  subs <- lapply(obj$Alternatives, function(o) o[[1]])
  # Check if all alternative subdivisions are NULL
  if (all(il <- sapply(subs, is.null) | sapply(subs,class) != "hespdiv"))
    stop("All alternative subdivisions are NULL or contained errors",
         call. = FALSE)
  # Remove NULL subdivisions or subdivisions with
  if (any(il))
    subs <- subs[-which(il)]
  connections <- matrix(NA, nrow = length(subs),ncol = length(obj$Basis$polygons.xy) -1 )
  colnames(connections) <- names(obj$Basis$polygons.xy[-1])
  rownames(connections) <- names(subs)
  jac.sim <- matrix(NA, nrow = length(subs),ncol = length(obj$Basis$polygons.xy) -1 )
  colnames(jac.sim) <- names(obj$Basis$polygons.xy[-1])
  rownames(jac.sim) <- names(subs)

    for (basis_pol in 2:length(obj$Basis$polygons.xy)){
      b.ids <- .get_ids(obj$Basis$polygons.xy[[basis_pol]],
                                  obj$Basis$call.info$Call_ARGS$xy.dat)
      l <- lapply(subs, function(o, bas.ids) {
        lapply(o$polygons.xy[-1],function(e, dat, bs.ids) {
          .jaccard.sim(.get_ids(e, dat), bs.ids)
        }, dat = obj$Basis$call.info$Call_ARGS$xy.dat, bs.ids = bas.ids)
      }, b.ids)

      jac.sim[,basis_pol-1] <- unlist(lapply(l,function(o) max(unlist(o))))
      connections[,basis_pol-1] <- unlist(lapply(l,function(o) as.numeric(names(
        which.max(unlist(o))))))
    }

  structure(list(jaccard.quantiles = apply(jac.sim,2, stats::quantile,
                                 probs = probs),
    jaccard.similarity = jac.sim, analog.clusters = connections),
    class = "hsa_quant")
}

#' @noRd
.jaccard.sim <- function(x,y){
  inter <- length(intersect(x,y))
  inter / (length(x) + length(y) - inter)
}

#' Binary search
#'
#' This function helps to find two adjacent indeces of a numeric vector, sorted in increasing order, that surround a value of interest.
#' @param X A numeric vector, sorted in increasing order.
#' @param x3 A value of interest.
#' @param l The highest possible lower index that surrounds a value of interest from left (bottom). Default value 1.
#' @param h The lowest possible higher index that surrounds a value of interest from right (top).Default value is the length of X.
#' @return Two adjacent X vector indeces of values that surrounds a value of interest.
#' @note If the value of interest (\code{x3}) is equal to some value of X, than the index of this value will be returned as a higher index.
#' Function is not intended to be used in cases of duplicated values.
#' @author Liudas Daumantas
#' @examples # simple case
#' .binar_search(X=1:100,x3=55.5)
#' #when x3 is equal to some value of X,
#' #index with x3 value in the returned vector will be the higher one.
#' .binar_search(X=1:100,x3=50)
#' @noRd
.binar_search <- function(X,x3,l=1,h=length(X)){

  if ( x3 < X[l]| x3 > X[h]){
    stop("`x3` must lie between `X[l]` and `X[h]`.", call. = FALSE)
  }
  return(.recurs(X,l,h,x3))
}

# helper function to perform recursive binary search
.recurs<-function(X,l,h,x3){
  mid_id <- round(l+(h-l)/2,0)
  if (mid_id==l|mid_id==h){
    return(c(l,h))
  }
  if (X[mid_id]>=x3){
    h <- mid_id
  } else {
    l <- mid_id
  }
  return(.recurs(X,l,h,x3))
}

#' Find a point on a line
#'
#' This function helps to find a missing coordinate (x or y) of a point on a straight line that joins two provided points.
#' @param x1 An x coordinate of the start of a line.
#' @param x2 An x coordinate of the end of a line.
#' @param y1 An y coordinate of the start of a line.
#' @param y2 An y coordinate of the end of a line.
#' @param x3 An x coordinates, where missing y coordinates on a line must be found.
#' @param y3 An y coordinates, where missing x coordinates on a line must be found.
#' @return A missing x or y coordinates on a line that joins (x1,y1) and (x2,y2) points.
#' @author Liudas Daumantas
#' @examples
#' #A line between (3,0) & (1,-1) points
#' plot(c(3,1),c(0,-1),type='l')
#' #Finding points at x = 1, 1.5, 2, 2.5, 3
#' x <- seq(1,3,0.5)
#' y <- .pt_on_line(x1 = 3,x2 = 1,y1 = 0,y2 = -1,x3 = x)
#' points(x,y,col=2,pch=19)
#' @noRd
.pt_on_line<-function(x1,x2,y1,y2,x3=NULL,y3=NULL){
  if (is.null(y3) & is.null(x3)){
    stop("Provide either `x3` or `y3`.", call. = FALSE)
  } else {
    if (!is.null(y3) & !is.null(x3)){
      message("Both y3 & x3 is provided, only x3 will be used. Returning a missing y coordinate: ")
    }
    if (is.null(y3)){
      return(y1+(x3-x1)*(y2-y1)/(x2-x1))
    } else {
      return(x1+(y3-y1)/((y2-y1)/(x2-x1)))
    }
  }
}

#' Pairing connectable polygon points
#'
#' This function pairs points that are located on a perimeter of a polygon and can be joined by a straight line that is strictly contained within the polygon.
#' @param per_pts a data frame of 4 columns, first item of list output of \code{perimeter_pts} function.
#' @param polygon A data frame containing coordinates of a polygon. The polygon can be either closed or open.
#' @return A data frame with 7 columns. Each row represents a connectable pair of points located on a perimeter of a polygon.
#' \itemize{
#'   \item \code{x1} - x coordinate of a point that has lower x coordinate.
#'   \item \code{y1} - y coordinate of a point that has lower x coordinate.
#'   \item \code{x2} - x coordinate of a point that has higher x coordinate.
#'   \item \code{y2} - y coordinate of a point that has higher x coordinate.
#'   \item \code{b} - slope of a straight line connecting the paired points.
#'   \item \code{id1} - relative location of a point that has lower x coordinate. This ID is derived from \code{per_pts} third column (\code{ID}).
#'   \item \code{id2} - relative location of a point that has higher x coordinate. This ID is derived from \code{per_pts} third column (\code{ID}).
#' }
#' @note If both points have the same x coordinate, a point with lower ID will be treated as the one with lower x coordinate.
#' @author Liudas Daumantas
#' @examples #Creating data.frame of a polygon
#' poly<-data.frame(X=c(3.38,3.30,1.70,0.78,-0.06,-2.30,-2.94,-3.97,-1.61,
#'                     -0.39,0.68,1.28,1.60,3.38),
#'                 Y=c(-0.12,-0.31,-2.73,-3.22,0,-2.19,-1.62,0.94,3.10,
#'                 3.00,2.91,2.49,2.20,-0.12))
#' plot(poly,type='o')
#' #Generating 10 points on a polygon perimeter
#' a <- .perimeter_pts(poly,n.pts = 15)
#' #location of points
#' points(a[[1]][,-3],col=2,pch=19)
#' #ID of points
#' text(x=a[[1]][,1],y=a[[1]][,2],a[[1]][,3],-0.3)
#' #pairing points
#' b <- .pair_pts(a[[1]],poly)
#' #all arrows point to the right and all lines are within the polygon.
#' arrows(x0=b[,1],y0=b[,2],x1=b[,3],y1=b[,4],length = 0.15,col=2)
#' @noRd
.pair_pts<-function(per_pts,polygon){
  pairs_pts<-data.frame(numeric())
  for (a in 1:(nrow(per_pts)-1)){
    k<-1
    while(per_pts[a+k,4]==per_pts[a,4] &(a+k)<nrow(per_pts)){
      k=k+1
    }
    if(c(a+k)<=c(nrow(per_pts))&per_pts[a+k,4]!=per_pts[a,4]){
      for (o in c(a+k):c(nrow(per_pts))){
        x<-per_pts[c(a,o),1]
        y<-per_pts[c(a,o),2]
        bad <- FALSE
        b1 <- (y[1]-y[2])/(x[1]-x[2])
        a1 <- y[1] - b1*x[1]
        mid.x <- mean(x)
        mid.y <- a1 + mid.x * b1
        if (.point.in.polygon(point.x = mid.x + 7e-14, point.y = mid.y+7e-14,
                                 pol.x = polygon[,1], pol.y = polygon[,2]
        ) == 1 &
        .point.in.polygon(point.x = mid.x - 7e-14, point.y = mid.y-7e-14,
                                      pol.x = polygon[,1], pol.y = polygon[,2]
        ) == 1 &
        .point.in.polygon(point.x = mid.x - 7e-14, point.y = mid.y+7e-14,
                                      pol.x = polygon[,1], pol.y = polygon[,2]
        ) == 1 &
        .point.in.polygon(point.x = mid.x + 7e-14, point.y = mid.y-7e-14,
                             pol.x = polygon[,1], pol.y = polygon[,2]
        ) == 1){
          for ( seg in 1:(nrow(polygon)-1) ){
            pol_seg <- polygon[seg:(seg+1),]
            if ( all(pol_seg[1,] == c(x[1],y[1])) |
                 all(pol_seg[1,] == c(x[2],y[2])) |
                 all(pol_seg[2,] == c(x[1],y[1])) |
                 all(pol_seg[2,] == c(x[2],y[2]))){
              next
            }
            X1 <- .in.range(range(x),pol_seg$x[1]) | .in.range(range(x),pol_seg$x[2])
            X2 <- .in.range(range(pol_seg$x),x[1]) | .in.range(range(pol_seg$x),x[2])
            Y1 <- .in.range(range(y),pol_seg$y[1]) | .in.range(range(y),pol_seg$y[2])
            Y2 <- .in.range(range(pol_seg$y),y[1]) | .in.range(range(pol_seg$y),y[2])

            if ( (X1 & Y1) | (X2 & Y2) | (X1 & Y2) | (X2 & Y1) ){
              b2 <- (pol_seg$y[1]-pol_seg$y[2])/(pol_seg$x[1]-pol_seg$x[2])
              a2 <- pol_seg$y[1] - b2*pol_seg$x[1]
              x.inters <- (a2 - a1) / (b1 - b2)
              if ( is.nan(x.inters) | is.infinite(x.inters) ){
                next
              }
              bad <- .in.range(range(x), x.inters) &
                .in.range(range(pol_seg$x),x.inters)
              if (bad) {
                break
              }
            }
          }
          if (!bad){
          id.min<-which.min(x)
          pairs_pts <- rbind(pairs_pts, data.frame(
            x1=x[id.min],
            y1=y[id.min],
            x2=x[-id.min],
            y2=y[-id.min],
            b=b1,
            id1=per_pts[(c(a,o)[id.min]),3],
            id2=per_pts[(c(a,o))[-id.min],3])
          )
          }
        }
      }} else{
        break
      }}
  pairs_pts
}
#' is x in range of x?
#' @noRd
.in.range <- function(range,x){
  range[1] < x & x < range[2]
}

#' Test significance of split-lines in spatial subdivision
#'
#' @description
#' Assess the statistical significance of each split-line (bioregion boundary) identified by `hespdiv` by comparing its observed performance to a null distribution. The null is generated by permuting the data `n` times and recomputing the split-line performance after each permutation. Multiple shuffling strategies are supported to probe the influence of spatial structure on delineation.
#'
#' @details
#' Two shuffling scopes are available:
#' \describe{
#'   \item{\strong{"all"}:}{Shuffle across the entire study area, ignoring polygonal subdivisions.}
#'   \item{\strong{"within"}:}{Shuffle only within each parent polygon (the region in which the split-line is nested), preserving local spatial structure.}
#' }
#'
#' Two shuffling types are available:
#' \describe{
#'   \item{\strong{"localities"}:}{Shuffle whole localities, preserving each site's assemblage (recommended, since occurrences within a locality are not independent).}
#'   \item{\strong{"occurrences"}:}{Shuffle individual occurrences across sites (use with caution; may violate within-locality independence).}
#' }
#'
#' For each split-line, the function reports the observed performance, the mean and standard deviation of the permuted (null) performances, an empirical one-sided p-value (proportion of permuted values \emph{as or more extreme} than observed; ties included), and a z-score quantifying departure from the null.
#'
#' @param obj An object of class \code{hespdiv}.
#' @param n Integer. Number of permutations used to form the null distribution.
#' @param shuffle.scope Character. Either \code{"all"} (shuffle across the full study area) or \code{"within"} (shuffle only within each parent polygon).
#' @param shuffle.type Character. Either \code{"localities"} (shuffle whole localities, preserving assemblages) or \code{"occurrences"} (shuffle individual occurrences). \code{"localities"} is generally preferred.
#' @param maintain.n Logical. Only honored when \code{shuffle.scope = "within"} and \code{shuffle.type = "localities"}. If \code{TRUE}, attempts to keep the randomized child polygon occurrence counts close to the observed (maximum discrepancy up to \code{max_n/2}, where \code{max_n} is the largest locality size). Ignored otherwise.
#'
#' @return Invisibly returns an object of class \code{nullhespdiv}, a list with:
#' \itemize{
#'   \item \code{$stats}, a data frame summarizing each split-line with:
#'   \itemize{
#'     \item \code{performance}: observed performance.
#'     \item \code{mean.random}: mean of null performances.
#'     \item \code{sd.random}: standard deviation of null performances.
#'     \item \code{p.val}: empirical one-sided p-value (ties included).
#'     \item \code{z.score.random}: standardized effect size.
#'   }
#'   \item \code{$null}, a matrix or data frame containing all null performance
#'   values for every split-line across permutations.
#' }
#'
#' @family functions for hespdiv results post-processing
#' @examples
#' # if split-line is strongly significant, the choice of parameters should not
#' # matter. For example (look at p-value, z.score.random, sd.random and
#' # mean.random):
#' (nulltest(example_hespdiv, maintain.n = FALSE, shuffle.type = "occurrences"))
#' (nulltest(example_hespdiv, maintain.n = FALSE, shuffle.type = "localities"))
#' (nulltest(example_hespdiv, maintain.n = TRUE, shuffle.type = "localities"))
#'
#' @export

nulltest <- function(obj, n = 999, maintain.n  = TRUE, shuffle.scope = "within",
                     shuffle.type = "localities"){

  # Validate shuffle.scope and shuffle.type arguments
  shuffle.scope <- .arg_check(name = "shuffle.scope", given = shuffle.scope,
                             NAMES = c("all", "within"))
  shuffle.type <- .arg_check(name = "shuffle.type", given = shuffle.type,
                              NAMES = c("localities", "occurrences"))

  if (isTRUE(maintain.n) && !(identical(shuffle.scope, "within") &&
                              identical(shuffle.type, "localities"))) {
    warning(
      "`maintain.n` is ignored unless `shuffle.scope = \"within\"` and ",
      "`shuffle.type = \"localities\"`. Disabling it.",
      call. = FALSE
    )
    maintain.n <- FALSE
  }

  # Ensure correct object type
  if (!inherits(obj, "hespdiv"))
    stop("`obj` must be a `hespdiv` object.", call. = FALSE)

  # Extract coordinates and data from object
  coords <- obj$call.info$Call_ARGS$xy.dat
  data <- obj$call.info$Call_ARGS$data
  N <- nrow(coords)  # Number of occurrences
  l <- length(obj$split.lines) # Number of split-lines

  # Prepare to store comparison and p-value results for each permutation
  comp.vals <- vector(mode = "list", length = n)
  p.vals <- vector(mode = "list", length = n)
  for (i in 1:n){
    comp.vals[[i]] <- numeric(l)
    p.vals[[i]] <- numeric(l)
  }

  # Define comparison operator depending on whether maximizing or minimizing
  # comparison value. Checks if random is as-or-more-extreme than observed.
  if (obj$call.info$Call_ARGS$maximize){
    pal <- function(x, criteria){ x >= criteria}
  } else {
    pal <- function(x, criteria){ x <= criteria}
  }

  # Determine the correct data slicing function based on data type
  if (is.data.frame(data) | is.matrix(data)){
    .slicer <- .slicer.table
  } else {
    if (is.list(data)) {
      .slicer <- .slicer.list
    } else {
      .slicer <- .slicer.vect
    }
  }

  # Main permutation loop
  if (shuffle.scope == "all" && shuffle.type == "occurrences") {
    # in each run mix all occurrence labels and re-sample coordinates
    # with them, then filter mixed coordinates with child polygons:
    for ( i in 1:n) {
      # mix all occurrence ids (labels)
      ids <- sample(1:N,N,replace = FALSE)
      # sample coordinates with mixed ids
      co <- coords[ids,]

      for (split.id in 1 : l){
        # identify child pol ids in poly.stats
        pol_ids <- which(obj$poly.stats$root.id ==
                           obj$split.stats$plot.id[split.id])
        # get ids of coordinates in each child polygon
        split.ids1 <- .get_ids(obj$polygons.xy[[as.character(obj$poly.stats$plot.id)[pol_ids[1]]]], co)
        split.ids2 <- .get_ids(obj$polygons.xy[[as.character(obj$poly.stats$plot.id)[pol_ids[2]]]], co)
        # slice data with these ids
        dat_pol1 <- .slicer(data, split.ids1)
        dat_pol2 <- .slicer(data, split.ids2)
        # compare data
        comp.vals[[i]][split.id] <- obj$call.info$Call_ARGS$compare.f(
          obj$call.info$Call_ARGS$generalize.f(dat_pol1),
          obj$call.info$Call_ARGS$generalize.f(dat_pol2))
        # compare null comp.val with original comp.val
        p.vals[[i]][split.id] <- pal(comp.vals[[i]][split.id],
                                     obj$split.stats$performance[split.id])
      }
    }
  }
  if (shuffle.scope == "all" && shuffle.type == "localities") {
    # in each run mix locality coordinates, then filter mixed coordinates with
    # child polygons:
    for ( i in 1:n) {
      co <- .shuffle_localities(coords)

      for (split.id in 1 : l){
        # identify child pol ids in poly.stats
        pol_ids <- which(obj$poly.stats$root.id ==
                           obj$split.stats$plot.id[split.id])
        # get ids of coordinates in each child polygon
        split.ids1 <- .get_ids(obj$polygons.xy[[as.character(obj$poly.stats$plot.id)[pol_ids[1]]]], co)
        split.ids2 <- .get_ids(obj$polygons.xy[[as.character(obj$poly.stats$plot.id)[pol_ids[2]]]], co)
        # slice data with these ids
        dat_pol1 <- .slicer(data, split.ids1)
        dat_pol2 <- .slicer(data, split.ids2)
        # compare data
        comp.vals[[i]][split.id] <- obj$call.info$Call_ARGS$compare.f(
          obj$call.info$Call_ARGS$generalize.f(dat_pol1),
          obj$call.info$Call_ARGS$generalize.f(dat_pol2))
        # compare null comp.val with original comp.val
        p.vals[[i]][split.id] <- pal(comp.vals[[i]][split.id],
                                     obj$split.stats$performance[split.id])
      }
    }
  }
  if (shuffle.scope == "within" && shuffle.type == "occurrences") {
    # in each run mix occurrences within parent polygon, then filter mixed
    # coordinates with child polygons:
    for ( i in 1:n) {
      for (split.id in 1 : l){
        # get labels of observations inside parent polygon
        ids <- .get_ids(obj$polygons.xy[[as.character(obj$split.stats$plot.id[split.id])]],
                        coords)
        # extract points inside parent polygon & mix their labels (rows)
        co <- coords[sample(ids),]
        # identify child pol ids in poly.stats
        pol_ids <- which(obj$poly.stats$root.id ==
                           obj$split.stats$plot.id[split.id])
        # get (mixed) coordinate labels in each child polygon
        local_idx1 <- .get_ids(obj$polygons.xy[[as.character(obj$poly.stats$plot.id)[pol_ids[1]]]], co)
        local_idx2 <- .get_ids(obj$polygons.xy[[as.character(obj$poly.stats$plot.id)[pol_ids[2]]]], co)
        # sample unpermuted order of points with permuted point ideces
        split.ids1 <- ids[local_idx1]
        split.ids2 <- ids[local_idx2]

        # slice data with these ids
        dat_pol1 <- .slicer(data, split.ids1)
        dat_pol2 <- .slicer(data, split.ids2)

        # compare data
        comp.vals[[i]][split.id] <- obj$call.info$Call_ARGS$compare.f(
          obj$call.info$Call_ARGS$generalize.f(dat_pol1),
          obj$call.info$Call_ARGS$generalize.f(dat_pol2))
        # compare null comp.val with original comp.val
        p.vals[[i]][split.id] <- pal(comp.vals[[i]][split.id],
                                     obj$split.stats$performance[split.id])
      }
    }
  }
  if (shuffle.scope == "within" && shuffle.type == "localities") {
    # in each run mix localities within parent polygon keeping similar or
    # different number of occurrences per polygon, then filter mixed
    # coordinates with child polygons:
    for ( i in 1:n) {
      for (split.id in 1 : l){
        # get labels of observations inside parent polygon
        ids <- .get_ids(obj$polygons.xy[[as.character(obj$split.stats$plot.id[split.id])]],
                        coords)
        # identify child pol ids in poly.stats
        pol_ids <- which(obj$poly.stats$root.id ==
                           obj$split.stats$plot.id[split.id])

        if (maintain.n){ # maintain similar number of occurrences in each
          # polygon while shuffling localities

          # get number of observation in polygon 1
          quota1 <- obj$poly.stats$n.obs[pol_ids[1]]

          # shuffle localities and obtain ids to sample data
          assignments <- .shuffle_localities_by_quota(ids, coords, quota1)
          # slice data
          dat_pol1 <- .slicer(data, assignments[[1]])
          dat_pol2 <- .slicer(data, assignments[[2]])
        } else {
          # shuffle localities irrespective of occurrence number variations.
          co <- .shuffle_localities(coords[ids,])
          # get ids of coordinates in each child polygon
          local_idx1 <- .get_ids(obj$polygons.xy[[as.character(obj$poly.stats$plot.id)[pol_ids[1]]]], co)
          local_idx2 <- .get_ids(obj$polygons.xy[[as.character(obj$poly.stats$plot.id)[pol_ids[2]]]], co)
          # map back to original row numbers of coords
          split.ids1 <- ids[local_idx1]
          split.ids2 <- ids[local_idx2]
          # slice data with these ids
          dat_pol1 <- .slicer(data, split.ids1)
          dat_pol2 <- .slicer(data, split.ids2)
        }
        # compare data
        comp.vals[[i]][split.id] <- obj$call.info$Call_ARGS$compare.f(
          obj$call.info$Call_ARGS$generalize.f(dat_pol1),
          obj$call.info$Call_ARGS$generalize.f(dat_pol2))
        # compare null comp.val with original comp.val
        p.vals[[i]][split.id] <- pal(comp.vals[[i]][split.id],
                                     obj$split.stats$performance[split.id])
      }
    }
  }


  p.vals <- (Reduce("+", p.vals)+1)/(n+1) # convert logical to p.val
  comp.vals <- t(as.data.frame(comp.vals))
  rownames(comp.vals) <- 1:n
  stats <- data.frame(ID = 1:l, performance = obj$split.stats$performance,
                      mean.random = apply(comp.vals,2,mean),
                      sd.random = apply(comp.vals,2,sd),
                      p.val = p.vals)
  stats$z.score.random <- (stats$performance - stats$mean.random)/stats$sd.random
  out <- structure(list(stats = stats, null = comp.vals), class = "nullhespdiv")
  return(invisible(out))
}

#' Shuffle locality positions but preserve within-locality assemblages
#'
#' @param coords A data.frame or matrix of coordinates (rows = occurrences, columns = coordinates).
#'               All rows with identical values are treated as the same locality.
#' @return A matrix/data.frame with the same dimensions as coords, where each group of
#'         rows belonging to the same locality has been reassigned to a new, randomly chosen locality's position.
#' @noRd
.shuffle_localities <- function(coords) {
  # Combine coordinate columns into a unique string per row (locality key)
  loc_keys <- apply(coords, 1, function(x) paste(x, collapse = "_"))

  # Encode each locality as a factor (grouping variable)
  loc_factor <- factor(loc_keys)

  # Unique locality keys (sorted alphabetically, as factor levels)
  loc_levels <- levels(loc_factor)

  # Unique coordinates, ordered to match loc_levels
  loc_coords <- coords[match(loc_levels, loc_keys), , drop = FALSE]

  # Shuffle localities (rows of loc_coords)
  shuffled_idx <- sample(seq_along(loc_levels))
  shuffled_coords <- loc_coords[shuffled_idx, , drop = FALSE]

  # For each occurrence, assign new (shuffled) locality coordinates based on their group
  co <- shuffled_coords[as.integer(loc_factor), , drop = FALSE]

  return(co)
}

##' Randomly assign whole localities to two polygons to fill a quota as closely as possible
#'
#' @param ids Integer vector: occurrence indices in the parent polygon
#' @param coords Data.frame: all coords (so `coords[ids,]` is valid)
#' @param quota Integer: desired number of occurrences in polygon 1
#' @return List: indices for polygon 1 and polygon 2 (by occurrence)
#' @noRd
.shuffle_localities_by_quota <- function(ids, coords, quota) {
  # Subset the coordinates for the relevant occurrences (in parent polygon)
  sub_coords <- coords[ids, , drop = FALSE]

  # Generate a unique key for each locality by concatenating all coordinate columns
  loc_keys <- apply(sub_coords, 1, function(x) paste(x, collapse = "_"))

  # Create a factor where each level represents a unique locality
  loc_factor <- factor(loc_keys)

  # Get all unique localities as factor levels, sorted
  locality_levels <- levels(loc_factor)

  # Count how many occurrences are in each locality (locality "size")
  locality_sizes <- table(loc_factor)

  # Shuffle the order of localities to randomise sampling
  shuffled_locs <- sample(locality_levels)
  cum_sum <- 0         # Current number of occurrences assigned

  # Add localities until quota is reached or just exceeded
  cumsums <- cumsum(locality_sizes[shuffled_locs ])
  ids_with <- 1 : which(cumsums >= quota)[1]
  ids_without <- which(cumsums < quota)

  if (length(ids_without) == 0){ # edge case: no localities assigned to polygon 1
    assigned_locs <- shuffled_locs[ids_with]
  } else {
    if (length(ids_with) == length(cumsums)){ # edge case: all localities assigned to polygon 1
      assigned_locs <- shuffled_locs[ids_without]
    } else {

      # Measure how close we are to the quota including and excluding the last locality
      discrepancy_with <- abs(cumsums[cumsums >= quota][1] - quota)
      discrepancy_without <- abs(cumsums[cumsums < quota][length(cumsums[cumsums < quota])] - quota)


      # If it's closer to quota without the last locality, remove it
      if (discrepancy_without < discrepancy_with) {
        assigned_locs <- shuffled_locs[ids_without]
      } else {
        assigned_locs <- shuffled_locs[ids_with]
      }
    }
  }
  # Get the indices of occurrences that belong to assigned localities
  in_poly1 <- which(loc_factor %in% assigned_locs)
  in_poly2 <- setdiff(seq_along(ids), in_poly1) # The rest go to polygon 2

  # Return the assignment as a list of indices for the two polygons
  return(list(
    poly1 = ids[in_poly1],
    poly2 = ids[in_poly2]
  ))
}

#' Get data points that lie inside a polygon
#'
#' This function extracts points from a provided dataset that lie inside a given polygon.
#' @param polygon A data frame of 2 columns ("x","y") that contain coordinates of polygon vertices.
#' @param xy.dat A data frame, containing "x" and "y" columns. Other columns may also be present.
#' @return A filtered data frame.
#' @note This function excludes points that are strictly outside the given
#' polygon. Points lying on the edges or vertices of the
#' polygon (if the polygon is not closed) will be included in the
#' filtered data frame.
#' @author Liudas Daumantas
#' @examples #Creating data.frame of a polygon
#' poly<- data.frame(c(3.38,3.30,1.70,0.78,-0.06,-2.30,-2.94,-3.97,-1.61,-0.39,0.68,1.28,1.60,3.38),
#' c(-0.12,-0.31,-2.73,-3.22,-3.29,-2.19,-1.62,0.94,3.10,3.00,2.91,2.49,2.20,-0.12))
#'
#' #Creating a data set of points
#' xy.dat<-data.frame(x=runif(250,-4,4),y=runif(250,-4,4))
#' plot(poly,type='l',xlab="X",ylab="Y")
#' points(xy.dat)
#'
#' #Extracting points that lie inside a polygon
#' points(get_data(poly,xy.dat),pch=19,col=2)
#' @export
get_data<-function(polygon, xy.dat){
  data.frame(xy.dat[.point.in.polygon(pol.x = polygon[,1],
                                   pol.y = polygon[,2],
                                   point.x = xy.dat$x,
                                   point.y = xy.dat$y)!=0,])
}

#' Close an open polygon
#'
#' Function closes an open polygon. If a line is provided, then it adds the line to the open ends of the polygon, else it appends the first point of the polygon at its end to close it.
#' @param open.poly A data frame of 2 columns (x,y), representing the vertices of an open polygon. E. g. an output of \emph{split_poly} function.
#' @param close.line A line to be used to close a given polygon. Must be a 2 column (x,y) data frame. The first coordinates of a line must correspond to the first point of the polygon, while the last coordinates - to the last point of the polygon. If close_line is not provided (the default), then the first point of the open polygon is used to close it.
#' @return A data frame of 2 columns (x,y), containing coordinates of a closed polygon.
#' @author Liudas Daumantas
#' @examples #A data.frame of an open polygon
#' open.pol <- data.frame(x=c(0,0,1,1),y=c(0,1,1,0))
#' plot(open.pol,type="l")
#'
#' #Closing the open polygon with its first point
#' poly <- .close_poly(open.pol)
#' plot(poly,type="l")
#'
#' #Closing the open polygon with a line
#' line <- data.frame(x=seq(0,1,length.out=10),y=c(0,.2,.1,.3,.6,.5,.3,.4,.2,0))
#' poly <- .close_poly(open.pol,line)
#' plot(poly,type="l")
#' @noRd

.close_poly<-function(open.poly, close.line=NULL){
  if (is.null(close.line)){
    return(rbind(open.poly,open.poly[1,]))
  } else {
    data.frame(x= c(open.poly[,1],rev(close.line[,1])[-1]), y=c(open.poly[,2],rev(close.line[,2])[-1]))
  }
}

#' Plot the stability of hespdiv clusters
#'
#' @description  This function visualizes the stability of basal subdivision
#' clusters obtained from \code{hsa_quant}
#' @param obj The output of \code{hsa_quant} (an object of class \code{hsa_quant}).
#' @param hist A Boolean value. If FALSE, EPDFs obtained with \code{plot(density(jac.sim))}
#' will be displayed instead of histograms.
#' @return None
#' @details The stability of each basal cluster is revealed by the distribution
#' of Jaccard similarity values with the 'analogues' clusters found in alternative
#' hespdiv subdivisions. For example, a unimodal distribution with a peak at
#' high similarity values (>0.8) indicates that the basal hespdiv cluster is
#' stable, even if the polygon boundaries are not. This situation may arise when
#' there is indeed a spatial structure within the data, but there are also wide
#' gaps between sampled regions (or more generally when there is limited spatial
#' data coverage). A unimodal distribution with a peak at medium values
#' (0.4-0.6) and a tail to higher values could also indicate a more persistent
#' spatial structure. On the other hand, a single peak at low values (<0.4)
#' indicates low cluster stability (e.g., bioregion does not exist). Finally,
#' uniform, bimodal, or other more complex distributions may indicate that the
#' stability and existence of the corresponding basal cluster depend on the
#' parameters used in alternative hespdiv calls.
#' @family functions for hespdiv sensitivity analysis
#' @family HespDiv visualization options
#' @family functions to evaluate hesdpiv cluster stability
#' @author Liudas Daumantas
#' @importFrom graphics plot hist layout
#' @importFrom stats density
#' @export
plot_hsa_q <- function(obj, hist = FALSE){
  if (!inherits(obj,"hsa_quant"))
    stop(
      "`obj` must be an `hsa_quant` object produced by `hsa_quant()`.",
      call. = FALSE
    )

  jac.sim <- obj[[2]]
  l <- ncol(jac.sim)
  if (!hist){
    n <- 5
    while(!l%%n == 0){
      n <- n-1
    }
    graphics::layout(1:n)
    for (i in 1:l)
      graphics::plot(stats::density(jac.sim[,i]),
                     main = paste0("Stability of the HespDiv cluster - ",
                                   colnames(jac.sim)[i]), xlim= c(0,1),
                     xlab = "Jaccard similarity with the analogue clusters")
    graphics::layout(1)
  } else {
    n <- 5
    while(!l%%n == 0){
      n <- n-1
    }
    graphics::layout(1:n)
    for (i in 1:l)
      graphics::hist(jac.sim[,i],
                     main = paste0("Stability of the HespDiv cluster - ",
                                   colnames(jac.sim)[i]),xlim= c(0,1),
                     xlab = "Jaccard similarity with the analogue clusters")
    graphics::layout(1)
  }
}

#' Detailed HespDiv Sensitivity Analysis
#'
#' This function is one of the two that perform HespDiv sensitivity analysis.
#' It creates and evaluates alternative hespdiv calls, according to the desired
#' changes in method, data and other subdivision criteria arguments. As a result,
#' it returns alternative hespdiv objects that can be directly compared with
#' the original hespdiv object and with each other using \code{plot.hsa} and
#' \code{hsa_quant} functions.
#'
#' @param obj An object of hespdiv class. The base object whose call will be
#' modified to produce alternative hespdiv objects.
#' @param comb.args A Boolean value. Do you want to combine the provided
#' argument values to make alternative hespdiv calls? If not, then at once only one
#' argument will be modified, trying all provided values for it one by one.
#' @param pick.n.args A numeric vector that controls how many arguments would
#' you like to change at once in hespdiv runs. Multiple values allowed.
#' @param comb.type A character determining how combinations of argument values
#' are selected. Possible values: "all", "random", or "handpicked".
#' @param n.combs An integer controlling how many argument value combinations
#' should be randomly selected from all possible combinations when comb.type
#' is "random".
#' @param paired Logical. Are the provided \code{data} and \code{xy.dat}
#'   arguments paired?
#' @param display A Boolean value. The value of the "display" argument in each
#' hespdiv call.
#' @param images.path  A path to an existing directory where PNG images of the
#' displayed results will be saved. If NULL (default), images won't be saved.
#' @param pnts.col The value of the "pnts.col" argument in each hespdiv call.
#' @param data A list containing matrices, time-series, lists, data frames,
#' vectors, or other data structures.
#' @param xy.dat,study.pol Lists of coordinate data frames. Each data frame
#'   must contain two columns named \code{x} and \code{y}.
#' @param same.n.split,c.fast.optim,use.chull,c.splits Lists with a Boolean value (if used, should be different from the
#' one in the basal hespdiv call).
#' @param n.split.pts,c.max.iter.no,N.crit,N.loc.crit,c.X.knots,c.Y.knots Lists with integer values.
#' @param N.rel.crit,N.loc.rel.crit,S.crit,S.rel.crit Lists with values between 0 and 1.
#' @param Q.crit,c.Q.crit,c.crit.improv Lists of numeric values.
#' @param c.corr.term  A list of numeric values between 0.01 and 0.2.
#' @param generalize.f,compare.f Lists of functions.
#' @param maximize A list of logical values with the same length as the
#'   \code{compare.f} list.
#' @param method A list of character values.
#' @return
#' An object of class \code{hsa}. The object is a list with two elements:
#' \describe{
#'   \item{\code{Alternatives}}{A list containing the alternative
#'   \code{hespdiv} objects produced by the sensitivity analysis.}
#'   \item{\code{Basis}}{The original \code{hespdiv} object whose call was
#'   modified to produce the alternative subdivisions.}
#' }
#' @importFrom utils combn
#' @importFrom grDevices dev.off png
#' @details
#' \subsection{Difference Between "hsa" And "hsa_detailed"}{
#' The major difference between "hsa_detailed" and "hsa" is that the former produces
#' all possible hespdiv calls from combinations of the provided hespdiv arguments.
#' Therefore, it samples a much smaller segment of the parameter space but more
#' densely, requiring much more computation time. Although such behavior may
#' be desired in some cases, the "hsa" function is generally more suitable for
#' performing hespdiv sensitivity analysis.
#'
#' Additionally, alternative values for hespdiv arguments in the "hsa_detailed"
#' function are provided in lists, whereas in the "hsa" function, they are
#' provided in vectors or lists (depending on the argument).
#'}
#' \subsection{Internally Set Default Argument Values}{
#' When comb.args is TRUE, the default value of comb.type is "all".
#'
#' When comb.args is TRUE and pick.n.args is NULL (default), the value of
#' pick.n.args will be changed to a vector 1:N, where N is the maximum possible
#' value of pick.n.args. The maximum possible value for pick.n.args depends on
#' the hespdiv arguments provided. Each hespdiv argument that
#' influences the results is counted as one, except for "data" and "xy.dat"
#' when paired is TRUE, and all four arguments ("method", "compare.f",
#' "generalize.f", and "maximize") that define the subdivision method, as the
#' pair/group of them is counted as one. Therefore, N can vary from 1 (single
#' argument provided) to 22 (all arguments provided and paired is FALSE). If
#' comb.args is FALSE, then pick.n.args should be NULL. Using pick.n.args = 1
#' is the same as setting comb.type to FALSE.
#'}
#' \subsection{Paired Arguments}{
#' If paired is TRUE, the "data" and "xy.dat" elements with the same index are
#' treated as one value of the same argument. Therefore, the provided lists of
#' "data" and "xy.dat" should be of the same length. Pairing of "data" and
#' "xy.dat" can be useful, for example, when you want to re-run hespdiv after
#' adding or removing some observations (these changes should be made in both
#' "xy.dat" and "data") to test how hespdiv results are influenced by some
#' extra observations or the number of observations in general. When paired
#' is FALSE, the number of observations in "data" and "xy.dat" must be the same
#' as it was in the call of the base hespdiv object. This option allows you to
#' re-run hespdiv after adding some noise to the object features (via changes
#' in "data") or coordinates (via changes in "xy.dat") to test how hespdiv
#' results are influenced by the data itself or localization.
#'
#' By default, arguments determining the custom method ("compare.f",
#' "generalize.f", "maximize") are paired, similar to how "data" and "xy.dat"
#' are paired when paired is TRUE. Thus, the lists of "compare.f",
#' "generalize.f", and "maximize" should be of the same length.
#' }
#' @family functions for hespdiv sensitivity analysis
#' @family functions for hespdiv results post-processing
#' @note Use "pnts.col" of length >1 only when the number of observations does
#' not change.
#'
#' If a particular call produced a warning or error, then a list of
#' length 2 will be returned for that call. If a warning was produced, then the
#' first element of the list will hold the created hespdiv object, and the
#' second element will contain the warning message. In the case of an error,
#' the first element will be a list of arguments used to produce the call,
#' and the second element will contain the error message.
#' @author Liudas Daumantas
#' @export
hsa_detailed <- function(
    obj,
    comb.args = TRUE,
    pick.n.args = NULL,
    comb.type = NULL,
    n.combs = NULL,
    display = TRUE,
    images.path = NULL,
    paired = NULL,
    pnts.col = 1,
    data = NULL,
    xy.dat = NULL,
    same.n.split = NULL,
    n.split.pts = NULL,
    N.crit = NULL,
    N.rel.crit = NULL,
    N.loc.crit = NULL,
    N.loc.rel.crit = NULL,
    S.crit = NULL,
    S.rel.crit = NULL,
    Q.crit = NULL,
    c.splits = NULL,
    c.Q.crit = NULL,
    c.crit.improv = NULL,
    c.X.knots = NULL,
    c.Y.knots = NULL,
    c.max.iter.no = NULL,
    c.fast.optim = NULL,
    c.corr.term = NULL,
    study.pol = NULL,
    use.chull = NULL,
    generalize.f = NULL,
    maximize = NULL,
    method = NULL,
    compare.f = NULL
) {
  use.chull <- .to.list(use.chull)
  c.splits <- .to.list(c.splits)
  same.n.split <- .to.list(same.n.split)
  c.fast.optim <- .to.list(c.fast.optim)
  if (!is.null(images.path)) {
    if (!display) {
      stop("If you wish to save images, set `display = TRUE`.", call. = FALSE)
    }
  }

  if (obj$call.info$METHOD$method.type == "preset") {
    obj$call.info$Call_ARGS[which(names(obj$call.info$Call_ARGS) ==
                                    "generalize.f")] <- list(NULL)
    obj$call.info$Call_ARGS[which(names(obj$call.info$Call_ARGS) ==
                                    "compare.f")] <- list(NULL)
  }

  if (comb.args && is.null(comb.type)) {
    comb.type <- "all"
  }

  if (!comb.args && !is.null(comb.type)) {
    stop(
      "Check `comb.args` and `comb.type` arguments.\n",
      "`comb.args` is FALSE, but `comb.type` is not NULL.",
      call. = FALSE
    )
  }

  if (comb.args) {
    comb.type <- .arg_check(
      name = "comb.type",
      given = comb.type,
      NAMES = c("all", "random", "handpicked")
    )

    if (comb.type == "random" && (is.null(n.combs) || length(n.combs) > 1)) {
      stop(
        "Check `n.combs`.\n",
        "When `comb.type = \"random\"`, `n.combs` should be a single numeric integer.",
        call. = FALSE
      )
    }
  }

  if (!is.null(n.combs)) {
    if (!comb.args) {
      stop(
        "Check `comb.args` and `n.combs` arguments.\n",
        "`comb.args` is FALSE, but `n.combs` is not NULL.",
        call. = FALSE
      )
    }

    if (comb.type != "random") {
      stop(
        "Check `comb.type` and `n.combs` arguments.\n",
        "If `n.combs` is not NULL, `comb.type` should be set to \"random\".",
        call. = FALSE
      )
    }
  }

  prov.args.names <- ls()[!ls() %in% c("obj", "comb.args","pick.n.args","paired",
                                       "display","images.path", "pnts.col", "comb.type",
                                       "n.combs")]
  c.pars <- any(c(!is.null(c.Q.crit),!is.null(c.crit.improv),!is.null(c.X.knots),
                  !is.null(c.Y.knots),!is.null(c.max.iter.no),!is.null(c.fast.optim),
                  !is.null(c.corr.term)))
  if (is.null(c.splits) && !obj$call.info$Call_ARGS$c.splits && c.pars) {
    stop(
      "Curve-generation arguments starting with `c.` were provided, but ",
      "`c.splits` is FALSE in the base `hespdiv` object and no alternative ",
      "value was provided. Use `c.splits = list(TRUE)`.",
      call. = FALSE
    )
  }

  prov.args.l <- sapply(prov.args.names, get, environment())
  ind <- which(!sapply(prov.args.l, is.null))

  if (length(ind) == 0) {
    stop(
      "All provided `hespdiv` arguments for sensitivity analysis are NULL.",
      call. = FALSE
    )
  }

  hes.names <- (names(obj$call.info$Call_ARGS)[
    !names(obj$call.info$Call_ARGS) %in%
      c("tracing", "display", "pnts.col","pacific.region")])[ind]
  hes.mods <- prov.args.l[hes.names]

  if ("data" %in% hes.names | "xy.dat" %in% hes.names){

    if (is.null(paired) && "data" %in% hes.names && "xy.dat" %in% hes.names) {
      stop(
        "`paired` is NULL, but both `data` and `xy.dat` were provided.\n",
        "Please set `paired` to either TRUE or FALSE.",
        call. = FALSE
      )
    }
    if ("data" %in% hes.names){
      if (
        any(!unlist(lapply(lapply(hes.mods[["data"]], class), identical,
                           class(obj$call.info$Call_ARGS$data)))) ||
        any(!unlist(lapply(lapply(hes.mods[["data"]], mode), identical,
                           mode(obj$call.info$Call_ARGS$data))))
      ) {
        warning(
          "Class or mode of the provided `data` differs from `data` in `obj`.",
          call. = FALSE
        )
      }
      dat_lens <- unlist(lapply(hes.mods[["data"]],function(o){
        if (is.data.frame(o) | is.matrix(o)){
          nrow(o)
        } else {length(o)}
      }))
    }
    if ("xy.dat" %in% hes.names){
      xy_lens <- unlist(lapply(hes.mods[["xy.dat"]], nrow))
    }
    if (paired) {
      if (!("data" %in% hes.names && "xy.dat" %in% hes.names)) {
        stop(
          "`paired = TRUE`, but `data` or `xy.dat` was not provided.",
          call. = FALSE
        )
      }

      if (length(hes.mods[["data"]]) != length(hes.mods[["xy.dat"]])) {
        stop(
          "`paired = TRUE`, but `data` and `xy.dat` do not have the same length.",
          call. = FALSE
        )
      }

      if (!all(dat_lens == xy_lens)) {
        stop(
          "`paired = TRUE`, but some paired `data` and `xy.dat` elements have ",
          "different numbers of observations.",
          call. = FALSE
        )
      }

    } else {
      if ("data" %in% hes.names) {
        if (!all(dat_lens == nrow(obj$call.info$Call_ARGS$xy.dat))) {
          stop(
            "Some elements of the provided `data` do not have the same length as ",
            "the number of coordinates in `obj`.",
            call. = FALSE
          )
        }
      }

      if ("xy.dat" %in% hes.names) {
        if (!all(xy_lens == nrow(obj$call.info$Call_ARGS$xy.dat))) {
          stop(
            "Some elements of the provided `xy.dat` do not have the same number ",
            "of rows as the coordinate data in `obj`.",
            call. = FALSE
          )
        }
      }
    }
  } else{

    if (!is.null(paired)) {
      stop(
        "`paired` should only be provided when both `data` and `xy.dat` are provided.",
        call. = FALSE
      )
    }

  }

  if (!all(unlist(lapply(hes.mods, is.list)) & unlist(lapply(hes.mods, class)) == "list" )){
    stop(
      "Arguments provided for `hespdiv` sensitivity analysis should be lists.",
      call. = FALSE
    )
  }

  if ("study.pol" %in% hes.names) {

    if (obj$call.info$Call_ARGS$use.chull &&
        !"use.chull" %in% hes.names) {
      stop(
        "If `use.chull = TRUE`, changes in `study.pol` cannot change the ",
        "subdivision results.\n",
        "If you want to test the impact of `study.pol` on subdivisions, ",
        "provide `use.chull = list(FALSE)`.",
        call. = FALSE
      )
    }

    if (any(unlist(lapply(hes.mods[["study.pol"]], is.null)))){
      stop(
        "`study.pol = list(NULL, ...)` is invalid because NULL polygon values ",
        "cannot change the results.",
        call. = FALSE
      )
    }
  }

  if ("method" %in% hes.names) {
    hes.mods[["method"]] <- lapply(
      hes.mods[["method"]],
      .arg_check,
      name = "metric",
      NAMES = names(.get_methods()[["biozonation"]])
    )
  }
  if (is.null(paired)){
    paired <- FALSE
  }

  if(paired){
    org_dataxy <- obj$call.info$Call_ARGS[c("data", "xy.dat")]
    names(org_dataxy) <- NULL
    given_dataxy <- vector(mode = "list", length = length(hes.mods[["data"]]))
    for (i in seq_along(given_dataxy)) {
      given_dataxy[[i]] <- list(
        hes.mods[["data"]][[i]],
        hes.mods[["xy.dat"]][[i]]
      )

      if (identical(given_dataxy[[i]], org_dataxy)) {
        stop(
          "Element ", i, " of the paired `data` and `xy.dat` datasets is not new ",
          "(it is the same as in the base `hespdiv` object).",
          call. = FALSE
        )
      }
    }

    if (any(duplicated(given_dataxy))) {
      stop(
        "There are duplicated paired `data` and `xy.dat` values. ",
        "Remove duplicated pairs.",
        call. = FALSE
      )
    }
  }


  cond <- c("compare.f", "generalize.f", "maximize") %in% hes.names

  if (any(cond) && !all(cond)) {
    stop(
      "If you provide a custom method, all required arguments must be provided: ",
      "`compare.f`, `generalize.f`, and `maximize`.\n",
      "Missing argument(s): ",
      paste(c("compare.f", "generalize.f", "maximize")[!cond], collapse = ", "),
      call. = FALSE
    )
  }
  if (all(cond)){
    if (length(unique(c(
      length(hes.mods[["compare.f"]]),
      length(hes.mods[["generalize.f"]]),
      length(hes.mods[["maximize"]])))) != 1)
      stop(
        "Custom method arguments are paired and must have the same number of elements.",
        call. = FALSE
      )
    org_method <-  obj$call.info$Call_ARGS[c("compare.f", "generalize.f",
                                             "maximize")]
    names(org_method) <- NULL
    given_methods <- vector(mode = "list",
                            length = length(hes.mods[["compare.f"]]))
    for (i in seq(length(given_methods))) {
      given_methods[[i]] <- list(hes.mods[["compare.f"]][[i]],
                                 hes.mods[["generalize.f"]][[i]],
                                 hes.mods[["maximize"]][[i]])
      if (identical(given_methods[[i]], org_method) &
          !is.null(org_method[[1]])){
        stop(
          "Element ", i, " of the paired custom method arguments is not new ",
          "(it is the same as in the base `hespdiv` object).",
          call. = FALSE
        )
      }
    }
    if (any(duplicated(given_methods))){
      stop(
        "There are duplicated values in the paired custom method arguments. ",
        "Remove duplicated custom methods.",
        call. = FALSE
      )
    }
  }


  for (var.n in hes.names){
    if ((!var.n %in%  c("data", "xy.dat") | !paired) & !var.n %in%
        c("compare.f", "generalize.f", "maximize")){
      if (any(
        unlist(
          lapply(
            hes.mods[[var.n]],
            identical,
            obj$call.info$Call_ARGS[[var.n]])
        )
      )) {
        stop(
          "All provided argument values for sensitivity analysis should be new.\n",
          "At least one value of `", var.n, "` is the same as in the base `hespdiv` object.",
          call. = FALSE
        )
      }
      if(any(duplicated(hes.mods[[var.n]]))){
        stop(
          "There are duplicated values in the provided `", var.n,
          "` argument. Remove duplicated values.",
          call. = FALSE
        )
      }
    }
    if (!var.n %in% c("data", "xy.dat", "study.pol")){
      if(any(unlist(lapply(hes.mods[[var.n]], length)) != 1)){
        stop(
          "Some elements of `", var.n, "` have length greater than 1, but should be atomic.",
          call. = FALSE
        )
      }
    }


  }

  base <- paste0("hespdiv(data = obj$call.info$Call_ARGS$data,
                n.split.pts = obj$call.info$Call_ARGS$n.split.pts,
                generalize.f = obj$call.info$Call_ARGS$generalize.f,
                maximize = obj$call.info$Call_ARGS$maximize,
                method = obj$call.info$Call_ARGS$method,
                same.n.split = obj$call.info$Call_ARGS$same.n.split,
                compare.f = obj$call.info$Call_ARGS$compare.f,
                N.crit = obj$call.info$Call_ARGS$N.crit,
                N.rel.crit = obj$call.info$Call_ARGS$N.rel.crit,
                N.loc.crit = obj$call.info$Call_ARGS$N.loc.crit,
                N.loc.rel.crit = obj$call.info$Call_ARGS$N.loc.rel.crit,
                S.crit = obj$call.info$Call_ARGS$S.crit,
                S.rel.crit = obj$call.info$Call_ARGS$S.rel.crit,
                Q.crit = obj$call.info$Call_ARGS$Q.crit,
                c.splits = obj$call.info$Call_ARGS$c.splits,
                c.Q.crit = obj$call.info$Call_ARGS$c.Q.crit,
                c.crit.improv = obj$call.info$Call_ARGS$c.crit.improv,
                c.X.knots = obj$call.info$Call_ARGS$c.X.knots,
                c.Y.knots = obj$call.info$Call_ARGS$c.Y.knots,
                xy.dat = obj$call.info$Call_ARGS$xy.dat,
                c.max.iter.no = obj$call.info$Call_ARGS$c.max.iter.no,
                c.fast.optim = obj$call.info$Call_ARGS$c.fast.optim,
                c.corr.term = obj$call.info$Call_ARGS$c.corr.term,
                study.pol = obj$call.info$Call_ARGS$study.pol,
                use.chull = obj$call.info$Call_ARGS$use.chull,
                tracing = NULL,
                pnts.col = ", pnts.col,",
                display = ", display,",
                pacific.region = obj$call.info$Call_ARGS$pacific.region)")

  if (any(c("data","xy.dat", "study.pol", "compare.f") %in% hes.names)){
    if ("compare.f" %in% hes.names){

      methods_ids <- which(hes.names %in% c("compare.f","generalize.f",
                                            "maximize"))
      name.list <- hes.mods
      name.list$method <- c(name.list$method,
                            paste0("Custom_method_",
                                   1:length(hes.mods[["compare.f"]])))
      name.list <- name.list[-methods_ids]
    } else {
      name.list <- hes.mods
    }
    if (any(c("data","xy.dat") %in% hes.names)) {
      if (paired) {
        name.list$pdataxy <- paste0("Dataset_",
                                    1:length(hes.mods[["data"]]))
        name.list <- name.list[-which(names(name.list) %in% c("data","xy.dat"))]
        if ("study.pol" %in% hes.names){
          name.list[["study.pol"]] <- as.list(paste0("study.pol_",
                                                     1:length(hes.mods[["study.pol"]])))
        }
      } else {
        df_ids <- which(hes.names %in% c("data","xy.dat","study.pol"))
        for (i in df_ids){
          name.part <- names(hes.mods[i])
          name.list[[i]] <- as.list(paste0(name.part,"_",
                                           1:length(hes.mods[[i]])))
        }
      }
    }
    vec.par.vals <- lapply(name.list,unlist)
  } else {
    vec.par.vals <- lapply(hes.mods,unlist)
  }


  if (!comb.args){
    model.names <- unlist(lapply(1:length(vec.par.vals),
                                 function(i) paste0(names(vec.par.vals[i]),
                                                    " = ",
                                                    vec.par.vals[[i]] )))
  } else {
    N <- ifelse(!"compare.f" %in% hes.names,length(hes.names),ifelse(
      !"method" %in% hes.names,length(hes.names)-2,length(hes.names)-3))
    N <- ifelse(paired, N-1, N)
    argum <- lapply(1:N, function(i) paste0(names(vec.par.vals[i]), " = ",
                                            vec.par.vals[[i]] ))
    if (is.null(pick.n.args)){
      pick.n.args <- 1:N
    } else {
      if (max(pick.n.args) > N){
        stop(
          "Some provided values of `pick.n.args` are higher than the possible maximum: ",
          N,
          call. = FALSE
        )
      }
    }
    arg.combs <- vector( mode = "list", length = length(pick.n.args))
    for ( i in seq(length(pick.n.args))){
      arg.combs[[i]] <- Reduce(rbind,
                               lapply(combn(1:N, m = sort(pick.n.args)[i],
                                            simplify = FALSE),
                                      function(id) expand.grid(argum[id])))
    }
    model.names <- unlist(lapply(arg.combs, apply, 1, paste, collapse = " & "))
    if (comb.type == "random") {
      if (n.combs >= length(model.names)){
        message(
          "The requested number of argument combinations (`n.combs = ", n.combs,
          "`) is higher than the number of available argument combinations (",
          length(model.names), ").\n",
          "Do you wish to proceed and test all available combinations?\n",
          "Choose: Yes or No"
        )
        answer <- readline(prompt = "")
        answer <- .answer_check(given = answer,NAMES = c("Yes", "No"))
        if (answer == 'No'){
          return(NULL)
        } else {
          comb.type <- "all"
          n.combs <- NULL
        }
      } else {
        model.names <- sample(model.names, n.combs, replace = FALSE)
      }
    } else {
      if (comb.type == "handpicked"){
        message(paste(model.names, collapse = "\n"))

        message(
          "\nFrom the vector printed above, choose the argument combinations ",
          "you would like to test.\n",
          "Then type a call that creates a vector of their indices, ",
          "e.g. c(43:55, 67:68, 101)."
        )
        cond <- TRUE
        while(cond) {
          modinds <- readline(prompt = "")
          modinds <- tryCatch({eval(parse(text = modinds))},
                              error = function(cond) {
                                message(cond)
                                message("\nRetype the call.")
                                return(NULL)
                              },
                              warning = function(cond) {
                                message(cond)
                                message("\nRetype the call.")
                                return(NULL)
                              })
          if (!is.null(modinds)){
            if (
              is.numeric(modinds) &&
              length(modinds) > 0 &&
              is.vector(modinds) &&
              !is.list(modinds) &&
              all(!is.na(modinds)) &&
              all(!is.nan(modinds))
            ) {
              selection <- tryCatch(
                {
                  model.names[modinds]
                },
                error = function(cond) {
                  message("Use of provided indices produced an error.")
                  message(conditionMessage(cond))
                  message("\nRetype the call.")
                  NULL
                },
                warning = function(cond) {
                  message("Use of provided indices issued a warning.")
                  message(conditionMessage(cond))
                  message("\nRetype the call.")
                  NULL
                }
              )

              if (!is.null(selection)) {
                if (any(is.nan(selection) | is.na(selection))) {
                  message("Use of provided indices produced NA or NaN values.")
                  message(paste(selection, collapse = "\n"))
                  message("Retype the call.")
                } else {
                  cond <- FALSE
                }
              }
            } else {
              message(
                "Call is invalid. The output is not a numeric integer vector. ",
                "Its values are:"
              )
              message(paste(modinds, collapse = "\n"))
              message("Retype the call.")
            }
          }
        }

        model.names <- selection

      }
    }
  }
  model.names <- .clean_names(model.names, obj, hes.names, c.splits, c.pars)
  message("Changes to be made in the base hespdiv call: ")
  message(paste(model.names, collapse = "\n"))

  l <- length(model.names)
  mods <- vector(mode = "list", length = l)
  names(mods) <- model.names

  for (mod.id in 1:l){
    message(
      "Calling `hespdiv()` alternative: [", mod.id, "] ",
      model.names[mod.id]
    )
    filt.names <- unlist(
      strsplit(gsub("= | &", "",  model.names[mod.id]), split = " ",
               perl = TRUE))
    f.id <- 1:length(filt.names)
    arg.names <- filt.names[f.id %% 2 == 1]
    arg.vals <- filt.names[f.id %% 2 == 0]
    mod.base <- base
    if (any(grepl("Custom_method_.*", arg.vals, fixed = F))){
      del.id <- which(grepl("Custom_method_", arg.vals, fixed = TRUE))
      vals.id <- as.numeric(strsplit(arg.vals[del.id], split = "_")[[1]][3])
      arg.names <- c(arg.names[-del.id],"compare.f","generalize.f",
                     "maximize")
      arg.vals <- c(arg.vals[-del.id], rep(vals.id,3))
      mod.base <- sub('obj\\$call.info\\$Call_ARGS\\$method',"NULL",
                      mod.base)
    } else {
      if (any(arg.names == "method")) {
        args <- c("compare.f","generalize.f", "maximize")
        for (arg in args)
          mod.base <- sub(paste0('obj\\$call.info\\$Call_ARGS\\$',arg),
                          "NULL",
                          mod.base)
      }
    }
    if ("pdataxy" %in% arg.names) {
      del.id <- which(arg.names == "pdataxy")
      vals.id <- as.numeric(strsplit(arg.vals[del.id], split = "_")[[1]][2])
      arg.names <- c(arg.names[-del.id],"data","xy.dat")
      arg.vals <- c(arg.vals[-del.id], rep(vals.id,2))
    }
    for (arg.id in 1:length(arg.names)){
      if (arg.names[arg.id] == "study.pol" |
          (arg.names[arg.id] %in%  c("data", "xy.dat") & !paired ) ){
        val.id <- as.numeric(strsplit(arg.vals[arg.id], split = "_")[[1]][2])
        arg.name <- arg.names[arg.id]
      } else {
        arg.name <- arg.names[arg.id]
        val.id <- ifelse(!arg.name %in% c("compare.f","generalize.f",
                                          "maximize", "data", "xy.dat"),
                         yes =
                           which(as.character(unlist(hes.mods[arg.name])) ==
                                   arg.vals[arg.id]),
                         no = as.numeric(arg.vals[arg.id]))
      }
      mod.base <- sub(paste0('obj\\$call.info\\$Call_ARGS\\$',
                             arg.name),
                      paste0("hes.mods[['",arg.name,"']][[",
                             val.id,"]]"),
                      mod.base)
    }
    if (!is.null(images.path))
      png(paste0(images.path,"\\",paste0(gsub(pattern = "&", replacement = "_",
                                              x = gsub(pattern = " ",
                                                       replacement = "",
                                                       x =  model.names[mod.id]))
                                         ,"_",mod.id,".png")))
    mods[[mod.id]] <- list(Subdivision = tryCatch({eval(parse(text = mod.base))},
                                                  error = function(cond) {
                                                    base::message(paste0(cond,"\n"))
                                                    return(cond)
                                                  },
                                                  warning = function(cond) {
                                                    base::message(paste0(cond,"\n"))
                                                    return(list(Subdivision =
                                                                  eval(parse(text = mod.base)),
                                                                Warning = cond))
                                                  }))#, Changes = model.names[mod.id]) removed so that hsa and hsa_detailed would be similar
    if (!is.null(images.path))
      dev.off()
  }
  structure(list(Alternatives = mods, Basis = obj), class = "hsa")
}
#' @noRd
.to.list <- function(x){
  if(!is.null(x)){
    if (is.list(x)){
      x <- unlist(x)
    }
    if (length(x) > 1){
      stop(
        "Alternative Boolean `hespdiv` arguments should each contain only one value.",
        call. = FALSE
      )
    }
    as.list(x)
  }
}
#' @noRd
.clean_names <- function(model.names, obj, hes.names, c.splits, c.pars){
  if("study.pol" %in% hes.names){
    ids <- which(
      (grepl("study.pol", model.names, fixed = TRUE) &
         grepl("use.chull = TRUE", model.names, fixed = TRUE)) |
        (grepl("study.pol", model.names, fixed = TRUE) &
           obj$call.info$Call_ARGS$use.chull &
           !grepl("use.chull = FALSE", model.names, fixed = TRUE))
    )
    if (length(ids) > 0)
      model.names <- model.names[-ids]
  }
  if (c.pars & !is.null(c.splits)){
    if (c.splits[[1]]){
      ids <- which(
        grepl("c.splits", model.names, fixed = TRUE) |
          !grepl("^c\\.| c\\.", model.names)
      )
      if (length(ids) > 0)
        model.names <- model.names[ids]
    } else {
      ids <- which(
        (!grepl("c.splits", model.names, fixed = TRUE) &
           grepl("^c\\.| c\\.", model.names)) |
          (grepl("c.splits", model.names, fixed = TRUE) &
             !grepl("^c\\.| c\\.",
                    sub(pattern = "c.splits", replacement = "",
                        x = model.names))) |
          !grepl("^c\\.| c\\.", model.names)
      )
      if (length(ids) > 0)
        model.names <- model.names[ids]
    }
  }
  model.names
}

#' @noRd
.answer_check <- function(given,NAMES){
  matched.i <- pmatch(tolower(given), tolower(NAMES))
  while(is.na(matched.i)){
    message("Invalid input: ", paste0('"', given,'".'),
            paste0("\nPlease select viable option: "),
            paste(NAMES,collapse = " or ",sep = "'"))
    given <- readline(prompt = "")
    matched.i <- pmatch(tolower(given), tolower(NAMES))
  }
  NAMES[matched.i]
}

#' Example hespdiv object
#'
#' A small \code{hespdiv} object created from simulated occurrence data with a
#' known boundary at \code{x = 0.45}. The object is intended for examples,
#' tests, and demonstrations of plotting and post-processing functions.
#'
#' @format A \code{hespdiv} object.
#'
#' @details
#' The object was generated using simulated coordinates and taxon names.
#' Occurrences on the left and right sides of the artificial boundary share
#' several common taxa but also include side-specific taxa.
#' The exact code used to simulate the dataset and run \code{hespdiv} is also
#' demonstrated in the examples of \code{hespdiv} function.
#'
#' @examples
#' data(example_hespdiv)
#' plot_hespdiv(example_hespdiv)
#'
#' @name example_hespdiv
#' @docType data
#' @keywords datasets
NULL

#' Split a polygon
#'
#' This function splits a polygon in two halfs by a split line that links two provided polygon vertices and returns one half of this polygon.
#'
#' @param polygon A data frame containing coordinates of a polygon that will be split. The polygon can be either closed or open.
#' @param split_ids A vector containing 2 integers. These integers should be indexes of polygon data frame rows that contain coordinates
#' of vertices that are used to split a polygon.
#' @param min_id (Optional) Integer (1 or 2) showing which of the polygon vertexes indicated by the \code{split_ids} has lower x coordinate
#' (or lower y coordinate, if split line is vertical).
#' @param trivial_side A logical argument indicating whether it matters which side of the divided polygon will be returned.
#' @param poli_side A logical argument. When \code{trivial_side} is \code{FALSE}, \code{poli_side = TRUE} returns the upper half
#' (or the left half, if split line is vertical) of a divided polygon and \code{poli_side = FALSE} returns the lower (the right) half.
#' @return A data frame containing coordinates of one half of a polygon divided by the split line. The resulting polygon is open.
#' @details  When \code{trivial_side = TRUE}, computations are much faster. The duration of computations also depend on the complexity of
#' polygons - they are longer when none of the sides of the divided polygon has all vertices on one side of the split line. The computations
#' are also faster, when split line is horizontal or vertical. The use of \code{min_id} only marginally reduces the amount of calculations.
#' @author Liudas Daumantas
#' @examples #Creating a data.frame of an irregular polygon
#' poly<-data.frame(X=c(0,-1,12,0,5,12,12,7,-5,
#' 11,-3,10,-10,-10,15,15,-1,13,13,-7,-7,0),
#' Y=c(0,5,10,4,4,3,-3,-10,-11,-4,-4,0,-2,20,20,-20,-20,-15,15,15,2,0))
#' plot(poly,type='o',main='Horizontal Split Line')
#' #Horizontal Split Line
#' lines(poly[c(1,12),1],poly[c(1,12),2],lty='dotted')
#' #Dividing a polygon
#' upper_half <- .split_poly(polygon = poly, split_ids = c(1,12), min_id = 1,
#' trivial_side = FALSE, poli_side = TRUE)
#' lower_half <- .split_poly(polygon = poly, split_ids = c(1,12), min_id = 1,
#' trivial_side = FALSE, poli_side = FALSE)
#' lines(upper_half,col=4)
#' lines(lower_half,col=2)
#'
#'
#' plot(poly,type='o',main='Vertical Split Line')
#' #Vertical Split Line
#' lines(poly[c(1,4),1],poly[c(1,4),2],lty='dotted')
#' #Dividing a polygon
#' left_half <- .split_poly(polygon = poly, split_ids = c(1,4), min_id = 1,
#' trivial_side = FALSE, poli_side = TRUE)
#' right_half <- .split_poly(polygon = poly, split_ids = c(1,4), min_id = 1,
#' trivial_side = FALSE, poli_side = FALSE)
#' lines(left_half,col=4)
#' lines(right_half,col=2)
#'
#' plot(poly,type='o',main='Inclined Split Line')
#' #Inclined Split Line
#' lines(poly[c(1,6),1],poly[c(1,6),2],lty='dotted')
#' #Dividing a polygon
#' upper_half <- .split_poly(polygon = poly, split_ids = c(1,6), min_id = 1,
#' trivial_side = FALSE, poli_side = TRUE)
#' lower_half <- .split_poly(polygon = poly, split_ids = c(1,6), min_id = 1,
#'  trivial_side = FALSE, poli_side = FALSE)
#' lines(upper_half,col=4)
#' lines(lower_half,col=2)
#' @noRd

.split_poly <- function(polygon,split_ids,min_id=NULL, trivial_side=TRUE, poli_side){
  if (is.null(min_id)){
    if (polygon[split_ids[1],1]==polygon[split_ids[2],1]){
      min_id<-which.min(c(polygon[split_ids[1],2],polygon[split_ids[2],2]))
    } else {
      min_id<-which.min(c(polygon[split_ids[1],1],polygon[split_ids[2],1]))
    }
  }
  min_x_id<-split_ids[min_id]
  max_x_id<-split_ids[-min_id]
  ids<-min_x_id:max_x_id
  if (any(polygon[1,]!=polygon[nrow(polygon),])){
    polygon<-.close_poly(polygon)
  }
  if (trivial_side==T){
    case <- TRUE
  } else {
    if (polygon[min_x_id,2]==polygon[max_x_id,2]){
      zero <- polygon[min_x_id,2]
      if ((all(polygon[ids,2]>=zero) | all(polygon[ids,2]<=zero)) | (all(polygon[-ids,2]>=zero) | all(polygon[-ids,2]<=zero))) {
        if (all(polygon[ids,2]>=zero) | all(polygon[ids,2]<=zero)){
          case<-all(polygon[ids,2]>=zero)
        } else {
          case<-all(polygon[-ids,2]<=zero)
        }
      } else {
        pol<-polygon[ids,]
        direction<-numeric()
        side<-numeric()
        mid_x<-polygon[max_x_id,1]/2
        i=1
        while (i < nrow(pol)){
          if (pol[i,1]>polygon[min_x_id,1] & pol[i,1] < polygon[max_x_id,1] & i>1){
            start=i-1
            i=i+1
            while (pol[i,1]>polygon[min_x_id,1] & pol[i,1] < polygon[max_x_id,1] & i < nrow(pol)){
              i=i+1
            }
            end=i
            i=i-1
            if (pol[start,1]<=polygon[min_x_id,1] & pol[end,1]>=polygon[max_x_id,1]) {
              direction <- c(direction,1)
            } else {
              if (pol[start,1]>=polygon[max_x_id,1] & pol[end,1]<=polygon[min_x_id,1]){
                direction <- c(direction,-1)
              } else {direction <- c(direction,0)}
            }

            side <- c(side,pol[i,2]>zero)
            i=end
          } else {
            if ((pol[i,1]<=polygon[min_x_id,1] | pol[i+1,1]<=polygon[min_x_id,1]) &
                (pol[i,1]>=polygon[max_x_id,1] | pol[i+1,1]>=polygon[max_x_id,1])){
              side <- c(side,.pt_on_line(x1=pol[i,1],x2=pol[i+1,1],y1=pol[i,2],
                                        y2=pol[i+1,2],x3=mid_x)>zero)
              if (pol[i,1]<pol[i+1,1]){
                direction<-c(direction, 1)
              } else{
                direction<-c(direction, -1)
              }
              i=i+1
            } else{
              i=i+1
            }
          }
        }
        case<-sum(direction[side==1])==1
      }
    } else {
      if (polygon[min_x_id,1]==polygon[max_x_id,1]){
        zero <- polygon[min_x_id,1]
        if ((all(polygon[ids,1]>=zero) | all(polygon[ids,1]<=zero)) | (all(polygon[-ids,1]>=zero) | all(polygon[-ids,1]<=zero))){
          if (all(polygon[ids,1]>=zero) | all(polygon[ids,1]<=zero)){
            case<-all(polygon[ids,1]<=zero)
          } else {
            case<-all(polygon[-ids,1]>=zero)
          }
        } else {
          pol<-polygon[ids,]
          direction<-numeric()
          side<-numeric()
          mid_y<-max(polygon[c(max_x_id,min_x_id),2])/2
          i=1
          while (i < nrow(pol)){
            if (pol[i,2]>polygon[min_x_id,2] & pol[i,2] < polygon[max_x_id,2] & i>1){
              start=i-1
              i=i+1
              while (pol[i,2]>polygon[min_x_id,2] & pol[i,2] < polygon[max_x_id,2]){
                i=i+1
              }
              end=i
              i=i-1
              if (pol[start,2]<=polygon[min_x_id,2] & pol[end,2]>=polygon[max_x_id,2]) {
                direction <- c(direction,1)
              } else {
                if (pol[start,2]>=polygon[max_x_id,2] & pol[end,2]<=polygon[min_x_id,2]){
                  direction <- c(direction,-1)
                } else {direction <- c(direction,0)}
              }

              side <- c(side,pol[i,1]<zero)
              i=end
            } else {
              if ((pol[i,2]<=polygon[min_x_id,2] | pol[i+1,2]<=polygon[min_x_id,2]) &
                  (pol[i,2]>=polygon[max_x_id,2] | pol[i+1,2]>=polygon[max_x_id,2])){
                side <- c(side,.pt_on_line(x1=pol[i,1],x2=pol[i+1,1],y1=pol[i,2],y2=pol[i+1,2],y3=mid_y)<zero)
                if (pol[i,2]<pol[i+1,2]){
                  direction<-c(direction, 1)
                } else{
                  direction<-c(direction, -1)
                }
                i=i+1
              } else{
                i=i+1
              }
            }
          }
          case<-sum(direction[side==1])==1
        }
      }  else {
        b<-(polygon[max_x_id,2]-polygon[min_x_id,2])/(polygon[max_x_id,1]-polygon[min_x_id,1])
        b1<-tan((atan(b)*180/pi+90)*pi/180)
        y_pts_on_line<-polygon[min_x_id,2]+b*polygon[ids,1]
        y2_pts_on_line<-polygon[min_x_id,2]+b*polygon[-ids,1]
        if ((all(polygon[ids,2]>=y_pts_on_line) | all(polygon[ids,2]<=y_pts_on_line)) |
            (all(polygon[-ids,2]>=y2_pts_on_line) | all(polygon[-ids,2]<=y2_pts_on_line)) ) {
          if (all(polygon[ids,2]>=y_pts_on_line) | all(polygon[ids,2]<=y_pts_on_line)) {
            case <- all(polygon[ids,2]>=y_pts_on_line)
          } else {
            case <- all(polygon[-ids,2]<=y2_pts_on_line)
          }
        } else {
          pol<-polygon[ids,]
          direction<-numeric()
          side<-numeric()
          mid_x<-polygon[min_x_id,1]+(polygon[max_x_id,1]-polygon[min_x_id,1])/2
          mid_y<-polygon[min_x_id,2]+b*mid_x
          i=1
          y_b1_pts<-polygon[min_x_id,2]+b1*pol[,1]
          y_b2_pts<-polygon[max_x_id,2]+b1*pol[,1]
          while (i < nrow(pol)){
            if (pol[i,2]>y_b1_pts[i] & pol[i,2] < y_b2_pts[i] & i>1){
              start=i-1
              i=i+1
              while (pol[i,2]>y_b1_pts[i] & pol[i,2] < y_b2_pts[i] & i < nrow(pol)){
                i=i+1
              }
              end=i
              i=i-1
              if (pol[start,2]<=y_b1_pts[i] & pol[end,2]>=y_b2_pts[i]) {
                direction <- c(direction,1)
              } else {
                if (pol[start,2]>=y_b2_pts[i] & pol[end,2]<=y_b1_pts[i]){
                  direction <- c(direction,-1)
                } else {direction <- c(direction,0)}
              }

              side <- c(side,pol[i,2]>y_pts_on_line[i])
              i=end
            } else {
              if ((pol[i,2]<=y_b1_pts[i] | pol[i+1,2]<=y_b1_pts[i+1]) &
                  (pol[i,2]>=y_b2_pts[i] | pol[i+1,2]>=y_b2_pts[i+1])){
                side <- c(side,.pt_on_line(x1=pol[i,1],x2=pol[i+1,1],y1=pol[i,2],y2=pol[i+1,2],x3=mid_x)>mid_y)
                if (pol[i,2]<=y_b1_pts[i]){
                  direction<-c(direction, 1)
                } else{
                  direction<-c(direction, -1)
                }
                i=i+1
              } else{
                i=i+1
              }
            }
          }
          case<-sum(direction[side==1])==1
        }
      }
    }
  }
  if (case){
    if (poli_side==T){
      return(polygon[ids,])
    } else {
      polygon<-cbind(polygon,1:nrow(polygon))
      polygon<-polygon[-ids[-c(1,length(ids))],]
      polygon<-polygon[-nrow(polygon),]
    }
  } else {
    if (poli_side==F){
      return(polygon[ids,])
    } else {
      polygon<-cbind(polygon,1:nrow(polygon))
      polygon<-polygon[-ids[-c(1,length(ids))],]
      polygon<-polygon[-nrow(polygon),]
    }
  }
  if (min_x_id<max_x_id){
    return(polygon[match(c(min_x_id:1,polygon[nrow(polygon),3]:max_x_id),polygon[,3]),1:2])
  } else {
    return(polygon[match(c(min_x_id:polygon[nrow(polygon),3],1:max_x_id),polygon[,3]),1:2])
  }
}

#' Draw hespdiv polygons in 3D space
#'
#' This function visualizes HespDiv polygons in 3D space. The height axis
#' corresponds to a chosen column from the "poly.stats" data frame.
#'
#' @param obj An object of the hespdiv class
#' @param height A character vector with default value \code{"mean"}. Determines
#'   which information from the \code{poly.stats} data frame is encoded as the
#'   height of the polygons. Options are \code{"mean"}, \code{"sd"},
#'   \code{"best"}, \code{"z.score"}, \code{"str.best"},
#'   \code{"str.z.score"}, and \code{"rank"}. Multiple values are allowed.
#' @param color.seed An integer that controls the colors of the polygons. Change
#' it to a different number if you want to get a different set of colors.
#' @param lines A Boolean value. Do you want split-lines to be displayed over
#' the top of the polygons?
#' @param obs A Boolean value. Do you want observations to be displayed over the top of
#' the polygons?
#' @param pnts.col A character or numeric vector. Color codes to be used for
#' displaying observations.
#' @return No return value. Called for the side effect of producing a 3D plot
#'   using the \pkg{rgl} graphics engine.
#' @details
#' The function opens an \pkg{rgl} device for each selected column of the
#' \code{poly.stats} data frame.
#'
#' Visualizing values from \code{poly.stats} as polygon height can provide
#' insight into spatial heterogeneity in the analyzed data and its hierarchical
#' spatial structure.
#'
#' The \code{height = "rank"} option provides a more intuitive way to understand
#' the position of each polygon in the spatial hierarchy than
#' \code{poly_scheme()}.
#'
#' Because higher-rank polygons are displayed above lower-rank polygons, they may
#' obscure the view. For this reason, \code{polypop(obj, height)} can be used
#' with the same arguments to interactively select unwanted polygons and remove
#' them from the actives plot.
#' @importFrom rgl open3d plot3d lines3d polygon3d persp3d points3d text3d
#' @importFrom pracma poly_center
#' @family HespDiv visualization options
#' @author Liudas Daumantas
#' @export
blok3d <- function(obj,height = "mean", color.seed=1, lines=TRUE, pnts.col = NULL,
                   obs = TRUE) {
  height <- as.vector(sapply(height, .arg_check, name = "height", NAMES = c("mean","sd","best","z.score",
                                                                            "str.best", "str.z.score","rank")))
  for (height in height){
    rgl::open3d()
    poly.stats <- obj$poly.stats
    if (obj$call.info$Call_ARGS$c.splits){
      if (height == "best"){
        poly.stats$best <- ifelse(is.na(obj$poly.stats$is.curve),
                                  obj$poly.stats$str.best,
                                  ifelse(obj$poly.stats$is.curve,
                                         obj$poly.stats$crv.best,
                                         obj$poly.stats$str.best))
      } else {
        if (height == "z.score" ){
          poly.stats$z.score <- ifelse(is.na(obj$poly.stats$is.curve),
                                       obj$poly.stats$str.z.score,
                                       ifelse(obj$poly.stats$is.curve,
                                              obj$poly.stats$crv.z.score,
                                              obj$poly.stats$str.z.score))
        }
      }
    } else {
      poly.stats$best <- obj$poly.stats$str.best
      poly.stats$z.score <- obj$poly.stats$str.z.score
    }
    xy.dat <- obj$call.info$Call_ARGS$xy.dat

    if (height == "rank") {
      ZZ <- data.frame(zmin = obj$poly.stats$rank-1, zmax = obj$poly.stats$rank)
      del.id <- numeric()
      Zoff <- 0.2
    } else {
      ZZ <- .Zcoords(poly.stats,height)
      del.id <- which(ZZ[,1]==ZZ[,2])
      Zoff <- mean(apply(ZZ[-del.id,], 1, function(x){x[2]-x[1]}))/5
    }

    if (height %in% c('z.score','str.z.score')){
      highest.z <- min(ZZ,na.rm = T)
    } else {
      highest.z <- max(ZZ,na.rm = T)
    }
    if (is.null(pnts.col)){
      pnts.col <- 1
    }
    # ZZ - is data.frame of z coordinates of blocks. Col 1 is z cord of the bottom
    # and coDl 2 is z cord of the ceiling. Rows correspond to polygons in
    # obj$poly.stats


    #NUPIESIAM RIBAS
    #sukuriam erdve, kurioj piesim blokus - min, max koordinates padedam, nematomi taskiukai plius
    #rgl.material(alpha = 0)
    #spalvu palete sukuriam
    n <- nrow(ZZ)-length(del.id)
    palete <- .generate_cols(n, seed=color.seed)


    # jei uzduoti duomenys, nupaisom erdves pavirsiuje duomenis, spalvu paletes paskutiniai elementai - rusiu spalvos
    if (obs){
      alfa <- 1
    } else {
      alfa <- 0
    }


    rgl::plot3d(x = xy.dat$x, y = xy.dat$y, z = highest.z,
                col = pnts.col , alpha = alfa, add = F,
                ylim = range(obj$polygons.xy$`1`$y),
                xlim = range(obj$polygons.xy$`1`$x),
                zlim = range(unlist(c(ZZ)),
                             na.rm = T),
                zlab = height, xlab = "x", ylab = "y")
    #
    #paisom kiekvieno bloko pavirsius
    if (length(del.id) == 0){
      IDs <- seq(nrow(ZZ))
    } else {
      IDs <- seq(nrow(ZZ))[-del.id]
    }
    for (i in seq(n)) {
      .draw_poly(obj,IDs[i],color=palete[i], ZZ, Zoff)
    }
    if (lines==TRUE){
      for (i in seq(nrow(ZZ))){
        rgl::lines3d(x=obj$polygons.xy[[i]]$x,y=obj$polygons.xy[[i]]$y,
                            z=highest.z,col=1)
      }
    }
  }
}


#' Iterative function that provides the id of the closest polygon parent
#' @noRd
.iterator<-function(poly.id,roots,id.list){
  id.list<-c(id.list,roots[poly.id])
  if (roots[poly.id]==0) {return(id.list)
  } else {
    poly.id <- roots[poly.id]
    return(.iterator(poly.id,roots,id.list))
  }
}
#' Return the ids of parent polygons for a given polygon id.
#' @noRd
.collector<-function(poly.id,roots){
  id.list <- roots[poly.id]
  if (id.list == 0) {return(0)
  } else {
    poly.id <- roots[poly.id]
  }
  rezas <- .iterator(poly.id,roots,id.list)
  return(rezas[-length(rezas)])
}
#' Return bottom z coordinate for a polygon of given poly.id
#' @noRd
.zmin <- function(poly.id, poly.stats, roots, height){

  ids <- .collector(poly.id = poly.id, roots = roots) # ids surenka iki duoto polio
  zmin <- sum(poly.stats[ids,height])
  return(zmin)
}
#' Obtain top and bottom Z coordinates for each polygon that has non-NA
#' performance scores.
#' @noRd
.Zcoords<-function(poly.stats,height){
  ids.relative <- 1:nrow(poly.stats)
  names(ids.relative) <- poly.stats$plot.id
  roots <- ids.relative[as.character(poly.stats$root.id)]
  roots[1] <- 0
  zmini <- numeric(length(ids.relative)) # bottom z coord. for each poly.
  zmaxi <- numeric(length(ids.relative)) # ceiling z coord. for each poly.
  for (poly.id in ids.relative){
    zmini[poly.id] <- .zmin(poly.id,poly.stats,roots, height)
    zmaxi[poly.id] <- zmini[poly.id]+
      ifelse(is.na(poly.stats[poly.id,height]),0,
             poly.stats[poly.id,height]) # [[3]]
  }
  return(data.frame(zmini,zmaxi))
}
# Add a block of polygon to the rgl device
#' @noRd
.draw_poly<-function(obj,i,color,ZZ, Zoff){
  #bloko koordinates
  if (any(duplicated(obj$polygons.xy[[i]][-nrow(obj$polygons.xy[[i]]),])))
    obj$polygons.xy[[i]] <- obj$polygons.xy[[i]][
      -which(duplicated(obj$polygons.xy[[i]][-nrow(obj$polygons.xy[[i]]),])),]

  x <- obj$polygons.xy[[i]]$x
  y <- obj$polygons.xy[[i]]$y

  zmini <- ZZ[i,1]
  zmaxi <- ZZ[i,2]
  #virsus
  rgl::polygon3d(x, y, z=rep(zmaxi,length(x)), col = color,add=TRUE,fill = TRUE)
  #apacia
  rgl::polygon3d(x, y, z=rep(zmini,length(x)), col =color, add=TRUE, fill = TRUE)
  #sonai
  xmat <- matrix(NA, 2, length(x))
  ymat <- matrix(NA, 2, length(x))
  zmat <- matrix(NA, 2, length(x))
  ZZZ<-c(zmini,zmaxi)
  for (a in 0:1) {
    xmat[a+1,] <- x
    ymat[a+1,] <- y
    zmat[a+1,] <- ZZZ[a+1]
  }
  rgl::persp3d(x=xmat,y=ymat,z=zmat,col=color,add = T)
  #ID

  centras <- pracma::poly_center(x,y)
  rgl::points3d(x=centras[1],y=centras[2],z=zmaxi,col=1,pch=19,labels=i)
  rgl::text3d(x=centras[1],y=centras[2],z=zmaxi+Zoff, text = c(i),
                 justify="left")

}

#' HespDiv Sensitivity Analysis
#'
#' This function performs sensitivity analysis of the \code{hespdiv} method.
#' Starting from a reference \code{hespdiv} object, it generates a specified
#' number of alternative \code{hespdiv} calls by randomly sampling new values
#' for selected arguments from user-provided sets.
#'
#' @param obj A \code{hespdiv} class object.
#'
#' @param n.runs Integer. The number of alternative \code{hespdiv} calls to evaluate.
#'
#' @param parallel Logical. If \code{TRUE}, runs are evaluated in parallel using
#'   \pkg{future.apply}. If \code{FALSE} (default), runs are evaluated sequentially.
#'
#' @param workers Integer. Number of parallel workers (CPU cores) to use when
#'   \code{parallel = TRUE}. At most this many \code{hespdiv} calls will be executed
#'   simultaneously. If \code{NULL}, the number of workers is determined
#'   automatically based on \code{RAM} and \code{load_prop}.
#'
#' @param RAM Integer. Approximate amount of available RAM (in GB) used to limit
#'   the number of parallel workers when \code{parallel = TRUE}.
#'
#' @param load_prop Numeric in (0, 1]. Proportion of available CPU cores to use
#'   when determining the number of parallel workers automatically. Defaults to 0.8.
#'
#' @param chunk_size Integer. Number of runs submitted per batch (chunk) to the
#'   parallel backend when \code{parallel = TRUE}. Chunking does not change how
#'   many runs execute simultaneously (controlled by \code{workers}). Smaller
#'   values reduce the number of queued futures at once (often lowering RAM
#'   overhead). Defaults to \code{workers*2}.
#'
#' @param future_seed Logical. Passed to \code{future.apply::future_lapply()}.
#'   If \code{TRUE}, RNG is managed by \pkg{future} to provide parallel-safe,
#'   reproducible random streams.
#'
#' @param data.paired Logical. Controls whether alternative values of \code{data}
#'   are paired with corresponding elements of \code{xy.dat}.
#'
#' @param display Logical. Controls the value of the \code{display} argument
#'   in each \code{hespdiv} call.
#'
#' @param images.path Character or \code{NULL}. Path to an existing directory
#'   where PNG images of displayed results will be saved. If \code{NULL}
#'   (default), images are not saved.
#'
#' @param pnts.col Value passed to the \code{pnts.col} argument of each
#'   \code{hespdiv} call. Can be provided as atomic vector,  vector or list of
#'   vectors.
#'
#' @param data A list of data objects (matrices, data frames, vectors, lists,
#'   or other supported data structures) used as alternative inputs for
#'   sensitivity analysis.
#'
#' @param xy.dat,study.pol Lists of data frames with two columns: \code{x} and \code{y}.
#'
#' @param same.n.split,c.fast.optim,use.chull,c.splits Logical vectors specifying
#'   alternative values for corresponding \code{hespdiv} arguments.
#'
#' @param n.split.pts,c.max.iter.no,N.crit,N.loc.crit,c.X.knots,c.Y.knots
#'   Integer vectors specifying alternative values for corresponding arguments.
#'
#' @param N.rel.crit,N.loc.rel.crit,S.crit,S.rel.crit Numeric vectors with values
#'   between 0 and 1.
#'
#' @param Q.crit,c.Q.crit,c.crit.improv Numeric vectors specifying alternative
#'   threshold or improvement criteria.
#'
#' @param c.corr.term Numeric vector with values between 0.01 and 0.2.
#'
#' @param generalize.f,compare.f Lists of functions defining custom similarity
#'   or generalization methods.
#'
#' @param maximize Logical vector of the same length as \code{compare.f},
#'   indicating whether the corresponding similarity metric should be maximized.
#'
#' @param method Character vector specifying predefined similarity metrics.
#'
#' @param .run.id Integer. Runs with indices less than or equal to this value
#'   will be skipped. This can be used to resume an interrupted analysis.
#'
#' @return An object of class \code{hsa}, containing:
#' \describe{
#'   \item{\bold{Alternatives}}{A list of alternative \code{hespdiv} results
#'   produced during the sensitivity analysis.}
#'   \item{\bold{Basis}}{The reference \code{hespdiv} object whose arguments
#'   were perturbed.}
#' }
#'
#' @details
#' \subsection{Difference Between \code{"hsa"} and other sensitivity analysis functions}{
#' The \code{hsa_detailed} function evaluates all combinations of provided
#' argument values, resulting in dense sampling of the parameter space at
#' substantial computational cost. In contrast, \code{hsa} samples the parameter
#' space stochastically and is generally more suitable for exploratory or
#' large-scale sensitivity analyses.
#'
#' In \code{hsa_detailed}, alternative argument values are provided as lists,
#' whereas in \code{hsa} they are supplied as vectors or lists depending on
#' the argument.
#'
#'
#' The \code{hsa_sample_constrained} function performs non-recursive \code{hespdiv}
#' runs for each split-line produced based on different data subsamples. Thus,
#' \code{hsa} is more general, as it allows to inspect sensitivity to other arguments.
#' }
#'
#' \subsection{Paired Arguments}{
#' When \code{data.paired = TRUE}, the same index is used to sample elements
#' of \code{data} and \code{xy.dat}, allowing sensitivity analysis across
#' datasets of differing size or composition. When \code{FALSE}, data and
#' coordinates are sampled independently, enabling analyses based on noise
#' addition or spatial shuffling.
#'
#' Arguments defining custom methods (\code{compare.f}, \code{generalize.f},
#' \code{maximize}) are always treated as paired and must therefore have
#' equal lengths.
#' }
#'
#' @note
#' If a particular run produces a warning or an error, the corresponding list
#' element will contain two components. In case of a warning, these are the
#' resulting \code{hespdiv} object and the warning message. In case of an error,
#' they are the arguments used for the call and the error message.
#'
#' @family functions for hespdiv sensitivity analysis
#' @family functions for hespdiv results post-processing
#'
#' @export
hsa <- function(obj,
                n.runs = 100,
                data.paired = TRUE,
                display = FALSE,
                images.path = NULL,
                pnts.col = 1,
                data = NULL,
                xy.dat = NULL,
                same.n.split = NULL,
                n.split.pts = NULL,
                N.crit = NULL,
                N.rel.crit = NULL,
                N.loc.crit = NULL,
                N.loc.rel.crit = NULL,
                S.crit = NULL,
                S.rel.crit = NULL,
                Q.crit = NULL,
                c.splits = NULL,
                c.Q.crit = NULL,
                c.crit.improv = NULL,
                c.X.knots = NULL,
                c.Y.knots = NULL,
                c.max.iter.no = NULL,
                c.fast.optim = NULL,
                c.corr.term = NULL,
                study.pol = NULL,
                use.chull = NULL,
                generalize.f = NULL,
                maximize = NULL,
                method = NULL,
                compare.f = NULL,
                .run.id = NULL,
                parallel = FALSE,
                RAM = NULL,
                load_prop = 0.8,
                chunk_size = workers*2,
                workers = NULL,
                future_seed = TRUE) {
  eval_one <- function(i, d) {

    if (!is.null(images.path)) {
      fn <- file.path(images.path, sprintf("hsa_run_%04d.png", i))
      png(fn)
      on.exit(dev.off(), add = TRUE)
    }

    call_args <- quote(
      hespdiv(
        data = data,
        xy.dat = xy.dat,
        same.n.split = same.n.split,
        n.split.pts = n.split.pts,
        generalize.f = generalize.f,
        maximize = maximize,
        compare.f = compare.f,
        method = method,
        N.crit = N.crit,
        N.rel.crit = N.rel.crit,
        N.loc.crit = N.loc.crit,
        N.loc.rel.crit = N.loc.rel.crit,
        S.crit = S.crit,
        S.rel.crit = S.rel.crit,
        Q.crit = Q.crit,
        c.splits = c.splits,
        c.Q.crit = c.Q.crit,
        c.crit.improv = c.crit.improv,
        c.X.knots = c.X.knots,
        c.Y.knots = c.Y.knots,
        c.max.iter.no = c.max.iter.no,
        c.fast.optim = c.fast.optim,
        c.corr.term = c.corr.term,
        use.chull = use.chull,
        study.pol = study.pol,
        tracing = NULL,
        pnts.col = pnts.col,
        display = display,
        pacific.region = pacific.region
      )
    )

    env <- list2env(c(d, list(display = display, pacific.region = pacific.region)),
                    parent = parent.frame())

    out <- tryCatch(
      eval(call_args, envir = env),
      warning = function(w) {
        list(Subdivision = suppressWarnings(eval(call_args, envir = env)),
             warning = w)
      },
      error = function(e) {
        list(Arguments = as.list(d), error = e)
      }
    )

    list(Subdivison = out, Arguments = as.list(d))
  }

  draw_one <- function(i) {
    p.id <- sample.int(dat_len, 1)

    local_use_m <- use_m
    if (m.compete) local_use_m <- sample(0:1, 1, prob = c(length(maximize), length(method)))

    if (!local_use_m) cm.id <- sample.int(len_cm, 1) else cm.id <- NA_integer_

    v1 <- data[[p.id]]
    v2 <- if (data.paired) xy.dat[[p.id]] else sample(xy.dat, 1)[[1]]
    v3 <- same.n.split[sample.int(length(same.n.split), 1)]
    v4 <- n.split.pts[sample.int(length(n.split.pts), 1)]
    v5 <- .ifelse(!local_use_m, generalize.f[[cm.id]], NULL)
    v6 <- .ifelse(!local_use_m, maximize[cm.id], NULL)
    v7 <- .ifelse(!local_use_m, compare.f[[cm.id]], NULL)
    v8 <- .ifelse(local_use_m, sample(method, 1), NULL)
    v9 <- N.crit[sample.int(length(N.crit), 1)]
    v10 <- N.rel.crit[sample.int(length(N.rel.crit), 1)]
    v11 <- N.loc.crit[sample.int(length(N.loc.crit), 1)]
    v12 <- N.loc.rel.crit[sample.int(length(N.loc.rel.crit), 1)]
    v13 <- S.crit[sample.int(length(S.crit), 1)]
    v14 <- S.rel.crit[sample.int(length(S.rel.crit), 1)]
    v15 <- .ifelse(local_use_m,
                   ifelse(v8 == "pielou",
                          (Q.crit[Q.crit < 1])[
                            sample.int(length(Q.crit[Q.crit < 1]), 1)
                          ],
                          (Q.crit[Q.crit > 0])[
                            sample.int(length(Q.crit[Q.crit > 0]), 1)
                          ]),
                   NULL)
    v16 <- c.splits[sample.int(length(c.splits), 1)]
    is.maxim <- ifelse(local_use_m, v8 == "pielou", v6)

    if (is.null(v15)) {
      v17 <- NULL
    } else {
      v17 <- .ifelse(v16,
                     ifelse(is.maxim, (c.Q.crit[c.Q.crit <= v15])[
                       sample.int(length(c.Q.crit[c.Q.crit <= v15]), 1)
                     ],
                     (c.Q.crit[c.Q.crit >= v15])[
                       sample.int(length(c.Q.crit[c.Q.crit >= v15]), 1)
                     ]),
                     NULL)
    }

    v18 <- .ifelse(v16, c.crit.improv[sample.int(length(c.crit.improv), 1)], NULL)
    v19 <- .ifelse(v16, c.X.knots[sample.int(length(c.X.knots), 1)], NULL)
    v20 <- .ifelse(v16, c.Y.knots[sample.int(length(c.Y.knots), 1)], NULL)
    v21 <- .ifelse(v16, c.max.iter.no[sample.int(length(c.max.iter.no), 1)], NULL)
    v22 <- .ifelse(v16, c.fast.optim[sample.int(length(c.fast.optim), 1)], NULL)
    v23 <- .ifelse(v16, c.corr.term[sample.int(length(c.corr.term), 1)], NULL)
    v24 <- use.chull[sample.int(length(use.chull), 1)]
    v25 <- .ifelse(v24, obj$call$Call_ARGS$study.pol, sample(study.pol, 1)[[1]])
    v26 <- .ifelse(display, .ifelse(is.pnts, pnts.col[[p.id]], pnts.col), NULL)

    list(data=v1, xy.dat=v2, same.n.split=v3, n.split.pts=v4, generalize.f=v5,
         maximize=v6, compare.f=v7, method=v8, N.crit=v9, N.rel.crit=v10,
         N.loc.crit=v11, N.loc.rel.crit=v12, S.crit=v13, S.rel.crit=v14,
         Q.crit=v15, c.splits=v16, c.Q.crit=v17, c.crit.improv=v18,
         c.X.knots=v19, c.Y.knots=v20, c.max.iter.no=v21, c.fast.optim=v22,
         c.corr.term=v23, use.chull=v24, study.pol=v25, pnts.col=v26)
  }

  if (!inherits(obj, "hespdiv")) stop("`obj` must be a `hespdiv` object.", call. = FALSE)

  .check_hsa_arg_structures(
    data = data,
    xy.dat = xy.dat,
    study.pol = study.pol,
    same.n.split = same.n.split,
    c.fast.optim = c.fast.optim,
    use.chull = use.chull,
    c.splits = c.splits,
    n.split.pts = n.split.pts,
    c.max.iter.no = c.max.iter.no,
    N.crit = N.crit,
    N.loc.crit = N.loc.crit,
    c.X.knots = c.X.knots,
    c.Y.knots = c.Y.knots,
    N.rel.crit = N.rel.crit,
    N.loc.rel.crit = N.loc.rel.crit,
    S.crit = S.crit,
    S.rel.crit = S.rel.crit,
    Q.crit = Q.crit,
    c.Q.crit = c.Q.crit,
    c.crit.improv = c.crit.improv,
    c.corr.term = c.corr.term,
    generalize.f = generalize.f,
    compare.f = compare.f,
    maximize = maximize,
    method = method
  )

  if (is.null(.run.id)) .run.id <- -1

  if (!is.numeric(n.runs) || n.runs < 1 || n.runs %% 1 || length(n.runs) != 1)
    stop("`n.runs` must be a positive integer.", call. = FALSE)
  if (!is.numeric(chunk_size) || chunk_size < 1 || chunk_size %% 1 || length(chunk_size) != 1)
    stop("`chunk_size` must be a positive integer.", call. = FALSE)

  if (!is.null(images.path) && !display) stop("Attempting to save empty images.", call. = FALSE)
  if (!is.null(images.path) && !dir.exists(images.path))
    stop("`images.path` must be an existing directory.", call. = FALSE)

  # ---  pairing checks ---
  if (data.paired) {
    if (length(data) != length(xy.dat)) {
      stop(
        "When `data` and `xy.dat` are paired, their lists must have the same length.",
        call. = FALSE
      )
    }
  } else {
    if (!is.null(xy.dat)) xy.len <- length(xy.dat)
  }

  # basal data
  if ((!list(obj$call.info$Call_ARGS$xy.dat) %in% xy.dat |
       !list(obj$call.info$Call_ARGS$data) %in% data)) {
    if (data.paired) {
      xy.dat <- c(list(obj$call.info$Call_ARGS$xy.dat), xy.dat)
      data   <- c(list(obj$call.info$Call_ARGS$data), data)
    } else {
      xy.dat <- .set("xy.dat", obj)
      data   <- .set("data", obj)
    }
  }

  same.n.split <- .set("same.n.split", obj)
  n.split.pts  <- .set("n.split.pts", obj)

  # Custom method bookkeeping
  is.custom <- (!is.null(maximize) | obj$call.info$METHOD$method.type == "custom")

  if (obj$call.info$METHOD$method.type == "custom" &
      (!list(obj$call.info$Call_ARGS$generalize.f) %in% generalize.f |
       !list(obj$call.info$Call_ARGS$compare.f) %in% compare.f |
       !list(obj$call.info$Call_ARGS$maximize) %in% maximize)) {
    generalize.f <- c(list(obj$call.info$Call_ARGS$generalize.f), generalize.f)
    compare.f    <- c(list(obj$call.info$Call_ARGS$compare.f), compare.f)
    maximize     <- c(obj$call.info$Call_ARGS$maximize, maximize)
  }

  if (!is.null(method)) {
    method <- sapply(method, .arg_check, name = "metric",
                     NAMES = names(.get_methods()[["biozonation"]]))
    names(method) <- NULL
  }
  method    <- .set("method", obj)
  is.method <- !is.null(method)

  # Pull defaults/augment vectors
  N.crit       <- .set("N.crit", obj)
  N.rel.crit   <- .set("N.rel.crit", obj)
  N.loc.crit   <- .set("N.loc.crit", obj)
  N.loc.rel.crit <- .set("N.loc.rel.crit", obj)
  S.crit       <- .set("S.crit", obj)
  S.rel.crit   <- .set("S.rel.crit", obj)
  Q.crit       <- .set("Q.crit", obj)
  c.Q.crit     <- .set("c.Q.crit", obj)
  c.splits     <- .set("c.splits", obj)
  c.crit.improv <- .set("c.crit.improv", obj)
  c.X.knots    <- .set("c.X.knots", obj)
  c.Y.knots    <- .set("c.Y.knots", obj)
  c.max.iter.no <- .set("c.max.iter.no", obj)
  c.fast.optim <- .set("c.fast.optim", obj)
  c.corr.term  <- .set("c.corr.term", obj)
  use.chull    <- .set("use.chull", obj)

  if (!is.null(study.pol)) {
    if (!is.null(obj$call$Call_ARGS$study.pol)) {
      if (!list(obj$call$Call_ARGS$study.pol) %in% study.pol) {
        study.pol <- c(list(obj$call$Call_ARGS$study.pol), study.pol)
      }
    }
  } else {
    study.pol <- list(obj$call$Call_ARGS$study.pol)
  }

  pnts.col <- .set("pnts.col", obj)
  is.pnts  <- is.list(pnts.col) & length(pnts.col) > 1L
  pacific.region <- obj$call.info$Call_ARGS$pacific.region

  if (any(!use.chull) & is.null(study.pol) & is.null(obj$call$Call_ARGS$study.pol)) {
    stop(
      "`study.pol` must be provided when it is missing in `obj` and any value of ",
      "`use.chull` is FALSE.",
      call. = FALSE
    )
  }

  m.compete <- ifelse(is.custom & is.method, TRUE, FALSE)
  if (!m.compete) use_m <- ifelse(is.method, TRUE, FALSE)

  if (is.method) {
    if ("pielou" %in% method) {
      if (!any(Q.crit < 1))    Q.crit   <- c(-Inf, Q.crit)
      if (!any(c.Q.crit < 1))  c.Q.crit <- c(-Inf, c.Q.crit)
    } else {
      if (!any(Q.crit > 0))    Q.crit   <- c(+Inf, Q.crit)
      if (!any(c.Q.crit > 0))  c.Q.crit <- c(+Inf, c.Q.crit)
    }
  }

  # ---- Serial or parallel execution shares the same draw logic ----
  len_cm  <- length(compare.f)
  dat_len <- length(data)
  run_ids <- seq_len(n.runs)
  active_runs <- run_ids[run_ids > .run.id]

  # Pre-size results (important for resume/skip)
  hes.res <- vector("list", length = n.runs)
  names(hes.res) <- as.character(run_ids)


  # =========================
  # SERIAL PATH
  # =========================
  if (!parallel) {
    for (i in active_runs) {
      message(paste0("Evaluating call: ", i))
      d <- draw_one(i)
      hes.res[[i]] <- eval_one(i, d)
    }

    check.warns(hes.res, .message = FALSE)
    check.errs(hes.res, .message = FALSE)
    return(structure(list(Alternatives = hes.res, Basis = obj), class = "hsa"))
  }

  # =========================
  # PARALLEL PATH
  # =========================
  # Decide workers
  if (is.null(workers)) {
    num_cores <- future::availableCores()
    if (is.null(load_prop)) load_prop <- 0.8
    if (!is.numeric(load_prop) || load_prop <= 0 || load_prop > 1)
      stop("`load_prop` must be a numeric value in the interval (0, 1].", call. = FALSE)
    safe_cores <- max(1, floor(num_cores * load_prop))

    if (is.null(RAM)) {
      RAM <- Inf
    } else {
      if (!is.numeric(RAM) || length(RAM) != 1 || RAM < 1)
        stop("`RAM` must be a positive integer if not NULL.", call. = FALSE)
    }
    safe_workers <- max(1, min(safe_cores, RAM))
  } else {
    if (!is.numeric(workers) || length(workers) != 1 || workers < 1)
      stop("`workers` must be a positive integer.", call. = FALSE)
    safe_workers <- workers
  }
  message("Using ", safe_workers, " parallel workers.")

  future::plan(future::multisession, workers = safe_workers)
  options(future.globals.maxSize = 1024^3 * 2)

  # Precompute draws for all runs so workers only evaluate
  draws <- vector("list", length = n.runs)
  for (i in active_runs) draws[[i]] <- draw_one(i)

  # Chunk submission (queue-size control)
  run_chunks <- split(active_runs, ceiling(seq_along(active_runs) / chunk_size))

  res_active <- unlist(
    lapply(run_chunks, function(chunk) {
      future.apply::future_lapply(
        chunk,
        function(i) eval_one(i, draws[[i]]),
        future.seed = future_seed
      )
    }),
    recursive = FALSE
  )

  # Write results back (preserve indexing)
  hes.res[active_runs] <- res_active

  check.warns(hes.res, .message = FALSE)
  check.errs(hes.res, .message = FALSE)

  structure(list(Alternatives = hes.res, Basis = obj), class = "hsa")
}


#' @noRd
.ifelse <- function(test,yes,no){
  if (test){
    yes
  } else  {
    no
  }
}
#' @noRd
check.warns <- function(hes.res, .message = TRUE){
  ids <- which(unlist(lapply(hes.res,function(o)length(o)==2)))

  if (length(ids) > 0) {
    ids <- ids[which(unlist(lapply(hes.res[ids],function(o) names(o)[2] ==
                                     "warning")))]
    if (length(ids) > 0) {
      message(
        "Warnings were detected in calls: ",
        paste(as.character(ids), collapse = ", ")
      )
    } else {
      message("No warnings detected.")
      return(NULL)
    }
  }else {
    message("No warnings detected.")
    return(NULL)
  }
  if (.message) sapply(ids, function(o) hes.res[[o]][[2]]$message)
}

#' @noRd
check.errs <- function(hes.res, .message = TRUE){
  ids <- which(unlist(lapply(hes.res,function(o)length(o)==2)))

  if (length(ids) > 0) {
    ids <- ids[which(unlist(lapply(hes.res[ids],function(o) names(o)[2] ==
                                     "error")))]
    if (length(ids) > 0) {
      message(
        "Errors were detected in calls: ",
        paste(as.character(ids), collapse = ", ")
      )
    } else {
      message("No errors detected.")
      return(NULL)
    }
  }else {
    message("No errors detected.")
    return(NULL)
  }
  if (.message) sapply(ids, function(o) hes.res[[o]][[2]]$message)
}
#' @noRd
.set <- function(name, obi){
  x <- get(name,envir = parent.frame())
  if (is.null(x)){
    x <- eval(parse( text = paste0("obi$call.info$Call_ARGS$",name)))
  } else {
    if (is.list(x)){
      y <- list(eval(parse( text = paste0("obi$call.info$Call_ARGS$",name))))
    } else {
      y <- eval(parse( text = paste0("obi$call.info$Call_ARGS$",name)))
    }
    if (!y %in% x) {
      x <- c(y, x)
    }
  }
  x
}

#' @noRd
.check_hsa_arg_structures <- function(
    data = NULL,
    xy.dat = NULL, study.pol = NULL,
    same.n.split = NULL, c.fast.optim = NULL, use.chull = NULL, c.splits = NULL,
    n.split.pts = NULL, c.max.iter.no = NULL, N.crit = NULL, N.loc.crit = NULL,
    c.X.knots = NULL, c.Y.knots = NULL,
    N.rel.crit = NULL, N.loc.rel.crit = NULL, S.crit = NULL, S.rel.crit = NULL,
    Q.crit = NULL, c.Q.crit = NULL, c.crit.improv = NULL,
    c.corr.term = NULL,
    generalize.f = NULL, compare.f = NULL,
    maximize = NULL,
    method = NULL
) {
  .bad <- function(name, msg) {
    stop(paste0("`", name, "`: ", msg), call. = FALSE)
  }

  .is_intish <- function(x) is.numeric(x) && all(is.finite(x)) && all(x %% 1 == 0)

  .chk_list <- function(x, name) {
    if (is.null(x)) return(invisible(TRUE))
    if (!is.list(x)) .bad(name, "must be a list or NULL.")
    invisible(TRUE)
  }

  .chk_vec_type <- function(x, name, type = c("logical","numeric","integer","character")) {
    if (is.null(x)) return(invisible(TRUE))
    type <- match.arg(type)
    ok <- switch(
      type,
      logical   = is.logical(x),
      numeric   = is.numeric(x),
      integer   = .is_intish(x),
      character = is.character(x)
    )
    if (!ok) .bad(name, paste0("must be a ", type, " vector or NULL."))
    invisible(TRUE)
  }

  .chk_range <- function(x, name, lo, hi) {
    if (is.null(x)) return(invisible(TRUE))
    if (!is.numeric(x)) .bad(name, "must be a numeric vector or NULL.")
    if (any(x < lo | x > hi, na.rm = TRUE)) {
      .bad(name, paste0("values must be in [", lo, ", ", hi, "]."))
    }
    invisible(TRUE)
  }

  .chk_xy_list <- function(x, name) {
    if (is.null(x)) return(invisible(TRUE))
    if (!is.list(x)) .bad(name, "must be a list of data.frames or NULL.")
    for (i in seq_along(x)) {
      d <- x[[i]]
      if (!is.data.frame(d)) {
        .bad(name, paste0("element ", i, " must be a data frame."))
      }

      if (ncol(d) != 2) {
        .bad(name, paste0("element ", i, " must have exactly two columns: `x` and `y`."))
      }

      if (!all(c("x", "y") %in% names(d))) {
        .bad(name, paste0("element ", i, " must have columns named `x` and `y`."))
      }

      if (!is.numeric(d[["x"]]) || !is.numeric(d[["y"]])) {
        .bad(name, paste0("element ", i, " columns `x` and `y` must be numeric."))
      }
    }
    invisible(TRUE)
  }

  .chk_fun_list <- function(x, name) {
    if (is.null(x)) return(invisible(TRUE))
    if (!is.list(x)) .bad(name, "must be a list of functions or NULL.")
    if (any(!vapply(x, is.function, logical(1)))) .bad(name, "must contain only functions.")
    invisible(TRUE)
  }

  # data: list
  .chk_list(data, "data")

  # xy.dat, study.pol: lists of 2-col data.frames (x,y)
  .chk_xy_list(xy.dat, "xy.dat")
  .chk_xy_list(study.pol, "study.pol")

  # logical vectors
  .chk_vec_type(same.n.split, "same.n.split", "logical")
  .chk_vec_type(c.fast.optim, "c.fast.optim", "logical")
  .chk_vec_type(use.chull,    "use.chull",    "logical")
  .chk_vec_type(c.splits,     "c.splits",     "logical")

  # integer vectors (integer-like numeric)
  .chk_vec_type(n.split.pts,   "n.split.pts",   "integer")
  .chk_vec_type(c.max.iter.no, "c.max.iter.no", "integer")
  .chk_vec_type(N.crit,        "N.crit",        "integer")
  .chk_vec_type(N.loc.crit,    "N.loc.crit",    "integer")
  .chk_vec_type(c.X.knots,     "c.X.knots",     "integer")
  .chk_vec_type(c.Y.knots,     "c.Y.knots",     "integer")

  # numeric vectors with bounds
  .chk_range(N.rel.crit,     "N.rel.crit",     0, 1)
  .chk_range(N.loc.rel.crit, "N.loc.rel.crit", 0, 1)
  .chk_range(S.crit,         "S.crit",         0, 1)
  .chk_range(S.rel.crit,     "S.rel.crit",     0, 1)

  # numeric vectors (no bounds stated)
  .chk_vec_type(Q.crit,        "Q.crit",        "numeric")
  .chk_vec_type(c.Q.crit,      "c.Q.crit",      "numeric")
  .chk_vec_type(c.crit.improv, "c.crit.improv", "numeric")

  # c.corr.term bounds
  .chk_range(c.corr.term, "c.corr.term", 0.01, 0.2)

  # lists of functions
  .chk_fun_list(generalize.f, "generalize.f")
  .chk_fun_list(compare.f,    "compare.f")

  # maximize: logical vector; if compare.f also provided, lengths must match
  .chk_vec_type(maximize, "maximize", "logical")
  if (!is.null(maximize) && !is.null(compare.f) && length(maximize) != length(compare.f)) {
    .bad(
      "maximize",
      paste0(
        "must have the same length as `compare.f` (",
        length(compare.f),
        ")."
      )
    )
  }

  # method: character vector
  .chk_vec_type(method, "method", "character")

  invisible(TRUE)
}

#' Check whether points fall inside a polygon
#'
#' This function is an identical replica of the \code{point.in.poly()} function
#' from the \pkg{sp} package.
#' @param point.x numeric vector of point x coordinates
#' @param point.y numeric vector of point y coordinates
#' @param pol.x numeric vector of polygon x coordinates
#' @param pol.y numeric vector of polygon y coordinates
#' @param mode.checked default FALSE, used internally to save time when all the other argument are known to be of storage mode double
#' @return integer array; values are: 0: point is strictly exterior to pol; 1: point is strictly interior to pol; 2: point lies on the relative interior of an edge of pol; 3: point is a vertex of pol.
#' @references
#' Uses the C function InPoly(). InPoly is Copyright (c) 1998 by Joseph O'Rourke. It may be freely redistributed in its entirety provided that this copyright notice is not removed
#' Pebesma E, Bivand R (2005). “Classes and methods for spatial data in R.” _R News_, *5*(2), 9-13. <https://CRAN.R-project.org/doc/Rnews/>.
#' Bivand R, Pebesma E, Gomez-Rubio V (2013). _Applied spatial data analysis with R, Second edition_. Springer, NY. <https://asdar-book.org/>.
#' @examples
#'#Creating data.frame of a polygon
#'# open polygon:
#'.point.in.polygon(1:10,1:10,c(3,5,5,3),c(3,3,5,5))
#'# closed polygon:
#'.point.in.polygon(1:10,rep(4,10),c(3,5,5,3,3),c(3,3,5,5,3))
#' @useDynLib hespdiv, R_point_in_polygon_sp
#' @noRd
.point.in.polygon = function(point.x, point.y, pol.x, pol.y,
    mode.checked=FALSE) {
    if (mode.checked) res <- .Call(R_point_in_polygon_sp, point.x,
        point.y, pol.x, pol.y)
    else res <- .Call(R_point_in_polygon_sp, as.numeric(point.x),
        as.numeric(point.y), as.numeric(pol.x), as.numeric(pol.y))
    res
}

#' Remove redundant polygon points
#'
#' This function finds and removes redundant polygon points that lie on the same straight line between 2 other points
#' @param xy a data frame with polygon coordinates
#' @return a data frame with cleaned polygon from redundant points
#' @author Liudas Daumantas
#' @noRd
.clean_poly <- function(xy){
  xy <- unique(xy)
  id <- numeric()
  for (i in 1:(nrow(xy) - 2)){
    y3 <- .pt_on_line(x1 = xy[i, 1],
                      x2 = xy[i+2, 1],
                      y1 = xy[i, 2],
                      y2 = xy[i+2, 2],
                      x3 = xy[i+1, 1])
    if (xy[i+1, 2] == y3)
      id <- c(id, i+1)
  }
  if (length(id) > 0)
    return(.close_poly(xy[-id,]))
  .close_poly(xy)
}

#' Taxon-level leave-out contributions for split-lines
#'
#' @description
#' For each unique taxon label in \code{obj$call.info$Call_ARGS$data}
#' (for example, species or genus), remove all of its occurrences from both
#' child polygons of every split, recompute split-line performance, and record
#' both the resulting value and its difference from the original.
#'
#' @details
#' Let \eqn{P} be the split-line performance stored in
#' \code{obj$split.stats$performance}. For a focal taxon \eqn{t}, we compute
#' \eqn{P^{-t}} by removing all occurrences of \eqn{t}. We report \eqn{\Delta}
#' so that \eqn{\Delta > 0} always indicates a \strong{positive contributor}
#' (that is, removal worsens performance):
#' \itemize{
#'   \item If \code{obj$call.info$Call_ARGS$maximize = FALSE} (lower is better,
#'   for example similarity), \eqn{\Delta = P^{-t} - P}.
#'   \item If \code{obj$call.info$Call_ARGS$maximize = TRUE} (higher is better,
#'   for example distance), \eqn{\Delta = P - P^{-t}}.
#' }
#' If a taxon is absent from a split's parent polygons, elimination is a no-op
#' and \eqn{\Delta = NA}.
#'
#' @param obj A \code{hespdiv} object.
#'
#' @return A list of class \code{taxon_effect_result} with:
#' \describe{
#'   \item{\code{elim.comp.vals}}{
#'   Performance after removing each taxon; dimension
#'   \code{[n_splits x n_taxa]}.
#'   }
#'   \item{\code{delta}}{
#'   Signed contribution \eqn{\Delta} as defined above; dimension
#'   \code{[n_splits x n_taxa]}.
#'   }
#'   \item{\code{n_per_pol}}{
#'   Counts of the focal taxon per split and polygon.
#'   }
#'   \item{\code{baseline}}{
#'   The original performance vector.
#'   }
#' }
#' @family functions for hespdiv results post-processing
#' @export
taxon_effect <- function(obj) {
  if (!inherits(obj, "hespdiv"))
    stop("`obj` must be a `hespdiv` object.", call. = FALSE)

  dat_in_obj <- obj$call.info$Call_ARGS$data
  xy_in_obj  <- obj$call.info$Call_ARGS$xy.dat
  baseline   <- obj$split.stats$performance
  maximize   <- isTRUE(obj$call.info$Call_ARGS$maximize)

  # choose slicer + extract taxon vector
  if (is.data.frame(dat_in_obj) || is.matrix(dat_in_obj)) {
    .slicer <- .slicer.table
    taxa_vec <- as.character(dat_in_obj[[1]])
  } else if (is.list(dat_in_obj)) {
    .slicer <- .slicer.list
    taxa_vec <- as.character(unlist(dat_in_obj, use.names = FALSE))
  } else {
    .slicer <- .slicer.vect
    taxa_vec <- as.character(dat_in_obj)
  }

  group <- factor(taxa_vec)
  taxa_levels <- levels(group)
  n_taxa   <- length(taxa_levels)
  n_splits <- length(obj$split.lines)

  elim.comp.vals <- matrix(NA_real_, nrow = n_splits, ncol = n_taxa,
                           dimnames = list(rownames(obj$split.stats), taxa_levels))

  N <- n_taxa * n_splits * 2L
  n_per_pol <- data.frame(
    split.id = integer(N),
    group    = character(N),
    pol.id   = integer(N),
    n        = integer(N),
    stringsAsFactors = FALSE
  )

  for (tax_id in seq_len(n_taxa)) {
    tx <- taxa_levels[tax_id]
    idx_tax <- which(group == tx)
    xy_tax  <- xy_in_obj[idx_tax, , drop = FALSE]

    for (split.id in seq_len(n_splits)) {
      pol_ids <- which(obj$poly.stats$root.id == obj$split.stats$plot.id[split.id])
      pol_id1 <- obj$poly.stats$plot.id[pol_ids[1]]
      pol_id2 <- obj$poly.stats$plot.id[pol_ids[2]]
      child_pol1 <- obj$polygons.xy[[as.character(pol_id1)]]
      child_pol2 <- obj$polygons.xy[[as.character(pol_id2)]]
      ids1_tx <- .get_ids(child_pol1, xy_tax)
      ids2_tx <- .get_ids(child_pol2, xy_tax)

      offset_tax   <- (tax_id - 1L) * (2L * n_splits)
      offset_split <- (split.id - 1L) * 2L
      row1 <- offset_tax + offset_split + 1L
      row2 <- offset_tax + offset_split + 2L
      n_per_pol[row1, ] <- list(split.id, tx, pol_id1, length(ids1_tx))
      n_per_pol[row2, ] <- list(split.id, tx, pol_id2, length(ids2_tx))

      ids1_all <- .get_ids(child_pol1, xy_in_obj)
      ids2_all <- .get_ids(child_pol2, xy_in_obj)

      ids1_nogr <- setdiff(ids1_all, idx_tax)
      ids2_nogr <- setdiff(ids2_all, idx_tax)

      if (length(ids1_tx) == 0L && length(ids2_tx) == 0L) {
        elim.comp.vals[split.id, tax_id] <- NA
      } else {
        elim.comp.vals[split.id, tax_id] <- obj$call.info$Call_ARGS$compare.f(
          obj$call.info$Call_ARGS$generalize.f(.slicer(dat_in_obj, ids1_nogr)),
          obj$call.info$Call_ARGS$generalize.f(.slicer(dat_in_obj, ids2_nogr))
        )
      }
    }
  }

  # Base Δ as elim - baseline, then flip sign if the metric is maximised
  delta <- sweep(elim.comp.vals, 1L, baseline, FUN = "-")
  if (maximize) delta <- -delta

  structure(list(
    elim.comp.vals = elim.comp.vals,
    delta          = delta,
    n_per_pol      = n_per_pol,
    baseline       = baseline
  ), class = "taxon_effect")
}

#' Change the basal subdivision in hsa
#'
#' @description  This function allows you to select a new basal subdivision
#' from the results of hespdiv sensitivity analysis by specifying its ID. It
#' provides a convenient way to switch between different basal subdivisions.
#' You can identify a more stable subdivision alternative by plotting the
#' hespdiv sensitivity analysis results using the \code{plot_hsa} function.
#' By selecting a new basal subdivision, you can observe how it affects the
#' results of polygon object cross-comparison and the stability of hespdiv
#' clusters and polygons.
#' @param obj A \code{hsa} class object.
#' @param id An index of an alternative subdivision to be used as a new basal
#' subdivision.
#' @return The \code{hsa} class object with a new basal subdivision.
#' @author Liudas Daumantas
#' @family functions for hespdiv sensitivity analysis
#' @export
change_base <- function(obj, id){
  if (!inherits(obj,"hsa"))
    stop(
      "`obj` should be of class `hsa`, as produced by `hsa()` or `hsa_detailed()`.",
      call. = FALSE
    )
  base <- obj[[1]][[id]][[1]]
  if (is.null(base)) stop("Selected base has no subdivisions.", call. = FALSE)
  obj[[1]][[id]][[1]] <- obj[[2]]
  obj[[1]][[id]][[2]] <- obj[[2]]$call.info$Call_ARGS

  structure(
    list(
      Alternatives = obj[[1]],
      Basis = base
    ),
    class = "hsa"
  )
}

#' Remove split-lines
#'
#' @description  Function returns hespdiv object, without split-lines of
#' specified id.
#'
#' @return hespdiv object
#' @param obj hespdiv object
#' @param split.id vector of split-line ids
#' @param depend.splits logical. Remove split-lines that depend on specified
#' split-lines? If FALSE, only end-nodes of spatial dendrogram are removed.
#' @examples
#' \donttest{
#' if (requireNamespace("HDData")) {
#'
#'   # Inspect the hespdiv object
#'   print(plot_hespdiv(HDData::hd))
#'
#'   # Remove weak split-lines
#'   weak_splits <- which(HDData::hd$split.stats$performance >= 0.3)
#'   performance_filtered <- remove_splits(obj = HDData::hd, split.id = weak_splits)
#'   print(plot_hespdiv(performance_filtered))
#'
#'   # Remove non-significant split-lines
#'   plot(HDData::nl)
#'   nsig_splits <- which(HDData::nl[[1]]$quantile >= 0.05)
#'   sig_filtered <- remove_splits(obj = HDData::hd, split.id = nsig_splits)
#'   print(plot_hespdiv(sig_filtered))
#'
#'   # Remove only if a split-line has no dependent split-lines
#'   unchanged_hd <- remove_splits(obj = HDData::hd, split.id = 4, depend.splits = FALSE)
#'   print(plot_hespdiv(unchanged_hd))
#'
#'   # Remove the split-lines indicated as well as all other split-lines
#'   # that structurally depend on them (default behavior)
#'   changed_hd <- remove_splits(obj = HDData::hd, split.id = 4, depend.splits = TRUE)
#'   print(plot_hespdiv(changed_hd))
#' }
#' }
#' @export
remove_splits <- function(obj, split.id, depend.splits = TRUE) {
  all.splits <- split.id
  all.pols <- numeric()
  for (id in split.id) {
    d <- depend_splits(obj, id)
    #browser()
    if (!depend.splits & length(d$splits) != 0 & any(!d$splits %in% split.id)){
      all.splits <- all.splits[all.splits != id]
    } else {
      all.splits <- c(all.splits, d$splits)
      all.pols <- c(all.pols, d$polygons)
    }

  }
  if (length(all.splits) > 0){
    all.splits <- unique(all.splits)
    all.pols <- unique(all.pols)
    plot.id <- obj$split.stats[all.splits,"plot.id"]
    obj$poly.stats[plot.id, "has.split"] <- FALSE
    obj$poly.stats <- obj$poly.stats[-all.pols,]

    obj$split.stats <- obj$split.stats[-all.splits,]
    obj$split.lines <- obj$split.lines[-all.splits]
    obj$polygons.xy <- obj$polygons.xy[-all.pols]
    obj$poly.obj <- obj$poly.obj[-all.pols]
    obj$str.difs <- obj$str.difs[-all.pols]
  }
  obj
}

#'
if(getRversion() >= "2.15.1"){

  utils::globalVariables(c("S.crit", ".slicer", ".comp", "c.sign", ".minormax", ".which_minormax",
                           "samp.xy", "rims", "poly.obj", "generalize.f", "e", "n.split.pts",
                           "dst.pts", "trace.object", "pnts.col", "xy.dat", "study.pol.viz",
                           "S.rel.crit", "c.Q.crit", "trace.level", ".get_ids_str", "N.crit",
                           "N.rel.crit", "N.loc.crit", "N.loc.rel.crit", "S.cond", "str.split.quals",
                           "sd", "c.splits", "c.X.knots", "c.Y.knots", "c.max.iter.no",
                           "c.fast.optim", "c.corr.term", "c.crit.improv", "Q.crit", "splits","testid"
  ))
  }

#' Visualize the stability of hespdiv polygons
#'
#' @description  The function uses hespdiv sensitivity analysis results to visually
#' demonstrate the stability of the basal hespdiv subdivision. This is achieved
#' by displaying both alternative and basal hespdiv subdivisions on the same plot.
#' @param obj An object of class \code{hsa}
#' @param alpha The alpha value for transparency of split lines. Default is 0.6.
#' @param split.col Color of alternative subdivision split-lines. Default is "gray20".
#' @param pnts.col The color of data points. Default is NULL.
#' @param pol.col The color of polygons. Default is "7".
#' @param type An integer indicating the type of plot. Default is 1.
#' @param max.lwd The maximum line width for split-lines. Default is 3.
#' @param min.lwd The minimum line width for split-lines. Default is 0.5.
#' @param basal.col The color of basal subdivision split-lines.
#' @param split.col.seed A seed for generating random colors for split lines. Default is NULL.
#' @param seperated Boolean. When \code{type} is >= 3, open a new graphical device for each rank?
#' @param newplot Create a plot in new device?
#' @return No return value, called for plotting sensitivity analysis results.
#' @details The \code{type} parameter determines the type of plot generated:
#' \describe{
#'   \item{\code{1}}{
#'   Basic plot: displays the alternative and basal hespdiv subdivisions on
#'   the same plot without split-line ranks or titles.
#'   }
#'
#'   \item{\code{2}}{
#'   Plot with split-line ranks: includes split-line ranks in the plot.
#'   Each split-line is assigned a different line width based on its rank.
#'   }
#'
#'   \item{\code{3}}{
#'   Plot with separate ranks: generates multiple plots, each representing
#'   split-line ranks up to a certain value.
#'   }
#'
#'   \item{\code{4}}{
#'   Plot with separate and isolated ranks: similar to mode 3 but isolates
#'   split-line ranks. Generates multiple plots, each representing a
#'   specific split-line rank.
#'   }
#' }
#'
#' If the alternative subdivisions spatially converge but the basal
#' subdivision lies far from the zone of convergence, you can use
#' \code{change_base} to select a more representative alternative
#' subdivision to serve as the basal subdivision. However, you should
#' verify that the arguments used in that subdivision are appropriate.
#' @author Liudas Daumantas
#' @family functions for hespdiv sensitivity analysis
#' @family HespDiv visualization options
#' @note \code{newplot} allows the legend to be rendered correctly in
#' types 2 and 3, and helps with line rendering in general when drawing
#' in an active device (use \code{broom} otherwise to delete devices).
#' @importFrom graphics plot lines points
#' @importFrom scales alpha
#' @importFrom grDevices dev.new
#' @export
plot_hsa <- function(obj, alpha = 0.6, split.col = "gray20", pnts.col = NULL,
                     pol.col = "7", type = 1, basal.col = 2,
                     max.lwd = 3, min.lwd = 0.5, split.col.seed = NULL,
                     newplot = TRUE, seperated = TRUE){
  oldpar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(oldpar), add = TRUE)
  if (!inherits(obj,"hsa"))
    stop(
      "`obj` must be an `hsa` object produced by `hsa()` or `hsa_detailed()`.",
      call. = FALSE
    )
  if ( length(pol.col) > 1){
    if ( length(pol.col) != length(subs)){
      stop(
        "Length of colors for study area polygons is not equal to the number of subdivisions.",
        call. = FALSE
      )
    }}
  if (newplot & !seperated & type == 3) {
    stop(
      "Change `seperated` to TRUE or `newplot` to FALSE to make all plots visible.",
      call. = FALSE
    )
  }
  # Combine the basal and alternative subdivisions
  subs <- c(list(obj$Basis),lapply(obj$Alternatives, function(o) o[[1]]))
  # Check if all alternative subdivisions are NULL
  if (all(il <- sapply(subs[-1], is.null) | sapply(subs[-1],class) != "hespdiv"))
    stop(
      "All alternative subdivisions are NULL or contain errors or warnings.",
      call. = FALSE
    )
  # Remove NULL subdivisions or subdivisions with
  if (any(il))
    subs <- subs[-(which(il) + 1)]
  # Determine the plot settings based on the type parameter
  if (type == 1){
    ranks <- separate <- FALSE
    title <- ""
  } else {
    if (newplot) {
      dev.new()
      if (type == 2 | type == 3){
        graphics::par(mar=c(5, 4, 4, 8)+0.1, xpd = TRUE)
      }
    }
    if (type == 2){
      ranks <- TRUE
      separate <- FALSE
      title <- ""
    } else {
      if (type == 3){
        ranks <- separate <- TRUE
        isolate.ranks <- FALSE
      } else {
        if (type == 4){
          ranks <- separate <- isolate.ranks <- TRUE
        }
      }
    }
  }
  # Plot the initial plots if not separate
  if (!separate){
    .initial_plots(subs = subs, pnts.col = pnts.col,max.lwd, title, basal.col)
  }
 # Plot the split lines
  if (!ranks) {
    for (i in 1:length(subs)){
      for (split in 1:length(subs[[i]][["split.lines"]]))
        graphics::lines(x=subs[[i]][["split.lines"]][[split]],
                        col=scales::alpha(split.col[1],alpha),
                        lwd=0.5)
    }
  } else {
    all_ranks <- range(sapply(subs,function(o) range(o[["split.stats"]]$rank)))
    if (is.null(split.col.seed)){
    if (length(split.col) < all_ranks[2] ){
      if (length(split.col) != 1)
        warning(
          "More split ranks than provided by `split.col`.\n",
          "Generating random colors.",
          call. = FALSE
        )
      split.col <- .generate_cols(all_ranks[2], sample(1:9999,1))
    }} else {
      split.col <- .generate_cols(all_ranks[2], split.col.seed)
    }
    if (separate){
      if (!isolate.ranks){
      for (rangas in all_ranks[1]:all_ranks[2]){
        if (seperated) {
          if (rangas !=1)
            grDevices::dev.new()
          if (!(rangas == 1 & newplot)){
          graphics::par(mar=c(5, 4, 4, 8)+0.1, xpd = TRUE)}
        }
        title <- paste0("Split-line Rank - up to ", rangas)
        .initial_plots(subs, pnts.col,max.lwd, title, basal.col)
        if (any(viz <- sapply(subs,function(o,rangas) any(
          o[["split.stats"]]$rank <= rangas), rangas ))){
          for (i in (1:length(subs))[viz]){
            for (split in 1:length(subs[[i]][["split.lines"]])){
              if (subs[[i]][["split.stats"]]$rank[split] <= rangas)
                graphics::lines(x=subs[[i]][["split.lines"]][[split]],
                                col=scales::alpha(split.col[
                                  subs[[i]][["split.stats"]]$rank[split]
                                ],alpha),
                                lwd=seq(max.lwd,min.lwd,length.out = all_ranks[2])[
                                  subs[[i]][["split.stats"]]$rank[split]])
            }
          }
        }
        graphics::legend("right", inset=c(-0.2,0),
                         legend=c(1:all_ranks[2]),
                         col = scales::alpha(split.col[1:all_ranks[2]],alpha),
                         lty = 1,
                         cex = 0.7,
                         lwd = seq(max.lwd,min.lwd,length.out = all_ranks[2]),
                         title="Rank")
        .final_plots(subs, pol.col, max.lwd)
      }} else {
        for (rangas in all_ranks[1]:all_ranks[2]){
          if (seperated)
            grDevices::dev.new()
          title <- paste0("Split-line Rank - ", rangas)
          .initial_plots(subs, pnts.col,max.lwd, title, basal.col)
          if (any(viz <- sapply(subs,function(o,rangas) any(
            o[["split.stats"]]$rank == rangas), rangas ))){
            for (i in (1:length(subs))[viz]){
              for (split in 1:length(subs[[i]][["split.lines"]])){
                if (subs[[i]][["split.stats"]]$rank[split] == rangas)
                  graphics::lines(x=subs[[i]][["split.lines"]][[split]],
                                  col=scales::alpha(split.col[
                                    subs[[i]][["split.stats"]]$rank[split]
                                  ],alpha),
                                  lwd=seq(max.lwd,min.lwd,length.out = all_ranks[2])[
                                    subs[[i]][["split.stats"]]$rank[split]])
              }
            }
          }
          .final_plots(subs, pol.col, max.lwd)
        }
      }
    } else{
      for (i in 1:length(subs)){
        for (split in 1:length(subs[[i]][["split.lines"]])){
          graphics::lines(subs[[i]][["split.lines"]][[split]],
                          col=scales::alpha(split.col[
                            subs[[i]][["split.stats"]]$rank[split]
                          ],alpha),
                          lwd=seq(max.lwd,min.lwd,length.out = all_ranks[2])[
                            subs[[i]][["split.stats"]]$rank[split]])
        }

      }
      graphics::legend("right", inset=c(-0.2,0),
             legend=c(1:all_ranks[2]),
             col = scales::alpha(split.col[1:all_ranks[2]],alpha),
             lty = 1,
             cex = 0.7,
             lwd = seq(max.lwd,min.lwd,length.out = all_ranks[2]),
             title="Rank")}
    }
  if (!separate){
    .final_plots(subs, pol.col, max.lwd)
  }
}
#' @noRd
.initial_plots <- function(subs, pnts.col,max.lwd, title, basal.col){
  graphics::plot(NULL,
                 ylim = range(sapply(subs,
                                     function(o) range(c(o$polygons.xy[[1]]$y,
                                                         o$call.info$Call_ARGS$study.pol$y)))),
                 xlim =  range(sapply(subs,
                                      function(o) range(c(o$polygons.xy[[1]]$x,
                                                          o$call.info$Call_ARGS$study.pol$x)))),
                 col = 0,
                 xlab = "x coordinate", ylab = "y coordinate", main = title)
  if (!is.null(pnts.col)){
    graphics::points(subs[[1]]$call.info$Call_ARGS$xy.dat,col=pnts.col,
                     pch = 19, cex = 0.5)
  }
  if (!is.null(subs[[1]]$call.info$Call_ARGS$study.pol)){
    if (!identical(subs[[1]]$call.info$Call_ARGS$study.pol, subs[[1]]$polygons.xy[[1]])){
      graphics::lines(subs[[1]]$call.info$Call_ARGS$study.pol, col = "lightyellow3")
    }
  }
  for (split in 1:length(subs[[1]][["split.lines"]])){
    graphics::lines(x=subs[[1]][["split.lines"]][[split]],
                    col = basal.col,
                    lwd = max.lwd)
  }
}
#' @noRd
.final_plots <- function(subs, pol.col, max.lwd){
  if ( length(pol.col) > 1){
    for (i in 1:length(subs)){
      lines(subs[[i]]$polygons.xy[[1]],
            col = pol.col[i],lwd = c(max.lwd,rep(1,length(subs)-1))[i])
    }
  } else {
    for (i in 1:length(subs)){
      lines(subs[[i]]$polygons.xy[[1]],
            col = pol.col[1],lwd = c(max.lwd,rep(1,length(subs)-1))[i])
    }
  }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/poly_pop.R
\name{polypop}
\alias{polypop}
\title{Remove polygons from rgl device}
\usage{
polypop(obj, height)
}
\arguments{
\item{obj}{The hespdiv object used to create the currently active rgl
device with the \code{blok3d} function.}

\item{height}{A character value that indicates the height co-ordinate.}
}
\value{
No return value. Called for the interactive modification of a
plot created by \code{blok3d}
}
\description{
This function allows you to interactively select and remove unwanted polygons
from a 3D plot created with the \code{blok3d} function.
}
\seealso{
Other HespDiv visualization options: 
\code{\link{blok3d}()},
\code{\link{create_gif}()},
\code{\link{dendro}()},
\code{\link{plot.nullhespdiv}()},
\code{\link{plot_cs_hsa}()},
\code{\link{plot_hespdiv}()},
\code{\link{plot_hsa}()},
\code{\link{plot_hsa_q}()},
\code{\link{poly_scheme}()}
}
\author{
Liudas Daumantas
}
\concept{HespDiv visualization options}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hsa_quant.R
\name{hsa_quant}
\alias{hsa_quant}
\title{Quantify the stability of hespdiv clusters}
\usage{
hsa_quant(obj, probs = c(0.05, 0.5, 0.95))
}
\arguments{
\item{obj}{An object of class \code{hsa}.}

\item{probs}{A numeric vector of probabilities with values in the range
\eqn{0 \le p \le 1}. This argument is used to calculate quantiles of Jaccard similarity
values.}
}
\value{
A list containing three data frames:
\describe{
\item{jaccard.quantiles}{
Quantiles of Jaccard similarities between the basal cluster and
the analog clusters from alternative subdivisions.
}
\item{jaccard.similarity}{
Jaccard similarity values between the basal cluster and the analog
cluster from each alternative subdivision.
}
\item{analog.clusters}{
IDs of the hespdiv polygons that produced the analog clusters
in each alternative subdivision.
}
}
}
\description{
This function evaluates the stability of the basal subdivision
clusters using hespdiv sensitivity analysis results. It does so by
calculating Jaccard similarities between the observations of basal
subdivision clusters and the observations of alternative subdivision
clusters. For each basal cluster, the function identifies the most similar
analog cluster within each alternative subdivision. The stability of
each basal cluster can be assessed by examining the distribution of
similarity values with their corresponding analog clusters. If a highly
similar cluster reappears in multiple alternative subdivisions, it
indicates that the basal cluster is stable.
}
\details{
If a base subdivision cluster obtains a distribution of high similarity
values, it is considered stable and existing. Low analog-cluster similarity
values may indicate that a base cluster is an artifact of the
\code{hespdiv()} computation.

The more technical description of how \code{hsa_quant} works:
\describe{
\item{Obtaining alternative hespdiv clusters:}{ The function filters the \code{xy.dat} coordinates of the basal subdivision using all the polygons of alternative subdivisions, obtaining alternative hespdiv clusters.}
\item{Quantifying Jaccard similarity:}{ The function measures the Jaccard overlap index between the observations of the basal subdivision clusters and the observations of the alternative clusters. }
\item{Identification of analog clusters and value assignments:}{ Each basal hespdiv cluster from each alternative subdivision is assigned the ID of the cluster that produced the maximum Jaccard similarity value, along with the corresponding similarity value.}
}
The purpose of the \code{hsa_quant} function is to address situations where hespdiv
polygons, despite having different geometry and location, may filter nearly
identical sets of observations, leading to similar hespdiv clusters. This can occur when the spatial
coverage of observations is incomplete and irregular, or when the boundaries
between hespdiv polygons are expected to be open, soft, or fuzzy, such as in
the case of boundaries between bioregions. In such cases, visual hespdiv
sensitivity analysis alone may show irregular and non-converging
distributions of split-lines. However, \code{hsa_quant} can reveal that these
irregular polygons are based on nearly identical clusters of observations,
indicating a strong spatial structure within the analyzed data. Conversely,
if the observations within these clusters significantly differ, it indicates
that the basal clusters are specific to the hespdiv parameters used and
likely lack ontological meaning.

Thus, by analyzing the similarity values between clusters of observations, \code{hsa_quant}
facilitates the assessment of the stability and reliability of basal
subdivision clusters, aiding in evaluating their significance.
}
\note{
You can use the hsa_quant function to track the evolution of hespdiv
subdivisions over time by providing correctly formatted input. For
instance, you can obtain the basal subdivision for time bin 1 using the
hespdiv function. Then, using the hsa function, you can specify the paired
xy.dat and data from time bin 2. The resulting hsa object can be inputted
into \code{hsa_quant} The \code{hsa_quant} result will then provide insights into
extinctions, speciations, fusions, and splits of hespdiv polygons/clusters
that occur between time bin 1 and 2. This allows for the analysis of changes
and dynamics in hespdiv subdivisions over time.
}
\seealso{
Other functions for hespdiv sensitivity analysis: 
\code{\link{change_base}()},
\code{\link{hsa}()},
\code{\link{hsa_detailed}()},
\code{\link{hsa_sample_constrained}()},
\code{\link{plot_cs_hsa}()},
\code{\link{plot_hsa}()},
\code{\link{plot_hsa_q}()}

Other functions to evaluate hesdpiv cluster stability: 
\code{\link{plot_hsa_q}()}

Other functions for hespdiv results post-processing: 
\code{\link{cross_comp}()},
\code{\link{hsa}()},
\code{\link{hsa_detailed}()},
\code{\link{hsa_sample_constrained}()},
\code{\link{nulltest}()},
\code{\link{taxon_effect}()}
}
\author{
Liudas Daumantas
}
\concept{functions for hespdiv results post-processing}
\concept{functions for hespdiv sensitivity analysis}
\concept{functions to evaluate hesdpiv cluster stability}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_hespdiv.R
\name{plot_hespdiv}
\alias{plot_hespdiv}
\title{Plot hespdiv results}
\usage{
plot_hespdiv(
  obj,
  type = "color",
  n.loc = FALSE,
  performance = TRUE,
  legend_title = NULL,
  title = NULL,
  subtitle = NULL,
  pnts.col = NULL,
  seed = 10
)
}
\arguments{
\item{obj}{A hespdiv object.}

\item{type}{A character. Either "width" or "color" (default "color").
Determines whether quality of split-lines is expressed by line width or
color.}

\item{n.loc}{A Boolean value. Would you like to visualize the number of observations
at each location? Only possible, when there are localities with more than
one observation. If the type is 'color,' the number of observations is
expressed through point sizes. Otherwise, they are expressed using color in
a logarithmic scale.}

\item{performance}{logical. TRUE - display split-line performance,
FALSE - rank. Displaying rank makes the spatial dendrogram clearer.}

\item{legend_title}{A character value that indicates the title of the legend
for the split-lines. The default is built according to the method information
available in "obj$call.info".}

\item{title}{A character that indicates the title of the plot.}

\item{subtitle}{A character that indicates the subtitle of the plot.}

\item{pnts.col}{A character or numeric vector providing color codes
for data points.}

\item{seed}{An integer value that indicates seed used to randomize the colors
of the split-lines. Only meaningful, when argument \code{type = "width"}.
Try setting a different value, if colors of parallel split-lines or nearby
labels look too similar or to increase the general appeal of the graph.}
}
\value{
A ggplot object.
}
\description{
This function is used to plot the results obtained with the \code{hespidv}
function. The plot showcases subdivisions of the study area by split-lines,
visualizing their performances or rank with colors or line widths. Additionally,
it can display the spatial distribution of observations and number of
observations in each location.
}
\details{
The return ggplot object can be edited as any other ggplot objects
by removing undesired elements, changing theme or overlying the plot
with additional elements.
}
\examples{
plot_hespdiv(example_hespdiv)
plot_hespdiv(example_hespdiv, type = "width")
plot_hespdiv(example_hespdiv, n.loc = TRUE)
}
\seealso{
Other HespDiv visualization options: 
\code{\link{blok3d}()},
\code{\link{create_gif}()},
\code{\link{dendro}()},
\code{\link{plot.nullhespdiv}()},
\code{\link{plot_cs_hsa}()},
\code{\link{plot_hsa}()},
\code{\link{plot_hsa_q}()},
\code{\link{poly_scheme}()},
\code{\link{polypop}()}
}
\author{
Liudas Daumantas
}
\concept{HespDiv visualization options}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print_nullhespdiv.R
\name{print.nullhespdiv}
\alias{print.nullhespdiv}
\title{Print the results of nullhespdiv object}
\usage{
\method{print}{nullhespdiv}(x, ...)
}
\arguments{
\item{x}{A nullhespdiv class object}

\item{...}{other arguments}
}
\value{
x
}
\description{
print method for nullhespdiv class object
}
\author{
Liudas Daumantas
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hsa.R
\name{hsa}
\alias{hsa}
\title{HespDiv Sensitivity Analysis}
\usage{
hsa(
  obj,
  n.runs = 100,
  data.paired = TRUE,
  display = FALSE,
  images.path = NULL,
  pnts.col = 1,
  data = NULL,
  xy.dat = NULL,
  same.n.split = NULL,
  n.split.pts = NULL,
  N.crit = NULL,
  N.rel.crit = NULL,
  N.loc.crit = NULL,
  N.loc.rel.crit = NULL,
  S.crit = NULL,
  S.rel.crit = NULL,
  Q.crit = NULL,
  c.splits = NULL,
  c.Q.crit = NULL,
  c.crit.improv = NULL,
  c.X.knots = NULL,
  c.Y.knots = NULL,
  c.max.iter.no = NULL,
  c.fast.optim = NULL,
  c.corr.term = NULL,
  study.pol = NULL,
  use.chull = NULL,
  generalize.f = NULL,
  maximize = NULL,
  method = NULL,
  compare.f = NULL,
  .run.id = NULL,
  parallel = FALSE,
  RAM = NULL,
  load_prop = 0.8,
  chunk_size = workers * 2,
  workers = NULL,
  future_seed = TRUE
)
}
\arguments{
\item{obj}{A \code{hespdiv} class object.}

\item{n.runs}{Integer. The number of alternative \code{hespdiv} calls to evaluate.}

\item{data.paired}{Logical. Controls whether alternative values of \code{data}
are paired with corresponding elements of \code{xy.dat}.}

\item{display}{Logical. Controls the value of the \code{display} argument
in each \code{hespdiv} call.}

\item{images.path}{Character or \code{NULL}. Path to an existing directory
where PNG images of displayed results will be saved. If \code{NULL}
(default), images are not saved.}

\item{pnts.col}{Value passed to the \code{pnts.col} argument of each
\code{hespdiv} call. Can be provided as atomic vector,  vector or list of
vectors.}

\item{data}{A list of data objects (matrices, data frames, vectors, lists,
or other supported data structures) used as alternative inputs for
sensitivity analysis.}

\item{xy.dat, study.pol}{Lists of data frames with two columns: \code{x} and \code{y}.}

\item{same.n.split, c.fast.optim, use.chull, c.splits}{Logical vectors specifying
alternative values for corresponding \code{hespdiv} arguments.}

\item{n.split.pts, c.max.iter.no, N.crit, N.loc.crit, c.X.knots, c.Y.knots}{Integer vectors specifying alternative values for corresponding arguments.}

\item{N.rel.crit, N.loc.rel.crit, S.crit, S.rel.crit}{Numeric vectors with values
between 0 and 1.}

\item{Q.crit, c.Q.crit, c.crit.improv}{Numeric vectors specifying alternative
threshold or improvement criteria.}

\item{c.corr.term}{Numeric vector with values between 0.01 and 0.2.}

\item{generalize.f, compare.f}{Lists of functions defining custom similarity
or generalization methods.}

\item{maximize}{Logical vector of the same length as \code{compare.f},
indicating whether the corresponding similarity metric should be maximized.}

\item{method}{Character vector specifying predefined similarity metrics.}

\item{.run.id}{Integer. Runs with indices less than or equal to this value
will be skipped. This can be used to resume an interrupted analysis.}

\item{parallel}{Logical. If \code{TRUE}, runs are evaluated in parallel using
\pkg{future.apply}. If \code{FALSE} (default), runs are evaluated sequentially.}

\item{RAM}{Integer. Approximate amount of available RAM (in GB) used to limit
the number of parallel workers when \code{parallel = TRUE}.}

\item{load_prop}{Numeric in (0, 1]. Proportion of available CPU cores to use
when determining the number of parallel workers automatically. Defaults to 0.8.}

\item{chunk_size}{Integer. Number of runs submitted per batch (chunk) to the
parallel backend when \code{parallel = TRUE}. Chunking does not change how
many runs execute simultaneously (controlled by \code{workers}). Smaller
values reduce the number of queued futures at once (often lowering RAM
overhead). Defaults to \code{workers*2}.}

\item{workers}{Integer. Number of parallel workers (CPU cores) to use when
\code{parallel = TRUE}. At most this many \code{hespdiv} calls will be executed
simultaneously. If \code{NULL}, the number of workers is determined
automatically based on \code{RAM} and \code{load_prop}.}

\item{future_seed}{Logical. Passed to \code{future.apply::future_lapply()}.
If \code{TRUE}, RNG is managed by \pkg{future} to provide parallel-safe,
reproducible random streams.}
}
\value{
An object of class \code{hsa}, containing:
\describe{
\item{\bold{Alternatives}}{A list of alternative \code{hespdiv} results
produced during the sensitivity analysis.}
\item{\bold{Basis}}{The reference \code{hespdiv} object whose arguments
were perturbed.}
}
}
\description{
This function performs sensitivity analysis of the \code{hespdiv} method.
Starting from a reference \code{hespdiv} object, it generates a specified
number of alternative \code{hespdiv} calls by randomly sampling new values
for selected arguments from user-provided sets.
}
\details{
\subsection{Difference Between \code{"hsa"} and other sensitivity analysis functions}{
The \code{hsa_detailed} function evaluates all combinations of provided
argument values, resulting in dense sampling of the parameter space at
substantial computational cost. In contrast, \code{hsa} samples the parameter
space stochastically and is generally more suitable for exploratory or
large-scale sensitivity analyses.

In \code{hsa_detailed}, alternative argument values are provided as lists,
whereas in \code{hsa} they are supplied as vectors or lists depending on
the argument.

The \code{hsa_sample_constrained} function performs non-recursive \code{hespdiv}
runs for each split-line produced based on different data subsamples. Thus,
\code{hsa} is more general, as it allows to inspect sensitivity to other arguments.
}

\subsection{Paired Arguments}{
When \code{data.paired = TRUE}, the same index is used to sample elements
of \code{data} and \code{xy.dat}, allowing sensitivity analysis across
datasets of differing size or composition. When \code{FALSE}, data and
coordinates are sampled independently, enabling analyses based on noise
addition or spatial shuffling.

Arguments defining custom methods (\code{compare.f}, \code{generalize.f},
\code{maximize}) are always treated as paired and must therefore have
equal lengths.
}
}
\note{
If a particular run produces a warning or an error, the corresponding list
element will contain two components. In case of a warning, these are the
resulting \code{hespdiv} object and the warning message. In case of an error,
they are the arguments used for the call and the error message.
}
\seealso{
Other functions for hespdiv sensitivity analysis: 
\code{\link{change_base}()},
\code{\link{hsa_detailed}()},
\code{\link{hsa_quant}()},
\code{\link{hsa_sample_constrained}()},
\code{\link{plot_cs_hsa}()},
\code{\link{plot_hsa}()},
\code{\link{plot_hsa_q}()}

Other functions for hespdiv results post-processing: 
\code{\link{cross_comp}()},
\code{\link{hsa_detailed}()},
\code{\link{hsa_quant}()},
\code{\link{hsa_sample_constrained}()},
\code{\link{nulltest}()},
\code{\link{taxon_effect}()}
}
\concept{functions for hespdiv results post-processing}
\concept{functions for hespdiv sensitivity analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_gif.R
\name{create_gif}
\alias{create_gif}
\title{Create a gif of blok3D}
\usage{
create_gif(output_file, frames = 90, angle_per_frame = 5, fps = 10)
}
\arguments{
\item{output_file}{name and path of output file}

\item{frames}{integer. Number of frames.}

\item{angle_per_frame}{numeric. Angle to rotate per frame in degrees.}

\item{fps}{integer. Frames per second.}
}
\value{
No return value. Called for the side effect of saving
a GIF of a rotating 3D plot.
}
\description{
Functions creates a gif of rotating polygons displayed in a currently actively rgl device produced with blok3D
}
\note{
You can adjust the size of rgl device window to control the size of gif.
}
\seealso{
Other HespDiv visualization options: 
\code{\link{blok3d}()},
\code{\link{dendro}()},
\code{\link{plot.nullhespdiv}()},
\code{\link{plot_cs_hsa}()},
\code{\link{plot_hespdiv}()},
\code{\link{plot_hsa}()},
\code{\link{plot_hsa_q}()},
\code{\link{poly_scheme}()},
\code{\link{polypop}()}
}
\author{
Liudas Daumantas
}
\concept{HespDiv visualization options}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hsa_sample_constrained.R
\name{hsa_sample_constrained}
\alias{hsa_sample_constrained}
\title{Constrained HespDiv Sensitivity Analysis by Subsampling}
\usage{
hsa_sample_constrained(
  obj,
  n.runs = 100,
  subsample_factor = 0.7,
  RAM = NULL,
  load_prop = NULL,
  chunk_size = 8,
  workers = NULL
)
}
\arguments{
\item{obj}{A \code{hespdiv} object.}

\item{n.runs}{Integer. The number of subsampling runs to perform (default: 100).}

\item{subsample_factor}{Numeric proportion of data to subsample within each polygon (0 to 1].
For example, 0.7 means 70\% of the data in each polygon are retained.}

\item{RAM}{Integer. Approximate amount of RAM in GB to guide how many parallel
workers to use. The function uses up to 80\\% of your available CPU cores
but also caps the number of workers at \code{RAM}.}

\item{load_prop}{Numeric value (0,1]. Specifies the proportion of available
CPU cores or RAM to be used for setting up parallel workers. For example,
\code{load_prop = 0.8} uses 80\% of the available resources. If both
\code{load_prop} and \code{RAM} are provided, the number of workers will
be the minimum based on the constraints imposed by both. Defaults to
\code{0.8} if not provided.}

\item{chunk_size}{Integer. Number of runs submitted per batch. Parallelism is
controlled by \code{workers}; chunking mainly controls memory/queue size.}

\item{workers}{A number of parrallel workers. Determines the number of hespdiv runs to be processed in parallel.}
}
\value{
A \code{hsa_constrained} class object, a list with two elements:
\itemize{
\item \strong{Alternatives}
A named list corresponding to each polygon where each entry is another
list of \code{hespdiv} results for each subsample run.

\item \strong{Basis}
The original \code{hespdiv} object (\code{obj}).
}
}
\description{
Conduct a constrained sensitivity analysis on a \code{hespdiv} object by
repeatedly subsampling observations within each polygon. Each subsample
is used to call \code{hespdiv} with recursion disabled (i.e., single-split only).
}
\details{
For each polygon in the \code{hespdiv} object, this function draws \code{subsample_factor}
of the data (by default 70\%), creating multiple random subsamples (\code{n.runs}).
These are processed in chunks (as given by \code{chunk_size}) and runs in each chunk being
parallelized to manage memory usage.
}
\seealso{
\code{\link{hespdiv}} for details on the main function.
\CRANpkg{future.apply}

Other functions for hespdiv sensitivity analysis: 
\code{\link{change_base}()},
\code{\link{hsa}()},
\code{\link{hsa_detailed}()},
\code{\link{hsa_quant}()},
\code{\link{plot_cs_hsa}()},
\code{\link{plot_hsa}()},
\code{\link{plot_hsa_q}()}

Other functions for hespdiv results post-processing: 
\code{\link{cross_comp}()},
\code{\link{hsa}()},
\code{\link{hsa_detailed}()},
\code{\link{hsa_quant}()},
\code{\link{nulltest}()},
\code{\link{taxon_effect}()}
}
\concept{functions for hespdiv results post-processing}
\concept{functions for hespdiv sensitivity analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/change_base.R
\name{change_base}
\alias{change_base}
\title{Change the basal subdivision in hsa}
\usage{
change_base(obj, id)
}
\arguments{
\item{obj}{A \code{hsa} class object.}

\item{id}{An index of an alternative subdivision to be used as a new basal
subdivision.}
}
\value{
The \code{hsa} class object with a new basal subdivision.
}
\description{
This function allows you to select a new basal subdivision
from the results of hespdiv sensitivity analysis by specifying its ID. It
provides a convenient way to switch between different basal subdivisions.
You can identify a more stable subdivision alternative by plotting the
hespdiv sensitivity analysis results using the \code{plot_hsa} function.
By selecting a new basal subdivision, you can observe how it affects the
results of polygon object cross-comparison and the stability of hespdiv
clusters and polygons.
}
\seealso{
Other functions for hespdiv sensitivity analysis: 
\code{\link{hsa}()},
\code{\link{hsa_detailed}()},
\code{\link{hsa_quant}()},
\code{\link{hsa_sample_constrained}()},
\code{\link{plot_cs_hsa}()},
\code{\link{plot_hsa}()},
\code{\link{plot_hsa_q}()}
}
\author{
Liudas Daumantas
}
\concept{functions for hespdiv sensitivity analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_data.R
\name{get_data}
\alias{get_data}
\title{Get data points that lie inside a polygon}
\usage{
get_data(polygon, xy.dat)
}
\arguments{
\item{polygon}{A data frame of 2 columns ("x","y") that contain coordinates of polygon vertices.}

\item{xy.dat}{A data frame, containing "x" and "y" columns. Other columns may also be present.}
}
\value{
A filtered data frame.
}
\description{
This function extracts points from a provided dataset that lie inside a given polygon.
}
\note{
This function excludes points that are strictly outside the given
polygon. Points lying on the edges or vertices of the
polygon (if the polygon is not closed) will be included in the
filtered data frame.
}
\examples{
#Creating data.frame of a polygon
poly<- data.frame(c(3.38,3.30,1.70,0.78,-0.06,-2.30,-2.94,-3.97,-1.61,-0.39,0.68,1.28,1.60,3.38),
c(-0.12,-0.31,-2.73,-3.22,-3.29,-2.19,-1.62,0.94,3.10,3.00,2.91,2.49,2.20,-0.12))

#Creating a data set of points
xy.dat<-data.frame(x=runif(250,-4,4),y=runif(250,-4,4))
plot(poly,type='l',xlab="X",ylab="Y")
points(xy.dat)

#Extracting points that lie inside a polygon
points(get_data(poly,xy.dat),pch=19,col=2)
}
\author{
Liudas Daumantas
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/depend_splits.R
\name{depend_splits}
\alias{depend_splits}
\title{Identify dependent split-lines and polygons}
\usage{
depend_splits(obj, id)
}
\arguments{
\item{obj}{hespdiv object}

\item{id}{id of a split-line}
}
\value{
a list:
\enumerate{
\item Dependent split-line IDs, if any
\item Dependent polygon IDs
}
}
\description{
For a provided split-line id, function returns a list of
all offspring polygons and split-lines, if there are any.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_cs_hsa.R
\name{plot_cs_hsa}
\alias{plot_cs_hsa}
\title{Visualize Constrained HespDiv Sensitivity Analysis Results}
\usage{
plot_cs_hsa(
  obj,
  type = 1,
  rank = NULL,
  col_basal = "gray20",
  main,
  col_boundary = 7,
  col_alternatives = "lightyellow3",
  max_lwd = 2.5,
  min_lwd = 0.75,
  alpha_alt = 0.6
)
}
\arguments{
\item{obj}{An object of class \code{hsa_constrained}, the output
of \code{\link{hsa_sample_constrained}}.}

\item{type}{An integer indicating the type of plot. Defaults to 1.
\describe{
\item{\code{1}}{A single plot overlaying all alternative lines
and the basal lines, colored or width-coded by rank.}
\item{\code{2}}{Multiple plots, each corresponding to a specific
rank. Alternative lines are displayed in a user-defined color
(default \code{"lightyellow3"}), with the basal line highlighted
in another color.}
}}

\item{rank}{Integer. Optional. When \code{type = 2}, if provided the function
will only generate plots for the specified rank. If \code{NULL} (default),
plots for all unique ranks will be produced.}

\item{col_basal}{Character or numeric specifying the color of basal split-lines
(default \code{"gray20"}).}

\item{main}{Character. Title for the plot(s).}

\item{col_boundary}{Character or numeric specifying the color of the outer
(first) polygon boundary (default \code{7}).}

\item{col_alternatives}{Character or numeric specifying the color of alternative
split-lines (default \code{"lightyellow3"}).}

\item{max_lwd}{Numeric. The maximum line width for the highest-ranked split-line
(default \code{2.5}).}

\item{min_lwd}{Numeric. The minimum line width for the lowest-ranked split-line
(default \code{0.75}).}

\item{alpha_alt}{Numeric in the range \code{[0, 1]}. The transparency of
alternative split-lines (default \code{0.6}).}
}
\value{
\code{NULL}. The function is called for its side effect of generating
one or more plots.
}
\description{
Displays the alternative (subsampled) \code{hespdiv} subdivisions and the
basal (original) \code{hespdiv} subdivision on one or multiple plots,
illustrating how the split-lines vary across different ranks. Additionally,
for each alternative split-line (which is defined by a start and end coordinate),
the function aggregates identical endpoints and overlays their counts on the plot.
}
\details{
\itemize{
\item In \code{type = 1}, the function creates a single plot showing all
alternative split-lines overlaid on the first polygon boundary,
plus all basal split-lines of the \code{hespdiv} basis object.
\item In \code{type = 2}, the function creates separate plots, each focusing
on polygons of a specific rank, drawing alternative lines in the user-specified
color (with transparency) and the basal line in another color or line width.
If a specific \code{rank} is provided, only that rank is plotted.
\item In both cases, after drawing the alternative split-lines the function
aggregates their endpoints (start and end coordinates) and overlays the
count at each unique coordinate using \code{text()}.
}
}
\seealso{
Other functions for hespdiv sensitivity analysis: 
\code{\link{change_base}()},
\code{\link{hsa}()},
\code{\link{hsa_detailed}()},
\code{\link{hsa_quant}()},
\code{\link{hsa_sample_constrained}()},
\code{\link{plot_hsa}()},
\code{\link{plot_hsa_q}()}

Other HespDiv visualization options: 
\code{\link{blok3d}()},
\code{\link{create_gif}()},
\code{\link{dendro}()},
\code{\link{plot.nullhespdiv}()},
\code{\link{plot_hespdiv}()},
\code{\link{plot_hsa}()},
\code{\link{plot_hsa_q}()},
\code{\link{poly_scheme}()},
\code{\link{polypop}()}
}
\concept{HespDiv visualization options}
\concept{functions for hespdiv sensitivity analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remove_splits.R
\name{remove_splits}
\alias{remove_splits}
\title{Remove split-lines}
\usage{
remove_splits(obj, split.id, depend.splits = TRUE)
}
\arguments{
\item{obj}{hespdiv object}

\item{split.id}{vector of split-line ids}

\item{depend.splits}{logical. Remove split-lines that depend on specified
split-lines? If FALSE, only end-nodes of spatial dendrogram are removed.}
}
\value{
hespdiv object
}
\description{
Function returns hespdiv object, without split-lines of
specified id.
}
\examples{
\donttest{
if (requireNamespace("HDData")) {

  # Inspect the hespdiv object
  print(plot_hespdiv(HDData::hd))

  # Remove weak split-lines
  weak_splits <- which(HDData::hd$split.stats$performance >= 0.3)
  performance_filtered <- remove_splits(obj = HDData::hd, split.id = weak_splits)
  print(plot_hespdiv(performance_filtered))

  # Remove non-significant split-lines
  plot(HDData::nl)
  nsig_splits <- which(HDData::nl[[1]]$quantile >= 0.05)
  sig_filtered <- remove_splits(obj = HDData::hd, split.id = nsig_splits)
  print(plot_hespdiv(sig_filtered))

  # Remove only if a split-line has no dependent split-lines
  unchanged_hd <- remove_splits(obj = HDData::hd, split.id = 4, depend.splits = FALSE)
  print(plot_hespdiv(unchanged_hd))

  # Remove the split-lines indicated as well as all other split-lines
  # that structurally depend on them (default behavior)
  changed_hd <- remove_splits(obj = HDData::hd, split.id = 4, depend.splits = TRUE)
  print(plot_hespdiv(changed_hd))
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_effect.R
\name{group_effect}
\alias{group_effect}
\title{Investigate Group Effects on Split-Line Performance}
\usage{
group_effect(obj, group, perm.n = 999, maxdif = NULL, plot = TRUE, ...)
}
\arguments{
\item{obj}{A \code{hespdiv} object.}

\item{group}{A factor (or coercible to factor) giving the group label for each observation (row) in
\code{obj$call.info$Call_ARGS$xy.dat}.}

\item{perm.n}{Integer (\eqn{>0}). Number of permutations in the locality-block permutation test. Default \code{999}.}

\item{maxdif}{Numeric. The performance value representing a maximal between-polygons difference for the chosen metric.
If \code{NULL} and the metric is one of \code{c("pielou","morisita","sorensen","horn.morisita")},
\code{maxdif} is set automatically. Otherwise, it must be provided.}

\item{plot}{Logical. If \code{TRUE} (default), stores a \code{\link{plot_hespdiv}} output per group with
\code{performance = TRUE} and a group-specific subtitle.}

\item{...}{Additional arguments passed to \code{\link{plot_hespdiv}}. Note that \code{obj},
\code{performance}, and \code{subtitle} are set internally.}
}
\value{
A list of class \code{"group_effect"} with elements:
\itemize{
\item \code{within = list(est, delta)} where \code{est} and \code{delta} are \code{[n_splits x n_groups]} matrices;
\item \code{elim   = list(est, delta)} as above (group removed);
\item \code{perm   = list(est, delta)} where each is a nested list \code{[[split]][[group]]} of numeric vectors (length \code{perm.n}).  If permutations are uninformative (e.g., one-sided/absent), the corresponding entries are \code{NA};
\item \code{baseline}: the original performance vector;
\item \code{n_per_pol}: \code{data.frame} with columns \code{split.id}, \code{group}, \code{pol.id}, \code{n};
\item \code{plots}: list of \code{plot_hespdiv} outputs by group (or \code{NULL} if \code{plot = FALSE}).
}
}
\description{
For every level of \code{group}, the function:
\enumerate{
\item subsets \code{data} and \code{xy.dat} to the focal group's observations;
\item recomputes split-line performance using \code{compare.f} and \code{generalize.f};
\item runs two contribution assessments:
\itemize{
\item \emph{group-removal}: removes the group's observations and recomputes performance;
\item \emph{group-permutation (locality-block shuffle)}: shuffles intact within-locality assemblages of the focal group
among the set of localities where the group occurs within the split’s parent polygons (non-group observations fixed).
}
\item optionally re-plots the object for that group via \code{\link{plot_hespdiv}} (with a group-specific subtitle).
}
}
\details{
Re-evaluates split-line performance within each level of a grouping factor and
tests how much each group influences the detected split-lines in a \code{hespdiv} object.

\strong{Within-group recomputation (agreement).}
For each split, the group's observations are partitioned by the two child polygons and
performance is recomputed as \code{compare.f(generalize.f(pol1), generalize.f(pol2))}.
If the group's points fall on only one side, \code{within$est} is set to \code{maxdif};
if absent from both sides, it is \code{NA}.

\strong{Elimination test.}
All observations of the focal group are removed from both polygons and performance is recomputed on the remaining data.
If the group's points fall on only one side, \code{elim$est} is computed normally;
if absent from both sides, it is \code{NA} (if desired, you could identify these cases afterwards from \code{n_per_pol}, changing to baseline performance and zero delta).

\strong{Permutation test (locality-block shuffle).}
Localities are defined by identical coordinate pairs among the focal group's occurrences.
For each permutation, whole localities (blocks) are reassigned via a one-to-one mapping within the split’s parent polygons.
For speed, polygon membership of unique locality coordinates is precomputed once per split and then locality memberships are shuffled in each permutation.
If the group's points fall on only one side, \code{perm$est} is set to NA;
if absent from both sides, it is also \code{NA}.

\strong{Baseline and deltas.}
The baseline vector is \code{baseline <- obj$split.stats$performance}.
Let \code{maximize <- obj$call.info$Call_ARGS$maximize}.
Deltas are defined so that \emph{positive values always indicate a positive contributor} (removal/permutation worsens performance):
\itemize{
\item if \code{maximize == FALSE} (lower is better; e.g. similarity): \code{delta = est - baseline};
\item if \code{maximize == TRUE}  (higher is better; e.g. distance):  \code{delta = baseline - est}.
}
}
\section{Notes}{

\itemize{
\item Localities are defined by exact duplicate coordinate pairs among the group's observations
(harmonise coordinates upstream if needed).
\item The locality-block permutation preserves the \emph{number of group-present localities per polygon};
only identities of the locality blocks are shuffled. The number of group occurrences per side can vary if block sizes differ.
}
}

\examples{
data <- example_hespdiv$call.info$Call_ARGS$data
# See the hespdiv() example; lowercase letters represent widespread taxa:
endemic_l <- !data \%in\% letters[1:5]
group <- factor(ifelse(endemic_l, "endemic", "widespread"))

gr_ef <- group_effect(obj = example_hespdiv, group = group, plot = FALSE)

barplot(
  gr_ef$within$est,
  main = "Split-line similarity within each group",
  ylab = "Performance"
)

barplot(
  gr_ef$within$delta,
  main = "Group contribution to similarity within each group",
  ylab = "Contribution"
)
abline(h = 0)

barplot(
  gr_ef$elim$est,
  main = "Split-line similarity with group eliminated",
  ylab = "Performance"
)

barplot(
  gr_ef$elim$delta,
  main = "Group contribution to similarity with group eliminated",
  ylab = "Contribution"
)
abline(h = 0)

boxplot(
  gr_ef$perm$est$`1`,
  main = "Split-line similarity with group spatially permuted",
  ylab = "Performance"
)

boxplot(
  gr_ef$perm$delta$`1`,
  main = "Group contribution to similarity with group spatially permuted",
  ylab = "Contribution"
)
abline(h = 0)
}
\seealso{
\code{\link{plot_hespdiv}}
}
\concept{functions for hespdiv post-processing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dendro.R
\name{dendro}
\alias{dendro}
\title{Plot hespdiv dendrogram}
\usage{
dendro(
  obj,
  poly.scheme = NULL,
  color = 1,
  performance.col = "blue",
  labels.col = 1,
  offset.factor = 1,
  arrange = TRUE,
  grob = TRUE,
  label.size = 0.5
)
}
\arguments{
\item{obj}{A hespdiv object.}

\item{poly.scheme}{ggplot2 object produced with poly_scheme function. Provide if you want identical colors for polygons in both plots.}

\item{color}{color vector used for dendrogram nodes and branches.}

\item{performance.col}{color vector used for text, displaying difference values between polygons}

\item{labels.col}{color vector used for text, displaying polygon IDs.}

\item{offset.factor}{numeric value used to scale the offset distance of displayed polygon IDs and performance values from a dendrogram node. Adjust experimentally, if you don't like the current distance.}

\item{arrange}{logical. Plot hespdiv dendrogram above polygon scheme?}

\item{grob}{logical. Convert plot to grob? Must be true, if you want to arrange polygon scheme and the dendrogram in a single plot.}

\item{label.size}{size of labels.}
}
\value{
A \code{grob} or \code{TableGrob} object if \code{grob = TRUE}, otherwise \code{NULL}.
}
\description{
This function displays a dendrogram of polygons produced by hespdiv split-lines. Branch length is proportional to difference. If performance of split-lines is a similarity measure, it is internally converted to difference.
}
\note{
If you want to transform similarity to difference externally, before applying dendro, change maximize to TRUE in the call info of obj.
}
\examples{
scheme <- poly_scheme(example_hespdiv)
# notice the colors in used in scheme:
scheme
# Dendrogram visualization of polygons, using colors from scheme
dendro(example_hespdiv, poly.scheme = scheme, arrange = FALSE, grob = FALSE)
}
\seealso{
Other HespDiv visualization options: 
\code{\link{blok3d}()},
\code{\link{create_gif}()},
\code{\link{plot.nullhespdiv}()},
\code{\link{plot_cs_hsa}()},
\code{\link{plot_hespdiv}()},
\code{\link{plot_hsa}()},
\code{\link{plot_hsa_q}()},
\code{\link{poly_scheme}()},
\code{\link{polypop}()}
}
\author{
Liudas Daumantas
}
\concept{HespDiv visualization options}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hespdiv.R
\name{hespdiv}
\alias{hespdiv}
\title{Hierarchically subdivide spatial data}
\usage{
hespdiv(
  data,
  xy.dat = NULL,
  n.split.pts = 15,
  same.n.split = TRUE,
  method = "horn.morisita",
  generalize.f = NULL,
  compare.f = NULL,
  maximize = NULL,
  N.crit = 1,
  N.rel.crit = 0.2,
  N.loc.crit = 1,
  N.loc.rel.crit = 0.2,
  S.crit = 0.05,
  S.rel.crit = 0.2,
  Q.crit = NULL,
  c.splits = TRUE,
  c.Q.crit = NULL,
  c.crit.improv = 0,
  c.X.knots = 5,
  c.Y.knots = 10,
  c.max.iter.no = +Inf,
  c.fast.optim = FALSE,
  c.corr.term = 0.05,
  study.pol = NULL,
  use.chull = TRUE,
  tracing = NULL,
  pnts.col = 1,
  display = FALSE,
  pacific.region = FALSE,
  .do_recurse = TRUE
)
}
\arguments{
\item{data}{An R object that contains the data to be analyzed. The required data
structure depends on the selected method (e.g., character vector of taxa names).}

\item{xy.dat}{A data.frame containing the coordinates for observations in the
\code{data}. This parameter can be ignored if \code{data} is a data frame or
matrix with columns \code{x} or \code{y}.}

\item{n.split.pts}{integer. The number of split-points - 1. These points are
used in creating straight split-lines (see details). The total number of straight
split-lines generated can be obtained by \code{sum(1:n.split.pts)}.
Increasing the value of \code{n.split.pts} leads to an increase in both the
computation time and the fit to the data.}

\item{same.n.split}{logical. Should the number of split-points
(\code{n.split.pts}) remain the same within each lower-ranked polygon?
Choosing \code{TRUE} would result in a higher density of straight split-lines
within lower-ranked polygons, whereas \code{FALSE} would preserve the same
split-line density. Thus, essentially you are choosing between cross-scale
and fixed-scale analysis.}

\item{method}{Character. Name or abbreviation of a preset method:
\code{"sorensen"}, \code{"pielou"}, \code{"morisita"}, or
\code{"horn.morisita"}. Internally determines values for
\code{compare.f}, \code{generalize.f}, and \code{maximize}.}

\item{generalize.f}{Function. Optional function used in custom methods to
prepare input for the \code{compare.f} function; see Details. It must have
a single argument, for example \code{x}, to which a spatially filtered
subset of \code{data} can be assigned; see Note for exceptions.}

\item{compare.f}{function. Only required in custom methods. Employed to
quantify the comparison value of a split-line. For this purpose, the
\code{compare.f} function requires two arguments where the outputs of the
\code{generalize.f} function can be assigned (see notes for exceptions).}

\item{maximize}{logical. Only required in custom methods. Determines whether
the split-line comparison value should be maximized or minimized during the
optimization process.}

\item{N.crit}{number. Minimum required number of observations that should
be present in each polygon obtained by a split-line in order for it to meet
this criterion.}

\item{N.rel.crit}{number from 0 to 0.5. Each polygon obtained with a
split-line must have at least such proportion of observations to pass this
criterion. Equation of the proportion:  (Number of observations in 1st/2nd
resulting polygon) / (Number of observations in the polygon being subdivided)}

\item{N.loc.crit}{number. Minimum required number of different locations that
should be present in each polygon obtained by a split-line in order for it to
meet this criterion.}

\item{N.loc.rel.crit}{number from 0 to 0.5. Each polygon obtained with a
split-line must have at least such proportion of different locations to pass this
criterion. Equation of the proportion:  (Number of different locations in 1st/2nd
resulting polygon) / (Number of different locations in the polygon being subdivided)}

\item{S.crit}{number from 0 to 1. Each polygon obtained with a
split-line must have at least such area proportion to pass this
criterion. Equation of the proportion:  (Area of 1st/2nd resulting polygon) /
(Area of the first polygon). The first polygon is the provided study area
or convex hull of observation locations.}

\item{S.rel.crit}{number from 0 to 0.5. Each polygon obtained with a
split-line must have at least such area proportion to pass this
criterion. Equation of the proportion:  (Area of 1st/2nd resulting polygon) /
(Area of the polygon being subdivided).}

\item{Q.crit}{number. The threshold for a split-line comparison value to be
considered acceptable for a subdivision. When \code{maximize = TRUE}, higher
values of the comparison value indicate a better subdivision. Conversely, when
\code{maximize = FALSE}, lower values of the comparison value indicate a
better subdivision.}

\item{c.splits}{logical. When set to TRUE, the algorithm will explore
nonlinear split-lines in addition to straight split-lines in order to
find the optimal subdivision.}

\item{c.Q.crit}{number.  The threshold for a split-line comparison value to
be considered acceptable for generating nonlinear split-lines. It is
recommended to use the default value, which does not impose a performance
requirement, unless you have a clear understanding of the potential
improvements that nonlinear split-lines can achieve over straight split-lines.
If \code{maximize = TRUE}, a suggested value for \code{c.Q.crit} is
\code{Q.crit} minus the maximum potential improvement.}

\item{c.crit.improv}{number. The threshold for the improvement in a
split-line comparison value required for a nonlinear split-line to be
selected instead of a straight split-line for subdivision. The default value
of 0 means that even if a nonlinear split-line performs equally to a
straight split-line, the straight split-line will still be chosen.}

\item{c.X.knots}{integer. Specifies the number of columns in a network of
spline knots used to generate nonlinear split-lines. These knots are evenly
distributed along the straight split-line. Adjusting the value of
\code{c.X.knots} controls the degree of wiggliness (number of turns) in the
resulting nonlinear  split-lines.}

\item{c.Y.knots}{integer. specifies the number of  rows in a network of
spline knots used to generate nonlinear split-lines. These knots are
distributed regularly along lines orthogonal to the straight
split-line. Adjusting the value of \code{c.Y.knots} controls the number
of amplitudes tested in each wiggle of a spline, influencing the shape of
nonlinear split-lines.}

\item{c.max.iter.no}{integer. The maximum number of iterations allowed through
the network of spline knots when searching for the optimal shape of a
nonlinear split-line. Setting a higher value, such as \code{+Inf}, increases
the chances of converging to the best possible curve. It is recommended to
use higher values when \code{c.fast.optim = FALSE}, as in this case, only a
single spline knot can be selected per iteration.}

\item{c.fast.optim}{logical. Determines when spline knots are selected. If
\code{TRUE}, the algorithm selects the first knot that generates a curve with
a better comparison value, ensuring faster convergence. If set to
\code{FALSE}, the algorithm completes a full iteration through the network
of spline knots, which may result in slower convergence but potentially
better overall performance.}

\item{c.corr.term}{number from 0.01 to 0.2. A correction term for nonlinear
split-lines that intersect the boundary of the polygon. Smaller values
(default is 0.05) are recommended, as they determine the extent to
which the outlying interval of the generated spline, which crosses the
polygon boundary, should be shifted away from the boundary and inside the
polygon in a direction orthogonal to the straight split-line. This shift is
specified as a proportion of the polygon height where the spline intersects
the polygon boundary.}

\item{study.pol}{A data frame with two columns, \code{x} and \code{y}. Rows
should correspond to vertices of a polygon that defines the study area
encompassing all observation locations in \code{xy.dat}. If not provided
(default is \code{NULL}), the convex hull of \code{xy.dat} is used as the
study area polygon.}

\item{use.chull}{logical. If \code{study.pol} is provided, you can use
\code{use.chull = TRUE} (which is default) to use it only for visualization.}

\item{tracing}{Optional character vector of length two controlling diagnostic
tracing output. The first element specifies the tracing level and must be
one of \code{"best"}, \code{"main"}, or \code{"all"}. The second element
specifies the traced object and must be one of \code{"curves"},
\code{"straight"}, or \code{"both"}. The default, \code{NULL}, produces no
tracing output.}

\item{pnts.col}{character or numeric. Specifies the color of observations
in a plot. The argument is used when \code{tracing} is not NULL. If
\code{pnts.col = NULL}, observations will not be displayed.}

\item{display}{logical. Display a simple plot of results at the end of
computations?}

\item{pacific.region}{logical (default is FALSE). When set to TRUE, indicates
that the study area is crossed by the 180th meridian, such as being within
the Pacific Ocean. In this case, the coordinates of  \code{xy.dat} and
\code{study.pol} are transformed to eliminate the artificial abrupt change
in x-coordinate values at the 180th meridian.}

\item{.do_recurse}{Logical. Controls recursion (for internal use only). Default is \code{TRUE}.
Normal users should not modify this parameter.}
}
\value{
A \code{hespdiv} object, which is a list with seven elements:
\describe{
\item{\code{poly.stats}}{
A data frame containing information about polygons established by the
selected split-lines. Its columns are:
\itemize{
\item \code{rank}: The rank of a polygon. It corresponds to the rank of the
split-line that produced the polygon and to the polygon's position in the
hierarchical structure of the subdivision.
\item \code{plot.id}: The ID assigned to the polygon. It corresponds to the
order in which the polygon was processed during the \code{hespdiv} analysis.
\item \code{root.id}: The ID of the parent polygon whose subdivision
resulted in the current polygon.
\item \code{n.splits}: The number of straight split-lines evaluated in an
attempt to subdivide the polygon. This count excludes split-lines that crossed
the polygon boundary or did not meet area, sample size, or location-number criteria.
\item \code{n.obs}: The number of observations inside the polygon.
\item \code{mean}: The average comparison value of the straight split-lines
used in the attempted subdivision of the polygon. This value reflects the
general spatial heterogeneity of the data.
\item \code{sd}: The standard deviation of the comparison values of the
straight split-lines used in the attempted subdivision. It indicates
the extent of anisotropy or variation in spatial heterogeneity within the polygon.
\item \code{str.best}: The comparison value of the best straight split-line
produced within the polygon.
\item \code{str.z.score}: The z-score of the comparison value of the best
straight split-line within the polygon. It indicates how outstanding the best
straight split-line is relative to other evaluated straight subdivisions.
\item \code{has.split}: Logical. Indicates whether a subdivision was established
in the polygon.
\item \code{is.curve}: Logical. Indicates whether the established subdivision
was obtained using a curve. If \code{has.split} is \code{FALSE}, this value is \code{NA}.
\item \code{crv.best}: The same as \code{str.best}, but for nonlinear split-lines.
\item \code{crv.z.score}: The same as \code{str.z.score}, but for nonlinear split-lines.
\item \code{c.improv}: The improvement in comparison value achieved by using
nonlinear split-lines instead of straight split-lines.
}
}
\item{\code{split.stats}}{
A data frame containing information about established split-lines.
Its columns can be interpreted similarly to those in \code{poly.stats},
but from the perspective of split-lines rather than polygons. For example,
\code{rank} is the rank of the split-line, not of the polygon.
The \code{performance} column contains the comparison value of the split-line
and corresponds to either \code{str.best} or \code{crv.best} in \code{poly.stats},
depending on the value of \code{is.curve}. The same applies to the
\code{z.score} column.
}
\item{\code{split.lines}}{
A list of data frames containing coordinates of established split-lines.
The order of split-lines in this list corresponds to their order in
\code{split.stats}.
}
\item{\code{polygons.xy}}{
A list of data frames containing coordinates of polygons established by
the split-lines. The order of polygons in this list corresponds to their
order in \code{poly.stats}.
}
\item{\code{poly.obj}}{
A list of polygon objects, that is, outputs of \code{generalize.f} for each polygon.
The order of elements in this list corresponds to the row order of
\code{poly.stats} and the polygon order in \code{polygons.xy}.
}
\item{\code{call.info}}{
Information about the \code{hespdiv} call, including the method and arguments used.
}
\item{\code{str.difs}}{
A list containing comparison values of evaluated straight split-lines for
each polygon. The elements of this list correspond to the rows of
\code{poly.stats}.
}
}
}
\description{
This function is an implementation of spatial data analysis method HespDiv.
It performs hierarchical spatial data subdivision by recursively
dividing the data using random split-lines, evaluating their comparison
values (how well they separate data), and using the best to perform
subdivisions.
}
\details{
\subsection{The Algorithm}{
\bold{1) Split-point Placement:} The function places a predetermined number of
split-points (\code{n.split.pts + 1}) along the perimeter of the study area
(\code{study.pol}) or the convex hull of observation locations
(\code{xy.dat}) if a study area polygon is not provided. These split-points
are evenly spaced, resulting in a distance between points equal to
1/(n.split.pts + 1) fraction (1/16 by default) of a polygon circumference.

\bold{2) Straight Split-lines:} Straight split-lines are generated by connecting
the split-points. The total number of straight split-lines generated is equal
to the value of \code{(n.split.pts + 1) * n.split.pts / 2} or to
\code{choose(n.split.pts + 1, 2)}. This holds true only in the first iteration when
\code{same.n.split} is set to \code{FALSE} or in all iterations when
\code{same.n.split} is set to \code{TRUE}. Note that the total number of
split-lines generated will not be equal to the number of split-lines
evaluated since split-lines that cross polygon boundary or do not pass
sample size, area or location number subdivision criteria are not evaluated.

\bold{3) Subdivisions:} Each split-line spatially divides the data and study area
polygon into two subsets.

\bold{4) Criteria:} Both subsets are then checked to see if they meet
sample size, area and location number criteria.

\bold{5) Obtaining comparison values:} Subsets that meet the criteria are
compared using the \code{generalize.f} and \code{compare.f} functions to
obtain a comparison value. First, each subset is passed to
\code{generalize.f} to obtain a generalization value, such as the Pielou
evenness index for the \code{"pielou"} method; see the \strong{Note} section
for clarification. These values are then used by \code{compare.f} to compare
the subsets. In this way, each split-line that passed all subdivision criteria
is assigned a comparison value.

\bold{6) Best Straight Split-line Selection:} The best performing straight
split-line is determined based on whether the \code{maximize} argument is
set to \code{TRUE} or \code{FALSE}. If \code{maximize} is \code{TRUE}, the
best split-line is the one with the highest comparison value; if
\code{maximize} is \code{FALSE}, the best split-line has the lowest
comparison value.

The combination of the \code{generalize.f}, \code{compare.f}, and
\code{maximize} arguments can be provided, enabling the creation of custom
methods, or it can be determined internally based on the chosen preset
subdivision methods (see below).

\bold{7) Nonlinear Split-lines:} If the best straight split-line meets
the quality criteria specified by the \code{c.Q.crit}, it serves as the basis
for generating variously shaped curves (nonlinear split-lines). These curves
are produced using splines, which are mathematical functions that can create
smooth and flexible curves.

To generate the curves, a number of knots (control points) for the splines are
distributed evenly along a set of lines orthogonal to the best straight
split-line. These orthogonal lines are also evenly distributed along the
straight split-line itself. The number of knots and lines used can be
adjusted through the parameters \code{c.Y.knots} and \code{c.X.knots},
respectively.

The algorithm then iterates through this network of knots, considering
different combinations of knots, to produce curves. By varying the selection
and arrangement of knots, different shapes of curves are generated.

\bold{8) Nonlinear Split-line Evaluation and Selection:} Curves are then
processed in the same manner as straight split-lines in steps 3 to 6.

\bold{9) Final Split-line Selection and Establishment of Subdivision:} If the best
curve outperforms the best straight split-line by a margin of
\code{c.crit.improv}, the best split-line becomes nonlinear. Otherwise, it
remains straight. The split-line must also satisfy the criteria established
by the \code{Q.crit} argument to be used for subdivision.

\bold{10) Recursive Iteration:} The process described above is iteratively applied,
resulting in a collection of selected split-lines with their performance
values. These split-lines hierarchically subdivide space and data, forming
polygons of various shapes.
}
\subsection{Preset Methods}{
Preset methods and their combinations of \code{compare.f},
\code{generalize.f}, and \code{maximize}:
\describe{
\item{\code{"sorensen"}}{
Functions calculate the Sorensen similarity index
(Sorensen 1948). \code{maximize = FALSE}. Values vary from 0 to 1.
}
\item{\code{"pielou"}}{
Functions calculate the mean proportional decrease in
Pielou evenness (Pielou 1966) that occurs after polygon subdivision.
\code{maximize = TRUE}. Values vary from 0 to 1.
}
\item{\code{"morisita"}}{
Functions calculate the Morisita overlap index
(Morisita 1959). \code{maximize = FALSE}. Values vary from 0 to 1.
}
\item{\code{"horn.morisita"}}{
Functions calculate the Morisita overlap index, as modified by
Horn (1966). \code{maximize = FALSE}. Values vary from 0 to 1.
}
}
All preset methods currently available are specifically designed for
bioregionalization purposes. These methods require two key inputs: the
coordinates of fossil taxon occurrences (\code{xy.dat}) and the names or IDs
of taxa (\code{data}). These names or IDs should be structured as
character or numeric vectors, with each element corresponding to a row in the
\code{xy.dat} data frame. Each method compares fossil communities on
opposite sides of a split-line, aiming to minimize similarity or
maximize difference. The output yields biogeographical provinces with a
hierarchical structure.
}
}
\note{
Please note that if you use the \code{method} argument, the arguments
\code{generalize.f}, \code{compare.f}, and \code{maximize} are determined
internally and should not be provided. Therefore, you should only assign
values to these arguments when using a custom method, not a predefined one.
Additionally, you can ignore the \code{generalize.f} argument even when
applying custom methods. If \code{generalize.f} is set to NULL (default), the
data remains unchanged, as \code{generalize.f} acts as an identity function.
Hence, \code{generalize.f} is only an optional argument that allows to omit
the transformation or generalization step in \code{compare.f} function,
simplifying it.

Both \code{generalize.f} and \code{compare.f} inherit environments from
parent functions: \code{hespdiv()} and \code{.spatial_div()}. This
allows them to use additional variables from those environments, such as
\code{xy.dat}, \code{samp.xy}, \code{id1}, and \code{id2}.

There is a small possibility that a nonlinear split-line may cross a polygon
boundary. If the result contains such a split-line, or if an error related to
this issue occurs, slightly change one of the arguments, such as
\code{n.split.pts}, and re-run \code{hespdiv()}.
}
\examples{
# Simulated data with a known boundary at x = 0.45 to illustrate boundary detection

study.area <- data.frame(
  x = c(0, 0.8, 1, 0.6, 0, 0),
  y = c(0, 0, 0.4, 1.1, 1.1, 0)
)

set.seed(852)
# Simulate 100 occurrence coordinates
N <- 100
xy.dat_arg <- data.frame(
  x = rpois(N, 4) / 10,
  y = rpois(N, 4) / 10
)

xy.dat_arg <- xy.dat_arg[order(xy.dat_arg$x), ]

# Simulate a boundary at x = 0.45
n_left <- sum(xy.dat_arg$x < 0.45)

set.seed(1)
common_data <- letters[1:5]

left_data <- sample(
  c(common_data, LETTERS[1:10]),
  size = n_left,
  replace = TRUE,
  # common-endemic probability ratio 3:4
  prob = c(rep(3, 5), rep(4, 10))
)

right_data <- sample(
  c(common_data, LETTERS[11:20]),
  size = N - n_left,
  replace = TRUE,
  # common-endemic probability ratio 3:4
  prob = c(rep(3, 5), rep(4, 10))
)

data_arg <- c(left_data, right_data)

# Apply hespdiv
r <- hespdiv(
  data = data_arg,
  xy.dat = xy.dat_arg,
  n.split.pts = 6,  # small value used here for illustration
  method = "sor",   # subdivision minimizing Sorensen-Dice similarity
  S.crit = 0.3,     # minimum area size is 30\% of study area
  study.pol = study.area,
  use.chull = FALSE
)

plot_hespdiv(r, n.loc = TRUE) + ggplot2::geom_vline(xintercept = 0.45)

# Detected split-line performance
r$split.stats$performance

# Sorensen-Dice similarity across the true simulated boundary
2 * length(intersect(left_data, right_data)) /
  (length(unique(left_data)) + length(unique(right_data)))
}
\references{
Horn, H. S. (1966). Measurement of" overlap" in comparative ecological studies. The American Naturalist, 100(914), 419-424.

Morisita, M. (1959). Measuring of interspecific association and similarity between assemblages. Mem Fac Sci Kyushu Univ Ser E Biol, 3, 65-80.

Pielou, E. C. (1966). The measurement of diversity in different types of biological collections. Journal of theoretical biology, 13, 131-144.

Sorensen, T. A. (1948). A method of establishing groups of equal amplitude in plant sociology based on similarity of species content and its application to analyses of the vegetation on Danish commons. Biol. Skar., 5, 1-34.
}
\author{
Liudas Daumantas
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cross_comp.R
\name{cross_comp}
\alias{cross_comp}
\title{Calculate a polygon-object cross-comparison matrix}
\usage{
cross_comp(obj)
}
\arguments{
\item{obj}{A \code{hespdiv} object.}
}
\value{
A numeric matrix containing pairwise comparison values among the
\code{hespdiv} polygon objects stored in \code{obj$poly.obj}.
}
\description{
Computes a cross-comparison matrix for polygon objects from a \code{hespdiv}
object. The matrix quantifies similarity or dissimilarity among polygon
objects and can be used for further analyses, such as clustering, either
directly or after transformation.
}
\details{
The \code{cross_comp()} function uses the \code{compare.f} function from
\code{obj$call.info$Call_ARGS} to perform pairwise comparisons of
\code{hespdiv} polygon objects stored in \code{obj$poly.obj}. The result is
a cross-comparison matrix.
}
\note{
Polygon cross-comparison is currently not available for the \code{"pielou"}
method. It is also not supported for custom methods whose \code{compare.f}
function relies on variables from environments other than the function's own
arguments.
}
\seealso{
Other functions for hespdiv results post-processing: 
\code{\link{hsa}()},
\code{\link{hsa_detailed}()},
\code{\link{hsa_quant}()},
\code{\link{hsa_sample_constrained}()},
\code{\link{nulltest}()},
\code{\link{taxon_effect}()}
}
\author{
Liudas Daumantas
}
\concept{functions for hespdiv results post-processing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print_hespdiv.R
\name{print.hespdiv}
\alias{print.hespdiv}
\title{Print a hespdiv object}
\usage{
\method{print}{hespdiv}(x, ...)
}
\arguments{
\item{x}{A \code{hespdiv} object.}

\item{...}{Additional arguments, currently ignored.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Formats and prints the rounded \code{split.stats} data frame from a
\code{hespdiv} object.
}
\author{
Liudas Daumantas
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_hsa_q.R
\name{plot_hsa_q}
\alias{plot_hsa_q}
\title{Plot the stability of hespdiv clusters}
\usage{
plot_hsa_q(obj, hist = FALSE)
}
\arguments{
\item{obj}{The output of \code{hsa_quant} (an object of class \code{hsa_quant}).}

\item{hist}{A Boolean value. If FALSE, EPDFs obtained with \code{plot(density(jac.sim))}
will be displayed instead of histograms.}
}
\value{
None
}
\description{
This function visualizes the stability of basal subdivision
clusters obtained from \code{hsa_quant}
}
\details{
The stability of each basal cluster is revealed by the distribution
of Jaccard similarity values with the 'analogues' clusters found in alternative
hespdiv subdivisions. For example, a unimodal distribution with a peak at
high similarity values (>0.8) indicates that the basal hespdiv cluster is
stable, even if the polygon boundaries are not. This situation may arise when
there is indeed a spatial structure within the data, but there are also wide
gaps between sampled regions (or more generally when there is limited spatial
data coverage). A unimodal distribution with a peak at medium values
(0.4-0.6) and a tail to higher values could also indicate a more persistent
spatial structure. On the other hand, a single peak at low values (<0.4)
indicates low cluster stability (e.g., bioregion does not exist). Finally,
uniform, bimodal, or other more complex distributions may indicate that the
stability and existence of the corresponding basal cluster depend on the
parameters used in alternative hespdiv calls.
}
\seealso{
Other functions for hespdiv sensitivity analysis: 
\code{\link{change_base}()},
\code{\link{hsa}()},
\code{\link{hsa_detailed}()},
\code{\link{hsa_quant}()},
\code{\link{hsa_sample_constrained}()},
\code{\link{plot_cs_hsa}()},
\code{\link{plot_hsa}()}

Other HespDiv visualization options: 
\code{\link{blok3d}()},
\code{\link{create_gif}()},
\code{\link{dendro}()},
\code{\link{plot.nullhespdiv}()},
\code{\link{plot_cs_hsa}()},
\code{\link{plot_hespdiv}()},
\code{\link{plot_hsa}()},
\code{\link{poly_scheme}()},
\code{\link{polypop}()}

Other functions to evaluate hesdpiv cluster stability: 
\code{\link{hsa_quant}()}
}
\author{
Liudas Daumantas
}
\concept{HespDiv visualization options}
\concept{functions for hespdiv sensitivity analysis}
\concept{functions to evaluate hesdpiv cluster stability}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/detailed_hsa.R
\name{hsa_detailed}
\alias{hsa_detailed}
\title{Detailed HespDiv Sensitivity Analysis}
\usage{
hsa_detailed(
  obj,
  comb.args = TRUE,
  pick.n.args = NULL,
  comb.type = NULL,
  n.combs = NULL,
  display = TRUE,
  images.path = NULL,
  paired = NULL,
  pnts.col = 1,
  data = NULL,
  xy.dat = NULL,
  same.n.split = NULL,
  n.split.pts = NULL,
  N.crit = NULL,
  N.rel.crit = NULL,
  N.loc.crit = NULL,
  N.loc.rel.crit = NULL,
  S.crit = NULL,
  S.rel.crit = NULL,
  Q.crit = NULL,
  c.splits = NULL,
  c.Q.crit = NULL,
  c.crit.improv = NULL,
  c.X.knots = NULL,
  c.Y.knots = NULL,
  c.max.iter.no = NULL,
  c.fast.optim = NULL,
  c.corr.term = NULL,
  study.pol = NULL,
  use.chull = NULL,
  generalize.f = NULL,
  maximize = NULL,
  method = NULL,
  compare.f = NULL
)
}
\arguments{
\item{obj}{An object of hespdiv class. The base object whose call will be
modified to produce alternative hespdiv objects.}

\item{comb.args}{A Boolean value. Do you want to combine the provided
argument values to make alternative hespdiv calls? If not, then at once only one
argument will be modified, trying all provided values for it one by one.}

\item{pick.n.args}{A numeric vector that controls how many arguments would
you like to change at once in hespdiv runs. Multiple values allowed.}

\item{comb.type}{A character determining how combinations of argument values
are selected. Possible values: "all", "random", or "handpicked".}

\item{n.combs}{An integer controlling how many argument value combinations
should be randomly selected from all possible combinations when comb.type
is "random".}

\item{display}{A Boolean value. The value of the "display" argument in each
hespdiv call.}

\item{images.path}{A path to an existing directory where PNG images of the
displayed results will be saved. If NULL (default), images won't be saved.}

\item{paired}{Logical. Are the provided \code{data} and \code{xy.dat}
arguments paired?}

\item{pnts.col}{The value of the "pnts.col" argument in each hespdiv call.}

\item{data}{A list containing matrices, time-series, lists, data frames,
vectors, or other data structures.}

\item{xy.dat, study.pol}{Lists of coordinate data frames. Each data frame
must contain two columns named \code{x} and \code{y}.}

\item{same.n.split, c.fast.optim, use.chull, c.splits}{Lists with a Boolean value (if used, should be different from the
one in the basal hespdiv call).}

\item{n.split.pts, c.max.iter.no, N.crit, N.loc.crit, c.X.knots, c.Y.knots}{Lists with integer values.}

\item{N.rel.crit, N.loc.rel.crit, S.crit, S.rel.crit}{Lists with values between 0 and 1.}

\item{Q.crit, c.Q.crit, c.crit.improv}{Lists of numeric values.}

\item{c.corr.term}{A list of numeric values between 0.01 and 0.2.}

\item{generalize.f, compare.f}{Lists of functions.}

\item{maximize}{A list of logical values with the same length as the
\code{compare.f} list.}

\item{method}{A list of character values.}
}
\value{
An object of class \code{hsa}. The object is a list with two elements:
\describe{
\item{\code{Alternatives}}{A list containing the alternative
\code{hespdiv} objects produced by the sensitivity analysis.}
\item{\code{Basis}}{The original \code{hespdiv} object whose call was
modified to produce the alternative subdivisions.}
}
}
\description{
This function is one of the two that perform HespDiv sensitivity analysis.
It creates and evaluates alternative hespdiv calls, according to the desired
changes in method, data and other subdivision criteria arguments. As a result,
it returns alternative hespdiv objects that can be directly compared with
the original hespdiv object and with each other using \code{plot.hsa} and
\code{hsa_quant} functions.
}
\details{
\subsection{Difference Between "hsa" And "hsa_detailed"}{
The major difference between "hsa_detailed" and "hsa" is that the former produces
all possible hespdiv calls from combinations of the provided hespdiv arguments.
Therefore, it samples a much smaller segment of the parameter space but more
densely, requiring much more computation time. Although such behavior may
be desired in some cases, the "hsa" function is generally more suitable for
performing hespdiv sensitivity analysis.

Additionally, alternative values for hespdiv arguments in the "hsa_detailed"
function are provided in lists, whereas in the "hsa" function, they are
provided in vectors or lists (depending on the argument).
}
\subsection{Internally Set Default Argument Values}{
When comb.args is TRUE, the default value of comb.type is "all".

When comb.args is TRUE and pick.n.args is NULL (default), the value of
pick.n.args will be changed to a vector 1:N, where N is the maximum possible
value of pick.n.args. The maximum possible value for pick.n.args depends on
the hespdiv arguments provided. Each hespdiv argument that
influences the results is counted as one, except for "data" and "xy.dat"
when paired is TRUE, and all four arguments ("method", "compare.f",
"generalize.f", and "maximize") that define the subdivision method, as the
pair/group of them is counted as one. Therefore, N can vary from 1 (single
argument provided) to 22 (all arguments provided and paired is FALSE). If
comb.args is FALSE, then pick.n.args should be NULL. Using pick.n.args = 1
is the same as setting comb.type to FALSE.
}
\subsection{Paired Arguments}{
If paired is TRUE, the "data" and "xy.dat" elements with the same index are
treated as one value of the same argument. Therefore, the provided lists of
"data" and "xy.dat" should be of the same length. Pairing of "data" and
"xy.dat" can be useful, for example, when you want to re-run hespdiv after
adding or removing some observations (these changes should be made in both
"xy.dat" and "data") to test how hespdiv results are influenced by some
extra observations or the number of observations in general. When paired
is FALSE, the number of observations in "data" and "xy.dat" must be the same
as it was in the call of the base hespdiv object. This option allows you to
re-run hespdiv after adding some noise to the object features (via changes
in "data") or coordinates (via changes in "xy.dat") to test how hespdiv
results are influenced by the data itself or localization.

By default, arguments determining the custom method ("compare.f",
"generalize.f", "maximize") are paired, similar to how "data" and "xy.dat"
are paired when paired is TRUE. Thus, the lists of "compare.f",
"generalize.f", and "maximize" should be of the same length.
}
}
\note{
Use "pnts.col" of length >1 only when the number of observations does
not change.

If a particular call produced a warning or error, then a list of
length 2 will be returned for that call. If a warning was produced, then the
first element of the list will hold the created hespdiv object, and the
second element will contain the warning message. In the case of an error,
the first element will be a list of arguments used to produce the call,
and the second element will contain the error message.
}
\seealso{
Other functions for hespdiv sensitivity analysis: 
\code{\link{change_base}()},
\code{\link{hsa}()},
\code{\link{hsa_quant}()},
\code{\link{hsa_sample_constrained}()},
\code{\link{plot_cs_hsa}()},
\code{\link{plot_hsa}()},
\code{\link{plot_hsa_q}()}

Other functions for hespdiv results post-processing: 
\code{\link{cross_comp}()},
\code{\link{hsa}()},
\code{\link{hsa_quant}()},
\code{\link{hsa_sample_constrained}()},
\code{\link{nulltest}()},
\code{\link{taxon_effect}()}
}
\author{
Liudas Daumantas
}
\concept{functions for hespdiv results post-processing}
\concept{functions for hespdiv sensitivity analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{example_hespdiv}
\alias{example_hespdiv}
\title{Example hespdiv object}
\format{
A \code{hespdiv} object.
}
\description{
A small \code{hespdiv} object created from simulated occurrence data with a
known boundary at \code{x = 0.45}. The object is intended for examples,
tests, and demonstrations of plotting and post-processing functions.
}
\details{
The object was generated using simulated coordinates and taxon names.
Occurrences on the left and right sides of the artificial boundary share
several common taxa but also include side-specific taxa.
The exact code used to simulate the dataset and run \code{hespdiv} is also
demonstrated in the examples of \code{hespdiv} function.
}
\examples{
data(example_hespdiv)
plot_hespdiv(example_hespdiv)

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_nullhespdiv.R
\name{plot.nullhespdiv}
\alias{plot.nullhespdiv}
\title{Plot a nullhespdiv object}
\usage{
\method{plot}{nullhespdiv}(x, ...)
}
\arguments{
\item{x}{A \code{nullhespdiv} object.}

\item{...}{Additional arguments passed to \code{graphics::boxplot()}.}
}
\value{
No return value. Called for the side effect of plotting a
\code{nullhespdiv} object.
}
\description{
Plot method for a \code{nullhespdiv} object.
}
\examples{
test_results <- nulltest(example_hespdiv)
plot(test_results)
}
\seealso{
Other HespDiv visualization options: 
\code{\link{blok3d}()},
\code{\link{create_gif}()},
\code{\link{dendro}()},
\code{\link{plot_cs_hsa}()},
\code{\link{plot_hespdiv}()},
\code{\link{plot_hsa}()},
\code{\link{plot_hsa_q}()},
\code{\link{poly_scheme}()},
\code{\link{polypop}()}
}
\author{
Liudas Daumantas
}
\concept{HespDiv visualization options}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/block3D.R
\name{blok3d}
\alias{blok3d}
\title{Draw hespdiv polygons in 3D space}
\usage{
blok3d(
  obj,
  height = "mean",
  color.seed = 1,
  lines = TRUE,
  pnts.col = NULL,
  obs = TRUE
)
}
\arguments{
\item{obj}{An object of the hespdiv class}

\item{height}{A character vector with default value \code{"mean"}. Determines
which information from the \code{poly.stats} data frame is encoded as the
height of the polygons. Options are \code{"mean"}, \code{"sd"},
\code{"best"}, \code{"z.score"}, \code{"str.best"},
\code{"str.z.score"}, and \code{"rank"}. Multiple values are allowed.}

\item{color.seed}{An integer that controls the colors of the polygons. Change
it to a different number if you want to get a different set of colors.}

\item{lines}{A Boolean value. Do you want split-lines to be displayed over
the top of the polygons?}

\item{pnts.col}{A character or numeric vector. Color codes to be used for
displaying observations.}

\item{obs}{A Boolean value. Do you want observations to be displayed over the top of
the polygons?}
}
\value{
No return value. Called for the side effect of producing a 3D plot
using the \pkg{rgl} graphics engine.
}
\description{
This function visualizes HespDiv polygons in 3D space. The height axis
corresponds to a chosen column from the "poly.stats" data frame.
}
\details{
The function opens an \pkg{rgl} device for each selected column of the
\code{poly.stats} data frame.

Visualizing values from \code{poly.stats} as polygon height can provide
insight into spatial heterogeneity in the analyzed data and its hierarchical
spatial structure.

The \code{height = "rank"} option provides a more intuitive way to understand
the position of each polygon in the spatial hierarchy than
\code{poly_scheme()}.

Because higher-rank polygons are displayed above lower-rank polygons, they may
obscure the view. For this reason, \code{polypop(obj, height)} can be used
with the same arguments to interactively select unwanted polygons and remove
them from the actives plot.
}
\seealso{
Other HespDiv visualization options: 
\code{\link{create_gif}()},
\code{\link{dendro}()},
\code{\link{plot.nullhespdiv}()},
\code{\link{plot_cs_hsa}()},
\code{\link{plot_hespdiv}()},
\code{\link{plot_hsa}()},
\code{\link{plot_hsa_q}()},
\code{\link{poly_scheme}()},
\code{\link{polypop}()}
}
\author{
Liudas Daumantas
}
\concept{HespDiv visualization options}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxon_effect.R
\name{taxon_effect}
\alias{taxon_effect}
\title{Taxon-level leave-out contributions for split-lines}
\usage{
taxon_effect(obj)
}
\arguments{
\item{obj}{A \code{hespdiv} object.}
}
\value{
A list of class \code{taxon_effect_result} with:
\describe{
\item{\code{elim.comp.vals}}{
Performance after removing each taxon; dimension
\code{[n_splits x n_taxa]}.
}
\item{\code{delta}}{
Signed contribution \eqn{\Delta} as defined above; dimension
\code{[n_splits x n_taxa]}.
}
\item{\code{n_per_pol}}{
Counts of the focal taxon per split and polygon.
}
\item{\code{baseline}}{
The original performance vector.
}
}
}
\description{
For each unique taxon label in \code{obj$call.info$Call_ARGS$data}
(for example, species or genus), remove all of its occurrences from both
child polygons of every split, recompute split-line performance, and record
both the resulting value and its difference from the original.
}
\details{
Let \eqn{P} be the split-line performance stored in
\code{obj$split.stats$performance}. For a focal taxon \eqn{t}, we compute
\eqn{P^{-t}} by removing all occurrences of \eqn{t}. We report \eqn{\Delta}
so that \eqn{\Delta > 0} always indicates a \strong{positive contributor}
(that is, removal worsens performance):
\itemize{
\item If \code{obj$call.info$Call_ARGS$maximize = FALSE} (lower is better,
for example similarity), \eqn{\Delta = P^{-t} - P}.
\item If \code{obj$call.info$Call_ARGS$maximize = TRUE} (higher is better,
for example distance), \eqn{\Delta = P - P^{-t}}.
}
If a taxon is absent from a split's parent polygons, elimination is a no-op
and \eqn{\Delta = NA}.
}
\seealso{
Other functions for hespdiv results post-processing: 
\code{\link{cross_comp}()},
\code{\link{hsa}()},
\code{\link{hsa_detailed}()},
\code{\link{hsa_quant}()},
\code{\link{hsa_sample_constrained}()},
\code{\link{nulltest}()}
}
\concept{functions for hespdiv results post-processing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_hsa.R
\name{plot_hsa}
\alias{plot_hsa}
\title{Visualize the stability of hespdiv polygons}
\usage{
plot_hsa(
  obj,
  alpha = 0.6,
  split.col = "gray20",
  pnts.col = NULL,
  pol.col = "7",
  type = 1,
  basal.col = 2,
  max.lwd = 3,
  min.lwd = 0.5,
  split.col.seed = NULL,
  newplot = TRUE,
  seperated = TRUE
)
}
\arguments{
\item{obj}{An object of class \code{hsa}}

\item{alpha}{The alpha value for transparency of split lines. Default is 0.6.}

\item{split.col}{Color of alternative subdivision split-lines. Default is "gray20".}

\item{pnts.col}{The color of data points. Default is NULL.}

\item{pol.col}{The color of polygons. Default is "7".}

\item{type}{An integer indicating the type of plot. Default is 1.}

\item{basal.col}{The color of basal subdivision split-lines.}

\item{max.lwd}{The maximum line width for split-lines. Default is 3.}

\item{min.lwd}{The minimum line width for split-lines. Default is 0.5.}

\item{split.col.seed}{A seed for generating random colors for split lines. Default is NULL.}

\item{newplot}{Create a plot in new device?}

\item{seperated}{Boolean. When \code{type} is >= 3, open a new graphical device for each rank?}
}
\value{
No return value, called for plotting sensitivity analysis results.
}
\description{
The function uses hespdiv sensitivity analysis results to visually
demonstrate the stability of the basal hespdiv subdivision. This is achieved
by displaying both alternative and basal hespdiv subdivisions on the same plot.
}
\details{
The \code{type} parameter determines the type of plot generated:
\describe{
\item{\code{1}}{
Basic plot: displays the alternative and basal hespdiv subdivisions on
the same plot without split-line ranks or titles.
}

\item{\code{2}}{
Plot with split-line ranks: includes split-line ranks in the plot.
Each split-line is assigned a different line width based on its rank.
}

\item{\code{3}}{
Plot with separate ranks: generates multiple plots, each representing
split-line ranks up to a certain value.
}

\item{\code{4}}{
Plot with separate and isolated ranks: similar to mode 3 but isolates
split-line ranks. Generates multiple plots, each representing a
specific split-line rank.
}
}

If the alternative subdivisions spatially converge but the basal
subdivision lies far from the zone of convergence, you can use
\code{change_base} to select a more representative alternative
subdivision to serve as the basal subdivision. However, you should
verify that the arguments used in that subdivision are appropriate.
}
\note{
\code{newplot} allows the legend to be rendered correctly in
types 2 and 3, and helps with line rendering in general when drawing
in an active device (use \code{broom} otherwise to delete devices).
}
\seealso{
Other functions for hespdiv sensitivity analysis: 
\code{\link{change_base}()},
\code{\link{hsa}()},
\code{\link{hsa_detailed}()},
\code{\link{hsa_quant}()},
\code{\link{hsa_sample_constrained}()},
\code{\link{plot_cs_hsa}()},
\code{\link{plot_hsa_q}()}

Other HespDiv visualization options: 
\code{\link{blok3d}()},
\code{\link{create_gif}()},
\code{\link{dendro}()},
\code{\link{plot.nullhespdiv}()},
\code{\link{plot_cs_hsa}()},
\code{\link{plot_hespdiv}()},
\code{\link{plot_hsa_q}()},
\code{\link{poly_scheme}()},
\code{\link{polypop}()}
}
\author{
Liudas Daumantas
}
\concept{HespDiv visualization options}
\concept{functions for hespdiv sensitivity analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/poly_scheme.R
\name{poly_scheme}
\alias{poly_scheme}
\title{Schematic plot of hespdiv polygons}
\usage{
poly_scheme(obj, segment = TRUE, id = TRUE, seed = 1)
}
\arguments{
\item{obj}{A hespdiv object.}

\item{segment}{A Boolean value. Display the punctuated lines joining the
polygon centroids with the split-lines?}

\item{id}{A Boolean value. Display the IDs of polygons?}

\item{seed}{An integer value that determines the random set of colors used
in visualization of split-lines and polygons.}
}
\value{
A ggplot object.
}
\description{
This function generates a schematic visualization of subdivided territory.
It highlights the location of each polygon by displaying their centroids,
ID labels, and punctuated lines that connect the polygon centroids with the
split-lines that created them. This visualization represents the spatial
arrangement of hespdiv polygons within the territory in 2D.
}
\note{
A much clearer way to visualize polygons is by using the \code{blok3D}
function, with \code{height = "rank"}. However, a 3D plot is less suitable
option for papers.
}
\examples{
poly_scheme(example_hespdiv)

}
\seealso{
Other HespDiv visualization options: 
\code{\link{blok3d}()},
\code{\link{create_gif}()},
\code{\link{dendro}()},
\code{\link{plot.nullhespdiv}()},
\code{\link{plot_cs_hsa}()},
\code{\link{plot_hespdiv}()},
\code{\link{plot_hsa}()},
\code{\link{plot_hsa_q}()},
\code{\link{polypop}()}
}
\author{
Liudas Daumantas
}
\concept{HespDiv visualization options}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hespdiv_nullltest.R
\name{nulltest}
\alias{nulltest}
\title{Test significance of split-lines in spatial subdivision}
\usage{
nulltest(
  obj,
  n = 999,
  maintain.n = TRUE,
  shuffle.scope = "within",
  shuffle.type = "localities"
)
}
\arguments{
\item{obj}{An object of class \code{hespdiv}.}

\item{n}{Integer. Number of permutations used to form the null distribution.}

\item{maintain.n}{Logical. Only honored when \code{shuffle.scope = "within"} and \code{shuffle.type = "localities"}. If \code{TRUE}, attempts to keep the randomized child polygon occurrence counts close to the observed (maximum discrepancy up to \code{max_n/2}, where \code{max_n} is the largest locality size). Ignored otherwise.}

\item{shuffle.scope}{Character. Either \code{"all"} (shuffle across the full study area) or \code{"within"} (shuffle only within each parent polygon).}

\item{shuffle.type}{Character. Either \code{"localities"} (shuffle whole localities, preserving assemblages) or \code{"occurrences"} (shuffle individual occurrences). \code{"localities"} is generally preferred.}
}
\value{
Invisibly returns an object of class \code{nullhespdiv}, a list with:
\itemize{
\item \code{$stats}, a data frame summarizing each split-line with:
\itemize{
\item \code{performance}: observed performance.
\item \code{mean.random}: mean of null performances.
\item \code{sd.random}: standard deviation of null performances.
\item \code{p.val}: empirical one-sided p-value (ties included).
\item \code{z.score.random}: standardized effect size.
}
\item \code{$null}, a matrix or data frame containing all null performance
values for every split-line across permutations.
}
}
\description{
Assess the statistical significance of each split-line (bioregion boundary) identified by \code{hespdiv} by comparing its observed performance to a null distribution. The null is generated by permuting the data \code{n} times and recomputing the split-line performance after each permutation. Multiple shuffling strategies are supported to probe the influence of spatial structure on delineation.
}
\details{
Two shuffling scopes are available:
\describe{
\item{\strong{"all"}:}{Shuffle across the entire study area, ignoring polygonal subdivisions.}
\item{\strong{"within"}:}{Shuffle only within each parent polygon (the region in which the split-line is nested), preserving local spatial structure.}
}

Two shuffling types are available:
\describe{
\item{\strong{"localities"}:}{Shuffle whole localities, preserving each site's assemblage (recommended, since occurrences within a locality are not independent).}
\item{\strong{"occurrences"}:}{Shuffle individual occurrences across sites (use with caution; may violate within-locality independence).}
}

For each split-line, the function reports the observed performance, the mean and standard deviation of the permuted (null) performances, an empirical one-sided p-value (proportion of permuted values \emph{as or more extreme} than observed; ties included), and a z-score quantifying departure from the null.
}
\examples{
# if split-line is strongly significant, the choice of parameters should not
# matter. For example (look at p-value, z.score.random, sd.random and
# mean.random):
(nulltest(example_hespdiv, maintain.n = FALSE, shuffle.type = "occurrences"))
(nulltest(example_hespdiv, maintain.n = FALSE, shuffle.type = "localities"))
(nulltest(example_hespdiv, maintain.n = TRUE, shuffle.type = "localities"))

}
\seealso{
Other functions for hespdiv results post-processing: 
\code{\link{cross_comp}()},
\code{\link{hsa}()},
\code{\link{hsa_detailed}()},
\code{\link{hsa_quant}()},
\code{\link{hsa_sample_constrained}()},
\code{\link{taxon_effect}()}
}
\concept{functions for hespdiv results post-processing}
